/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <mapiext.h>

#include "archivectrl.h"
#include "archiver-session.h"
#include "storehelper.h"
#include "operations.h"

#include "ECLogger.h"
#include "ECConfig.h"
#include "restrictionutil.h"
#include "ECRestriction.h"
#include "stringutil.h"

#include <iostream>
#include <algorithm>
using namespace std;

#include "mapi_ptr.h"

/**
 * Create a new Archive object.
 *
 * @param[in]	lpSession
 *					Pointer to the Session.
 * @param[in]	lpConfig
 *					Pointer to an ECConfig object that determines the operational options.
 * @param[in]	lpLogger
 *					Pointer to an ECLogger object that's used for logging.
 * @param[out]	lpptrArchiver
 *					Pointer to a ArchivePtr that will be assigned the address of the returned object.
 *
 * @return HRESULT
 */
eResult ArchiveControlImpl::Create(Session *lpSession, ECConfig *lpConfig, ECLogger *lpLogger, ArchiveControlPtr *lpptrArchiveControl)
{
	HRESULT hr = hrSuccess;
	std::auto_ptr<ArchiveControlImpl> ptrArchiveControl;
	
	try {
		ptrArchiveControl.reset(new ArchiveControlImpl(lpSession, lpConfig, lpLogger));
	} catch (bad_alloc &) {
		hr = MAPI_E_NOT_ENOUGH_MEMORY;
		goto exit;
	}
	
	hr = ptrArchiveControl->Init();
	if (hr != hrSuccess)
		goto exit;
		
	*lpptrArchiveControl = ptrArchiveControl;	// transfers ownership
	
exit:
	return MAPIErrorToArchiveError(hr);
}

/**
 * Constructor
 *
 * @param[in]	lpSession
 *					Pointer to the Session.
 * @param[in]	lpConfig
 *					Pointer to an ECConfig object that determines the operational options.
 * @param[in]	lpLogger
 *					Pointer to an ECLogger object that's used for logging.
 *
 * @return HRESULT
 */
ArchiveControlImpl::ArchiveControlImpl(Session *lpSession, ECConfig *lpConfig, ECLogger *lpLogger)
: m_lpSession(lpSession)
, m_lpConfig(lpConfig)
, m_lpLogger(lpLogger)
, m_bArchiveEnable(true)
, m_ulArchiveAfter(30)
, m_bDeleteEnable(false)
, m_bDeleteUnread(false)
, m_ulDeleteAfter(0)
, m_bStubEnable(false)
, m_bStubUnread(false)
, m_ulStubAfter(0)
{ }

/**
 * Destructor
 */
ArchiveControlImpl::~ArchiveControlImpl()
{ }

/**
 * Initialize the Archiver object.
 */
HRESULT ArchiveControlImpl::Init()
{
	m_bArchiveEnable = parseBool(m_lpConfig->GetSetting("archive_enable", "", "no"));
	m_ulArchiveAfter = atoi(m_lpConfig->GetSetting("archive_after", "", "30"));

	m_bDeleteEnable = parseBool(m_lpConfig->GetSetting("delete_enable", "", "no"));
	m_bDeleteUnread = parseBool(m_lpConfig->GetSetting("delete_unread", "", "no"));
	m_ulDeleteAfter = atoi(m_lpConfig->GetSetting("delete_after", "", "0"));

	m_bStubEnable = parseBool(m_lpConfig->GetSetting("stub_enable", "", "no"));
	m_bStubUnread = parseBool(m_lpConfig->GetSetting("stub_unread", "", "no"));
	m_ulStubAfter = atoi(m_lpConfig->GetSetting("stub_after", "", "0"));

	m_bPurgeEnable = parseBool(m_lpConfig->GetSetting("purge_enable", "", "no"));
	m_ulPurgeAfter = atoi(m_lpConfig->GetSetting("purge_after", "", "0"));
	
	GetSystemTimeAsFileTime(&m_ftCurrent);
	
	return hrSuccess;
}

/**
 * Archive messages for all users. Optionaly only user that have their store on the server
 * to which the archiver is connected will have their messages archived.
 *
 * @param[in]	bLocalOnly
 *					If set to true only  messsages for users that have their store on the local server
 *					will be archived.
 *
 * @return HRESULT
 */
eResult ArchiveControlImpl::ArchiveAll(bool bLocalOnly)
{
	HRESULT hr = hrSuccess;
	ABContainerPtr ptrGAL;
	ABContainerPtr ptrABContainer;
	MAPITablePtr ptrFolderTable;
	mapi_rowset_ptr ptrFolderRows;
	ULONG ulType = 0;
	
	SizedSPropTagArray(1, sptaFolderProps) = {1, {PR_ENTRYID}};
	
	hr = m_lpSession->GetGAL(&ptrGAL);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open the GAL. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	// Process root of GAL
	hr = ArchiveAllFromContainer(ptrGAL, bLocalOnly);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to process GAL. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	
	// Process subfolders
	hr = ptrGAL->GetHierarchyTable(CONVENIENT_DEPTH|MAPI_DEFERRED_ERRORS, &ptrFolderTable);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get GAL hierarchy table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrFolderTable->SetColumns((LPSPropTagArray)&sptaFolderProps, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to select folder table columns. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	while (true) {
		hr = ptrFolderTable->QueryRows(50, 0, &ptrFolderRows);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get rows from folder table. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		for (ULONG i = 0; i < ptrFolderRows.size(); ++i) {
			hr = ptrGAL->OpenEntry(ptrFolderRows[i].lpProps[0].Value.bin.cb, (LPENTRYID)ptrFolderRows[i].lpProps[0].Value.bin.lpb, &ptrABContainer.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrABContainer);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open folder. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}

			hr = ArchiveAllFromContainer(ptrABContainer, bLocalOnly);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to process folder. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}
		}

		if (ptrFolderRows.size() < 50)
			break;
	}
	
exit:
	return MAPIErrorToArchiveError(hr);
}

HRESULT ArchiveControlImpl::ArchiveAllFromContainer(ABContainerPtr &ptrABContainer, bool bLocalOnly)
{
	HRESULT hr = hrSuccess;
	MAPITablePtr ptrUserTable;
	mapi_rowset_ptr ptrUserRows;
	string strLocalServername;
	
	SRestrictionPtr ptrUserRestriction;
	SPropValue		sObjTypePropVal = {0};
	SPropValue		sDispTypePropVal = {0};
	
	SizedSPropTagArray(2, sptaUserProps) = {2, {PR_ACCOUNT_A, PR_EC_HOMESERVER_NAME}};
	
	sObjTypePropVal.ulPropTag = PR_OBJECT_TYPE;
	sObjTypePropVal.Value.l = MAPI_MAILUSER;

	sDispTypePropVal.ulPropTag = PR_DISPLAY_TYPE;
	sDispTypePropVal.Value.l = DT_MAILUSER;

	hr = ECAndRestriction(
			ECPropertyRestriction(RELOP_EQ, PR_OBJECT_TYPE, &sObjTypePropVal, ECRestriction::Cheap) +
			ECPropertyRestriction(RELOP_EQ, PR_DISPLAY_TYPE, &sDispTypePropVal, ECRestriction::Cheap)
		).CreateMAPIRestriction(&ptrUserRestriction, ECRestriction::Cheap);
	if (hr != hrSuccess)
		goto exit;

	hr = ptrABContainer->GetContentsTable(MAPI_DEFERRED_ERRORS, &ptrUserTable);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get GAL contents table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrUserTable->SetColumns((LPSPropTagArray)&sptaUserProps, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to select table columns. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrUserTable->Restrict(ptrUserRestriction, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to restrict table columns. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	while (true) {
		bool bSkip = false;
	
		hr = ptrUserTable->QueryRows(50, 0, &ptrUserRows);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get rows from table. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		for (ULONG j = 0; j < ptrUserRows.size(); ++j) {
			if (bLocalOnly) {
				if (strLocalServername.empty()) {
					bool bIsLocal = false;
				
					hr = m_lpSession->ServerIsLocal(ptrUserRows[j].lpProps[1].Value.lpszA, &bIsLocal);
					if (hr != hrSuccess) {
						m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to determine if server '%s' is locale. (hr=%s)", ptrUserRows[j].lpProps[1].Value.lpszA, stringify(hr, true).c_str());
						goto exit;
					}
					
					if (bIsLocal)
						strLocalServername = ptrUserRows[j].lpProps[1].Value.lpszA;
					else
						bSkip = true;
				} else if (strLocalServername.compare(ptrUserRows[j].lpProps[1].Value.lpszA) != 0)
					bSkip = true;
			}
		
			if (bSkip) {
				m_lpLogger->Log(EC_LOGLEVEL_INFO, "Skipping user '%s' from server '%s'.", ptrUserRows[j].lpProps[0].Value.lpszA, ptrUserRows[j].lpProps[1].Value.lpszA);
				continue;
			}
			
			hr = Archive(ptrUserRows[j].lpProps[0].Value.lpszA);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to archive user %s. (hr=%s)", ptrUserRows[j].lpProps[0].Value.lpszA, stringify(hr, true).c_str());
				continue;
			}
		}
		
		if (ptrUserRows.size() < 50)
			break;
	}	



	
exit:
	return hr;
}

/**
 * Archive the messages of a particular user.
 *
 * @param[in]	lpszUser
 *					The username for which to archive the messages.
 *
 * @return HRESULT
 */
eResult ArchiveControlImpl::Archive(const char *lpszUser)
{
	HRESULT hr = hrSuccess;
	MsgStorePtr ptrUserStore;
	StoreHelperPtr ptrStoreHelper;
	MAPIFolderPtr ptrSearchArchiveFolder;
	MAPIFolderPtr ptrSearchDeleteFolder;
	MAPIFolderPtr ptrSearchStubFolder;
	ObjectEntryList lstArchives;
	
	SizedSPropTagArray(3, sptaExcludeProps) = {3, {PROP_ARCHIVE_STORE_ENTRYIDS, PROP_ARCHIVE_ITEM_ENTRYIDS, PROP_STUBBED}};

	if (lpszUser == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Archiving store for user %s", lpszUser);
	
	hr = m_lpSession->OpenStoreByName(lpszUser, &ptrUserStore);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open store. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	PROPMAP_INIT_NAMED_ID(ARCHIVE_STORE_ENTRYIDS, PT_MV_BINARY, PSETID_Archive, dispidStoreEntryIds)
	PROPMAP_INIT_NAMED_ID(ARCHIVE_ITEM_ENTRYIDS, PT_MV_BINARY, PSETID_Archive, dispidItemEntryIds)
	PROPMAP_INIT_NAMED_ID(STUBBED, PT_BOOLEAN, PSETID_Archive, dispidStubbed)
	PROPMAP_INIT(ptrUserStore)

	hr = StoreHelper::Create(ptrUserStore, &ptrStoreHelper);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create store helper. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrStoreHelper->GetArchiveList(&lstArchives);
	if (hr != hrSuccess) {
		if (hr == MAPI_E_CORRUPT_DATA) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "List of archives is corrupt for user %s, skipping user.", lpszUser);
			hr = hrSuccess;
		} else
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get list of archives. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	if (lstArchives.empty()) {
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "%s has no attached archives", lpszUser);
		goto exit;
	}
	
	hr = ptrStoreHelper->GetSearchFolders(&ptrSearchArchiveFolder, &ptrSearchDeleteFolder, &ptrSearchStubFolder);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get the search folders. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	
	if (m_bArchiveEnable) {
		// Archive all unarchived messages that are old enough
		Copier sCopier(m_lpSession, m_lpConfig, m_lpLogger, lstArchives, (LPSPropTagArray)&sptaExcludeProps);
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Archiving messages");
		hr = ProcessFolder(ptrSearchArchiveFolder, &sCopier, m_ulArchiveAfter, true);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to archive messages. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done archiving messages");
	}
	
	
	if (m_bDeleteEnable) {
		// First delete all messages that are elegible for deletion, so we don't unneccesary stub them first
		Deleter sDeleter(m_lpLogger);
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Deleting old messages");
		hr = ProcessFolder(ptrSearchDeleteFolder, &sDeleter, m_ulDeleteAfter, m_bDeleteUnread);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to delete old messages. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done deleting messages");
	}
	
	
	if (m_bStubEnable) {
		// Now stub the remaing messages (if they're old enough)
		Stubber sStubber(m_lpLogger, PROP_STUBBED);
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Stubbing messages");
		hr = ProcessFolder(ptrSearchStubFolder, &sStubber, m_ulStubAfter, m_bStubUnread);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to stub messages. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done stubbing messages");
	}

	if (m_bPurgeEnable) {
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Purging archive(s)");
		hr = PurgeArchives(lstArchives);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to purge archive(s). (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done purging archive(s)");
	}
	
exit:
	return MAPIErrorToArchiveError(hr);
}

/**
 * Process a search folder and place an additional restriction on it to get the messages
 * that should really be archived.
 *
 * @param[in]	ptrFolder
 *					A MAPIFolderPtr that points to the search folder to be processed.
 * @param[in]	lpArchiveOperation
 *					The pointer to a IArchiveOperation derived object that's used to perform
 *					the actual processing.
 * @param[in]	ulAge
 *					The age in days since the message was delivered, that a message must be before
 *					it will be processed.
 * @param[in]	bProcessUnread
 *					If set to true, unread messages will also be processed. Otherwise unread message
 *					will be left untouched.
 *
 * @return HRESULT
 */
HRESULT ArchiveControlImpl::ProcessFolder(MAPIFolderPtr &ptrFolder, IArchiveOperation *lpArchiveOperation, int ulAge, bool bProcessUnread)
{
	HRESULT hr = hrSuccess;
	MAPITablePtr ptrTable;
	LPSRestriction lpRestriction = NULL;
	SSortOrderSetPtr ptrSortOrder;
	SPropValue sPropCreationTime;
	ULARGE_INTEGER li;
	mapi_rowset_ptr ptrRowSet;
	MessagePtr ptrMessage;
	
	SizedSPropTagArray(3, sptaProps) = {3, {PR_ENTRYID, PR_PARENT_ENTRYID, PR_STORE_ENTRYID}};
	
	if (ulAge < 0)
		goto exit;
	
	hr = ptrFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &ptrTable);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get search folder contents table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrTable->SetColumns((LPSPropTagArray)&sptaProps, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set columns on table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
		
	CREATE_RESTRICTION(lpRestriction);
	if (bProcessUnread) {
		DATA_RES_PROPERTY_CHEAP(lpRestriction, *lpRestriction, RELOP_LT, PR_MESSAGE_DELIVERY_TIME, &sPropCreationTime);
	} else {
		CREATE_RES_AND(lpRestriction, lpRestriction, 2);
			DATA_RES_PROPERTY_CHEAP(lpRestriction, lpRestriction->res.resAnd.lpRes[0], RELOP_LT, PR_MESSAGE_DELIVERY_TIME, &sPropCreationTime);
			DATA_RES_BITMASK(lpRestriction, lpRestriction->res.resAnd.lpRes[1], BMR_NEZ, PR_MESSAGE_FLAGS, MSGFLAG_READ);
	}
		
	li.LowPart = m_ftCurrent.dwLowDateTime;
	li.HighPart = m_ftCurrent.dwHighDateTime;
	
	li.QuadPart -= (ulAge * _DAY);
		
	sPropCreationTime.ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	sPropCreationTime.Value.ft.dwLowDateTime = li.LowPart;
	sPropCreationTime.Value.ft.dwHighDateTime = li.HighPart;
		
	hr = ptrTable->Restrict(lpRestriction, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set restriction on table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
		
	hr = MAPIAllocateBuffer(CbNewSSortOrderSet(1), &ptrSortOrder);
	if (hr != hrSuccess)
		goto exit;
	
	ptrSortOrder->cSorts = 1;
	ptrSortOrder->cCategories = 0;
	ptrSortOrder->cExpanded = 0;
	ptrSortOrder->aSort[0].ulPropTag = PR_PARENT_ENTRYID;
	ptrSortOrder->aSort[0].ulOrder = TABLE_SORT_ASCEND ;
	
	hr = ptrTable->SortTable(ptrSortOrder, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to sort table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;	
	}
		
	while (true) {
		hr = ptrTable->QueryRows(50, 0, &ptrRowSet);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get rows from table. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Processing batch of %u messages", ptrRowSet.size());
		for (ULONG i = 0; i < ptrRowSet.size(); ++i) {
			hr = lpArchiveOperation->ProcessEntry(ptrFolder, ptrRowSet[i].cValues, ptrRowSet[i].lpProps);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to process entry. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}
		}
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Done processing batch");
	
		if (ptrRowSet.size() < 50)
			break;
	}
		
exit:
	if (lpRestriction)
		MAPIFreeBuffer(lpRestriction);

	return hr;
}

HRESULT ArchiveControlImpl::PurgeArchives(const ObjectEntryList &lstArchives)
{
	HRESULT hr = hrSuccess;
	ObjectEntryList::const_iterator iArchive;
	bool bErrorOccurred = false;
	LPSRestriction lpRestriction = NULL;
	SPropValue sPropCreationTime;
	ULARGE_INTEGER li;
	mapi_rowset_ptr ptrRowSet;

	SizedSPropTagArray(1, sptaFolderProps) = {1, {PR_ENTRYID}};

	// Create the common restriction that determines which messages are old enough to purge.
	CREATE_RESTRICTION(lpRestriction);
	DATA_RES_PROPERTY_CHEAP(lpRestriction, *lpRestriction, RELOP_LT, PR_MESSAGE_DELIVERY_TIME, &sPropCreationTime);
		
	li.LowPart = m_ftCurrent.dwLowDateTime;
	li.HighPart = m_ftCurrent.dwHighDateTime;
	
	li.QuadPart -= (m_ulPurgeAfter * _DAY);
		
	sPropCreationTime.ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	sPropCreationTime.Value.ft.dwLowDateTime = li.LowPart;
	sPropCreationTime.Value.ft.dwHighDateTime = li.HighPart;

	for (iArchive = lstArchives.begin(); iArchive != lstArchives.end(); ++iArchive) {
		MsgStorePtr ptrArchiveStore;
		MAPIFolderPtr ptrArchiveRoot;
		ULONG ulType = 0;
		MAPITablePtr ptrFolderTable;
		mapi_rowset_ptr ptrFolderRows;

		hr = m_lpSession->OpenStore(iArchive->sStoreEntryId, &ptrArchiveStore);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open archive. (entryid=%s, hr=%s)", iArchive->sStoreEntryId.tostring().c_str(), stringify(hr, true).c_str());
			bErrorOccurred = true;
			continue;
		}

		// Purge root of archive
		hr = PurgeArchiveFolder(ptrArchiveStore, iArchive->sItemEntryId, lpRestriction);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to purge archive root. (entryid=%s, hr=%s)", iArchive->sItemEntryId.tostring().c_str(), stringify(hr, true).c_str());
			bErrorOccurred = true;
			continue;
		}

		// Get all subfolders and purge those as well.
		hr = ptrArchiveStore->OpenEntry(iArchive->sItemEntryId.size(), iArchive->sItemEntryId, &ptrArchiveRoot.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrArchiveRoot);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open archive root. (entryid=%s, hr=%s)", iArchive->sItemEntryId.tostring().c_str(), stringify(hr, true).c_str());
			bErrorOccurred = true;
			continue;
		}

		hr = ptrArchiveRoot->GetHierarchyTable(CONVENIENT_DEPTH|MAPI_DEFERRED_ERRORS, &ptrFolderTable);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to get archive hierarchy table. (hr=%s)", stringify(hr, true).c_str());
			bErrorOccurred = true;
			continue;
		}
		
		hr = ptrFolderTable->SetColumns((LPSPropTagArray)&sptaFolderProps, TBL_BATCH);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to select folder table columns. (hr=%s)", stringify(hr, true).c_str());
			bErrorOccurred = true;
			continue;
		}

		while (true) {
			hr = ptrFolderTable->QueryRows(50, 0, &ptrFolderRows);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get rows from folder table. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}
			
			for (ULONG i = 0; i < ptrFolderRows.size(); ++i) {
				hr = PurgeArchiveFolder(ptrArchiveStore, ptrFolderRows[i].lpProps[0].Value.bin, lpRestriction);
				if (hr != hrSuccess) {
					m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to purge archive folder. (entryid=%s, hr=%s)", bin2hex(ptrFolderRows[i].lpProps[0].Value.bin.cb, ptrFolderRows[i].lpProps[0].Value.bin.lpb).c_str(), stringify(hr, true).c_str());
					bErrorOccurred = true;
				}
			}

			if (ptrFolderRows.size() < 50)
				break;
		}
	}

exit:
	if (lpRestriction)
		MAPIFreeBuffer(lpRestriction);

	if (hr == hrSuccess && bErrorOccurred)
		hr = MAPI_W_ERRORS_RETURNED;

	return hr;
}

HRESULT ArchiveControlImpl::PurgeArchiveFolder(MsgStorePtr &ptrArchive, const entryid_t &folderEntryID, const LPSRestriction lpRestriction)
{
	HRESULT hr = hrSuccess;
	ULONG ulType = 0;
	MAPIFolderPtr ptrFolder;
	MAPITablePtr ptrContentsTable;
	list<entryid_t>::const_iterator iEntryId;
	list<entryid_t> lstEntries;
	mapi_rowset_ptr ptrRows;
	EntryListPtr ptrEntryList;
	ULONG ulIdx = 0;
	
	SizedSPropTagArray(1, sptaTableProps) = {1, {PR_ENTRYID}};

	hr = ptrArchive->OpenEntry(folderEntryID.size(), folderEntryID, &ptrFolder.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrFolder);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open archive folder. (entryid=%s, hr=%s)", folderEntryID.tostring().c_str(), stringify(hr, true).c_str());
		goto exit;
	}

	hr = ptrFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &ptrContentsTable);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open contents table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	hr = ptrContentsTable->SetColumns((LPSPropTagArray)&sptaTableProps, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to select table columns. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrContentsTable->Restrict(lpRestriction, TBL_BATCH);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to restrict contents table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}


	while (true) {
		hr = ptrContentsTable->QueryRows(50, 0, &ptrRows);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get rows from contents table. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		for (ULONG i = 0; i < ptrRows.size(); ++i)
			lstEntries.push_back(ptrRows[i].lpProps[0].Value.bin);

		if (ptrRows.size() < 50)
			break;
	}


	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Purging %lu messaged from archive folder", lstEntries.size());
	
	hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), &ptrEntryList);
	if (hr != hrSuccess)
		goto exit;
	
	hr = MAPIAllocateMore(lstEntries.size() * sizeof(SBinary), ptrEntryList, (LPVOID*)&ptrEntryList->lpbin);
	if (hr != hrSuccess)
		goto exit;
		
	ptrEntryList->cValues = lstEntries.size();
	for (iEntryId = lstEntries.begin(); iEntryId != lstEntries.end(); ++iEntryId, ++ulIdx) {
		ptrEntryList->lpbin[ulIdx].cb = iEntryId->size();
		ptrEntryList->lpbin[ulIdx].lpb = *iEntryId;
	}
	
	hr = ptrFolder->DeleteMessages(ptrEntryList, 0, NULL, 0);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to delete %u messages. (hr=%s)", ptrEntryList->cValues, stringify(hr, true).c_str());
		goto exit;
	}		
	
exit:
	return hr;
}
