/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <platform.h>
#include "archivemanage.h"
#include "archiver-session.h"
#include "archivehelper.h"
#include "storehelper.h"

#include <string>
#include <assert.h>
#include <mapiutil.h>
#include "ECLogger.h"
#include "stringutil.h"

#include "stringutil.h"

using namespace std;

/**
 * Create an ArchiveManageImpl object.
 *
 * @param[in]	lpSession
 *					Pointer to a Session object.
 * @param[in]	lpszUser
 *					The username of the user for which to create the archive manager.
 * @param[in]	lpLogger
 *					Pointer to an ECLogger object to which message will be logged.
 * @param[out]	lpptrArchiveManager
 *					Pointer to an ArchiveManagePtr that will be assigned the address of the returned object.
 *
 * @return HRESULT
 */
eResult ArchiveManageImpl::Create(Session *lpSession, const char *lpszUser, ECLogger *lpLogger, ArchiveManagePtr *lpptrArchiveManage)
{
	HRESULT hr = hrSuccess;
	std::auto_ptr<ArchiveManageImpl> ptrArchiveManage;

	if (lpszUser == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	try {
		ptrArchiveManage.reset(new ArchiveManageImpl(lpSession, lpszUser, lpLogger));
	} catch (bad_alloc &) {
		hr = MAPI_E_NOT_ENOUGH_MEMORY;
		goto exit;
	}
	
	hr = ptrArchiveManage->Init();
	if (hr != hrSuccess)
		goto exit;
	
	*lpptrArchiveManage = ptrArchiveManage;	// transfers ownership
	
exit:
	return MAPIErrorToArchiveError(hr);
}

/**
 * Constructor
 *
 * @param[in]	lpSession
 *					Pointer to a Session object.
 * @param[in]	lpszUser
 *					The username of the user for which to create the archive manager.
 * @param[in]	lpLogger
 *					Pointer to an ECLogger object to which message will be logged.
 */
ArchiveManageImpl::ArchiveManageImpl(Session *lpSession, const std::string &strUser, ECLogger *lpLogger)
: m_lpSession(lpSession)
, m_strUser(strUser)
, m_lpLogger(lpLogger)
{ }

/**
 *  Destructor
 */
ArchiveManageImpl::~ArchiveManageImpl()
{ }

/**
 * Initialize the ArchiveManager object.
 */
HRESULT ArchiveManageImpl::Init()
{
	HRESULT hr = hrSuccess;

	hr = m_lpSession->OpenStoreByName(m_strUser, &m_ptrUserStore);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open user store '%s' (hr=%s).", m_strUser.c_str(), stringify(hr, true).c_str());
		goto exit;
	}

exit:
	return hr;
}

/**
 * Attach an archive to the store of the user for which the ArchiveManger was created.
 *
 * @param[in]	lpszArchiveServer
 * 					If not NULL, this argument specifies the path of the server on which to create the archive.
 * @param[in]	lpszArchive
 *					The username of the non-active user that's the placeholder for the archive.
 * @param[in]	lpszFolder
 *					The name of the folder that will be used as the root of the archive. If ATT_USE_IPM_SUBTREE is passed
 *					in the ulFlags argument, the lpszFolder argument is ignored. If this argument is NULL, the username of
 *					the user is used as the root foldername.
 * @param[in]	ulFlags
 *					@ref flags specifying the options used for attaching the archive. 
 * @return HRESULT
 *
 * @section flags Flags
 * @li \b ATT_USE_IPM_SUBTREE	Use the IPM subtree of the archive store as the root of the archive.
 * @li \b ATT_WRITABLE			Make the archive writable for the user.
 */
eResult ArchiveManageImpl::AttachTo(const char *lpszArchiveServer, const char *lpszArchive, const char *lpszFolder, unsigned ulFlags)
{
	HRESULT	hr = hrSuccess;
	MsgStorePtr ptrArchiveStore;
	ArchiveHelperPtr ptrArchiveHelper;
	entryid_t sUserEntryId;
	string strFoldername;
	entryid_t sAttachedUserEntryId;
	StoreHelperPtr ptrStoreHelper;
	ObjectEntryList lstArchives;
	SObjectEntry SObjectEntry;
	bool bEqual = false;
	ArchiveType aType = UndefArchive;
	Session *lpArchiveSession = m_lpSession;
	Session *lpRemoteSession = NULL;
	
	// Resolve the requested user.
	hr = m_lpSession->GetUserInfo(m_strUser, &sUserEntryId, &strFoldername);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to resolve user information for '%s' (hr=%s).", m_strUser.c_str(), stringify(hr, true).c_str());
		goto exit;
	}
	
	
	if ((ulFlags & UseIpmSubtree) == UseIpmSubtree)
		strFoldername.clear();	// Empty folder name indicates tje IPM subtree.
	else if (lpszFolder)
		strFoldername.assign(lpszFolder);
		
	if (lpszArchiveServer) {
		hr = m_lpSession->CreateRemote(lpszArchiveServer, m_lpLogger, &lpRemoteSession);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to connect to archive server '%s' (hr=%s).", lpszArchiveServer, stringify(hr, true).c_str());
			goto exit;
		}
		
		lpArchiveSession = lpRemoteSession;
	}
	
	// Find the requested archive.
	hr = lpArchiveSession->OpenStoreByName(lpszArchive, &ptrArchiveStore);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open archive store '%s' (hr=%s).", lpszArchive, stringify(hr, true).c_str());
		goto exit;
	}
	


	hr = ArchiveHelper::Create(ptrArchiveStore, strFoldername, lpszArchiveServer, &ptrArchiveHelper);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create archive helper (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}

	// Check if the archive is usable for the requested type
	hr = ptrArchiveHelper->GetArchiveType(&aType);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive type (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Archive Type: %d", (int)aType);

	if (aType == SingleArchive && !strFoldername.empty()) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Attempted to create an archive folder in an archive store that has an archive in its root.");
		hr = MAPI_E_COLLISION;
		goto exit;
	} else if (aType == MultiArchive && strFoldername.empty()) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Attempted to create an archive in the root of an archive store that has archive folders.");
		hr = MAPI_E_COLLISION;
		goto exit;
	}

	// Check if the archive is attached yet.
	hr = ptrArchiveHelper->GetAttachedUser(&sAttachedUserEntryId);
	if (hr == MAPI_E_NOT_FOUND)
		hr = hrSuccess;

	else if ( hr == hrSuccess && (!sAttachedUserEntryId.empty() && sAttachedUserEntryId != sUserEntryId)) {
		string strUser;
		string strFullname;

		hr = m_lpSession->GetUserInfo(sAttachedUserEntryId, &strUser, &strFullname);
		if (hr == hrSuccess)
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Archive is already used by %s (%s).", strUser.c_str(), strFullname.c_str());
		else
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Archive is already used (user entry: %s).", sAttachedUserEntryId.tostring().c_str());

		hr = MAPI_E_COLLISION;
		goto exit;
	} else if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get attached user for the requested archive (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}
		
		

	
	// Check if we're not trying to attach a store to itself.
	hr = m_lpSession->CompareStoreIds(m_ptrUserStore, ptrArchiveStore, &bEqual);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to compare user and archive store (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}
	if (bEqual) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "User and archive store are the same.");
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
		
	// Add new archive to list of archives.
	hr = ptrArchiveHelper->GetArchiveEntry(&SObjectEntry);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive entry (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}
	
	hr = StoreHelper::Create(m_ptrUserStore, &ptrStoreHelper);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create store helper (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrStoreHelper->GetArchiveList(&lstArchives);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive list (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}
	
	lstArchives.push_back(SObjectEntry);
	lstArchives.sort();
	lstArchives.unique();
	
	hr = ptrStoreHelper->SetArchiveList(lstArchives);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to update archive list (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}
		
		
	hr = ptrArchiveHelper->SetAttachedUser(sUserEntryId);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to mark archive used (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}

	if (aType == UndefArchive) {
		hr = ptrArchiveHelper->SetArchiveType(strFoldername.empty() ? SingleArchive : MultiArchive);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set archive type to %d (hr=%s).", (int)(strFoldername.empty() ? SingleArchive : MultiArchive), stringify(hr, true).c_str());
			goto exit;
		}
	}
	
	// Update permissions
	if (!lpszArchiveServer) {	// No need to set permissions on a remote archive.
		hr = ptrArchiveHelper->SetPermissions(sUserEntryId, (ulFlags & Writable) == Writable);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set permissions on archive (hr=%s).", stringify(hr, true).c_str());
			goto exit;	
		}
	}

	
	// Create search folder
	hr = ptrStoreHelper->UpdateSearchFolders();
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set search folders (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Successfully attached archive.");

exit:
	if (lpRemoteSession)
		delete lpRemoteSession;

	return MAPIErrorToArchiveError(hr);
}

/**
 * Detach an archive from a users store.
 *
 * @param[in]	lpszArchiveServer
 * 					If not NULL, this argument specifies the path of the server on which to create the archive.
 * @param[in]	lpszArchive
 *					The username of the non-active user that's the placeholder for the archive.
 * @param[in]	lpszFolder
 *					The name of the folder that's be used as the root of the archive. If this paramater
 *					is set to NULL and the user has only one archive in the archive store, which 
 *					is usualy the case, that archive will be detached. If a user has multiple archives
 *					in the archive store, the exact folder need to be specified.
 *					If the archive root was placed in the IPM subtree of the archive store, this parameter
 *					must be set to NULL.
 *
 * @return HRESULT
 */
eResult ArchiveManageImpl::DetachFrom(const char *lpszArchiveServer, const char *lpszArchive, const char *lpszFolder)
{
	HRESULT hr = hrSuccess;
	entryid_t sUserEntryId;
	StoreHelperPtr ptrStoreHelper;
	ObjectEntryList lstArchives;
	ObjectEntryList::iterator iArchive;
	MsgStorePtr ptrArchiveStore;
	ArchiveHelperPtr ptrArchiveHelper;
	SObjectEntry SObjectEntry;
	SPropValuePtr ptrArchiveStoreEntryId;
	MAPIFolderPtr ptrArchiveFolder;
	SPropValuePtr ptrDisplayName;
	ULONG ulType = 0;
	Session *lpArchiveSession = m_lpSession;
	Session *lpRemoteSession = NULL;


	hr = StoreHelper::Create(m_ptrUserStore, &ptrStoreHelper);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create store helper (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrStoreHelper->GetArchiveList(&lstArchives);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive list (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}



	if (lpszArchiveServer) {
		hr = m_lpSession->CreateRemote(lpszArchiveServer, m_lpLogger, &lpRemoteSession);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to connect to archive server '%s' (hr=%s).", lpszArchiveServer, stringify(hr, true).c_str());
			goto exit;
		}
		
		lpArchiveSession = lpRemoteSession;
	}

	hr = lpArchiveSession->OpenStoreByName(lpszArchive, &ptrArchiveStore);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open archive store '%s' (hr=%s).", lpszArchive, stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = HrGetOneProp(ptrArchiveStore, PR_ENTRYID, &ptrArchiveStoreEntryId);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive entryid (hr=%s).", stringify(hr, true).c_str());
		goto exit;
	}


	// Find an archives on the passed store.
	iArchive = find_if(lstArchives.begin(), lstArchives.end(), StoreCompare(ptrArchiveStoreEntryId->Value.bin));
	if (iArchive == lstArchives.end()) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "%s has no archive on %s", m_strUser.c_str(), lpszArchive);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	// If no folder name was passed and there are more archives for this user on this archive, we abort.
	if (lpszFolder == NULL) {
		ObjectEntryList::iterator iNextArchive(iArchive);
		iNextArchive++;

		if (find_if(iNextArchive, lstArchives.end(), StoreCompare(ptrArchiveStoreEntryId->Value.bin)) != lstArchives.end()) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "%s has multiple archives on %s", m_strUser.c_str(), lpszArchive);
			hr = MAPI_E_COLLISION;
			goto exit;
		}
	}
	
	// If a folder name was passed, we need to find the correct folder.
	if (lpszFolder) {
		while (iArchive != lstArchives.end()) {
			hr = ptrArchiveStore->OpenEntry(iArchive->sItemEntryId.size(), iArchive->sItemEntryId, &ptrArchiveFolder.iid, MAPI_DEFERRED_ERRORS, &ulType, &ptrArchiveFolder);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open archive folder (hr=%s).", stringify(hr, true).c_str());
				goto exit;
			}
			
			hr = HrGetOneProp(ptrArchiveFolder, PR_DISPLAY_NAME_A, &ptrDisplayName);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get archive folder name (hr=%s).", stringify(hr, true).c_str());
				goto exit;
			}
			
			if (strcmp(ptrDisplayName->Value.lpszA, lpszFolder) == 0)
				break;
				
			iArchive = find_if(++iArchive, lstArchives.end(), StoreCompare(ptrArchiveStoreEntryId->Value.bin));
		}
		
		if (iArchive == lstArchives.end()) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "%s has no archive named %s on %s", m_strUser.c_str(), lpszFolder, lpszArchive);
			hr = MAPI_E_NOT_FOUND;
			goto exit;				
		}
	}
	
	assert(iArchive != lstArchives.end());
	lstArchives.erase(iArchive);
	
	hr = ptrStoreHelper->SetArchiveList(lstArchives);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to update archive list (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}

	// Update search folders
	hr = ptrStoreHelper->UpdateSearchFolders();
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set search folders (hr=%s).", stringify(hr, true).c_str());
		goto exit;	
	}

	m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Successfully detached archive.");
	
exit:
	return MAPIErrorToArchiveError(hr);
}

/**
 * List the attached archives for a user.
 *
 * @param[in]	ostr
 *					The std::ostream to which the list will be outputted.
 *
 * @return HRESULT
 */
eResult ArchiveManageImpl::ListArchives(ostream &ostr)
{
	HRESULT hr = hrSuccess;
	MsgStorePtr ptrUserStore;
	StoreHelperPtr ptrStoreHelper;
	ObjectEntryList lstArchives;
	ObjectEntryList::const_iterator iArchive;
	MsgStorePtr ptrArchiveStore;
	MAPIFolderPtr ptrArchiveFolder;
	ULONG ulType = 0;
	ULONG ulIdx = 0;

	hr = m_lpSession->OpenStoreByName(m_strUser, &ptrUserStore);
	if (hr != hrSuccess)
		goto exit;
		
	hr = StoreHelper::Create(ptrUserStore, &ptrStoreHelper);
	if (hr != hrSuccess)
		goto exit;
	
	hr = ptrStoreHelper->GetArchiveList(&lstArchives);
	if (hr != hrSuccess)
		goto exit;
		
	ostr << "User '" << m_strUser << "' has " << lstArchives.size() << " attached archives:" << endl;
	for (iArchive = lstArchives.begin(); iArchive != lstArchives.end(); ++iArchive) {
		HRESULT hrTmp = hrSuccess;
		SPropValuePtr ptrPropValue;
		
		ostr << "\t" << ulIdx++ << ": ";
	
		hrTmp= m_lpSession->OpenStore(iArchive->sStoreEntryId, &ptrArchiveStore);
		if (hrTmp != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open store (hr=%s)", stringify(hrTmp, true).c_str());
			ostr << "Error, check log for details" << endl;
			continue;
		}
		
		hrTmp = ptrArchiveStore->OpenEntry(iArchive->sItemEntryId.size(), iArchive->sItemEntryId, &ptrArchiveFolder.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrArchiveFolder);
		if (hrTmp != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open folder (hr=%s)", stringify(hrTmp, true).c_str());
			ostr << "Error, check log for details" << endl;
			continue;
		}
		
		ostr << "Store: ";
		hrTmp = HrGetOneProp(ptrArchiveStore, PR_DISPLAY_NAME_A, &ptrPropValue);
		if (hrTmp != hrSuccess)
			ostr << "Unknown (0x" << stringify(hrTmp, true) << ")";
		else
			ostr << "'" << ptrPropValue->Value.lpszA << "'";
			
		ostr << ", Folder: ";
		hrTmp = HrGetOneProp(ptrArchiveFolder, PR_DISPLAY_NAME_A, &ptrPropValue);
		if (hrTmp != hrSuccess)
			ostr << "Unknown (0x" << stringify(hrTmp, true) << ")";
		else
			ostr << "'" << ptrPropValue->Value.lpszA << "'";
		
		ostr << endl;
	}
	
exit:
	return MAPIErrorToArchiveError(hr);
}

eResult ArchiveManageImpl::ListArchives(ArchiveList *lplstArchives)
{
	HRESULT hr = hrSuccess;
	StoreHelperPtr ptrStoreHelper;
	ObjectEntryList lstArchives;
	ObjectEntryList::const_iterator iArchive;
	MsgStorePtr ptrArchiveStore;
	MAPIFolderPtr ptrArchiveFolder;
	ULONG ulType = 0;
	ArchiveList lstEntries;

	hr = StoreHelper::Create(m_ptrUserStore, &ptrStoreHelper);
	if (hr != hrSuccess)
		goto exit;
	
	hr = ptrStoreHelper->GetArchiveList(&lstArchives);
	if (hr != hrSuccess)
		goto exit;
		
	for (iArchive = lstArchives.begin(); iArchive != lstArchives.end(); ++iArchive) {
		HRESULT hrTmp = hrSuccess;
		ULONG cStoreProps = 0;
		SPropArrayPtr ptrStoreProps;
		SPropValuePtr ptrPropValue;
		ArchiveEntry entry;

		SizedSPropTagArray(2, sptaStoreProps) = {2, {PR_DISPLAY_NAME_A, PR_MAILBOX_OWNER_ENTRYID}};
		enum {IDX_DISPLAY_NAME, IDX_MAILBOX_OWNER_ENTRYID};
		
		hrTmp = m_lpSession->OpenStore(iArchive->sStoreEntryId, &ptrArchiveStore);
		if (hrTmp != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open store (hr=%s)", stringify(hrTmp, true).c_str());
			entry.StoreName = "Failed id=" + iArchive->sStoreEntryId.tostring() + ", hr=" + stringify(hrTmp, true);
			lstEntries.push_back(entry);
			continue;
		}
		
		hrTmp = ptrArchiveStore->GetProps((LPSPropTagArray)&sptaStoreProps, 0, &cStoreProps, &ptrStoreProps);
		if (FAILED(hrTmp))
			entry.StoreName = entry.StoreOwner = "Unknown (" + stringify(hrTmp, true) + ")";
		else {
			if (ptrStoreProps[IDX_DISPLAY_NAME].ulPropTag == PR_DISPLAY_NAME_A)
				entry.StoreName = ptrStoreProps[IDX_DISPLAY_NAME].Value.lpszA;
			else
				entry.StoreName = "Unknown (" + stringify(ptrStoreProps[IDX_DISPLAY_NAME].Value.err, true) + ")";
				
			if (ptrStoreProps[IDX_MAILBOX_OWNER_ENTRYID].ulPropTag == PR_MAILBOX_OWNER_ENTRYID) {
				MAPIPropPtr ptrOwner;

				hrTmp = m_lpSession->OpenMAPIProp(ptrStoreProps[IDX_MAILBOX_OWNER_ENTRYID].Value.bin.cb,
												  (LPENTRYID)ptrStoreProps[IDX_MAILBOX_OWNER_ENTRYID].Value.bin.lpb,
												  &ptrOwner);
				if (hrTmp == hrSuccess)
					hrTmp = HrGetOneProp(ptrOwner, PR_ACCOUNT_A, &ptrPropValue);

				if (hrTmp == hrSuccess)
					entry.StoreOwner = ptrPropValue->Value.lpszA;
				else
					entry.StoreOwner = "Unknown (" + stringify(hrTmp, true) + ")";
			} else
				entry.StoreOwner = "Unknown (" + stringify(ptrStoreProps[IDX_MAILBOX_OWNER_ENTRYID].Value.err, true) + ")";
		}

		hrTmp = ptrArchiveStore->OpenEntry(iArchive->sItemEntryId.size(), iArchive->sItemEntryId, &ptrArchiveFolder.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrArchiveFolder);
		if (hrTmp != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open folder (hr=%s)", stringify(hrTmp, true).c_str());
			entry.FolderName = "Failed id=" + iArchive->sStoreEntryId.tostring() + ", hr=" + stringify(hrTmp, true);
			lstEntries.push_back(entry);
			continue;
		}		
			
		hrTmp = HrGetOneProp(ptrArchiveFolder, PR_DISPLAY_NAME_A, &ptrPropValue);
		if (hrTmp != hrSuccess)
			entry.FolderName = "Unknown (" + stringify(hrTmp, true) + ")";
		else
			entry.FolderName =  ptrPropValue->Value.lpszA ;
		
		lstEntries.push_back(entry);
	}

	lplstArchives->swap(lstEntries);
	
exit:
	return MAPIErrorToArchiveError(hr);
}
