/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "archiver.h"
#include "archiver-session.h"
#include "archivectrl.h"
#include "archivemanage.h"

#include "mapix.h"

#include "ECConfig.h"
#include "ECLogger.h"

class ArchiverImpl : public Archiver
{
public:
	ArchiverImpl();
	~ArchiverImpl();

	eResult Init(const char *lpszAppName, const char *lpszConfig, unsigned int ulFlags);

	eResult GetControl(ArchiveControlPtr *lpptrControl);
	eResult GetManage(const char *lpszUser, ArchiveManagePtr *lpptrManage);

private:
	bool		m_bInitialized;
	ECConfig	*m_lpsConfig;
	ECLogger	*m_lpLogger;
	Session 	*m_lpSession;
};



eResult Archiver::Create(auto_ptr_type *lpptrArchiver)
{
	eResult r = Success;
	auto_ptr_type ptrArchiver;

	if (lpptrArchiver == NULL) {
		r = InvalidParameter;
		goto exit;
	}

	try {
		ptrArchiver.reset(new ArchiverImpl());
	} catch (std::bad_alloc &) {
		r = OutOfMemory;
		goto exit;
	}

	*lpptrArchiver = ptrArchiver;

exit:
	return r;
}


ArchiverImpl::ArchiverImpl()
: m_bInitialized(false)
, m_lpsConfig(NULL)
, m_lpLogger(NULL)
, m_lpSession(NULL)
{
}

ArchiverImpl::~ArchiverImpl()
{
	if (m_lpSession)
		delete m_lpSession;
	
	delete m_lpLogger;
	
	if (m_lpsConfig)
		delete m_lpsConfig;

	if (m_bInitialized)
		MAPIUninitialize();
}

eResult ArchiverImpl::Init(const char *lpszAppName, const char *lpszConfig, unsigned int ulFlags)
{
	eResult r = Success;

	MAPIINIT_0 sMapiInit = {MAPI_INIT_VERSION, MAPI_MULTITHREAD_NOTIFICATIONS};
	
	const configsetting_t lpDefaults[] = {
		// Connect settings
		{ "server_socket",	"" },
		{ "sslkey_file",	"" },
		{ "sslkey_pass",	"" },
		
		// Archive settings
		{ "archive_enable",	"yes" },
		{ "archive_after", 	"30" },

		{ "stub_enable",	"no" },
		{ "stub_unread",	"no" },
		{ "stub_after", 	"0" },

		{ "delete_enable",	"no" },
		{ "delete_unread",	"no" },
		{ "delete_after", 	"0" },
		
		{ "purge_enable",	"no" },
		{ "purge_after", 	"0" },

		{ "track_history",	"no" },
		
		// Log options
		{ "log_method",		"file" },
		{ "log_file",		"-" },
		{ "log_level",		"2", CONFIGSETTING_RELOADABLE },
		{ "log_timestamp",	"1" },

		{ NULL, NULL },
	};
	
	m_lpsConfig = new ECConfig(lpDefaults);

	if (!m_lpsConfig->LoadSettings(lpszConfig) && (ulFlags & RequireConfig)) {
		r = FileNotFound;
		goto exit;
	}
	if (!m_lpsConfig->LoadSettings(lpszConfig)) {
		if ((ulFlags & RequireConfig)) {
			r = FileNotFound;
			goto exit;
		}
	} else if (m_lpsConfig->HasErrors()) {
		if (!(ulFlags & InhibitErrorLogging)) {
			ECLogger *lpLogger = new ECLogger_File(EC_LOGLEVEL_FATAL, 0, "-");
			LogConfigErrors(m_lpsConfig, lpLogger);
			delete lpLogger;
		}
		
		r = InvalidConfig;
		goto exit;
	}

	m_lpLogger = CreateLogger(m_lpsConfig, (char*)lpszAppName, "");
	if (ulFlags & InhibitErrorLogging) {
		// We need to check if we're logging to stderr. If so we'll replace
		// the logger with a NULL logger.
		ECLogger_File *lpFileLogger = dynamic_cast<ECLogger_File*>(m_lpLogger);
		if (lpFileLogger) {
			delete m_lpLogger;
			m_lpLogger = new ECLogger_Null();
		}
	} else if (ulFlags & AttachStdErr) {
		// We need to check if the current logger isn't logging to the console
		// as that would give duplicate messages
		delete m_lpLogger;
		m_lpLogger = new ECLogger_File(EC_LOGLEVEL_ERROR, 0, "-", false);
	}

	if (m_lpsConfig->HasWarnings())
		LogConfigErrors(m_lpsConfig, m_lpLogger);

	if (MAPIInitialize(&sMapiInit) != hrSuccess) {
		r = Failure;
		goto exit;
	}
	
	if (Session::Create(m_lpsConfig, m_lpLogger, &m_lpSession) != hrSuccess) {
		r = Failure;
		MAPIUninitialize();
		goto exit;
	}

	m_bInitialized = true;

exit:
	return r;
}

eResult ArchiverImpl::GetControl(ArchiveControlPtr *lpptrControl)
{
	if (!m_bInitialized)
		return Uninitialized;
		
	return ArchiveControlImpl::Create(m_lpSession, m_lpsConfig, m_lpLogger, lpptrControl);
}

eResult ArchiverImpl::GetManage(const char *lpszUser, ArchiveManagePtr *lpptrManage)
{
	if (!m_bInitialized)
		return Uninitialized;
		
	return ArchiveManageImpl::Create(m_lpSession, lpszUser, m_lpLogger, lpptrManage);
}
