/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <platform.h>
#include "ECConfig.h"
#include "operations.h"
#include "mapiprophelper.h"
#include "archivehelper.h"
#include "archiver-session.h"

#include <mapiutil.h>
#include <mapiext.h>

#include "Util.h"
#include "stringutil.h"
#include "mapi_ptr.h"

#include <algorithm>
using namespace std;


/**
 * Constructor.
 * @param[in]	lpLogger
 *					Pointer to an ECLogger object that's used for logging.
 */
ArchiveOperationBase::ArchiveOperationBase(ECLogger *lpLogger)
: m_lpLogger(lpLogger)
{ }

/**
 * Process the message that exists in the search folder ptrFolder and deletegate the work to the derived class.
 * The property array lpProps contains at least the entryid of the message to open and the entryid of the parent
 * folder of the message to open.
 *
 * @param[in]	ptrFolder
 *					A MAPIFolderPtr that points to the parent folder.
 * @param[in]	cProps
 *					The number op properties pointed to by lpProps.
 * @param[in]	lpProps
 *					Pointer to an array of properties that are used by the Operation object.
 *
 * @return HRESULT
 */
HRESULT ArchiveOperationBase::ProcessEntry(MAPIFolderPtr &ptrFolder, ULONG cProps, const LPSPropValue &lpProps)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpFolderEntryId;
	bool bReloadFolder = false;
	ULONG ulType = 0;
	
	lpFolderEntryId = PpropFindProp(lpProps, cProps, PR_PARENT_ENTRYID);
	if (lpFolderEntryId == NULL) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "PR_PARENT_ENTRYID missing");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	if (!m_ptrCurFolderEntryId.is_null()) {
		int nResult = 0;
		hr = Util::CompareProp(m_ptrCurFolderEntryId, lpFolderEntryId, &nResult);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to compare current and new entryid. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		if (nResult != 0) {
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Leaving folder (%s)", bin2hex(m_ptrCurFolderEntryId->Value.bin.cb, m_ptrCurFolderEntryId->Value.bin.lpb).c_str());
			hr = LeaveFolder();
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to leave folder. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}
				
			bReloadFolder = true;
		}
	}
	
	if (m_ptrCurFolderEntryId.is_null() || bReloadFolder) {	
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Opening folder (%s)", bin2hex(lpFolderEntryId->Value.bin.cb, lpFolderEntryId->Value.bin.lpb).c_str());
	
		hr = ptrFolder->OpenEntry(lpFolderEntryId->Value.bin.cb, (LPENTRYID)lpFolderEntryId->Value.bin.lpb, &m_ptrCurFolder.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &m_ptrCurFolder);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open folder. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		hr = MAPIAllocateBuffer(sizeof(SPropValue), &m_ptrCurFolderEntryId);
		if (hr != hrSuccess)
			goto exit;
		
		hr = Util::HrCopyProperty(m_ptrCurFolderEntryId, lpFolderEntryId, m_ptrCurFolderEntryId);
		if (hr != hrSuccess)
			goto exit;
	}
	
	hr = DoProcessEntry(cProps, lpProps);
	
exit:
	return hr;
}

/**
 * Constructor
 *
 * @param[in]	lpSession
 *					Pointer to the session.
 * @param[in]	lpLogger
 *					Pointer to the logger.
 * @param[in]	lstArchives
 *					The list of attached archives for this store.
 * @param[in]	lpExcludeProps
 *					The list of properties that will not be copied during the archive operation.
 *
 * @return HRESULT
 */
Copier::Copier(Session *lpSession, ECConfig *lpConfig, ECLogger *lpLogger, const ObjectEntryList &lstArchives, LPSPropTagArray lpExcludeProps)
: ArchiveOperationBase(lpLogger)
, m_lpSession(lpSession)
, m_lpConfig(lpConfig)
, m_lstArchives(lstArchives)
{
	MAPIAllocateBuffer(CbNewSPropTagArray(lpExcludeProps->cValues), &m_ptrExcludeProps);
	memcpy(m_ptrExcludeProps, lpExcludeProps, CbNewSPropTagArray(lpExcludeProps->cValues));
}

HRESULT Copier::LeaveFolder()
{
	// Just clear the map with archive folders, so new folders need to be resolved.
	m_mapArchiveFolders.clear();
	return hrSuccess;
}

HRESULT Copier::DoProcessEntry(ULONG cProps, const LPSPropValue &lpProps)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpEntryId = NULL;
	LPSPropValue lpStoreEntryId = NULL;
	SObjectEntry refObjectEntry;
	MessagePtr ptrMessage;
	ULONG ulType = 0;
	MAPIPropHelperPtr ptrMsgHelper;
	bool bDirty = false;
	ObjectEntryList lstMsgArchives;
	ObjectEntryList::iterator iArchive;
	
	lpEntryId = PpropFindProp(lpProps, cProps, PR_ENTRYID);
	if (lpEntryId == NULL) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "PR_ENTRYID missing");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	lpStoreEntryId = PpropFindProp(lpProps, cProps, PR_STORE_ENTRYID);
	if (lpStoreEntryId == NULL) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "PR_STORE_ENTRYID missing");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	// Set the reference to the original message
	refObjectEntry.sStoreEntryId.assign(lpStoreEntryId->Value.bin);
	refObjectEntry.sItemEntryId.assign(lpEntryId->Value.bin);

	Logger()->Log(EC_LOGLEVEL_DEBUG, "Opening message (%s)", bin2hex(lpEntryId->Value.bin.cb, lpEntryId->Value.bin.lpb).c_str());
	hr = CurrentFolder()->OpenEntry(lpEntryId->Value.bin.cb, (LPENTRYID)lpEntryId->Value.bin.lpb, &IID_IECMessageRaw, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrMessage);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to open message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = MAPIPropHelper::Create(ptrMessage.as<MAPIPropPtr>(), &ptrMsgHelper);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to create prop helper. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	hr = ptrMsgHelper->IsDirty(&bDirty);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to determine if message is dirty. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrMsgHelper->GetArchiveList(&lstMsgArchives);
	if (hr != hrSuccess) {
		if (hr == MAPI_E_CORRUPT_DATA) {
			Logger()->Log(EC_LOGLEVEL_ERROR, "Existing list of archives is corrupt, skipping message.");
			hr = hrSuccess;
		} else
			Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to get list of archives. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
		
	for (iArchive = m_lstArchives.begin(); iArchive != m_lstArchives.end(); ++iArchive) {
		ObjectEntryList::iterator iArchivedMsg = find_if(lstMsgArchives.begin(), lstMsgArchives.end(), StoreCompare(*iArchive));
		// Check if this message is archived to the current archive.
		if (iArchivedMsg == lstMsgArchives.end()) {
			MessagePtr ptrNewMessage;
			SPropValuePtr ptrEntryId;
			SObjectEntry objectEntry;
			
			Logger()->Log(EC_LOGLEVEL_DEBUG, "Message not yet archived to archive (%s)", iArchive->sStoreEntryId.tostring().c_str());

			// It's possible to have a dirty message that has not been archived to the current archive if at the time of the
			// previous archiver run, this archive was unavailable for some reason.
			// If this happens, we'll can't do much more than just archive the current version of the message to the archive.
			//
			// Alternatively we could get the previous version from another archive and copy that to this archive if the
			// configuration dictates that old archives are linked. We won't do that though!

			hr = CreateArchivedMessage(ptrMessage, *iArchive, refObjectEntry, &ptrNewMessage);
			if (hr != hrSuccess)
				goto exit;
			
			hr = ptrNewMessage->SaveChanges(KEEP_OPEN_READONLY);
			if (hr != hrSuccess) {
				Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to save archive message. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}
				
			hr = HrGetOneProp(ptrNewMessage, PR_ENTRYID, &ptrEntryId);
			if (hr != hrSuccess) {
				Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to get entry id of archive message. (hr=%s)", stringify(hr, true).c_str());
				goto exit;
			}			
			
			// Update the list of archive messages for this message.
			objectEntry.sStoreEntryId.assign(iArchive->sStoreEntryId);
			objectEntry.sItemEntryId.assign(ptrEntryId->Value.bin);
			
			lstMsgArchives.push_back(objectEntry);
		} else if (bDirty) {
			Logger()->Log(EC_LOGLEVEL_INFO, "Found archive for dirty message.");

			// We found an archived version for the current message. However, the message is marked dirty...
			// There are two things we can do:
			// 1. Update the current archived version.
			// 2. Make a new archived message and link that to the old version, basically tracking history.
			//
			// This will be configurable through the track_history option.
			if (parseBool(m_lpConfig->GetSetting("track_history"))) {

				MessagePtr ptrNewMessage;
				MAPIPropHelperPtr ptrMsgHelper;
				SPropValuePtr ptrEntryId;
				SObjectEntry objectEntry;

				Logger()->Log(EC_LOGLEVEL_DEBUG, "Creating new archived message.");

				hr = CreateArchivedMessage(ptrMessage, *iArchive, refObjectEntry, &ptrNewMessage);
				if (hr != hrSuccess)
					goto exit;

				hr = MAPIPropHelper::Create(MAPIPropPtr(ptrNewMessage, true), &ptrMsgHelper);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to create prop helper. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = ptrMsgHelper->ReferencePrevious(*iArchivedMsg);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to set reference to original message. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}
				
				hr = ptrNewMessage->SaveChanges(KEEP_OPEN_READONLY);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to save archive message. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}
					
				hr = HrGetOneProp(ptrNewMessage, PR_ENTRYID, &ptrEntryId);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to get entry id of archive message. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}			
				
				// Update the list of archive messages for this message.
				objectEntry.sStoreEntryId.assign(iArchive->sStoreEntryId);
				objectEntry.sItemEntryId.assign(ptrEntryId->Value.bin);
				
				lstMsgArchives.push_back(objectEntry);

				// Remove the old entry (so we don't also point to the old version)
				lstMsgArchives.erase(iArchivedMsg);

			} else {
				MsgStorePtr ptrArchiveStore;
				MessagePtr ptrArchivedMsg;
				SPropTagArrayPtr ptrPropList;

				Logger()->Log(EC_LOGLEVEL_DEBUG, "Updating archived message.");

				hr = m_lpSession->OpenStore(iArchivedMsg->sStoreEntryId, &ptrArchiveStore);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to open archive store. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = ptrArchiveStore->OpenEntry(iArchivedMsg->sItemEntryId.size(), iArchivedMsg->sItemEntryId, &IID_IECMessageRaw, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrArchivedMsg);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to open existing archived message. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = ptrArchivedMsg->GetPropList(fMapiUnicode, &ptrPropList);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to get property list. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = ptrArchivedMsg->DeleteProps(ptrPropList, NULL);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to delete properties. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = Util::HrDeleteAttachments(ptrArchivedMsg);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to delete attachments. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = Util::HrDeleteRecipients(ptrArchivedMsg);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to delete recipients. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}

				hr = ArchiveMessage(ptrMessage, refObjectEntry, ptrArchivedMsg);
				if (hr != hrSuccess)
					goto exit;
				
				hr = ptrArchivedMsg->SaveChanges(0);
				if (hr != hrSuccess) {
					Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to save archive message. (hr=%s)", stringify(hr, true).c_str());
					goto exit;
				}
			}			
		} else
			Logger()->Log(EC_LOGLEVEL_WARNING, "Ignoring already archived message.");
	}

	if (bDirty) {
		hr = ptrMsgHelper->SetClean();
		if (hr != hrSuccess) {
			Logger()->Log(EC_LOGLEVEL_WARNING, "Failed to unmark message as dirty.");
			goto exit;
		}
	}
	
	lstMsgArchives.sort();
	lstMsgArchives.unique();
	
	hr = ptrMsgHelper->SetArchiveList(lstMsgArchives, true);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to save list of archives for this message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

exit:
	return hr;
}

HRESULT Copier::CreateArchivedMessage(LPMESSAGE lpSource, const SObjectEntry &objectEntry, const SObjectEntry &refObjectEntry, LPMESSAGE *lppArchivedMsg)
{
	HRESULT hr = hrSuccess;
	MAPIFolderPtr ptrArchiveFolder;
	MessagePtr ptrNewMessage;
	
	hr = GetArchiveFolder(objectEntry, &ptrArchiveFolder);
	if (hr != hrSuccess)
		goto exit;

	hr = ptrArchiveFolder->CreateMessage(&ptrNewMessage.iid, MAPI_DEFERRED_ERRORS, &ptrNewMessage);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to create archive message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	hr = ArchiveMessage(lpSource, refObjectEntry, ptrNewMessage);
	if (hr != hrSuccess)
		goto exit;

	hr = ptrNewMessage->QueryInterface(IID_IMessage, (LPVOID*)lppArchivedMsg);

exit:
	return hr;
}

HRESULT Copier::GetArchiveFolder(const SObjectEntry &objectEntry, LPMAPIFOLDER *lppArchiveFolder)
{
	HRESULT hr = hrSuccess;
	ArchiveFolderMap::const_iterator iArchiveFolder;
	MAPIFolderPtr ptrArchiveFolder;

	if (lppArchiveFolder == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	iArchiveFolder = m_mapArchiveFolders.find(objectEntry.sStoreEntryId);
	if (iArchiveFolder == m_mapArchiveFolders.end()) {
		ULONG ulType = 0;
		MsgStorePtr ptrArchiveStore;
		MAPIFolderPtr ptrArchiveRootFolder;
		ArchiveHelperPtr ptrArchiveHelper;

		Logger()->Log(EC_LOGLEVEL_DEBUG, "Archive folder not found in cache");
	
		// Find the associated archive folder 
		hr = m_lpSession->OpenStore(objectEntry.sStoreEntryId, &ptrArchiveStore);
		if (hr != hrSuccess) {
			Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to open archive store. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		hr = ptrArchiveStore->OpenEntry(objectEntry.sItemEntryId.size(), objectEntry.sItemEntryId, &ptrArchiveRootFolder.iid, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS, &ulType, &ptrArchiveRootFolder);
		if (hr != hrSuccess) {
			Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to open archive root folder. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		// We pass a NULL to the lpszServerPath argument, indicating that it's local. However, it was already
		// remotly opened by m_lpSession->OpenStore(). Effectively this causes ptrArchiveHelper->GetArchiveEntry()
		// to malfunction (it won't wrap the entryid with the serverpath). This is not an issue as we don't use
		// that here anyway.
		hr = ArchiveHelper::Create(ptrArchiveStore, ptrArchiveRootFolder, NULL, &ptrArchiveHelper);
		if (hr != hrSuccess) {
			Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to create archive helper. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		hr = ptrArchiveHelper->GetArchiveFolderFor(CurrentFolder(), m_lpSession, &ptrArchiveFolder);
		if (hr != hrSuccess) {
			Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to get archive folder. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}

		m_mapArchiveFolders.insert(ArchiveFolderMap::value_type(objectEntry.sStoreEntryId, ptrArchiveFolder)); 			
	} else {
		Logger()->Log(EC_LOGLEVEL_DEBUG, "Archive folder found in cache");
		ptrArchiveFolder = iArchiveFolder->second;
	}


	hr = ptrArchiveFolder->QueryInterface(IID_IMAPIFolder, (LPVOID*)lppArchiveFolder);

exit:
	return hr;
}

HRESULT Copier::ArchiveMessage(LPMESSAGE lpSource, const SObjectEntry &refObjectEntry, LPMESSAGE lpDest)
{
	HRESULT hr = hrSuccess;
	MAPIPropHelperPtr ptrMsgHelper;
	SPropValuePtr ptrEntryId;

	if (lpSource == NULL || lpDest == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	hr = lpSource->CopyTo(0, NULL, m_ptrExcludeProps, 0, NULL, &IID_IMessage, lpDest, 0, NULL);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to copy message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	hr = MAPIPropHelper::Create(MAPIPropPtr(lpDest, true), &ptrMsgHelper);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to create prop helper. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	hr = ptrMsgHelper->SetReference(refObjectEntry);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to set reference to original message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

exit:
	return hr;
}


/**
 * Constructor
 *
 * @param[in]	lpLogger
 *					Pointer to the logger.
 *
 * @return HRESULT
 */
Deleter::Deleter(ECLogger *lpLogger)
: ArchiveOperationBase(lpLogger)
{ }

/**
 * Destructor
 */
Deleter::~Deleter()
{
	PurgeQueuedMessages();
}

HRESULT Deleter::LeaveFolder()
{
	// Folder is still available, purge now!
	return PurgeQueuedMessages();
}

HRESULT Deleter::DoProcessEntry(ULONG cProps, const LPSPropValue &lpProps)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpEntryId = NULL;

	lpEntryId = PpropFindProp(lpProps, cProps, PR_ENTRYID);
	if (lpEntryId == NULL) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "PR_ENTRYID missing");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
	if (m_lstEntryIds.size() >= 50) {
		hr = PurgeQueuedMessages();
		if (hr != hrSuccess)
			goto exit;
	}

	m_lstEntryIds.push_back(lpEntryId->Value.bin);
	
exit:
	return hr;
}

/**
 * Delete the messages that are queued for deletion.
 * @return HRESULT
 */
HRESULT Deleter::PurgeQueuedMessages()
{
	HRESULT hr = hrSuccess;
	EntryListPtr ptrEntryList;
	list<entryid_t>::const_iterator iEntryId;
	ULONG ulIdx = 0;
	
	if (m_lstEntryIds.empty())
		goto exit;
	
	hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), &ptrEntryList);
	if (hr != hrSuccess)
		goto exit;
	
	hr = MAPIAllocateMore(m_lstEntryIds.size() * sizeof(SBinary), ptrEntryList, (LPVOID*)&ptrEntryList->lpbin);
	if (hr != hrSuccess)
		goto exit;
		
	ptrEntryList->cValues = m_lstEntryIds.size();
	for (iEntryId = m_lstEntryIds.begin(); iEntryId != m_lstEntryIds.end(); ++iEntryId, ++ulIdx) {
		ptrEntryList->lpbin[ulIdx].cb = iEntryId->size();
		ptrEntryList->lpbin[ulIdx].lpb = *iEntryId;
	}
	
	hr = CurrentFolder()->DeleteMessages(ptrEntryList, 0, NULL, 0);
	if (hr != hrSuccess) {
		Logger()->Log(EC_LOGLEVEL_FATAL, "Failed to delete %u messages. (hr=%s)", ptrEntryList->cValues, stringify(hr, true).c_str());
		goto exit;
	}
	
	m_lstEntryIds.clear();
	
exit:
	return hr;
}

/**
 * Constructor
 * 
 * @param[in]	lpLogger
 *					Pointer to the logger.
 * @param[in]	ulptStubbed
 *					The proptag of the stubbed property {72e98ebc-57d2-4ab5-b0aad50a7b531cb9}/stubbed
 *
 * @return HRESULT
 */
Stubber::Stubber(ECLogger *lpLogger, ULONG ulptStubbed)
: m_lpLogger(lpLogger)
, m_ulptStubbed(ulptStubbed)
{ }

HRESULT Stubber::ProcessEntry(MAPIFolderPtr &ptrFolder, ULONG cProps, const LPSPropValue &lpProps)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpEntryId = NULL;
	MessagePtr ptrMessage;
	ULONG ulType = 0;
	SPropValue sProps[2];
	SPropValue sProp = {0};
	MAPITablePtr ptrAttTable;
	mapi_rowset_ptr ptrRowSet;
	AttachPtr ptrAttach;
	ULONG ulAttachNum = 0;
	MAPIPropHelperPtr ptrMsgHelper;
	ObjectEntryList lstMsgArchives;

	SizedSPropTagArray(1, sptaDelProps) = {1, {PR_ICON_INDEX}};
	SizedSPropTagArray(1, sptaTableProps) = {1, {PR_ATTACH_NUM}};

	lpEntryId = PpropFindProp(lpProps, cProps, PR_ENTRYID);
	if (lpEntryId == NULL) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "PR_ENTRYID missing");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Opening message (%s)", bin2hex(lpEntryId->Value.bin.cb, lpEntryId->Value.bin.lpb).c_str());
	hr = ptrFolder->OpenEntry(lpEntryId->Value.bin.cb, (LPENTRYID)lpEntryId->Value.bin.lpb, &IID_IECMessageRaw, MAPI_BEST_ACCESS, &ulType, &ptrMessage);
	if (hr == MAPI_E_NOT_FOUND) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Failed to open message. This can happen if the search folder is lagging.");
		hr = hrSuccess;
		goto exit;
	} else if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to open message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	
	// Verify if we have at least one archive that's in the current multi-server cluster.
	hr = MAPIPropHelper::Create(ptrMessage.as<MAPIPropPtr>(), &ptrMsgHelper);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create prop helper. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrMsgHelper->GetArchiveList(&lstMsgArchives);
	if (hr != hrSuccess) {
		if (hr == MAPI_E_CORRUPT_DATA) {
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Existing list of archives is corrupt, skipping message.");
			hr = hrSuccess;
		} else
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get list of archives. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	if (find_if(lstMsgArchives.begin(), lstMsgArchives.end(), IsNotWrapped()) == lstMsgArchives.end()) {
		m_lpLogger->Log(EC_LOGLEVEL_WARNING, "Message has no archives that are directly accessible, message will not be stubbed.");
		goto exit;
	}
	
	
		
	sProps[0].ulPropTag = m_ulptStubbed;
	sProps[0].Value.b = 1;
	
	sProps[1].ulPropTag = PR_BODY;
	sProps[1].Value.lpszA = "This message is archived...";

	hr = ptrMessage->SetProps(2, sProps, NULL);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set properties. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}

	// Remove the PR_ICON_INDEX property so our forms icon will override
	hr = ptrMessage->DeleteProps((LPSPropTagArray)&sptaDelProps, NULL);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to remove properties. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = ptrMessage->GetAttachmentTable(MAPI_DEFERRED_ERRORS, &ptrAttTable);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get attachment table. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	hr = HrQueryAllRows(ptrAttTable, (LPSPropTagArray)&sptaTableProps, NULL, NULL, 0, &ptrRowSet);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to get attachment numbers. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
	if (!ptrRowSet.empty()) {
		m_lpLogger->Log(EC_LOGLEVEL_INFO, "Removing %u attachments", ptrRowSet.size());
		for (ULONG i = 0; i < ptrRowSet.size(); ++i) {
			hr = ptrMessage->DeleteAttach(ptrRowSet[i].lpProps[0].Value.ul, 0, NULL, 0);
			if (hr != hrSuccess) {
				m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to delete attachment %u. (hr=%s)", i, stringify(hr, true).c_str());
				goto exit;
			}
		}
		
		m_lpLogger->Log(EC_LOGLEVEL_ANY, "Adding placeholder attachment");		
		hr = ptrMessage->CreateAttach(&ptrAttach.iid, 0, &ulAttachNum, &ptrAttach);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to create attachment. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		sProp.ulPropTag = PR_ATTACH_FILENAME;
		sProp.Value.lpszA = "dummy";
		hr = ptrAttach->SetProps(1, &sProp, NULL);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to set attachment properties. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
		
		hr = ptrAttach->SaveChanges(0);
		if (hr != hrSuccess) {
			m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to save attachment. (hr=%s)", stringify(hr, true).c_str());
			goto exit;
		}
	}
	
	hr = ptrMessage->SaveChanges(0);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_FATAL, "Failed to save stubbed message. (hr=%s)", stringify(hr, true).c_str());
		goto exit;
	}
	
exit:
	return hr;

}
