/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "CalDAV.h"
#include "Http.h"
#include "CalDavUtil.h"
#include "iCal.h"
#include "WebDav.h"
#include "CalDavProto.h"

#include <signal.h>

#include <iostream>
#include <string>

#include "ECLogger.h"
#include "ECChannel.h"
#include "my_getopt.h"
#include "ecversion.h"
#include "CommonUtil.h"
#include "SSLUtil.h"

#include <execinfo.h>
#include "UnixUtil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


struct HandlerArgs {
    ECChannel *lpChannel;
	bool bUseSSL;
};

bool g_bDaemonize = true;
bool g_bQuit = false;
bool g_bThreads = false;
ECLogger *g_lpLogger = NULL;
ECConfig *g_lpConfig = NULL;
pthread_t mainthread;
int nChildren = 0;

#define KEEP_ALIVE_TIME 300

void sigterm(int) {
	g_bQuit = true;
}

void sighup(int) {
	// In Win32, the signal is sent in a seperate, special signal thread. So this test is
	// not needed or required.
	if (g_bThreads && pthread_equal(pthread_self(), mainthread)==0)
		return;
	if (g_lpConfig)
		g_lpConfig->ReloadSettings();

	if (g_lpLogger) {
		if (g_lpConfig) {
			char *ll = g_lpConfig->GetSetting("log_level");
			int new_ll = ll ? atoi(ll) : 2;
			g_lpLogger->SetLoglevel(new_ll);
		}

		g_lpLogger->Reset();
		g_lpLogger->Log(EC_LOGLEVEL_WARNING, "Log connection was reset");
	}
}


void sigchld(int) {
	int stat;
	while (waitpid (-1, &stat, WNOHANG) > 0) nChildren--;
}

void sigsegv(int signr)
{
	void *bt[64];
	int i, n;
	char **btsymbols;

	if (!g_lpLogger)
		goto exit;

	g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Pid %d caught SIGSEGV (%d), traceback:", getpid(), signr);

	n = backtrace(bt, 64);

	btsymbols = backtrace_symbols(bt, n);

	for (i = 0; i < n; i++) {
		if (btsymbols)
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p %s", bt[i], btsymbols[i]);
		else
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "%016p", bt[i]);
	}

	g_lpLogger->Log(EC_LOGLEVEL_FATAL, "When reporting this traceback, please include Linux distribution name, system architecture and Zarafa version.");

exit:
	kill(getpid(), signr);
}

void PrintHelp(char *name) {
	cout << "Usage:\n" << endl;
	cout << name << " [-h] [-F] [-V] [-c <configfile>]" << endl;
	cout << "  -F\t\tDo not run in the background" << endl;
	cout << "  -h\t\tShows this help." << endl;
	cout << "  -V\t\tPrint version info." << endl;
	cout << "  -c filename\tUse alternate config file (e.g. /etc/zarafa/ical.cfg)\n\t\tDefault: /etc/zarafa/ical.cfg" << endl;
	cout << endl;
}

void PrintVersion() {
	cout << "Product version:\t"  <<  PROJECT_VERSION_CALDAV_STR << endl << "File version:\t\t" << PROJECT_SVN_REV_STR << endl;
}


int main(int argc, char **argv) {
	HRESULT hr = hrSuccess;
	int ulListenCalDAV = 0;
	int ulListenCalDAVs = 0;
	int i;

	ECChannel *lpChannel = NULL;
    stack_t st = {0};
    struct sigaction act = {{0}};

	// Configuration
	char opt = '\0';
	char *lpszCfg = (char*)"/etc/zarafa/ical.cfg";
	const configsetting_t lpDefaults[] = {
		{ "run_as_user", "" },
		{ "run_as_group", "" },
		{ "pid_file", "/var/run/zarafa-ical.pid" },
		{ "running_path", "/" },
		{ "process_model", "fork" },
		{ "server_bind", "0.0.0.0" },
		{ "ical_port", "8080" },
		{ "ical_enable", "yes" },
		{ "icals_port", "8443" },
		{ "icals_enable", "no" },
		{ "server_socket", "http://localhost:236/zarafa" },
		{ "server_timezone","Europe/Amsterdam"},
		{ "default_charset","utf-8"},
		{ "log_method", "file" },
		{ "log_file", "/var/log/zarafa/ical.log" },
		{ "log_level", "3", CONFIGSETTING_RELOADABLE },
		{ "log_timestamp", "1" },
        { "ssl_private_key_file", "/etc/zarafa/ical/privkey.pem" },
        { "ssl_certificate_file", "/etc/zarafa/ical/cert.pem" },
        { "ssl_verify_client", "no" },
        { "ssl_verify_file", "" },
        { "ssl_verify_path", "" },
		{ NULL, NULL },
	};
	struct option long_options[] = {
		{"help", no_argument, NULL, 'h'},
		{"config", required_argument, NULL, 'c'},
		{"version", no_argument, NULL, 'v'},
		{"foreground", no_argument, NULL, 'F'},
		{NULL, 0, NULL, 0}
	};

	while (1) {
		opt = my_getopt_long(argc, argv, "Fhc:V", long_options, NULL);

		if (opt == -1)
			break;

		switch (opt) {
			case 'c': lpszCfg = my_optarg; break;
			case 'F': g_bDaemonize = false; break;
			case 'V': PrintVersion(); goto exit;
			case 'h':
			default:  PrintHelp(argv[0]); goto exit;
		}
	}
	
	// init xml parser
	xmlInitParser();

	g_lpConfig = new ECConfig(lpDefaults);
	if (!g_lpConfig->LoadSettings(lpszCfg) || g_lpConfig->HasErrors()) {
		g_lpLogger = new ECLogger_File(1, 0, "-");
		LogConfigErrors(g_lpConfig, g_lpLogger);
		goto exit;
	}

	g_lpLogger = CreateLogger(g_lpConfig, argv[0], "ZarafaICal");

	if (strncmp(g_lpConfig->GetSetting("process_model"), "thread", strlen("thread")) == 0)
		g_bThreads = true;

	if (!g_lpLogger) {
		fprintf(stderr, "Error loading configuration or parsing commandline arguments.\n");
		goto exit;
	}

	// initialize SSL threading
    ssl_threading_setup();

	hr = MAPIInitialize(NULL);
	if (hr != hrSuccess) {
		fprintf(stderr, "Messaging API could not be initialized.");
		goto exit;
	}

	hr = HrSetupListners(&ulListenCalDAV, &ulListenCalDAVs);
	if (hr != hrSuccess)
		goto exit;


	// setup signals
	signal(SIGTERM, sigterm);
	signal(SIGINT, sigterm);
	signal(SIGHUP, sighup);
	signal(SIGCHLD, sigchld);
	signal(SIGPIPE, SIG_IGN);

    memset(&st, 0, sizeof(st));
    memset(&act, 0, sizeof(act));

    st.ss_sp = malloc(65536);
    st.ss_flags = 0;
    st.ss_size = 65536;

    act.sa_handler = sigsegv;
    act.sa_flags = SA_ONSTACK | SA_RESETHAND;

    sigaltstack(&st, NULL);
    sigaction(SIGSEGV, &act, NULL);
	sigaction(SIGBUS, &act, NULL);
	sigaction(SIGABRT, &act, NULL);

	// fork if needed and drop privileges as requested.
	// this must be done before we do anything with pthreads
	if (g_bDaemonize && unix_daemonize(g_lpConfig, g_lpLogger))
		goto exit;
	if (!g_bDaemonize)
		setsid();
	unix_create_pidfile(argv[0], g_lpConfig, g_lpLogger);
	if (unix_runas(g_lpConfig, g_lpLogger))
		goto exit;

	if (g_bThreads == false)
		g_lpLogger = StartLoggerProcess(g_lpConfig, g_lpLogger);
	else
		g_lpLogger->SetLogprefix(LP_TID);

	if (g_bThreads)
		mainthread = pthread_self();

	g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Starting zarafa-ical version " PROJECT_VERSION_CALDAV_STR " (" PROJECT_SVN_REV_STR "), pid %d", getpid());

	hr = HrProcessConnections(ulListenCalDAV, ulListenCalDAVs);
	if (hr != hrSuccess)
		goto exit;


	g_lpLogger->Log(EC_LOGLEVEL_FATAL, "CalDAV Gateway will now exit");

	// in forked mode, send all children the exit signal
	if (g_bThreads == false) {
		signal(SIGTERM, SIG_IGN);
		kill(0, SIGTERM);
		i = 30;						// wait max 30 seconds
		while (nChildren && i) {
			if (i % 5 == 0)
				g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Waiting for %d processes to exit", nChildren);
			sleep(1);
			i--;
		}

		if (nChildren)
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Forced shutdown with %d procesess left", nChildren);
		else
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "CalDAV Gateway shutdown complete");
	}

exit:

	if(st.ss_sp)
		free(st.ss_sp);

	ECChannel::HrFreeCtx();

	if (lpChannel)
		delete lpChannel;

	if (g_lpConfig)
		delete g_lpConfig;

	DeleteLogger(g_lpLogger);

	MAPIUninitialize();

	// Cleanup ssl parts
	ssl_threading_cleanup();

	// Cleanup libxml2 library
	xmlCleanupParser();

	return hr;

}

HRESULT HrSetupListners(int *lpulNormal, int *lpulSecure)
{
	HRESULT hr = hrSuccess;
	bool bListen = false;
	bool bListenSecure = false;
	int ulPortICal = 0;
	int ulPortICalS = 0;
	int ulNormalSocket = 0;
	int ulSecureSocket = 0;

	// setup sockets
	bListenSecure = (stricmp(g_lpConfig->GetSetting("icals_enable"), "yes") == 0);
	bListen = (stricmp(g_lpConfig->GetSetting("ical_enable"), "yes") == 0);

	if (!bListen && !bListenSecure) {
		g_lpLogger->Log(EC_LOGLEVEL_FATAL, "No ports to open for listening.");
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ulPortICal = atoi(g_lpConfig->GetSetting("ical_port"));
	ulPortICalS = atoi(g_lpConfig->GetSetting("icals_port"));

	// start listening on normal port
	if (bListen) {
		hr = HrListen(g_lpLogger, g_lpConfig->GetSetting("server_bind"), ulPortICal, &ulNormalSocket);
		if (hr != hrSuccess) {
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not listen on port %d. (0x%08X)", ulPortICal, hr);
			bListen = false;
		} else {
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Listening on port %d.", ulPortICal);
		}
	}

	// start listening on secure port
	if (bListenSecure) {
		hr = ECChannel::HrSetCtx(g_lpConfig, g_lpLogger);
		if (hr == hrSuccess) {
			hr = HrListen(g_lpLogger, g_lpConfig->GetSetting("server_bind"), ulPortICalS, &ulSecureSocket);
			if (hr != hrSuccess) {
				g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not listen on secure port %d. (0x%08X)", ulPortICalS, hr);
				bListenSecure = false;
			}
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Listening on secure port %d.", ulPortICalS);
		} else {
			g_lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not listen on secure port %d. (0x%08X)", ulPortICalS, hr);
			bListenSecure = false;
		}
	}

	if (!bListen && !bListenSecure) {
		g_lpLogger->Log(EC_LOGLEVEL_FATAL, "No ports have been opened for listening, exiting.");
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = hrSuccess;
	*lpulNormal = ulNormalSocket;
	*lpulSecure = ulSecureSocket;

exit:
	return hr;
}

/**
 * Listen to the passed sockets and calls HrStartHandlerClient for
 * every incoming connection.
 *
 * @param[in]	ulNormalSocket	Listening socket of incoming HTTP connections
 * @param[in]	ulSecureSocket	Listening socket of incoming HTTPS connections
 * @retval MAPI error code
 */
HRESULT HrProcessConnections(int ulNormalSocket, int ulSecureSocket)
{
	HRESULT hr = hrSuccess;
	fd_set readfds = {{0}};
	int err = 0;
	bool bUseSSL;
	struct timeval timeout = {0};
	ECChannel *lpChannel = NULL;
	int nCloseFDs = 0;
	int pCloseFDs[2] = {0};

	if (ulNormalSocket)
		pCloseFDs[nCloseFDs++] = ulNormalSocket;
	if (ulSecureSocket)
		pCloseFDs[nCloseFDs++] = ulSecureSocket;

	// main program loop
	while (!g_bQuit) {
		FD_ZERO(&readfds);
		if (ulNormalSocket)
			FD_SET(ulNormalSocket, &readfds);
		if (ulSecureSocket)
			FD_SET(ulSecureSocket, &readfds);

		timeout.tv_sec = 10;
		timeout.tv_usec = 0;

		// Check whether there are incoming connections.
		err = select(max(ulNormalSocket, ulSecureSocket) + 1, &readfds, NULL, NULL, &timeout);
		if (err < 0) {
			if (errno != EINTR) {
				g_lpLogger->Log(EC_LOGLEVEL_FATAL, "An unknown socket error has occurred.");
				g_bQuit = true;
				hr = MAPI_E_NETWORK_ERROR;
			}
			continue;
		} else if (err == 0) {
			continue;
		}

		if (g_bQuit) {
			hr = hrSuccess;
			break;
		}

		// Check if a normal connection is waiting.
		if (ulNormalSocket && FD_ISSET(ulNormalSocket, &readfds)) {
			g_lpLogger->Log(EC_LOGLEVEL_INFO, "Connection waiting on port %d.", atoi(g_lpConfig->GetSetting("ical_port")));
			bUseSSL = false;
			hr = HrAccept(g_lpLogger, ulNormalSocket, &lpChannel);
			if (hr != hrSuccess) {
				g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not accept incoming connection on port %d. (0x%08X)", atoi(g_lpConfig->GetSetting("ical_port")), hr);
				continue;
			}
		// Check if a secure connection is waiting.
		} else if (ulSecureSocket && FD_ISSET(ulSecureSocket, &readfds)) {
			g_lpLogger->Log(EC_LOGLEVEL_INFO, "Connection waiting on secure port %d.", atoi(g_lpConfig->GetSetting("icals_port")));
			bUseSSL = true;
			hr = HrAccept(g_lpLogger, ulSecureSocket, &lpChannel);
			if (hr != hrSuccess) {
				g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not accept incoming secure connection on port %d. (0x%08X)", atoi(g_lpConfig->GetSetting("ical_port")), hr);
				continue;
			}
		} else {
			continue;
		}

		hr = HrStartHandlerClient(lpChannel, bUseSSL, nCloseFDs, pCloseFDs);
		if (hr != hrSuccess) {
			delete lpChannel;	// destructor closes sockets
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Handling client connection failed. (0x%08X)", hr);
			continue;
		}
		if (g_bThreads == false)
			delete lpChannel;	// always cleanup channel in main process
	}

	return hr;
}

/**
 * Starts a new thread or forks a child to process the incoming
 * connection.
 *
 * @param[in]	lpChannel	The accepted connection in ECChannel object
 * @param[in]	bUseSSL		The ECChannel object is an SSL connection
 * @param[in]	nCloseFDs	Number of FDs in pCloseFDs, used on forks only
 * @param[in]	pCloseFDs	Array of FDs to close in child process
 * @retval E_FAIL when thread of child did not start
 */
HRESULT HrStartHandlerClient(ECChannel *lpChannel, bool bUseSSL, int nCloseFDs, int *pCloseFDs)
{
	HRESULT hr = hrSuccess;
	pthread_attr_t pThreadAttr;
	pthread_t pThread;
	HandlerArgs *lpHandlerArgs = new HandlerArgs;

	lpHandlerArgs->lpChannel = lpChannel;
	lpHandlerArgs->bUseSSL = bUseSSL;

	if (g_bThreads) {
		pthread_attr_init(&pThreadAttr);

		if (pthread_attr_setdetachstate(&pThreadAttr, PTHREAD_CREATE_DETACHED) != 0) {
			g_lpLogger->Log(EC_LOGLEVEL_WARNING, "Could not set thread attribute to detached.");
		}

		if (pthread_create(&pThread, &pThreadAttr, HandlerClient, lpHandlerArgs) != 0) {
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not create thread.");
			hr = E_FAIL;
			goto exit;
		}
	} else {
		if (unix_fork_function(HandlerClient, lpHandlerArgs, nCloseFDs, pCloseFDs) < 0) {
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not create process.");
			hr = E_FAIL;
			goto exit;
		}
		nChildren++;
	}

exit:
	if (hr != hrSuccess && lpHandlerArgs)
		delete lpHandlerArgs;

	return hr;
}

void *HandlerClient(void *lpArg)
{
	HRESULT hr = hrSuccess;
	HandlerArgs *lpHandlerArgs = (HandlerArgs *) lpArg;
	ECChannel *lpChannel = lpHandlerArgs->lpChannel;
	bool bUseSSL = lpHandlerArgs->bUseSSL;	
	IMAPISession *lpSession = NULL;

	delete lpHandlerArgs;

	if (bUseSSL) {
		if (lpChannel->HrEnableTLS() != hrSuccess) {
			g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to negotiate SSL connection");
			goto exit;
		}
    }

	while(1)
	{
		hr = lpChannel->HrSelect(KEEP_ALIVE_TIME);
		if (hr != hrSuccess) {
			g_lpLogger->Log(EC_LOGLEVEL_INFO, "Request timeout, closing connection");
			break;
		}

		//Save mapi session between Requests
		hr = HrHandleRequest(lpChannel, &lpSession);
		if (hr != hrSuccess) {
			g_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request done: 0x%08X", hr);
			break;
		}
	}

exit:
	if (hr == hrSuccess)
		g_lpLogger->Log(EC_LOGLEVEL_INFO, "Connection closed");
		
	if(lpChannel)
		delete lpChannel;
	
	if(lpSession)
		lpSession->Release();

	return NULL;
}

HRESULT HrHandleRequest(ECChannel *lpChannel, IMAPISession **lpSessionSave)
{
	HRESULT hr = hrSuccess;
	std::string strBuffer;
	std::string strUser;
	std::string strPass;
	std::string strCalPath;
	std::string strService;
	std::string strSrvPath = g_lpConfig->GetSetting("server_socket");
	std::string strMsg;
	std::string strCalName;
	std::string strMethod;
	std::string strServerTZ = g_lpConfig->GetSetting("server_timezone");
	std::string strDefaultCharset = g_lpConfig->GetSetting("default_charset");
	std::string strCharset;

	Http *lpRequest = new Http(lpChannel,g_lpLogger);
	CalDAV *lpCalDav = NULL;
	iCal *lpiCal = NULL;
	IMAPISession *lpSession = NULL;

	if(lpSessionSave)
		lpSession = *lpSessionSave;

	g_lpLogger->Log(EC_LOGLEVEL_WARNING, "New Request");

	hr = lpRequest->HrReadHeaders();
	if(hr != hrSuccess)
	{
		hr = MAPI_E_USER_CANCEL; // connection is closed by client no data to be read
		goto exit;
	}

	hr = lpRequest->HrReadBody();
	//ignore Empty Body

	hr = lpRequest->HrValidateReq();
	if(hr != hrSuccess) {
		g_lpLogger->Log(EC_LOGLEVEL_ERROR, "Invalid HTTP request: %08X", hr);
		lpRequest->HrResponseHeader(400, "Bad Request");
		lpRequest->HrResponseBody("\nBad Request");
		goto exit;
	}

	hr = lpRequest->HrGetCharSet(&strCharset);
	if (hr != hrSuccess)
	{
		// TODO: may be parsed from the xml body, since it contains: encoding="utf-8" in the xml header
		strCharset = strDefaultCharset;
		g_lpLogger->Log(EC_LOGLEVEL_DEBUG, "No charset specified in http request. Using default charset: %s", strCharset.c_str());
	}

	hr = lpRequest->HrGetUser(&strUser);
	//ignore Empty User field.

	hr = lpRequest->HrGetPass(&strPass);
	//ignore Empty Password field

	hr = lpRequest->HrGetMethod(&strMethod);
	if(hr != hrSuccess)
	{
		g_lpLogger->Log(EC_LOGLEVEL_INFO,"Request Method not found");
		goto exit;
	}

	lpRequest->HrSetKeepAlive(KEEP_ALIVE_TIME);

	if(!strMethod.compare("OPTIONS"))
	{
		lpRequest->HrResponseHeader(200,"OK");
		lpRequest->HrResponseHeader("Allow", "OPTIONS, GET, HEAD, POST, PUT, DELETE, TRACE, COPY, MOVE");
		lpRequest->HrResponseHeader("Allow", "PROPFIND, PROPPATCH, LOCK, UNLOCK, REPORT");
		goto exit;
	}

	hr = lpRequest->HrGetService(&strService);
	if(hr != hrSuccess || ( strService.compare("ical") && strService.compare("caldav") ))
	{
		//Hack to bypass some Requests for Free-busy Info
		if (!strMethod.compare("PROPFIND") || !strMethod.compare("PROPPATCH"))
		{
			lpRequest->HrResponseHeader(200,"Ok");
			hr = hrSuccess;
		}
		else
		{
			g_lpLogger->Log(EC_LOGLEVEL_WARNING, "Invalid client request. Service: %s, Method: %s", strService.c_str(), strMethod.c_str());
			lpRequest->HrResponseHeader(501,"Bad Request");
			lpRequest->HrResponseBody("Bad Request");
		}
		goto exit;
	}

	if(!lpSession && (strUser.empty() || strPass.empty() || HrAuthenticate(strUser, strPass, strSrvPath, &lpSession) != hrSuccess )) {
		g_lpLogger->Log(EC_LOGLEVEL_INFO, "Sending authentication request");
		if(!strService.compare("ical"))
			strMsg = "Zarafa iCal Gateway";
		else
			strMsg = "Zarafa CalDav Gateway";
		lpRequest->HrRequestAuth(strMsg);
		hr = hrSuccess; //keep connection open.
		goto exit;
	}
	//PROCESS GET & ical Requests
	if( !strMethod.compare("GET") || (!strService.compare("ical") && strMethod.compare("PROPFIND")) )
	{
		lpiCal = new iCal(lpRequest,lpSession,g_lpLogger, strServerTZ, strCharset);

		if (!strMethod.compare("GET"))
			hr = lpiCal->HrHandleIcalGet();
		else if (!strMethod.compare("PUT"))
			hr = lpiCal->HrHandleIcalPost(strServerTZ);
		else if (!strMethod.compare("DELETE"))
			hr = lpiCal->HrDelFolder();
		else
			goto exit;
	}
	////PROCESS CALDAV Requests
	else if(!strService.compare("caldav") || ( !strMethod.compare("PROPFIND") && !strService.compare("ical")))
	{
		lpCalDav = new CalDAV(lpRequest, lpSession, g_lpLogger, strServerTZ, strCharset);
		hr = lpCalDav->HrHandleCommand(strMethod);
	}	

exit:
	if(hr != hrSuccess && !strMethod.empty())
		g_lpLogger->Log(EC_LOGLEVEL_ERROR,"Error processing %s request",strMethod.c_str());

	if ( lpRequest && hr != MAPI_E_USER_CANCEL ) // do not send response to client if connection closed by client.
		hr = lpRequest->HrFinalize();

	g_lpLogger->Log(EC_LOGLEVEL_WARNING, "End Of Request");

	if(lpRequest)
		delete lpRequest;
	
	if (lpiCal)
		delete lpiCal;

	if(lpSession)
		*lpSessionSave = lpSession;

	if(lpCalDav)
		delete lpCalDav;

	return hr;
}
