/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "PublishFreeBusy.h"
#include "CalDavProto.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/**
 * Maping of caldav properties to Mapi properties
 */
struct sMymap {
	unsigned int ulPropTag;
	char *name;
} sPropMap[] = {
	{ PR_LOCAL_COMMIT_TIME_MAX, "getctag" },
	{ PR_LAST_MODIFICATION_TIME, "getetag" },
	{ PR_DISPLAY_NAME, "displayname" },
	{ PR_CONTAINER_CLASS, "resourcetype" },
	{ PR_DISPLAY_NAME, "owner" },
	{ PR_ENTRYID, "calendar-data" },
	{ PR_COMMENT, "calendar-description" },
	{ PR_DISPLAY_TYPE, "calendar-user-type" },
	{ PR_SMTP_ADDRESS, "email-address-set" },
	{ PR_SMTP_ADDRESS, "calendar-user-address-set" },
	{ PR_DISPLAY_NAME, "first-name" },
	{ PR_DISPLAY_TYPE, "record-type" }
};

/**
 * Retrieves the Mapi Property from caldav property
 * @param[in]	name	ical property name
 * @return		ULONG	Proptery tag of corresponding mapi property
 */
ULONG GetPropmap(char *name)
{
	for(unsigned int i = 0; i < sizeof(sPropMap)/sizeof(sPropMap[0]); i++)
	{
		if (strcmp(name, sPropMap[i].name) == 0)
			return sPropMap[i].ulPropTag;
	}

	return 0;
}
/**
 * Default constructor
 * @param[in]	lpRequest	Pointer to Http class object
 * @param[in]	lpSession	Pointer to Mapi session object
 * @param[in]	lpLogger	Pointer to ECLogger object
 * @param[in]	strSrvTz	String specifying the server timezone, set in ical.cfg
 * @param[in]	strCharset	String specifying the default charset of the http response
 */
CalDAV::CalDAV(Http *lpRequest, IMAPISession *lpSession, ECLogger *lpLogger, std::string strSrvTz, std::string strCharset ) : WebDav(lpRequest, lpSession, lpLogger)
{
	m_lpUsrFld = NULL;
	m_lpRootFld = NULL;
	m_lpDefStore = NULL;
	m_lpAddrBook = NULL;
	m_lpSharedStore = NULL;
	m_lpPubStore = NULL;
	m_lpActiveStore = NULL;
	m_lpImailUser = NULL;
	m_lpNamedProps = NULL;
	m_blFolderAccess = true;
	m_ulFolderFlag = 0;
	m_strSrvTz = strSrvTz;
	m_strCharset = strCharset;

	m_lpConvResponse = new ECIConv("utf-8", "windows-1252");	
	m_lpConvRequest = new ECIConv("windows-1252", m_strCharset);
}
/**
 * Default Destructor
 */
CalDAV::~CalDAV()
{
	if (m_lpNamedProps)
		MAPIFreeBuffer(m_lpNamedProps);

	if (m_lpConvRequest)
		delete m_lpConvRequest;

	if (m_lpConvResponse)
		delete m_lpConvResponse;

	if (m_lpImailUser)
		m_lpImailUser->Release();

	if (m_lpRootFld)
		m_lpRootFld->Release();

	if (m_lpUsrFld)
		m_lpUsrFld->Release();

	if (m_lpAddrBook)
		m_lpAddrBook->Release();

	if (m_lpDefStore)
		m_lpDefStore->Release();

	if (m_lpSharedStore)
		m_lpSharedStore->Release();

	if(m_lpPubStore)
		m_lpPubStore->Release();
}

/**
 * Process all the caldav requests
 * @param[in]	strMethod	Name of the http request(e.g PROPFIND, REPORT..)
 * @return		HRESULT 
 */
HRESULT CalDAV::HrHandleCommand(std::string strMethod)
{
	HRESULT hr = hrSuccess;

	hr = HrGetFolder();
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Folder named '%s' not found, error code : 0x%08X", m_strFldName.c_str(), hr);
		goto exit;
	}
	
	// force check  for folder for the following requests and ignore folders for rest
	if(!m_lpUsrFld &&  (!strMethod.compare("PUT") || !strMethod.compare("DELETE") || !strMethod.compare("MOVE")))
	{
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Folder named '%s' not found, error code : 0x%08X", m_strFldName.c_str(), hr);
		goto exit;
	}

	if(!strMethod.compare("PROPFIND")) {
		hr = HrPropfind();
	}
	else if(!strMethod.compare("REPORT")) {
		hr = HrReport();
	}
	else if(!strMethod.compare("PUT")) {
		hr = HrPut();
	}
	else if(!strMethod.compare("DELETE")) {
		hr = HrDelete();
	}
	else if(!strMethod.compare("MKCALENDAR")) {
		hr = HrMkCalendar();
	}
	else if(!strMethod.compare("PROPPATCH")) {
		hr = HrPropPatch();
	}
	else if (!strMethod.compare("POST"))
	{
		hr = HrPost();
	}
	else
	{
		m_lpRequest->HrResponseHeader(501, "Not Implemented");
	}

exit:
	if (hr != hrSuccess)
		m_lpRequest->HrResponseHeader(400, "Bad Request");

	return hr;
}

/**
 * Open Store and folders required for the Request
 *
 * @return	MAPI error code
 * @retval	MAPI_E_NO_ACCESS	Not enough permissions to open the folder or store
 * @retval	MAPI_E_NOT_FOUND	Folder requested by client not found
 */
HRESULT CalDAV::HrGetFolder()
{
	HRESULT hr = hrSuccess;
	std::string strMethod;
	std::string strFldName;	
	std::string strService;
	size_t ulFound = 0;
	ULONG ulDepth;
	LPSPropValue lpDefaultProp = NULL;
	LPSPropValue lpFldProp = NULL;
	IMAPIFolder *lpRootCont = NULL;
	bool blCreateIFNf = false;

	m_lpRequest->HrGetMethod(&strMethod);
	m_lpRequest->HrGetFldPath(&strFldName);
	m_lpRequest->HrGetDepth(&ulDepth);

	m_lpRequest->HrGetUser(&m_strUser);
	m_lpRequest->HrGetFldOwner(&m_strFldOwner);
	m_strUser = U2W(m_strUser);
	strFldName = U2W(strFldName);

	hr = m_lpSession->OpenAddressBook(0, NULL, 0, &m_lpAddrBook);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening addressbook, error code : 0x%08X", hr);
		goto exit;
	}

	hr = HrOpenDefaultStore(m_lpSession, &m_lpDefStore);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening default store of user %s, error code : 0x%08X", m_strUser.c_str(), hr);
		goto exit;
	}
	
	m_lpActiveStore = m_lpDefStore;
	m_lpRequest->HrGetService(&strService);
	
	if(m_lpRequest->IsPublic())
	{
		hr = HrOpenECPublicStore(m_lpSession, &m_lpPubStore);
		if (hr != hrSuccess)
			goto exit;
				
		m_lpActiveStore = m_lpPubStore;
		hr = OpenSubFolder(m_lpActiveStore, NULL, '/', m_lpLogger, true, false, &m_lpRootFld);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening IPM_SUBTREE Folder of user %s, error code : 0x%08X", m_strUser.c_str(), hr);
			goto exit;
		}
	}

	// Retrive named properties
	hr = HrLookupNames(m_lpActiveStore, &m_lpNamedProps);
	if (hr != hrSuccess)
		goto exit;

	if ( (!strMethod.compare("PUT")) && (!strService.compare("ical")) )
		blCreateIFNf = true;

	if(!strService.compare("ical"))
	{
		ulFound = strFldName.rfind(".ics");
		if (ulFound != string::npos)
			strFldName = strFldName.substr(0, ulFound);
	}else if(!strMethod.compare("PUT") || !strMethod.compare("DELETE") || !strMethod.compare("GET")) {
		// PUT & DELETE method have calendar entry uid in the request so remove that
		ulFound = strFldName.rfind("/");

		if (ulFound != string::npos)
		{
			if(ulFound == string::npos)
				strFldName = strFldName;
			else
				strFldName = strFldName.substr( 0, ulFound);
		}

		// when using default url '/caldav' then '/' is not present in url
		ulFound = strFldName.rfind(".ics");
		if (ulFound != string::npos)
			 strFldName = "";

	}

	m_strFldName = strFldName;
	if(!m_strFldOwner.empty() && m_strFldOwner != m_strUser && !m_lpRequest->IsPublic())
	{
		hr = HrGetSharedFolder(m_lpSession, m_lpDefStore, m_strFldOwner, m_strFldName, m_blIsCLMAC, false, m_lpLogger, &m_ulFolderFlag, &m_lpSharedStore, &m_lpUsrFld);
		if(m_lpSharedStore)
		{
			m_lpActiveStore = m_lpSharedStore;
			hr = OpenSubFolder(m_lpActiveStore, NULL, '/', m_lpLogger, false, false, &m_lpRootFld);
			if(hr != hrSuccess)
			{
				hr = hrSuccess;
				m_lpRootFld = NULL;
			}
		}

		if(hr == MAPI_E_NO_ACCESS)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Not enough permissions to user %s to access Folder for user %s", m_strUser.c_str(), m_strFldOwner.c_str());
			goto exit;
		}
		else if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while accessing shared folder by user %s, error code : 0x%08X", m_strUser.c_str(), hr);
			goto exit;
		}
		
	}	
	else if(m_blIsCLMAC ) //For Ical Client The folders are opened according to Folder_ID.
	{
		if(!strFldName.empty())
		{
			hr = HrFindFolder(m_lpSession, m_lpActiveStore, m_lpNamedProps, m_lpLogger, m_lpRequest->IsPublic(), strFldName, &m_lpUsrFld);
			if(hr != hrSuccess)
			{
				hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, blCreateIFNf, m_lpRequest->IsPublic(), m_lpLogger, &m_lpUsrFld);
				if(hr == hrSuccess) // do not exit , need to set the m_lpRootfolder if not set.
					m_ulFolderFlag |= SINGLE_FOLDER;
			}
		}
	}
	//For other Clients the Folder name is used to refer to folders.
	else
	{
		if (strMethod.compare("MKCALENDAR"))
		{
			hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, blCreateIFNf, m_lpRequest->IsPublic(), m_lpLogger, &m_lpUsrFld);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening folder of user %s, error code : 0x%08X", m_strUser.c_str(), hr);
				goto exit;
			}
		}
	}
	
	hr = HrGetUserInfo(m_lpSession, m_lpDefStore, &m_strUserEmail, &m_strUserFullName, &m_lpImailUser);
	if(hr != hrSuccess)
		goto exit;

	if(!m_lpRootFld)
	{
		//Open IPM_SUBTREE
		hr = OpenSubFolder(m_lpDefStore, NULL, '/', m_lpLogger, false, false, &m_lpRootFld);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening IPM_SUBTREE Folder of user %s, error code : 0x%08X", m_strUser.c_str(), hr);
			goto exit;
		}
	}

	//Set Limited Access to Default Calendar Folder
	//Block Rename & Deletion of this folder.
	if( (!strMethod.compare("PROPPATCH") || !strMethod.compare("DELETE") || !strMethod.compare("PUT")) && !m_lpRequest->IsPublic())
	{
		int ulCmp = 0;
		ULONG ulType = 0;

		hr = m_lpActiveStore->OpenEntry(0, NULL, NULL, 0, &ulType, (LPUNKNOWN*)&lpRootCont);
		if (hr != hrSuccess || ulType != MAPI_FOLDER)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening root Container of user %s, error code: (0x%08X)", m_strUser.c_str(), hr);
			goto exit;
		}

		m_blFolderAccess = CheckFolderAccess(m_lpActiveStore, lpRootCont, m_lpUsrFld);

		hr = HrGetOneProp(lpRootCont, PR_IPM_APPOINTMENT_ENTRYID, &lpDefaultProp);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving Entry id of Default calendar for user %s, error code: 0x%08X", m_strUser.c_str(), hr);
			goto exit;
		}

		if(m_lpUsrFld)
			hr = HrGetOneProp(m_lpUsrFld, PR_ENTRYID, &lpFldProp);
		else
		{
			m_blFolderAccess = false;
			goto exit;
		}

		if(hr != hrSuccess)
			goto exit;
		
		
		ulCmp = memcmp(lpDefaultProp->Value.bin.lpb, lpFldProp->Value.bin.lpb, lpFldProp->Value.bin.cb);
		if(!ulCmp) {
			m_blFolderAccess = false;
			m_ulFolderFlag |= DEFAULT_FOLDER;
		}
		
		MAPIFreeBuffer(lpFldProp);
		lpFldProp = NULL;

		hr = HrGetOneProp(m_lpUsrFld, PR_SUBFOLDERS, &lpFldProp);
		if(hr != hrSuccess)
			goto exit;

		if(lpFldProp->Value.b == (unsigned short)true && !strMethod.compare("DELETE"))
			m_blFolderAccess = false;
	}

exit:
	if(lpFldProp)
		MAPIFreeBuffer(lpFldProp);

	if(lpDefaultProp)
		MAPIFreeBuffer(lpDefaultProp);

	if(lpRootCont)
		lpRootCont->Release();

	return hr;
}

/**
 * Process the PROPFIND request in general
 * @return	HRESULT
 */
HRESULT CalDAV::HrPropfind()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrPropfind();
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

/**
 * Handles the PROPFIND request, identifies the type of PROPFIND request
 *
 * @param[in]	lpsDavProp			Pointer to structure cotaining info about the PROPFIND request
 * @param[out]	lpsDavMulStatus		Response generated for the PROPFIND request
 * @return		HRESULT
 */
HRESULT CalDAV::HrHandlePropfind(WEBDAVREQSTPROPS *lpsDavProp, WEBDAVMULTISTATUS *lpsDavMulStatus)
{
	HRESULT hr = hrSuccess;
	std::string strFldPath;
	ULONG ulDepth = 0;
	
	m_lpRequest->HrGetDepth(&ulDepth);
	m_lpRequest->HrGetFldPath(&strFldPath);

	if(ulDepth == 0 || ( strFldPath == "Inbox" ) || ( strFldPath == "Outbox"))
	{
		// Retrieve Properties
		hr = HrHandlePropfindRoot(lpsDavProp, lpsDavMulStatus);
	}
	else if( (strFldPath.empty() || !strFldPath.compare("public") ||( m_ulFolderFlag & SINGLE_FOLDER)) && ulDepth == 1 && m_blIsCLMAC)
	{	
		// Retrieve list of calendars
		hr = HrListCalendar(&(lpsDavProp->sProp), lpsDavMulStatus);
	}
	else
	{
		// Retrieve the Calendar entries list
		hr = HrListCalEntries(lpsDavProp, lpsDavMulStatus);
	}

	return hr;
}

/**
 * Handles the Depth 0 PROPFIND request
 *
 * The client requets for information about the user,store and folder by using this request
 *
 * @param[in]	sDavReqstProps		Pointer to structure cotaining info about properties requested by client
 * @param[in]	lpsDavMulStatus		Pointer to structure cotaining response to the request
 * @return		HRESULT
 */
HRESULT CalDAV::HrHandlePropfindRoot(WEBDAVREQSTPROPS *sDavReqstProps, WEBDAVMULTISTATUS *lpsDavMulStatus)
{
	HRESULT hr = hrSuccess;
	std::string strFldPath;
	std::string strReqPath;
	std::list<WEBDAVPROPERTY>::iterator iter;		
	WEBDAVPROP *lpsDavProp = NULL;
	WEBDAVRESPONSE sDavResp;
	IMAPIProp *lpMapiProp = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSPropValue lpSpropVal = NULL;
	ULONG cbsize = 0;

	lpsDavProp = &(sDavReqstProps->sProp);

	// number of properties requested by client.
	cbsize = lpsDavProp->lstProps.size();
	
	m_lpRequest->HrGetFldPath(&strFldPath);
	m_lpRequest->HrGetReqPath(&strReqPath);

	if(m_lpRequest->IsPublic())
		m_strCalHome = "/caldav/public/" + strFldPath;
	else
		m_strCalHome = "/caldav/" + W2U(m_strFldOwner.empty()? m_strUser : m_strFldOwner) + "/" + strFldPath;
	
	if (m_lpRequest->IsPublic() && strFldPath.empty())
		lpMapiProp = m_lpDefStore;
	else if(m_blIsCLMAC && (strFldPath.empty() || strFldPath == "Inbox" || strFldPath == "Outbox") )
		lpMapiProp = m_lpActiveStore;
	else if(m_lpUsrFld)
		lpMapiProp = m_lpUsrFld;
	else
	{
		m_lpLogger->Log( EC_LOGLEVEL_ERROR,"Error Folder named '%s' of user '%s' not found, error code 0x%08X", strFldPath.c_str(), m_strUser.c_str(), hr);
		goto exit;
	}
	
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **) &lpPropTagArr);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	lpPropTagArr->cValues = cbsize;

	// Get corresponding mapi properties.
	iter = lpsDavProp->lstProps.begin();
	for(int i = 0; iter != lpsDavProp->lstProps.end(); iter++, i++)
	{
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)iter->sPropName.strPropname.c_str());
	}

	hr = lpMapiProp->GetProps((LPSPropTagArray)lpPropTagArr, 0, &cbsize, &lpSpropVal);
	if (hr == MAPI_W_ERRORS_RETURNED)
		hr = hrSuccess;		// ignoring some items
	else if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetProps for user %s, error code : 0x%08X", m_strUser.c_str(), hr);
		goto exit;
	}

	HrSetDavPropName(&(sDavResp.sPropName), "response", CALDAVNSDEF);

	HrSetDavPropName(&(sDavResp.sHRef.sPropName), "href", CALDAVNSDEF);
	sDavResp.sHRef.strValue = strReqPath;

	// map values and properties in WEBDAVRESPONSE structure.
	hr = HrMapValtoStruct(lpSpropVal, cbsize, NULL, PROPFIND_ROOT, &(lpsDavProp->lstProps), &sDavResp);
	if (hr != hrSuccess)
		goto exit;
	
	HrSetDavPropName(&(lpsDavMulStatus->sPropName), "multistatus", CALDAVNSDEF);
	lpsDavMulStatus->lstResp.push_back(sDavResp);

exit:

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpSpropVal)
		MAPIFreeBuffer(lpSpropVal);

	return hr;
}

/**
 * Process the REPORT request
 *
 * REPORT has two parts CALENDAR-QUERY and MULTIGET.
 *
 * @return HRESULT
 */
HRESULT CalDAV::HrReport()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrReport();

	return hr;
}

/** 
 * Retrieve list of entries in the calendar folder
 *
 * The function handles REPORT(calendar-query) and PROPFIND(depth 1) request,
 * REPORT method is used by mozilla clients and mac ical.app uses PROPFIND request
 *
 * @param[in]	lpsWebRCalQry	Pointer to structure containing the list of properties requested by client
 * @param[out]	lpsWebMStatus	Pointer to structure containing the response
 * @return		HRESULT
 */
HRESULT CalDAV::HrListCalEntries(WEBDAVREQSTPROPS *lpsWebRCalQry, WEBDAVMULTISTATUS *lpsWebMStatus)
{
	HRESULT hr = hrSuccess;
	std::string strConvVal;
	std::string strReqPath;
	std::string strTrail;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPITable *lpTable = NULL;
	LPSRowSet lpRowSet = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSPropTagArray lpPropTags = NULL;
	LPSPropValue lpsPropVal = NULL;
	MapiToICal *lpMtIcal = NULL;
	ULONG cbsize = 0;
	ULONG ulTagGOID = 0;
	ULONG ulTagTsRef = 0;
	ULONG ulTagPrivate = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVPROP sDavProp;
	WEBDAVPROPERTY sDavProperty;
	WEBDAVRESPONSE sWebResponse;
	bool blCensorPrivate = false;
	ULONG ulCensorFlag = 0;

	ulTagGOID = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	ulTagTsRef = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTTSREF], PT_STRING8);
	ulTagPrivate = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);
	
	hr = m_lpRequest->HrGetReqPath(&strReqPath);
	if(hr != hrSuccess)
		goto exit;

	strTrail = strReqPath.at(strReqPath.length() - 1);
	if(strTrail.compare("/"))
		strReqPath = strReqPath + "/";

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	HrSetDavPropName(&(lpsWebMStatus->sPropName), "multistatus", CALDAVNSDEF);
	HrSetDavPropName(&(sWebResponse.sPropName), "response", CALDAVNSDEF);
	HrSetDavPropName(&(sWebResponse.sHRef.sPropName), "href", CALDAVNSDEF);
		
	sDavProp = lpsWebRCalQry->sProp;

	if(m_lpUsrFld)
		lpUsrFld = m_lpUsrFld;
	else
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error user's folder not found, cannot list calendar entries");
		goto exit;
	}

	hr = HrGetOneProp(m_lpUsrFld, PR_CONTAINER_CLASS, &lpsPropVal);
	if (hr != hrSuccess)
		goto exit;

	
	if (!lpsWebRCalQry->sFilter.lstFilters.empty())
	{
		if (lpsWebRCalQry->sFilter.lstFilters.back() == "VTODO"
			&& strncmp(lpsPropVal->Value.lpszA, "IPF.Task", strlen("IPF.Task")))
		{
				goto exit;
		}
	
		if (lpsWebRCalQry->sFilter.lstFilters.back() == "VEVENT"
			&& strncmp(lpsPropVal->Value.lpszA, "IPF.Appointment", strlen("IPF.Appointment")))
		{
			goto exit;
		}
	}

	hr = lpUsrFld->GetContentsTable(0, &lpTable);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetContentsTable, error code : 0x%08X", hr);
		goto exit;
	}

	// +4 to add GlobalObjid, dispidApptTsRef , PR_ENTRYID and private in SetColumns along with requested data.
	cbsize = (ULONG)sDavProp.lstProps.size() + 4;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	lpPropTagArr->cValues = cbsize;
	lpPropTagArr->aulPropTag[0] = ulTagTsRef;
	lpPropTagArr->aulPropTag[1] = ulTagGOID;
	lpPropTagArr->aulPropTag[2] = PR_ENTRYID;
	lpPropTagArr->aulPropTag[3] = ulTagPrivate;
	
	iter = sDavProp.lstProps.begin();

	//mapi property mapping for requested properties.
	//FIXME what if the property mapping is not found.
	for (int i = 4; iter != sDavProp.lstProps.end(); iter++, i++)
	{
		sDavProperty = *iter;
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)sDavProperty.sPropName.strPropname.c_str());
	}

	hr = lpTable->SetColumns((LPSPropTagArray)lpPropTagArr, 0);
	if(hr != hrSuccess)
		goto exit;
	
	CreateMapiToICal(m_lpAddrBook, "utf-8", &lpMtIcal);
	if (!lpMtIcal)
	{
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	while(1)
	{
		hr = lpTable->QueryRows(50, 0, &lpRowSet);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving rows of table");
			goto exit;
		}

		if(lpRowSet->cRows == 0)
			break;

		//add data from each requested property.
		for (ULONG ulRowCntr = 0; ulRowCntr < lpRowSet->cRows; ulRowCntr++)
		{
			ulCensorFlag = (ULONG)blCensorPrivate; // reset Censor flag for next message.

			//For entries created in Evolution client use Timestamp as URL ref
			if(lpRowSet->aRow[ulRowCntr].lpProps[0].ulPropTag == ulTagTsRef && m_blIsCLEVO)
				strConvVal = lpRowSet->aRow[ulRowCntr].lpProps[0].Value.lpszA;
			else if (lpRowSet->aRow[ulRowCntr].lpProps[1].ulPropTag == ulTagGOID)
				strConvVal = SPropValToString((SPropValue *)&(lpRowSet->aRow[ulRowCntr].lpProps[1]));
			else
				strConvVal.clear();

			// On some items, outlook never created the uid, so we need to create one for ical
			if (strConvVal.empty())
			{
				hr = CreateAndGetGuid(lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin, ulTagGOID, &strConvVal);
				if(hr == E_ACCESSDENIED)
				{
					// use PR_ENTRYID since we couldn't create a new guid for the item
					strConvVal = bin2hex(lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin.cb, lpRowSet->aRow[ulRowCntr].lpProps[2].Value.bin.lpb);
					hr = hrSuccess;
				}
				else if(hr != hrSuccess)
					continue;
			}

			strConvVal = strConvVal + ".ics";
			strConvVal = strReqPath + strConvVal;

			sWebResponse.sHRef.strValue = strConvVal;

			if (blCensorPrivate && (lpRowSet->aRow[ulRowCntr].lpProps[3].ulPropTag != ulTagPrivate || lpRowSet->aRow[ulRowCntr].lpProps[3].Value.b))
				ulCensorFlag |= M2IC_CENSOR_PRIVATE;
			else
				ulCensorFlag = 0;

			hr = HrMapValtoStruct(lpRowSet->aRow[ulRowCntr].lpProps, lpRowSet->aRow[ulRowCntr].cValues, lpMtIcal, ulCensorFlag, &(lpsWebRCalQry->sProp.lstProps), &sWebResponse);
			
			lpsWebMStatus->lstResp.push_back(sWebResponse);
			sWebResponse.lstsPropStat.clear();
		}

		if(lpRowSet)
		{
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}
	}

exit:

	if (lpsPropVal)
		MAPIFreeBuffer(lpsPropVal);

	if(lpMtIcal)
		delete lpMtIcal;

	if(lpPropTags)
		MAPIFreeBuffer(lpPropTags);

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpTable)
		lpTable->Release();

	if(lpRowSet)
		FreeProws(lpRowSet);
	
	return hr;
}

/**
 * Handles Report (calendar-multiget) caldav request.
 *
 * Sets values of requested caldav properties in WEBDAVMULTISTATUS structure.
 *
 * @param[in]	sWebRMGet		structure that contains the list of calendar entries and properties requested.
 * @param[out]	sWebMStatus		structure that values of requested properties.
 * @retval		HRESULT
 * @return		S_OK			always returns S_OK.
 *
 */
HRESULT CalDAV::HrHandleReport(WEBDAVRPTMGET *sWebRMGet, WEBDAVMULTISTATUS *sWebMStatus)
{
	HRESULT hr = hrSuccess;
	std::string strConvVal;
	std::string strReqPath;	
	IMAPIFolder *lpUsrFld = NULL;
	IMAPITable *lpTable = NULL;
	SRestriction * lpsRoot = NULL;
	LPSPropTagArray lpPropTags = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	MapiToICal *lpMtIcal = NULL;
	ULONG cbsize = 0;
	ULONG ulPropTag = 0;
	ULONG ulTagPrivate = 0;
	ULONG ulCensorFlag = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVPROP sDavProp;
	WEBDAVPROPERTY sDavProperty;
	WEBDAVRESPONSE sWebResponse;
	bool blCensorPrivate = false;
	
	HrSetDavPropName(&(sWebResponse.sPropName), "response", CALDAVNSDEF);
	HrSetDavPropName(&sWebMStatus->sPropName, "multistatus", CALDAVNSDEF);

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	if(m_lpUsrFld)
		lpUsrFld = m_lpUsrFld;
	else
	{
		hr = MAPI_E_NOT_FOUND;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error user's folder not found, cannot return ical entries");
		goto exit;
	}

	hr = lpUsrFld->GetContentsTable(0, &lpTable);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error in GetContentsTable, error code : 0x%08X", hr);
		goto exit;
	}

	sDavProp = sWebRMGet->sProp;

	//Add GUID in Setcolumns.
	cbsize = (ULONG)sDavProp.lstProps.size() + 2;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error allocating memory, error code : 0x%08X", hr);
		goto exit;
	}
	
	ulTagPrivate = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);
	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	
	lpPropTagArr->cValues = cbsize;
	lpPropTagArr->aulPropTag[0] = ulPropTag;
	lpPropTagArr->aulPropTag[1] = ulTagPrivate;
	
	iter = sDavProp.lstProps.begin();
	for(int i = 2; iter != sDavProp.lstProps.end(); iter++, i++)
	{
		sDavProperty = *iter;
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)sDavProperty.sPropName.strPropname.c_str());
	}

	hr = lpTable->SetColumns((LPSPropTagArray)lpPropTagArr, 0);
	if(hr != hrSuccess)
		goto exit;

	cbsize = (ULONG)sWebRMGet->lstWebVal.size();
	
	CreateMapiToICal(m_lpAddrBook, "utf-8", &lpMtIcal);
	if (!lpMtIcal)
	{
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	for(ULONG i = 0; i < cbsize; i++)
	{
		WEBDAVVALUE sWebDavVal;
		SRowSet *lpValRows = NULL;
		std::string strBinString;

		ulCensorFlag = (ULONG)blCensorPrivate; // reset bool for next message.
	
		sWebDavVal = sWebRMGet->lstWebVal.front();
		sWebRMGet->lstWebVal.pop_front();

		//assign href;
		m_lpRequest->HrGetReqPath(&strReqPath);
		strReqPath = strReqPath + sWebDavVal.strValue;
		strReqPath = strReqPath + ".ics";
		
		sWebResponse.sHRef = sWebDavVal;
		sWebResponse.sHRef.strValue = W2U(strReqPath);
		
		hr = HrMakeRestriction(sWebDavVal.strValue, m_lpNamedProps, &lpsRoot);
		if (hr != hrSuccess)
			goto next;
		
		hr = lpTable->FindRow(lpsRoot, BOOKMARK_BEGINNING, 0);
		if (hr != hrSuccess)
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Entry not found (%s), error code : 0x%08X", sWebDavVal.strValue.c_str(), hr);

		// conversion if everthing goes ok, otherwise, add empty item with failed status field
		// we need to return all items requested in the multistatus reply, otherwise sunbird will stop, displaying nothing to the user.

		if (hr == hrSuccess) {
			hr = lpTable->QueryRows(1, TBL_NOADVANCE, &lpValRows); // TODO: what if we get multiple items ?
		
			if(hr != hrSuccess || lpValRows->cRows != 1)
				goto exit;

			if (blCensorPrivate && (lpValRows->aRow[0].lpProps[1].ulPropTag == ulTagPrivate && lpValRows->aRow[0].lpProps[1].Value.b))
				ulCensorFlag |= M2IC_CENSOR_PRIVATE;
			else
				ulCensorFlag = 0;
		}

		if(hr == hrSuccess)
			hr = HrMapValtoStruct(lpValRows->aRow[0].lpProps, lpValRows->aRow[0].cValues, lpMtIcal, ulCensorFlag, &sDavProp.lstProps, &sWebResponse);
	
		sWebMStatus->lstResp.push_back(sWebResponse);
		sWebResponse.lstsPropStat.clear();
next:
		if (lpsRoot)
			FREE_RESTRICTION(lpsRoot);

		if(lpValRows)
			FreeProws(lpValRows);
		lpValRows = NULL;

	}

	hr = hrSuccess;

exit:
	
	if(lpMtIcal)
		delete lpMtIcal;

	if(lpsRoot)
		FREE_RESTRICTION(lpsRoot);

	if(lpTable)
		lpTable->Release();

	if(lpPropTags)
		MAPIFreeBuffer(lpPropTags);
	
	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);
	
	return hr;
}

/**
 * Generates response to Property search set request
 *
 * The request is to list the properties that can be requested by client,
 * while requesting for attendee suggestions
 *
 * @param[out]	lpsWebMStatus	Response structure returned
 *
 * @return		HRESULT			Always returns hrSuccess
 */
HRESULT CalDAV::HrHandlePropertySearchSet(WEBDAVMULTISTATUS *lpsWebMStatus)
{
	HRESULT hr = hrSuccess;
	WEBDAVRESPONSE sDavResponse;
	WEBDAVPROPSTAT sDavPropStat;
	WEBDAVPROPERTY sWebProperty;
	WEBDAVPROP sWebProp;

	HrSetDavPropName(&(lpsWebMStatus->sPropName), "principal-search-property-set", CALDAVNSDEF);

	HrSetDavPropName(&sDavResponse.sPropName, "principal-search-property", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sPropName, "prop", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "displayname", CALDAVNSDEF);
	sDavResponse.lstsPropStat.push_back(sDavPropStat);
	HrSetDavPropName(&sDavResponse.sHRef.sPropName, "description", "xml:lang", "en", CALDAVNSDEF);	
	sDavResponse.sHRef.strValue = "Display Name";	
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "", "");

	lpsWebMStatus->lstResp.push_back(sDavResponse);	
	sDavResponse.lstsPropStat.clear();

	HrSetDavPropName(&sDavResponse.sPropName, "principal-search-property", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sPropName, "prop", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "calendar-user-type", CALDAVNSDEF);
	sDavResponse.lstsPropStat.push_back(sDavPropStat);
	HrSetDavPropName(&sDavResponse.sHRef.sPropName, "description", "xml:lang", "en", CALDAVNSDEF);	
	sDavResponse.sHRef.strValue = "Calendar user type";	
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "", "");

	lpsWebMStatus->lstResp.push_back(sDavResponse);	
	sDavResponse.lstsPropStat.clear();

	HrSetDavPropName(&sDavResponse.sPropName, "principal-search-property", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sPropName, "prop", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "calendar-user-address-set", CALDAVNSDEF);
	sDavResponse.lstsPropStat.push_back(sDavPropStat);
	HrSetDavPropName(&sDavResponse.sHRef.sPropName, "description", "xml:lang", "en", CALDAVNSDEF);	
	sDavResponse.sHRef.strValue = "Calendar User Address Set";	
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "", "");

	lpsWebMStatus->lstResp.push_back(sDavResponse);	
	sDavResponse.lstsPropStat.clear();

	HrSetDavPropName(&sDavResponse.sPropName, "principal-search-property", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sPropName, "prop", CALDAVNSDEF);
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "email-address-set", "http://calendarserver.org/ns/");
	sDavResponse.lstsPropStat.push_back(sDavPropStat);
	HrSetDavPropName(&sDavResponse.sHRef.sPropName, "description", "xml:lang", "en", CALDAVNSDEF);	
	sDavResponse.sHRef.strValue = "Email Address";
	HrSetDavPropName(&sDavPropStat.sProp.sPropName, "", "");
	

	lpsWebMStatus->lstResp.push_back(sDavResponse);
	sDavResponse.lstsPropStat.clear();
	return hr;
}
/**
 * Handles attedee suggestion list request
 * 
 * @param[in]	sWebRMGet		Pointer to WEBDAVRPTMGET structure containing user to search in global address book
 * @param[out]	sWebMStatus		Pointer to WEBDAVMULTISTATUS structure cotaning attndees list matching the request
 *
 * @return		HRESULT			Always returns hrSuccess 
 */
HRESULT CalDAV::HrHandlePropertySearch(WEBDAVRPTMGET *sWebRMGet, WEBDAVMULTISTATUS *sWebMStatus)
{
	HRESULT hr = hrSuccess;
	IABContainer *lpAbCont = NULL;
	IMAPITable *lpTable = NULL;	
	SRestriction * lpsRoot = NULL;
	SRowSet *lpValRows = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSPropValue lpsPropVal = NULL;
	MapiToICal *lpMtIcal = NULL;
	ULONG cbsize = 0;
	ULONG ulPropTag = 0;
	ULONG ulTagPrivate = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	std::list<WEBDAVVALUE>::iterator iterWebVal;
	SBinary sbEid = {0};
	WEBDAVPROP sDavProp;
	WEBDAVPROPERTY sDavProperty;
	WEBDAVRESPONSE sWebResponse;
	ULONG ulObjType = 0;

	// Open Global Address book
	hr = m_lpAddrBook->GetDefaultDir(&sbEid.cb, (LPENTRYID*)&sbEid.lpb);
	if (hr != hrSuccess)
		goto exit;

	hr = m_lpSession->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, 0, &ulObjType, (LPUNKNOWN *)&lpAbCont);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpAbCont->GetContentsTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->GetRowCount(0, &ulObjType);
	if (hr != hrSuccess)
		goto exit;

	// create restriction
	CREATE_RESTRICTION(lpsRoot);
	CREATE_RES_OR(lpsRoot, lpsRoot, sWebRMGet->lstWebVal.size()); // max: or guid or raw or entryid
	iterWebVal = sWebRMGet->lstWebVal.begin();

	for (int i = 0; i < (int)sWebRMGet->lstWebVal.size(); i++, iterWebVal++)
	{
		hr = MAPIAllocateMore(sizeof(SPropValue), lpsRoot, (void**)&lpsPropVal);
		if (hr != hrSuccess)
			goto exit;

		hr = MAPIAllocateMore(iterWebVal->strValue.length() +1, lpsRoot, (void**)&lpsPropVal->Value.lpszA);
		if (hr != hrSuccess)
			goto exit;

		lpsPropVal->ulPropTag = GetPropmap((char *)iterWebVal->sPropName.strPropname.c_str());
		memcpy(lpsPropVal->Value.lpszA, iterWebVal->strValue.c_str(), iterWebVal->strValue.size()+ 1);

		DATA_RES_CONTENT(lpsRoot,lpsRoot->res.resOr.lpRes[i],FL_SUBSTRING | FL_IGNORECASE, lpsPropVal->ulPropTag, lpsPropVal);
		lpsPropVal = NULL;
	}

	// create proptagarray.
	sDavProp = sWebRMGet->sProp;

	//Add GUID in Setcolumns.
	cbsize = (ULONG)sDavProp.lstProps.size() + 3;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cbsize), (void **)&lpPropTagArr);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error allocating memory, error code : 0x%08X", hr);
		goto exit;
	}
	
	ulTagPrivate = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN);
	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_GOID], PT_BINARY);
	
	lpPropTagArr->cValues = cbsize;
	lpPropTagArr->aulPropTag[0] = ulPropTag;
	lpPropTagArr->aulPropTag[1] = ulTagPrivate;
	lpPropTagArr->aulPropTag[2] = PR_ACCOUNT;

	iter = sDavProp.lstProps.begin();
	for(int i = 3; iter != sDavProp.lstProps.end(); iter++, i++)
	{
		sDavProperty = *iter;
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)sDavProperty.sPropName.strPropname.c_str());
	}

	hr = lpTable->SetColumns((LPSPropTagArray)lpPropTagArr, 0);
	if (hr != hrSuccess)
		goto exit;
	
	// restrict table
	hr = lpTable->Restrict(lpsRoot, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->GetRowCount(0, &ulObjType);
	if (hr != hrSuccess)
		goto exit;

	HrSetDavPropName(&(sWebResponse.sPropName), "response", CALDAVNSDEF);
	HrSetDavPropName(&sWebMStatus->sPropName, "multistatus", CALDAVNSDEF);
	
	// set rows into Dav structures
	while (1) {
		hr = lpTable->QueryRows(50, 0, &lpValRows); // TODO: what if we get multiple items ?
		if (hr != hrSuccess || lpValRows->cRows == 0)
			break;

		for(ULONG i = 0; i < lpValRows->cRows; i++)
		{
			WEBDAVVALUE sWebDavVal;
			std::string strReqPath;

			m_lpRequest->HrGetReqPath(&strReqPath);
		
			lpsPropVal = PpropFindProp( lpValRows->aRow[i].lpProps, lpValRows->aRow[i].cValues, PR_ACCOUNT);
			if (lpsPropVal)
				strReqPath = strReqPath + lpsPropVal->Value.lpszA;
			HrSetDavPropName(&(sWebResponse.sHRef.sPropName), "href", CALDAVNSDEF);
			sWebResponse.sHRef.strValue = strReqPath;
		

			if (hr == hrSuccess)
				hr = HrMapValtoStruct(lpValRows->aRow[i].lpProps, lpValRows->aRow[i].cValues, NULL, 0, &sDavProp.lstProps, &sWebResponse);

			sWebMStatus->lstResp.push_back(sWebResponse);
			sWebResponse.lstsPropStat.clear();

		}
		if (lpValRows)
		{
			FreeProws(lpValRows);
			lpValRows = NULL;
		}
	}

	hr = hrSuccess;

exit:
	
	if (lpMtIcal)
		delete lpMtIcal;

	if (lpsRoot)
		FREE_RESTRICTION(lpsRoot);

	if (lpTable)
		lpTable->Release();

	if (lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	return hr;
}


/**
 * Handles deletion of folder & message
 * @return	HRESULT
 * @retval	MAPI_E_NO_ACCESS	Insufficient permissions on folder or request to delete default folder.
 * @retval	MAPI_E_NOT_FOUND	Message or folder to be deleted not found.
 *
 * @todo	Move Code To HrHandleDelete().
 */
HRESULT CalDAV::HrDelete()
{
	HRESULT hr = hrSuccess;
	std::string strGuid;
	std::string strPath;
	std::string strFldPath;
	std::string strIfMatch;
	size_t ulFound = 0;
	SBinary sbEid = {0,0};
	ULONG ulObjType = 0;
	ULONG cValues = 0;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPIFolder *lpRootFld = NULL;
	IMAPIFolder *lpWastBoxFld = NULL;
	LPSPropValue lpProp = NULL;
	LPSPropValue lpPropMsgVal = NULL;
	IMessage *lpMessage = NULL;
	LPENTRYLIST lpEntryList= NULL;
	bool isFolder = false;
	bool blCensorPrivate = false;
	bool blModified = false;
	SizedSPropTagArray(2, lpPropTagArr) = {2, {PR_ENTRYID, PR_LAST_MODIFICATION_TIME}};

	m_lpRequest->HrGetFldPath(&strPath);
	m_lpRequest->HrGetIfMatch(&strIfMatch);

	// @todo it's not really the HTTP class it's place to have such information
	isFolder = m_lpRequest->IsFolderRequest();

	lpRootFld = m_lpRootFld;
	lpUsrFld = m_lpUsrFld;

	if (!m_blFolderAccess && isFolder)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;
	
	hr = HrGetOneProp(m_lpDefStore, PR_IPM_WASTEBASKET_ENTRYID, &lpProp);
	if(hr != hrSuccess)
		goto exit;
	
	hr = m_lpDefStore->OpenEntry(lpProp->Value.bin.cb, (LPENTRYID)lpProp->Value.bin.lpb, NULL, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpWastBoxFld);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening \"Deleted items\" folder, error code : 0x%08X", hr);
		goto exit;
	}

	MAPIFreeBuffer(lpProp);
	lpProp= NULL;

	if (isFolder)
	{
		hr = HrGetOneProp(lpUsrFld, PR_ENTRYID, &lpProp);
		if(hr != hrSuccess)
			goto exit;

		sbEid = lpProp[0].Value.bin;
	}
	else
	{
		//remove trailing chars '.ics' from guid value.
		const std::string::const_iterator end = strPath.end() - 4;

		ulFound = strPath.rfind('/');
		if (ulFound != string::npos)
			ulFound++;	// Skip the '/'

		strGuid.reserve(strPath.size() - ulFound - 4);

		// The guid should contain only ascii data.
		std::string::const_iterator i;
		for (i = strPath.begin() + ulFound; i != end; ++i) {
			ASSERT((*i & ~0x7f) == 0);
			strGuid.append(1, (char)(*i & 0x7f));
		}

		if(ulFound == 0)
			strFldPath.clear(); //default folder
		else
			strFldPath = strPath.substr(0, ulFound - 1);
	
		//Find Entry With Particular Guid
		hr = HrFindAndGetMessage(strGuid, lpUsrFld,  m_lpNamedProps, &lpMessage);
		if (hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Entry to be deleted not found : 0x%08X", hr);
			goto exit;
		}

		if (blCensorPrivate && IsPrivate(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN)))
			blCensorPrivate = true;
		else
			blCensorPrivate = false;

		if (blCensorPrivate)
		{
			hr = MAPI_E_NO_ACCESS;
			goto exit;
		}

		hr = lpMessage->GetProps((LPSPropTagArray)&lpPropTagArr, 0, &cValues, &lpPropMsgVal);
		if (hr != hrSuccess && cValues != 2)
			goto exit;

		if (!strIfMatch.empty() && strIfMatch != SPropValToString(&lpPropMsgVal[1]))
		{
			blModified = true;
			goto exit;
		}
		sbEid = lpPropMsgVal[0].Value.bin;
	}
	
	//Create Entrylist
	hr = MAPIAllocateBuffer(sizeof(ENTRYLIST), (void**)&lpEntryList);
	if (hr != hrSuccess)
		goto exit;

	lpEntryList->cValues = 1;

	hr = MAPIAllocateMore(sizeof(SBinary), lpEntryList, (void**)&lpEntryList->lpbin);
	if (hr != hrSuccess)
		goto exit;
	
	hr = MAPIAllocateMore(sbEid.cb, lpEntryList, (void**)&lpEntryList->lpbin[0].lpb);
	if (hr != hrSuccess)
		goto exit;

	lpEntryList->lpbin[0].cb = sbEid.cb;
	memcpy((void*)lpEntryList->lpbin[0].lpb, (const void*)sbEid.lpb, sbEid.cb);

	if (isFolder)
	{
		hr = lpRootFld->CopyFolder(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, lpWastBoxFld, NULL, 0, NULL, MAPI_MOVE);
		if (hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Deleting Folder error code : 0x%08X", hr);
			goto exit;
		}
	}
	else
	{
		hr = lpUsrFld->CopyMessages(lpEntryList, NULL, lpWastBoxFld, 0, NULL, MAPI_MOVE);
		if (hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Deleting Entry");
			goto exit;
		}
	}
	
	// publish freebusy for default folder
	if (!isFolder && m_ulFolderFlag & DEFAULT_FOLDER)
		hr = HrPublishFreeBusy(m_lpSession, m_lpDefStore, time(NULL), FB_PUBLISH_DURATION, m_lpLogger);

	if ( hr != hrSuccess){
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Publishing Freebusy, error code : 0x%08X", hr);
		hr = hrSuccess;
	}

exit:
	if ( hr == MAPI_E_NO_ACCESS )
	{
		m_lpRequest->HrResponseHeader(403, "Forbidden");
		m_lpRequest->HrResponseBody("This folder cannot be deleted");
	}
	else if (blModified)
	{
		m_lpRequest->HrResponseHeader(412, "Precondition Failed");
	}
	else if (hr != hrSuccess)
	{
		m_lpRequest->HrResponseHeader(404, "Not Found");
		m_lpRequest->HrResponseBody("Item to be Deleted not found");
	}
	else
		m_lpRequest->HrResponseHeader(204, "No Content");
	
	if (lpPropMsgVal)
		MAPIFreeBuffer(lpPropMsgVal);

	if (lpWastBoxFld)
		lpWastBoxFld->Release();

	if (lpMessage)
		lpMessage->Release();

	if (lpProp)
		MAPIFreeBuffer(lpProp);
		
	if (lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	return hr;
}

/**
 * Handles adding new entries and modify existing entries in folder
 *
 * @return	HRESULT
 * @retval	MAPI_E_INVALID_PARAMETER	invalid folder specified in URL
 * @retval	MAPI_E_NO_ACCESS			insufficient permissions on folder or message
 * @retval	MAPI_E_INVALID_OBJECT		no message in ical data.
 */
HRESULT CalDAV::HrPut()
{
	HRESULT hr = hrSuccess;
	std::string strGuid;
	std::string strPath;
	std::string strFldPath;
	std::string strIcal;
	std::string strIcalMod;
	std::string strCharSet;
	std::string strModTime;
	std::string strIfMatch;
	LPSPropValue lpPropOldModtime = NULL;
	LPSPropValue lpPropNewModtime = NULL;
	LPSPropValue lpsPropVal = NULL;
	eIcalType etype = VEVENT;
	SBinary sbUid;
	time_t ttLastModTime = 0;
	size_t ulFound = 0;
	ULONG ulItemCount = 0;
	ULONG ulPropTag = 0;
	IMAPIFolder *lpUsrFld = NULL;
	IMessage *lpMessage = NULL;
	ICalToMapi *lpICalToMapi = NULL;
	bool blCensorPrivate = false;
	bool blNewEntry = false;
	bool blModified = false;
	char lpCharQuoted[50] = "";

	m_lpRequest->HrGetFldPath(&strPath);
	m_lpRequest->HrGetIfMatch(&strIfMatch);
	ulFound = strPath.rfind("/") + 1;

	lpUsrFld = m_lpUsrFld;
	
	if(!lpUsrFld)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving calendar folders in PUT request, error code : 0x%08X", hr);
		goto exit;
	}
	
	if ((m_ulFolderFlag & SHARED_FOLDER) && !HasDelegatePerm(m_lpDefStore, m_lpActiveStore))
		blCensorPrivate = true;

	strGuid = StripGuid(strPath);
	if(ulFound == 0)
		strFldPath = ""; //default Folder
	else
		strFldPath = strPath.substr(0, ulFound - 1);
	
	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_APPTTSREF], PT_STRING8);	

	//Find the Entry with particular Guid
	hr = HrFindAndGetMessage(strGuid, lpUsrFld, m_lpNamedProps, &lpMessage);
	if(hr != hrSuccess)
	{		
		//new entry.
		hr = lpUsrFld->CreateMessage(NULL, 0, &lpMessage);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating new message, error code : 0x%08X", hr);
			goto exit;
		}

		if(m_blIsCLEVO)
		{
			hr = HrAddProperty(lpMessage, ulPropTag, false, &strGuid);
			if(hr != hrSuccess)
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error adding property to new message, error code : 0x%08X", hr);
				goto exit;
			}
		}
		blCensorPrivate = false; // all add of new private items.
		blNewEntry = true;
	}
	else
	{
		if (blCensorPrivate && IsPrivate(lpMessage, CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_PRIVATE], PT_BOOLEAN)))
			blCensorPrivate = true;
		else
			blCensorPrivate = false;
	}
	
	if (blCensorPrivate)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	//save Ical data to mapi.
	CreateICalToMapi(m_lpDefStore, m_lpAddrBook, false, &lpICalToMapi);
	
	m_lpRequest->HrGetBody(&strIcal);

	hr = lpICalToMapi->ParseICal(strIcal, m_strCharset, m_strSrvTz, m_lpImailUser, 0);
	if(hr!=hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Parsing ical data in PUT request, error code : 0x%08X", hr);
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Error Parsing ical data : %s", strIcal.c_str());
		goto exit;
	}

	ulItemCount = lpICalToMapi->GetItemCount();
	
	if(ulItemCount != 1 )
	{
		hr = MAPI_E_INVALID_OBJECT;
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "No message in ical data in PUT request, error code : 0x%08X", hr);
		goto exit;
	}
	
	hr = HrGetOneProp (lpUsrFld, PR_CONTAINER_CLASS, &lpsPropVal);
	if (hr != hrSuccess)
		goto exit;

	hr = lpICalToMapi->GetItemInfo(0, &etype, &ttLastModTime, &sbUid);
	if(hr != hrSuccess)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	if ((etype == VEVENT && strncmp(lpsPropVal->Value.lpszA, "IPF.Appointment", strlen("IPF.Appointment")))
		|| (etype == VTODO && strncmp(lpsPropVal->Value.lpszA, "IPF.Task", strlen("IPF.Task"))))
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	// check if entry is modified on server
	if (!strIfMatch.empty() && !blNewEntry)
	{
		hr = HrGetOneProp(lpMessage, PR_LAST_MODIFICATION_TIME, &lpPropOldModtime);
		if (hr == hrSuccess && strIfMatch != SPropValToString(lpPropOldModtime))
		{
			blModified = true;
			goto exit;
		}
	}

	hr = lpICalToMapi->GetItem(0, 0, lpMessage);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error converting ical data to Mapi message in PUT request, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SaveChanges(0);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error saving Mapi message in PUT request, error code : 0x%08X", hr);
		goto exit;
	}

	// new modification time
	hr = HrGetOneProp(lpMessage, PR_LAST_MODIFICATION_TIME, &lpPropNewModtime);
	if (hr == hrSuccess)
		strModTime =  SPropValToString(lpPropNewModtime);
	
	if (strModTime.empty())
		goto exit;

	sprintf((char*)&lpCharQuoted, "\"%s\"", strModTime.c_str());
	strModTime = lpCharQuoted;	

	// Publish freebusy only for default Calendar
	if(m_ulFolderFlag & DEFAULT_FOLDER)
		hr = HrPublishFreeBusy(m_lpSession, m_lpDefStore, time(NULL), FB_PUBLISH_DURATION, m_lpLogger);

	if ( hr != hrSuccess){
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error Publishing Freebusy, error code : 0x%08X", hr);
		hr = hrSuccess;
	}
	
exit:
	if (hr == hrSuccess && blNewEntry)
		m_lpRequest->HrResponseHeader(201, "Created");
	else if (hr == hrSuccess && blModified && (m_blIsCLMAC || m_blIsCLEVO)) // entry is modifid on server (mac ical & evolution)
		m_lpRequest->HrResponseHeader(409, "Conflict");
	else if (hr == hrSuccess && blModified)
		m_lpRequest->HrResponseHeader(412, "Precondition Failed"); // entry is modifid on server (sunbird & TB)
	else if (hr == hrSuccess)
		m_lpRequest->HrResponseHeader(204, "No Content");
	else if( hr == MAPI_E_NOT_FOUND)
		m_lpRequest->HrResponseHeader(404, "Not Found");
	else if(hr == MAPI_E_NO_ACCESS)
		m_lpRequest->HrResponseHeader(403, "Forbidden");
	else
		m_lpRequest->HrResponseHeader(400, "Bad Request");

	if (hr == hrSuccess && !strModTime.empty())
		m_lpRequest->HrResponseHeader("Etag", strModTime);
	
	if (lpsPropVal)
		MAPIFreeBuffer(lpsPropVal);

	if(lpPropNewModtime)
		MAPIFreeBuffer(lpPropNewModtime);

	if(lpPropOldModtime)
		MAPIFreeBuffer(lpPropOldModtime);
	
	if(lpICalToMapi)
		delete lpICalToMapi;

	if(lpMessage)
		lpMessage->Release();

	return hr;
}

/**
 * Creates a new guid in the message and returns it
 * 
 * @param[in]	sbEid		EntryID of the message
 * @param[in]	ulPropTag	Property tag of the Guid property
 * @param[out]	lpstrGuid	The newly created guid is returned
 *
 * @retrun		HRESULT 
 */
HRESULT CalDAV::CreateAndGetGuid(SBinary sbEid, ULONG ulPropTag, string *lpstrGuid)
{
	HRESULT hr = hrSuccess;
	string strGuid;
	LPMESSAGE lpMessage = NULL;
	ULONG ulObjType = 0;
	LPSPropValue lpProp = NULL;

	hr = m_lpActiveStore->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpMessage);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening message to add Guid, error code : 0x%08X", hr);
		goto exit;
	}

	hr = HrCreateGlobalID(ulPropTag, NULL, &lpProp);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating Guid, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SetProps(1, lpProp, NULL);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while adding Guid to message, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMessage->SaveChanges(0);
	if (hr != hrSuccess)
		goto exit;

	*lpstrGuid = bin2hex(lpProp->Value.bin.cb, lpProp->Value.bin.lpb);

exit:
	if (lpMessage)
		lpMessage->Release();

	if (lpProp)
		MAPIFreeBuffer(lpProp);

	return hr;
}
/**
 * Processes the MKCALENDAR request
 * @return	HRESULT
 * @retval		MAPI_E_NO_ACCESS	Unable to create folder, while accessing single calendar
 * @retval		MAPI_E_COLLISION	Folder with same name already exists
 */
HRESULT CalDAV::HrMkCalendar()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrMkCalendar();
	
	return hr;
}

/**
 * Creates new calendar folder
 * 
 * @param[in]	lpsDavProp		Pointer to WEBDAVPROP structure, contains properite to be set on new folder
 * @return		HRESULT
 * @retval		MAPI_E_NO_ACCESS	Unable to create folder, while accessing single calendar
 * @retval		MAPI_E_COLLISION	Folder with same name already exists
 */
HRESULT CalDAV::HrHandleMkCal(WEBDAVPROP *lpsDavProp)
{
	HRESULT hr = hrSuccess;
	std::string strFldName;
	std::string strFldComment;
	std::string strUser;
	std::string strFldOwner;
	IMAPIFolder *lpUsrFld = NULL;
	IMAPIFolder *lpRootFld = NULL;
	LPSPropValue lpsPropValSet = NULL;
	size_t ulFound = 0;
	ULONG ulPropTag = 0;

	ulFound = m_strFldName.find("/");

	// block mkcalendar when accessing single folders.
	// single folder mkcalendar folder-name : temp/NEW-FOLDER-ID
	if(!m_lpRootFld || ulFound != string::npos)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}
	else
		lpRootFld = m_lpRootFld;

	strFldName = U2W(lpsDavProp->lstProps.front().strValue);
	hr = HrOpenUserFld(m_lpSession, m_lpActiveStore, strFldName, false, m_lpRequest->IsPublic(), m_lpLogger, &lpUsrFld);
	if(hr == hrSuccess)
	{
		hr = MAPI_E_COLLISION;
		goto exit;
	}

	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);
	
	hr = lpRootFld->CreateFolder(FOLDER_GENERIC, (LPSTR)strFldName.c_str(), (LPSTR)strFldComment.c_str(), NULL, 0, &lpUsrFld);
	if( hr != hrSuccess)
		goto exit;
	
	MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpsPropValSet);
	lpsPropValSet->ulPropTag = PR_CONTAINER_CLASS;
	lpsPropValSet->Value.lpszA = "IPF.Appointment";

	hr = lpUsrFld->SetProps(1, lpsPropValSet, NULL);
	if(hr != hrSuccess)
		goto exit;
	
	m_lpRequest->HrGetFldPath(&strFldName);

	ulPropTag = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_FLDID], PT_STRING8);
	hr = HrAddProperty(lpUsrFld, ulPropTag, true, &strFldName);
	if(hr != hrSuccess)
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot Add named property, error code : 0x%08X", hr);

exit:
	if(lpsPropValSet)
		MAPIFreeBuffer(lpsPropValSet);

	if(lpUsrFld)
		lpUsrFld->Release();
	return hr;
}

/**
 * Retrieves the list of calendar and sets it in WEBDAVMULTISTATUS structure
 *
 * @param[in]	lpsDavProp		Pointer to WEBDAVPROP structure , gives info about requested properties of folders
 * @param[out]	lpsMulStatus	Pointer to WEBDAVMULTISTATUS structure, the calendar list and its properites are set in this structure
 *
 * @return		HRESULT
 * @retval		MAPI_E_BAD_VALUE	Method called by a non mac client
 */
HRESULT CalDAV::HrListCalendar(WEBDAVPROP *lpsDavProp, WEBDAVMULTISTATUS *lpsMulStatus)
{
	HRESULT hr = hrSuccess;	
	IMAPITable *lpHichyTable = NULL;
	IMAPITable *lpDelHichyTable = NULL;
	IMAPIFolder *lpWasteBox = NULL;
	IMAPIFolder *lpSingleFld = NULL;
	SRestriction *lpRestrict = NULL;
	LPSPropValue lpSpropWbEID = NULL;
	LPSPropValue lpsPropSingleFld = NULL;
	LPSPropTagArray lpPropTagArr = NULL;
	LPSRowSet lpRowsALL = NULL;
	LPSRowSet lpRowsDeleted = NULL;
	size_t cbsize = 0;
	ULONG ulPropTagFldId = 0;
	ULONG ulObjType = 0;
	ULONG ulCmp = 0;
	ULONG ulDelEntries = 0;
	std::list<WEBDAVPROPERTY>::iterator iter;
	WEBDAVRESPONSE sDavResponse;
	std::string strUser;
	std::string strFldOwner;
	bool blSingleFld = false;

	m_lpRequest->HrGetUser(&strUser);
	m_lpRequest->HrGetFldOwner(&strFldOwner);

	if(!m_blIsCLMAC)
	{
		hr = MAPI_E_BAD_VALUE;
		goto exit;
	}

	if(m_ulFolderFlag & DEFAULT_FOLDER || m_ulFolderFlag & SINGLE_FOLDER)
	{
		blSingleFld = true;
		if(m_lpUsrFld)
			lpSingleFld = m_lpUsrFld;
		else
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error User's Folder not found, while accessing single folder");
			goto exit;
		}
	}

	if(!blSingleFld )
	{		
		hr = HrGetSubCalendars(m_lpSession, m_lpRootFld, NULL, &lpHichyTable);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for IPM_Subtree, error code : 0x%08X", hr);
			goto exit;
		}

		if(m_lpRequest->IsPublic())
			hr = HrGetOneProp(m_lpDefStore, PR_IPM_WASTEBASKET_ENTRYID, &lpSpropWbEID);
		else
			hr = HrGetOneProp(m_lpActiveStore, PR_IPM_WASTEBASKET_ENTRYID, &lpSpropWbEID);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for IPM_Subtree, error code : 0x%08X", hr);
			goto exit;
		}
		
		if(m_lpRequest->IsPublic())
			hr = m_lpDefStore->OpenEntry(lpSpropWbEID->Value.bin.cb, (LPENTRYID)lpSpropWbEID->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpWasteBox);
		else
			hr = m_lpActiveStore->OpenEntry(lpSpropWbEID->Value.bin.cb, (LPENTRYID)lpSpropWbEID->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpWasteBox);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening deleted items folder, error code : 0x%08X", hr);
			goto exit;
		}
		
		hr = HrGetSubCalendars(m_lpSession, lpWasteBox, NULL, &lpDelHichyTable);
		if(hr != hrSuccess)
		{
			m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error retrieving subcalendars for Deleted items folder, error code : 0x%08X", hr);
			goto exit;
		}
	}

	cbsize = lpsDavProp->lstProps.size() + 2;

	hr = MAPIAllocateBuffer(CbNewSPropTagArray((ULONG)cbsize), (void **)&lpPropTagArr);
	if(hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Cannot allocate memory");
		goto exit;
	}

	ulPropTagFldId = CHANGE_PROP_TYPE(m_lpNamedProps->aulPropTag[PROP_FLDID], PT_STRING8);
	//add PR_ENTRYID & FolderID in setcolumns along with requested data.
	lpPropTagArr->cValues = (ULONG)cbsize;
	lpPropTagArr->aulPropTag[0] = PR_ENTRYID;
	lpPropTagArr->aulPropTag[1] = ulPropTagFldId;

	iter = lpsDavProp->lstProps.begin();
	for(int i = 2; iter != lpsDavProp->lstProps.end(); iter++, i++)	
	{
		lpPropTagArr->aulPropTag[i] = GetPropmap((char*)iter->sPropName.strPropname.c_str());
	}

	if(!blSingleFld)
	{
		hr = lpHichyTable->SetColumns(lpPropTagArr, 0);
		if(hr != hrSuccess)
			goto exit;

		hr = lpDelHichyTable->SetColumns(lpPropTagArr, 0);
		if(hr != hrSuccess)
			goto exit;
	}
	else
	{
		hr = lpSingleFld->GetProps(lpPropTagArr, 0, (ULONG *)&cbsize, &lpsPropSingleFld);
	}

	if (FAILED(hr))
		goto exit;

	hr = hrSuccess;

	HrSetDavPropName(&(lpsMulStatus->sPropName), "multistatus", lpsDavProp->sPropName.strNS);
	while(1)
	{
		ULONG ulEntryCount = 0;
		if(!blSingleFld)
		{
			hr = lpHichyTable->QueryRows(50, 0, &lpRowsALL);
			if(hr != hrSuccess || lpRowsALL->cRows == 0)
				break;
		
			hr = lpDelHichyTable->QueryRows(50, 0, &lpRowsDeleted);
			if(hr != hrSuccess)
				break;
		}
		if(blSingleFld)
			ulEntryCount = 1;
		else
			ulEntryCount = lpRowsALL->cRows;

		for(ULONG i = 0; i < ulEntryCount; i++)
		{
			std::string strFldPath;
			std::string strPath;
			std::string strTrail;
			WEBDAVPROPSTAT sWebPropStat;
			WEBDAVPROP sPropNF;
			WEBDAVPROP sPropOK;
			
			if(!blSingleFld && lpRowsDeleted->cRows != 0 && ulDelEntries != lpRowsDeleted->cRows )
			{
				ulCmp = memcmp((const void *)lpRowsALL->aRow[i].lpProps[0].Value.bin.lpb,
							   (const void *)lpRowsDeleted->aRow[ulDelEntries].lpProps[0].Value.bin.lpb, lpRowsALL->aRow[i].lpProps[0].Value.bin.cb);
				if(ulCmp == 0)
				{
					ulDelEntries++;
					continue;
				}
			}

			HrSetDavPropName(&(sDavResponse.sPropName), "response", lpsDavProp->sPropName.strNS);
			
			m_lpRequest->HrGetReqPath(&strPath);
			strTrail = strPath.at(strPath.length() -1);

			if(blSingleFld && lpsPropSingleFld && lpsPropSingleFld[1].ulPropTag == ulPropTagFldId)
				strFldPath = lpsPropSingleFld[1].Value.lpszA;
			else if (lpRowsALL && lpRowsALL && lpRowsALL->aRow[i].lpProps[1].ulPropTag == ulPropTagFldId)
				strFldPath = lpRowsALL->aRow[i].lpProps[1].Value.lpszA;
			else
				strFldPath.clear();
			
			if(blSingleFld && strFldPath.empty() && lpsPropSingleFld && lpsPropSingleFld[0].ulPropTag == PR_ENTRYID )
				hr = HrAddProperty(m_lpActiveStore, lpsPropSingleFld[0].Value.bin, ulPropTagFldId, true, &strFldPath);
			else if (strFldPath.empty() && lpRowsALL && lpRowsALL->aRow[i].lpProps[0].ulPropTag == PR_ENTRYID )
				hr = HrAddProperty(m_lpActiveStore, lpRowsALL->aRow[i].lpProps[0].Value.bin, ulPropTagFldId, true, &strFldPath);

			if( hr != hrSuccess || strFldPath.empty())
			{
				m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error adding Folder id property, error code : 0x%08X", hr);
				continue;
			}
			if(strFldOwner.empty() && !m_lpRequest->IsPublic())
				strFldPath = W2U(strUser) + "/" + FOLDER_PREFIX + strFldPath + "/";
			else
				strFldPath = FOLDER_PREFIX + strFldPath+ "/";

			strPath = strPath + strFldPath;

			HrSetDavPropName(&(sDavResponse.sHRef.sPropName), "href", lpsDavProp->sPropName.strNS);
			sDavResponse.sHRef.strValue = strPath;

			iter = lpsDavProp->lstProps.begin();
			
			if (blSingleFld)
				HrMapValtoStruct(lpsPropSingleFld, cbsize, NULL, 0, &lpsDavProp->lstProps, &sDavResponse);
			else
				HrMapValtoStruct(lpRowsALL->aRow[i].lpProps, lpRowsALL->aRow[i].cValues, NULL, 0, &lpsDavProp->lstProps, &sDavResponse);

			lpsMulStatus->lstResp.push_back(sDavResponse);
			sDavResponse.lstsPropStat.clear();
		}

		if(blSingleFld)
			break;

		if(lpRowsALL)
		{
			FreeProws(lpRowsALL);
			lpRowsALL = NULL;
		}

		if(lpRowsDeleted)
		{
			FreeProws(lpRowsDeleted);
			lpRowsDeleted = NULL;
		}
	}

exit:
	if(lpWasteBox)
		lpWasteBox->Release();

	if(lpHichyTable)
		lpHichyTable->Release();
	
	if(lpDelHichyTable)
		lpDelHichyTable->Release();

	if(lpsPropSingleFld)
		MAPIFreeBuffer(lpsPropSingleFld);

	if(lpSpropWbEID)
		MAPIFreeBuffer(lpSpropWbEID);

	if(lpPropTagArr)
		MAPIFreeBuffer(lpPropTagArr);

	if(lpRowsALL)
		FreeProws(lpRowsALL);
	
	if(lpRowsDeleted)
		FreeProws(lpRowsDeleted);

	if(lpRestrict)
		FREE_RESTRICTION(lpRestrict);

	return hr;
}

/**
 * Processes the PROPPATCH request
 *
 * This method is usually used to rename folder or edit comments on folder.
 *
 * @return	HRESULT
 */
HRESULT CalDAV::HrPropPatch()
{
	HRESULT hr = hrSuccess;

	hr = WebDav::HrPropPatch();

	return hr;
}

/**
 * Sets the values on mapi object for PROPPATCH request
 *
 * @param[in]	lpsDavProp	WEBDAVPROP structure containing properties to be modified and its values
 *
 * @return		HRESULT
 * @retval		MAPI_E_NO_ACCESS	Error returned while renaming the default calendar
 */
HRESULT CalDAV::HrHandlePropPatch(WEBDAVPROP *lpsDavProp)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpsPropVal = NULL;
	std::list<WEBDAVPROPERTY>::iterator iter;
	std::string strFldPath;
	std::string strFldName;
	std::string strConvProp;
	IMAPIFolder *lpUsrFld = NULL;
	ULONG ulCntr = 0;
	
	m_lpRequest->HrGetFldPath(&strFldPath);
	
	hr = MAPIAllocateBuffer( (ULONG)lpsDavProp->lstProps.size() * sizeof(SPropValue), (LPVOID*)&lpsPropVal);
	if (hr != hrSuccess)
		goto exit;

	for(iter = lpsDavProp->lstProps.begin(); iter != lpsDavProp->lstProps.end(); iter++)
	{
		if (iter->sPropName.strPropname == "displayname" )
			strFldName = iter->strValue;
		else if (iter->sPropName.strPropname == "calendar-free-busy-set")
			goto exit;
		lpsPropVal[ulCntr].ulPropTag = GetPropmap((char*)iter->sPropName.strPropname.c_str());
		
		if (lpsPropVal[ulCntr].ulPropTag == 0) // GetPropMap may not find Property Tag.
			continue;

		strConvProp = U2W(iter->strValue);
		hr = MAPIAllocateMore(strConvProp.length() + 1, lpsPropVal, (void**)&lpsPropVal[ulCntr].Value.lpszA);
		if (hr != hrSuccess)
			goto exit;
		strncpy(lpsPropVal[ulCntr].Value.lpszA ,(char*)strConvProp.c_str(),strConvProp.length() + 1);

		ulCntr++;
	}
	
	// deny rename of defalt Calendar
	if (!strFldName.empty() && !m_blFolderAccess)
	{
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	if (ulCntr != 0 && m_lpUsrFld)
		hr = m_lpUsrFld->SetProps(ulCntr, lpsPropVal, NULL);

exit:
	if(lpsPropVal)
		MAPIFreeBuffer(lpsPropVal);

	if(lpUsrFld)
		lpUsrFld->Release();

	return hr;
}
/**
 * Processes the POST request from caldav client
 *
 * POST is used to request freebusy info of attendees or send
 * meeting request to attendees(used by mac ical.app only)
 *
 * @return	HRESULT
 */
HRESULT CalDAV::HrPost ()
{
	HRESULT hr = hrSuccess;
	std::string strIcal;
	ICalToMapi *lpIcalToMapi = NULL;

	hr = m_lpRequest->HrGetBody(&strIcal);
	if (hr != hrSuccess)
		goto exit;
	
	CreateICalToMapi(m_lpDefStore, m_lpAddrBook, false, &lpIcalToMapi);
	
	if (!lpIcalToMapi)
	{
		hr = MAPI_E_BAD_VALUE;
		goto exit;
	}

	hr = lpIcalToMapi->ParseICal(strIcal, m_strCharset, m_strSrvTz, m_lpImailUser, 0);
	if (hr != hrSuccess)
		goto exit;
	
	if (lpIcalToMapi->GetItemCount() == 1)
		hr = HrHandleMeeting(lpIcalToMapi);
	else
		hr = HrHandleFreebusy(lpIcalToMapi);

exit:
	
	if (lpIcalToMapi)
		delete lpIcalToMapi;

	return hr;
}

/**
 * Handles the caldav clients's request to view freebusy information
 * of attendees
 *
 * @param[in]	lpIcalToMapi	The ical to mapi conversion object
 * @return		HRESULT
 */
HRESULT CalDAV::HrHandleFreebusy(ICalToMapi *lpIcalToMapi)
{
	HRESULT hr = hrSuccess;	
	ECFreeBusySupport*	lpecFBSupport = NULL;	
	IFreeBusySupport* lpFBSupport = NULL;
	IABContainer *lpAbCont = NULL;
	IMAPITable *lpGabTable = NULL;
	MapiToICal *lpMapiToIcal = NULL;
	time_t tStart  = 0;
	time_t tEnd = 0;
	std::list<std::string> *lstUsers = NULL;
	std::string strIcal;
	std::string strUID;
	std::string strOrganiser;
	SBinary sbEid;
	WEBDAVFBINFO sWebFbInfo;
	ULONG ulObjType = 0;

	hr = lpIcalToMapi->GetItemInfo(&tStart ,&tEnd, &strUID, &lstUsers);	
	if (hr != hrSuccess)
		goto exit;

	CreateMapiToICal(m_lpAddrBook, "utf-8", &lpMapiToIcal);
	
	hr = m_lpAddrBook->GetDefaultDir(&sbEid.cb, (LPENTRYID*) &sbEid.lpb);
	if (hr != hrSuccess)
		goto exit;

	hr = m_lpSession->OpenEntry(sbEid.cb, (LPENTRYID)sbEid.lpb, NULL, 0, &ulObjType, (LPUNKNOWN *)&lpAbCont);
	if (hr != hrSuccess)
		goto exit;
	
	hr = lpAbCont->GetContentsTable(0, &lpGabTable);
	if (hr != hrSuccess)
		goto exit;
	
	hr = ECFreeBusySupport::Create(&lpecFBSupport);
	if (hr != hrSuccess)
		goto exit;

	hr = lpecFBSupport->QueryInterface(IID_IFreeBusySupport, (void**)&lpFBSupport);
		if (hr != hrSuccess)
			goto exit;
	
	hr = lpecFBSupport->Open(m_lpSession, m_lpDefStore,true);
	if (hr != hrSuccess)
		goto exit;
	
	hr = HrGetUserInfo(m_lpSession, m_lpDefStore, &strOrganiser, NULL, NULL);
	if (hr != hrSuccess)
		goto exit;

	sWebFbInfo.strOrganiser = strOrganiser;
	sWebFbInfo.tStart = tStart;
	sWebFbInfo.tEnd = tEnd;
	sWebFbInfo.strUID = strUID;

	hr = HrGetFreebusy(lpMapiToIcal, lpFBSupport , lpGabTable, lstUsers, &sWebFbInfo);
	if (hr != hrSuccess)
		goto exit;
	
	hr = WebDav::HrPost(&sWebFbInfo);
	if (hr != hrSuccess)
		goto exit;

exit:

	if (sbEid.lpb)
		MAPIFreeBuffer(sbEid.lpb);

	if (lpFBSupport)
		lpFBSupport->Release();

	if (lpGabTable)
		lpGabTable->Release();
	
	if (lpAbCont)
		lpAbCont->Release();
	
	if (lpecFBSupport)
		lpecFBSupport->Release();
	
	if (lpMapiToIcal)
		delete lpMapiToIcal;

	return hr;
}

/**
 * Handles Mac ical.app clients request to send a meeting request
 * to the attendee.
 *
 * @param[in]	lpIcalToMapi	ical to mapi conversion object
 * @return		HRESULT
 */
HRESULT CalDAV::HrHandleMeeting(ICalToMapi *lpIcalToMapi)
{
	HRESULT hr = hrSuccess;	
	LPSPropValue lpsGetPropVal = NULL;
	LPMAPIFOLDER lpOutbox = NULL;
	LPMESSAGE lpNewMsg = NULL;
	LPSPropTagArray lpsPropArray = NULL;
	LPSPropValue lpsSetPropVals = NULL;
	ULONG cValues = 0;
	ULONG ulObjType = 0;
	time_t tModTime = 0;
	SBinary sbEid = {0};
	eIcalType etype = VEVENT;	

	hr = lpIcalToMapi->GetItemInfo( 0, &etype, &tModTime, &sbEid);
	if ( hr != hrSuccess || etype != VEVENT)
	{
		hr = hrSuccess; // skip VFREEBUSY
		goto exit;
	}

	cValues = 2;
	hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void **)&lpsPropArray);
	if (hr != hrSuccess)
		goto exit;

	lpsPropArray->cValues = cValues;
	lpsPropArray->aulPropTag[0] = PR_IPM_OUTBOX_ENTRYID;
	lpsPropArray->aulPropTag[1] = PR_IPM_SENTMAIL_ENTRYID;

	hr = m_lpDefStore->GetProps(lpsPropArray, 0, &cValues, &lpsGetPropVal);
	if (hr != hrSuccess && cValues != 2)
		goto exit;

	hr = m_lpDefStore->OpenEntry(lpsGetPropVal[0].Value.bin.cb, (LPENTRYID) lpsGetPropVal[0].Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN*)&lpOutbox);
	if (hr != hrSuccess)
		goto exit;

	hr = lpOutbox->CreateMessage(NULL, MAPI_BEST_ACCESS, &lpNewMsg);
	if (hr != hrSuccess)
		goto exit;

	hr = lpIcalToMapi->GetItem( 0, 0, lpNewMsg);
	if (hr != hrSuccess)
		goto exit;

	cValues = 2;
	hr = MAPIAllocateBuffer(cValues * sizeof(SPropValue), (void**)&lpsSetPropVals);
	if (hr != hrSuccess)
		goto exit;

	lpsSetPropVals[0].ulPropTag = PR_SENTMAIL_ENTRYID;
	lpsSetPropVals[0].Value.bin = lpsGetPropVal[1].Value.bin;

	lpsSetPropVals[1].ulPropTag = PR_DELETE_AFTER_SUBMIT;
	lpsSetPropVals[1].Value.b = false;

	hr = lpNewMsg->SetProps(cValues, lpsSetPropVals, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpNewMsg->SaveChanges(KEEP_OPEN_READWRITE);
	if (hr != hrSuccess)
		goto exit;

	hr = lpNewMsg->SubmitMessage(0);
	if (hr != hrSuccess)
		goto exit;

exit:
	
	if(hr == hrSuccess)
		m_lpRequest->HrResponseHeader(200, "Ok");
	else
		m_lpRequest->HrResponseHeader(400, "Bad Request");

	if (lpNewMsg)
		lpNewMsg->Release();

	if (lpOutbox)
		lpOutbox->Release();

	if (lpsGetPropVal)
		MAPIFreeBuffer(lpsGetPropVal);

	if (lpsSetPropVals)
		MAPIFreeBuffer(lpsSetPropVals);

	return hr;
}

/**
 * converts windows-1252 to utf-8
 * @param[in]	strWinChar	source string(windows-1252) to be converted
 * @return		string		converted string (utf-8)
 */
std::string CalDAV::W2U(std::string strWinChar)
{
	return m_lpConvResponse->convert(strWinChar);
}

/**
 * converts utf-8 to windows-1252
 * @param[in]	strUtfChar	source string(utf-8) to be converted
 * @return		string		converted string (windows-1252)
 */
std::string CalDAV::U2W(std::string strUtfChar)
{
	return m_lpConvRequest->convert(strUtfChar);
}

/**
 * Converts the mapi message specified by EntryID to ical string.
 *
 * @param[in]	lpEid		EntryID of the mapi msg to be converted
 * @param[in]	lpMtIcal	mapi to ical conversion object
 * @param[in]	ulFlags		Flags used for mapi to ical conversion
 * @param[out]	strIcal		ical string output
 * @return		HRESULT
 */
HRESULT CalDAV::HrConvertToIcal(LPSPropValue lpEid, MapiToICal *lpMtIcal, ULONG ulFlags, std::string &strIcal)
{
	HRESULT hr = hrSuccess;
	IMessage *lpMessage = NULL;
	ULONG ulObjType = 0;

	hr = m_lpActiveStore->OpenEntry(lpEid->Value.bin.cb, (LPENTRYID)lpEid->Value.bin.lpb, NULL, MAPI_BEST_ACCESS, &ulObjType, (LPUNKNOWN *)&lpMessage);
	if (hr != hrSuccess && ulObjType == MAPI_MESSAGE)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening calendar entry, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMtIcal->AddMessage(lpMessage, m_strSrvTz, ulFlags);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error converting mapi message to ical, error code : 0x%08X", hr);
		goto exit;
	}

	hr = lpMtIcal->Finalize(0, NULL, &strIcal);
	if (hr != hrSuccess)
	{
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error creating ical data, error code : 0x%08X", hr);
		goto exit;
	}

	lpMtIcal->ResetObject();

exit:
	
	if(lpMessage)
		lpMessage->Release();
	
	return hr;
}

/**
 * Set Values for properties requested by caldav client
 * 
 * @param[in]	lpProps			SpropValue array containing values of requested properties
 * @param[in]	ulPropCount		Count of propety values present in lpProps
 * @param[in]	lpMtIcal		mapi to ical conversion object pointer
 * @param[in]	ulFlags			Flags used during mapi to ical conversion
 * @param[in]	lstDavProps		Pointer to structure containing properties requested by client
 * @param[out]	lpsResponse		Pointer to Response structure
 * @return		HRESULT
 * @retval		hrSuccess		Always returns hrSuccess
 */
HRESULT CalDAV::HrMapValtoStruct(LPSPropValue lpProps, ULONG ulPropCount, MapiToICal *lpMtIcal, ULONG ulFlags, std::list<WEBDAVPROPERTY> *lstDavProps, WEBDAVRESPONSE *lpsResponse)
{
	HRESULT hr = hrSuccess;
	std::list<WEBDAVPROPERTY>::iterator iterProperty;
	WEBDAVPROPERTY sWebProperty;
	std::string strProperty;
	std::string strIcal;
	std::string strFolderPath;
	std::string strPrincipalURL;
	LPSPropValue lpFoundProp = NULL;
	WEBDAVPROP sWebProp;
	WEBDAVPROP sWebPropNotFound;
	WEBDAVPROPSTAT sPropStat;
	ULONG ulDepth = 0;
	ULONG ulFolderType = 0;

	m_lpRequest->HrGetDepth(&ulDepth);
	m_lpRequest->HrGetFldPath(&strFolderPath);
	
	lpFoundProp = PpropFindProp(lpProps, ulPropCount, PR_CONTAINER_CLASS);
	if (lpFoundProp && !strncmp (lpFoundProp->Value.lpszA, "IPF.Appointment", strlen("IPF.Appointment")))
		ulFolderType = CALENDAR_FOLDER;
	else if (lpFoundProp && !strncmp (lpFoundProp->Value.lpszA, "IPF.Task", strlen("IPF.Task")))
		ulFolderType = TASKS_FOLDER;
	else
		ulFolderType = OTHER_FOLDER;

	if (!m_lpRequest->IsPublic() && strFolderPath.empty()) {
		strPrincipalURL = ("/caldav/" + ( m_strFldOwner.empty() 
											? m_strUser
											: m_strFldOwner ) + "/");

	} else {
		strPrincipalURL = lpsResponse->sHRef.strValue;
	}

	HrSetDavPropName(&(sWebProp.sPropName), "prop", CALDAVNSDEF);
	HrSetDavPropName(&(sWebPropNotFound.sPropName), "prop", CALDAVNSDEF);
	for (iterProperty = lstDavProps->begin(); iterProperty != lstDavProps->end() ; iterProperty++)
	{
		WEBDAVVALUE sWebVal;

		sWebProperty.lstItems.clear();
		sWebProperty.lstValues.clear();
		
		sWebProperty = *iterProperty;
		strProperty = sWebProperty.sPropName.strPropname;

		lpFoundProp = PpropFindProp(lpProps, ulPropCount, GetPropmap((char*)strProperty.c_str()));
		if (strProperty == "resourcetype") {
			
			// do not set resourcetype for REPORT request(ical data)
			if(!lpMtIcal){
				HrSetDavPropName(&(sWebVal.sPropName), "collection", CALDAVNSDEF);
				sWebProperty.lstValues.push_back(sWebVal);
			}
			
			if(lpFoundProp && (!strcmp(lpFoundProp->Value.lpszA ,"IPF.Appointment") || !strcmp(lpFoundProp->Value.lpszA , "IPF.Task"))) {
				HrSetDavPropName(&(sWebVal.sPropName), "calendar", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);
			}else if (m_strFldName == "Inbox") {
				HrSetDavPropName(&(sWebVal.sPropName), "schedule-inbox", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);
			}else if(m_strFldName == "Outbox") {
				HrSetDavPropName(&(sWebVal.sPropName), "schedule-outbox", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);
			}

		} else if (strProperty == "displayname") {
			
			if (m_blIsCLMAC && ((ulFlags & PROPFIND_ROOT) == PROPFIND_ROOT))
				sWebProperty.strValue = W2U(m_strUserFullName);
			else
				sWebProperty.strValue = W2U(SPropValToString(lpFoundProp));
			
		} else if (strProperty == "calendar-user-address-set") {
			
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			if ((ulFlags & PROPFIND_ROOT) == PROPFIND_ROOT)
				sWebVal.strValue = "mailto:" + m_strUserEmail;
			else
				sWebVal.strValue = "mailto:" + SPropValToString(lpFoundProp);
			sWebProperty.lstValues.push_back(sWebVal);
			sWebVal.strValue = m_strCalHome;
			if (!m_strCalHome.empty())
				sWebProperty.lstValues.push_back(sWebVal);

		} else if (strProperty == "acl" || strProperty == "current-user-privilege-set") {
			
			HrBuildACL(&sWebProperty);

		} else if (strProperty == "supported-report-set") {
			
			HrBuildReportSet(&sWebProperty);

		} else if (lpFoundProp && 
					(strProperty == "calendar-description" || strProperty == "getctag" || 
					 strProperty == "getetag" || strProperty == "last-name" || 
					 strProperty == "first-name")
					) {
			
			sWebProperty.strValue = SPropValToString(lpFoundProp);

		} else if (strProperty == "email-address-set") {

			HrSetDavPropName(&(sWebVal.sPropName), "email-address", CALDAVNSDEF);
			sWebVal.strValue = ((ulFlags & PROPFIND_ROOT) == PROPFIND_ROOT)? m_strUserEmail : SPropValToString(lpFoundProp);
			sWebProperty.lstValues.push_back(sWebVal);

		} else if (strProperty == "schedule-inbox-URL") {
			
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = ("/caldav/" + W2U(m_strFldOwner.empty()? m_strUser : m_strFldOwner) + "/Inbox");
			sWebProperty.lstValues.push_back(sWebVal);

		} else if (strProperty == "schedule-outbox-URL") {
			
			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = ("/caldav/" + W2U(m_strFldOwner.empty()? m_strUser : m_strFldOwner) + "/Outbox");
			sWebProperty.lstValues.push_back(sWebVal);

		} else if (strProperty == "supported-calendar-component-set") {
			
			if (ulFolderType == CALENDAR_FOLDER) {
				HrSetDavPropName(&(sWebVal.sPropName), "comp","name", "VEVENT", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);

				// actually even only for the standard calender folder
				HrSetDavPropName(&(sWebVal.sPropName), "comp","name", "VFREEBUSY", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);
			}
			else if (ulFolderType == TASKS_FOLDER) {
				HrSetDavPropName(&(sWebVal.sPropName), "comp","name", "VTODO", CALDAVNS);
				sWebProperty.lstValues.push_back(sWebVal);
			}

			HrSetDavPropName(&(sWebVal.sPropName), "comp","name", "VTIMEZONE", CALDAVNS);
			sWebProperty.lstValues.push_back(sWebVal);

		} else if (lpFoundProp && lpMtIcal && strProperty == "calendar-data") {
			
			hr = HrConvertToIcal(lpFoundProp, lpMtIcal, ulFlags, sWebProperty.strValue);
			if (hr != hrSuccess || sWebProperty.strValue.empty()){
				// ical data is empty so discard this calendar entry
				HrSetDavPropName(&(lpsResponse->sStatus.sPropName), "status", CALDAVNSDEF);
				lpsResponse->sStatus.strValue = "HTTP/1.1 404 Not Found";
				goto exit;
			}
			
			strIcal.clear();

		} else if (strProperty == "calendar-order" && ulFolderType == CALENDAR_FOLDER) {
			
			lpFoundProp = PpropFindProp(lpProps, ulPropCount, PR_ENTRYID);
			if (lpFoundProp) {
				if (HrGetCalendarOrder(lpFoundProp->Value.bin, sWebProperty.strValue) != hrSuccess) {
					sWebPropNotFound.lstProps.push_back(sWebProperty);
					continue;
				}
			}
		} else if (strProperty == "getcontenttype") {
			sWebProperty.strValue = "text/calendar";

		} else if (strProperty == "principal-collection-set") {

			sWebProperty.strValue = "/caldav/";

		} else if (strProperty == "current-user-principal" || strProperty == "owner") {

			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = ("/caldav/" + W2U(m_strFldOwner.empty()? m_strUser : m_strFldOwner) + "/");
			sWebProperty.lstValues.push_back(sWebVal);

		}else if (strProperty == "calendar-home-set" || strProperty == "principal-URL") {

			HrSetDavPropName(&(sWebVal.sPropName), "href", CALDAVNSDEF);
			sWebVal.strValue = strPrincipalURL;
			sWebProperty.lstValues.push_back(sWebVal);

		} else if (strProperty == "calendar-user-type") {
			if (SPropValToString(lpFoundProp) == "0")
				sWebProperty.strValue = "INDIVIDUAL";
		} else if (strProperty == "record-type"){

			sWebProperty.strValue = "users";

		} else {
			sWebPropNotFound.lstProps.push_back(sWebProperty);
			continue;
		}

		sWebProp.lstProps.push_back(sWebProperty);
	}
	
	HrSetDavPropName(&(sPropStat.sPropName), "propstat", CALDAVNSDEF);
	HrSetDavPropName(&(sPropStat.sStatus.sPropName), "status", CALDAVNSDEF);

	if( !sWebProp.lstProps.empty()) {
		sPropStat.sStatus.strValue = "HTTP/1.1 200 OK";
		sPropStat.sProp = sWebProp;
		lpsResponse->lstsPropStat.push_back (sPropStat);
	}
	
	if( !sWebPropNotFound.lstProps.empty()) {
		sPropStat.sStatus.strValue = "HTTP/1.1 404 Not Found";
		sPropStat.sProp = sWebPropNotFound;
		lpsResponse->lstsPropStat.push_back (sPropStat);
	}
exit:
	return hr;
}

/** 
 * Checks in the given store if the given folder may be removed, or
 * should be treated as a special folder.
 * 
 * @param[in] lpStore Store of a user or the public
 * @param[in] lpRootFolder Root Folder in lpStore, can be NULL
 * @param[in] lpCurrentFolder Folder in lpStore to check for protection
 * 
 * @todo should we ignore some protected folders because it can be the public store?
 *
 * @return true if the lpCurrentFolder may be deleted, false if not
 */
bool CalDAV::CheckFolderAccess(IMsgStore *lpStore, IMAPIFolder *lpRootFolder, IMAPIFolder *lpCurrentFolder) const
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpFolderEntryID = NULL;
	ULONG ulObjType;
	IMAPIFolder *lpRoot = NULL;

	// folders which cannot be deleted
	SizedSPropTagArray(3, sPropsStore) = { 3,
		{ PR_IPM_OUTBOX_ENTRYID, PR_IPM_SENTMAIL_ENTRYID, PR_IPM_WASTEBASKET_ENTRYID } };
	const char* sNamesStore[] = {"Outbox", "Sentmail", "Wastebasket"};
	SizedSPropTagArray(6, sPropsRoot) = { 6,
		{ PR_IPM_APPOINTMENT_ENTRYID, PR_IPM_CONTACT_ENTRYID, PR_IPM_DRAFTS_ENTRYID,
		  PR_IPM_JOURNAL_ENTRYID, PR_IPM_NOTE_ENTRYID, PR_IPM_TASK_ENTRYID } };
	const char* sNamesRoot[] = {"Calendar", "Contacts", "Drafts", "Journal", "Notes", "Tasks"};

	ULONG cbFolders;
	LPSPropValue lpProtectedFolders1 = NULL;
	LPSPropValue lpProtectedFolders2 = NULL;
	BOOL bFound = FALSE;
	ULONG ulPos = 0;

	if (!lpStore || !lpCurrentFolder) {
		hr = MAPI_E_NO_ACCESS;
		goto exit;
	}

	hr = HrGetOneProp(lpCurrentFolder, PR_ENTRYID, &lpFolderEntryID);
	if (hr != hrSuccess)
		goto exit;

	// Find protected folder from store level
	hr = lpStore->GetProps((LPSPropTagArray)&sPropsStore, 0, &cbFolders, &lpProtectedFolders1);
	if (FAILED(hr) || cbFolders < sPropsStore.cValues) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while retrieving protected folder list");
		goto exit;
	}

	hr = Util::HrFindEntryIDs(lpFolderEntryID->Value.bin.cb, (LPENTRYID)lpFolderEntryID->Value.bin.lpb, cbFolders, lpProtectedFolders1, &bFound, &ulPos);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while testing for protected folders");
		goto exit;
	}
	if (bFound == TRUE) {
		hr = MAPI_E_NO_ACCESS;
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Accessing Default folder named %s", sNamesStore[ulPos]); 
		goto exit;
	}

	// Find protected folder from inbox level
	if (lpRootFolder) {
		hr = lpRootFolder->QueryInterface(IID_IMAPIFolder, (void**)&lpRoot);
		ulObjType = MAPI_FOLDER;
	} else {
		hr = lpStore->OpenEntry(0, NULL, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRoot);
	}
	if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while retrieving root container");
		goto exit;
	}

	hr = lpRoot->GetProps((LPSPropTagArray)&sPropsRoot, 0, &cbFolders, &lpProtectedFolders2);
	if (FAILED(hr) || cbFolders < sPropsRoot.cValues) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while retrieving protected folder list");
		goto exit;
	}

	hr = Util::HrFindEntryIDs(lpFolderEntryID->Value.bin.cb, (LPENTRYID)lpFolderEntryID->Value.bin.lpb, cbFolders, lpProtectedFolders2, &bFound, &ulPos);
	if (hr != hrSuccess) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error while testing for protected folders");
		goto exit;
	}
	if (bFound == TRUE) {
		hr = MAPI_E_NO_ACCESS;
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Accessing Default folder named %s folder", sNamesRoot[ulPos]);
		goto exit;
	}

exit:
	if (lpFolderEntryID)
		MAPIFreeBuffer(lpFolderEntryID);

	if (lpRoot)
		lpRoot->Release();

	if (lpProtectedFolders1)
		MAPIFreeBuffer(lpProtectedFolders1);

	if (lpProtectedFolders2)
		MAPIFreeBuffer(lpProtectedFolders2);

	return hr == hrSuccess;
}

/**
 * Sets the Calendar Order to 1 for default calendar folder
 *
 * The function checks if the folder is the default calendar folder, if true sets
 * the calendar-order to 1
 *
 * The calendar-order property is set to show the default calendar first in 
 * the calendar list in ical.app(Mac). This makes the default zarafa calendar default
 * in ical.app too.
 *
 * @param[in]	sbEid				Entryid of the Folder to be checked
 * @param[out]	wstrCalendarOrder	string output in which the calendar order is set
 * @return		mapi error codes
 * @retval		MAPI_E_CALL_FAILED	the calendar-order is not set for this folder
 */
HRESULT CalDAV::HrGetCalendarOrder(SBinary sbEid, std::string &strCalendarOrder)
{
	HRESULT hr = hrSuccess;
	LPMAPIFOLDER lpRootCont = NULL;
	LPSPropValue lpProp = NULL;
	ULONG ulObjType = 0;
	ULONG ulResult = 0;
	
	hr = m_lpActiveStore->OpenEntry(0, NULL, NULL, 0, &ulObjType, (LPUNKNOWN*)&lpRootCont);
	if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
		m_lpLogger->Log(EC_LOGLEVEL_ERROR, "Error opening root Container of user %s, error code: (0x%08X)", m_strUser.c_str(), hr);
		goto exit;
	}

	// get default calendar folder entry id from root container
	hr = HrGetOneProp(lpRootCont, PR_IPM_APPOINTMENT_ENTRYID, &lpProp);
	if (hr != hrSuccess)
		goto exit;

	hr = m_lpActiveStore->CompareEntryIDs(sbEid.cb, (LPENTRYID)sbEid.lpb, lpProp->Value.bin.cb, (LPENTRYID)lpProp->Value.bin.lpb, 0, &ulResult);
	if (hr == hrSuccess && ulResult == TRUE)
		strCalendarOrder = "1";
	else
		hr = MAPI_E_CALL_FAILED;
	
exit:
	if (lpProp)
		MAPIFreeBuffer(lpProp);

	if (lpRootCont)
		lpRootCont->Release();

	return hr;
}
