/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <stdio.h>
#include <algorithm>
#include <assert.h>
#include <limits.h>

#include "stringutil.h"
#include "ECConfig.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const directive_t ECConfig::s_sDirectives[] = {
	{ "include",	&ECConfig::HandleInclude },
	{ "propmap",	&ECConfig::HandlePropMap },
	{ NULL }
};

// Configuration file parser

ECConfig::ECConfig(const configsetting_t *lpDefaults, const char **lpszDirectives)
{
	pthread_rwlock_init(&m_settingsRWLock, NULL);

	m_lpDefaults = lpDefaults;

	// allowed directives in this config object
	for (int i = 0; lpszDirectives[i] != NULL; i++)
		m_lDirectives.push_back(lpszDirectives[i]);

	InitDefaults(LOADSETTING_INITIALIZING | LOADSETTING_UNKNOWN | LOADSETTING_OVERWRITE);
}

bool ECConfig::LoadSettings(const char *szFilename)
{
	m_szConfigFile = szFilename;	
	return InitConfigFile(LOADSETTING_OVERWRITE);
}

bool ECConfig::ReloadSettings()
{
	InitDefaults(LOADSETTING_OVERWRITE_RELOAD);
	return InitConfigFile(LOADSETTING_OVERWRITE_RELOAD);
}

void ECConfig::AddSetting(const char *szName, const char *szValue, const unsigned int ulGroup)
{
	configsetting_t sSetting;

	sSetting.szName = szName;
	sSetting.szValue = szValue;
	sSetting.ulFlags = 0;
	sSetting.ulGroup = ulGroup;

	AddSetting(&sSetting, ulGroup ? LOADSETTING_OVERWRITE_GROUP : LOADSETTING_OVERWRITE);
}

void freeSettings(settingmap_t::value_type entry)
{
	free(entry.second); // free() because of strdup()
}

void ECConfig::CleanupMap(settingmap_t *lpMap)
{
	if (!lpMap->empty())
		for_each(lpMap->begin(), lpMap->end(), freeSettings);
}

ECConfig::~ECConfig()
{
	pthread_rwlock_wrlock(&m_settingsRWLock);

	CleanupMap(&m_mapSettings);
	CleanupMap(&m_mapAliases);

	pthread_rwlock_unlock(&m_settingsRWLock);

	pthread_rwlock_destroy(&m_settingsRWLock);
}

char *ECConfig::GetMapEntry(settingmap_t *lpMap, const char *szName)
{
	settingmap_t::iterator iterSettings;
	settingkey_t s;
	char *retval = NULL;

	memset(&s, 0, sizeof(s));
	strcpy(s.s, szName);

	pthread_rwlock_rdlock(&m_settingsRWLock);

	iterSettings = lpMap->find(s);
	if(iterSettings != lpMap->end())
		retval = iterSettings->second;

	pthread_rwlock_unlock(&m_settingsRWLock);

	return retval;
}

char *ECConfig::GetSetting(const char *szName)
{
	return GetMapEntry(&m_mapSettings, szName);
}

char *ECConfig::GetAlias(const char *szName)
{
	return GetMapEntry(&m_mapAliases, szName);
}

char* ECConfig::GetSetting(const char *szName, char *equal, char *other)
{
	char *value = this->GetSetting(szName);
	if ((value == equal) || (value && equal && !strcmp(value, equal)))
		return other;
	else
		return value;
}

list<configsetting_t> ECConfig::GetSettingGroup(unsigned int ulGroup)
{
	list<configsetting_t> lGroup;
	configsetting_t sSetting;

	for (settingmap_t::iterator iter = m_mapSettings.begin(); iter != m_mapSettings.end(); iter++) {
		if ((iter->first.ulGroup & ulGroup) == ulGroup) {
			if (CopyConfigSetting(&iter->first, iter->second, &sSetting))
				lGroup.push_back(sSetting);
		}
	}

	return lGroup;
}

bool ECConfig::InitDefaults(unsigned int ulFlags)
{
	unsigned int i = 0;

	/* throw error? this is unacceptable! useless object, since it won't set new settings */
	if (!m_lpDefaults)
		return false;

	while (m_lpDefaults[i].szName != NULL) {
		if (m_lpDefaults[i].ulFlags & CONFIGSETTING_ALIAS) {
			/* Aliases are only initialized once */
			if (ulFlags & LOADSETTING_INITIALIZING)
				AddAlias(&m_lpDefaults[i]);
		} else
			AddSetting(&m_lpDefaults[i], ulFlags);
		i++;
	}

	return true;
}

bool ECConfig::InitConfigFile(unsigned int ulFlags)
{
	bool bResult = false;

	assert(m_readFiles.empty());
	assert(m_pathStack.empty());

	if (!m_szConfigFile)
		return false;

	bResult = ReadConfigFile(m_szConfigFile, ulFlags);

	m_readFiles.clear();
	assert(m_pathStack.empty());

	return bResult;
}

bool ECConfig::ReadConfigFile(const char *szFilename, unsigned int ulFlags, unsigned int ulGroup)
{
	FILE *fp;
	bool bReturn = false;
	char cPathBuf[PATH_MAX] = {0};
	char cBuffer[MAXLINELEN] = {0};
	string strFilename;
	char *lpszPath = NULL;
	char *lpszBaseName = NULL;
	char *lpszDirName = NULL;
	string strLine;
	string strName;
	string strValue;
	size_t pos;

	assert(szFilename);

	lpszBaseName = strdup(szFilename);
	lpszDirName = strdup(szFilename);

	if (SetPathTo(dirname(lpszDirName)) < 0) {
		errors.push_back("Unable to change path for config file '" + string(szFilename) + "'");
		goto exit;
	}

	lpszPath = getcwd(cPathBuf, PATH_MAX);
	strFilename = string(lpszPath).append(1, PATH_SEPARATOR) + string(basename(lpszBaseName));

	/* Check if we read this file before. */
	if (m_readFiles.find(strFilename) != m_readFiles.end()) {
		bReturn = true;
		goto exit;
	}

    m_readFiles.insert(strFilename);

	if(!(fp = fopen(strFilename.c_str(), "rt"))) {
		errors.push_back("Unable to open config file '" + strFilename + "'");
		goto exit;
	}

	while (!feof(fp)) {
		memset(&cBuffer, 0, sizeof(cBuffer));

		if (!fgets(cBuffer, sizeof(cBuffer), fp))
			continue;

		strLine = string(cBuffer);

		/* Skip empty lines any lines which start with # */
		if (strLine.empty() || strLine[0] == '#')
 			continue;

		/* Handle special directives which start with '!' */
		if (strLine[0] == '!') {
			if (!HandleDirective(strLine, ulFlags))
				goto exit;
			continue;
		}

		/* Get setting name */
		pos = strLine.find('=');
		if (pos != string::npos) {
			strName = strLine.substr(0, pos);
			strValue = strLine.substr(pos + 1);
		} else
			continue;

		/*
		 * The line is build up like this:
		 * config_name = bla bla
		 *
		 * Whe should clean it in such a way that it resolves to:
		 * config_name=bla bla
		 *
		 * Be carefull _not_ to remove any whitespace characters
		 * within the configuration value itself
		 */
		strName = trim(strName, " \t\r\n");
		strValue = trim(strValue, " \t\r\n");

		if(!strName.empty()) {
			// Save it
			configsetting_t setting = { strName.c_str(), strValue.c_str(), 0, ulGroup };
			AddSetting(&setting, ulFlags);
		}
	}

	fclose(fp);

	bReturn = true;

exit:
	RestorePath();

	if (lpszBaseName)
		free(lpszBaseName);

	if (lpszDirName)
		free(lpszDirName);

	return bReturn;
}

bool ECConfig::HandleDirective(string &strLine, unsigned int ulFlags)
{
	size_t pos = strLine.find_first_of(" \t", 1);
	string strName = strLine.substr(1, pos - 1);

	/* Check if this directive is known */
	for (int i = 0; s_sDirectives[i].lpszDirective != NULL; ++i) {
		if (strName.compare(s_sDirectives[i].lpszDirective) == 0) {
			/* Check if this directive is supported */
			list<string>::iterator f = find(m_lDirectives.begin(), m_lDirectives.end(), strName);
			if (f != m_lDirectives.end())
				return (this->*s_sDirectives[i].fExecute)(strLine.substr(pos).c_str(), ulFlags);

			warnings.push_back("Unsupported directive '" + strName + "' found!");
			return true;
		}
	}

	warnings.push_back("Unknown directive '" + strName + "' found!");
	return true;
}

bool ECConfig::HandleInclude(const char *lpszArgs, unsigned int ulFlags)
{
	string	strValue;
	bool	bResult;

	strValue = trim(lpszArgs, " \t\r\n");
	bResult = ReadConfigFile(strValue.c_str(), ulFlags);

	return bResult;
}

bool ECConfig::HandlePropMap(const char *lpszArgs, unsigned int ulFlags)
{
	string	strValue;
	bool	bResult;

	strValue = trim(lpszArgs, " \t\r\n");
	bResult = ReadConfigFile(strValue.c_str(), LOADSETTING_UNKNOWN | LOADSETTING_OVERWRITE_GROUP, CONFIGGROUP_PROPMAP);

	return bResult;
}

bool ECConfig::CopyConfigSetting(const configsetting_t *lpsSetting, settingkey_t *lpsKey)
{
	if (lpsSetting->szName == NULL || lpsSetting->szValue == NULL)
		return false;

	memset(lpsKey, 0, sizeof(*lpsKey));
	strncpy(lpsKey->s, lpsSetting->szName, sizeof(lpsKey->s));
	lpsKey->ulFlags = lpsSetting->ulFlags;
	lpsKey->ulGroup = lpsSetting->ulGroup;

	return true;
}

bool ECConfig::CopyConfigSetting(const settingkey_t *lpsKey, const char *szValue, configsetting_t *lpsSetting)
{
	if (strlen(lpsKey->s) == 0 || szValue == NULL)
		return false;

	lpsSetting->szName = lpsKey->s;
	lpsSetting->szValue = szValue;
	lpsSetting->ulFlags = lpsKey->ulFlags;
	lpsSetting->ulGroup = lpsKey->ulGroup;

	return true;
}

void ECConfig::AddSetting(const configsetting_t *lpsConfig, unsigned int ulFlags)
{
	settingmap_t::iterator iterSettings;
	settingkey_t s;
	char *szAlias = NULL;

	if (!CopyConfigSetting(lpsConfig, &s))
		return;

	// Lookup name as alias
	szAlias = GetAlias(lpsConfig->szName);
	if (szAlias) {
		if (!(ulFlags & LOADSETTING_INITIALIZING))
			warnings.push_back("Option '" + string(lpsConfig->szName) + "' is deprecated! New name for option is '" + szAlias + "'.");
		strncpy(s.s, szAlias, sizeof(s.s));
	}

	pthread_rwlock_wrlock(&m_settingsRWLock);

	iterSettings = m_mapSettings.find(s);

	if (iterSettings == m_mapSettings.end()) {
		// new items from file are illegal, add error
		if (!(ulFlags & LOADSETTING_UNKNOWN)) {
			errors.push_back("Unknown option '" + string(lpsConfig->szName) + "' found!");
			goto exit;
		}
	} else {
		// Check for permissions before overwriting
		if (ulFlags & LOADSETTING_OVERWRITE_GROUP) {
			if (iterSettings->first.ulGroup != lpsConfig->ulGroup) {
				errors.push_back("option '" + string(lpsConfig->szName) + "' cannot be overridden (different group)!");
				goto exit;
			}
		} else if (ulFlags & LOADSETTING_OVERWRITE_RELOAD) {
			if (!(iterSettings->first.ulFlags & CONFIGSETTING_RELOADABLE))
				goto exit;
		} else if (!(ulFlags & LOADSETTING_OVERWRITE)) {
			errors.push_back("option '" + string(lpsConfig->szName) + "' cannot be overridden!");
			goto exit;
		}

		if (!(ulFlags & LOADSETTING_INITIALIZING) &&
			(iterSettings->first.ulFlags & CONFIGSETTING_UNUSED))
				warnings.push_back("Option '" + string(lpsConfig->szName) + "' is not used anymore.");

		// free() because of strdup()
		free(iterSettings->second);
	}

	m_mapSettings[s] = strdup(lpsConfig->szValue);

exit:
	pthread_rwlock_unlock(&m_settingsRWLock);
}

void ECConfig::AddAlias(const configsetting_t *lpsAlias)
{
	settingkey_t s;

	if (!CopyConfigSetting(lpsAlias, &s))
		return;

	pthread_rwlock_wrlock(&m_settingsRWLock);

	m_mapAliases[s] = strdup(lpsAlias->szValue);

	pthread_rwlock_unlock(&m_settingsRWLock);
}

bool ECConfig::HasWarnings() {
	return !warnings.empty();
}

list<string>* ECConfig::GetWarnings() {
	return &warnings;
}

bool ECConfig::HasErrors() {
	settingmap_t::iterator iterSettings;

	/* First validate the configuration settings */
	pthread_rwlock_rdlock(&m_settingsRWLock);

	for (iterSettings = m_mapSettings.begin(); iterSettings != m_mapSettings.end(); iterSettings++) {
		if (iterSettings->first.ulFlags & CONFIGSETTING_NONEMPTY) {
			if (!iterSettings->second || strlen(iterSettings->second) == 0)
				errors.push_back("option '" + string(iterSettings->first.s) + "' cannot be empty!");
		}
	}

	pthread_rwlock_unlock(&m_settingsRWLock);

	return !errors.empty();
}

list<string>* ECConfig::GetErrors() {
	return &errors;
}

int ECConfig::SetPathTo(const char *lpszDir)
{
	char *lpszCwd = getcwd(NULL, 0);
	int ret;

	m_pathStack.push_back(lpszCwd);
	ret = chdir(lpszDir);
	free(lpszCwd);
	return ret;
}

void ECConfig::RestorePath()
{
	assert (!m_pathStack.empty());
	chdir(m_pathStack.back().c_str());
	m_pathStack.pop_back();
}
