/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// From http://www.wischik.com/lu/programmer/mapi_utils.html

#include "platform.h"
#include <iostream>
#include "codepage.h"
#include "CommonUtil.h"
#include "Util.h"

#include "rtfutil.h"

#include <string> 
#include <sstream>

using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Charsets used in \fcharsetXXX (from http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnrtfspec/html/rtfspec_6.asp )
// charset "" is the ANSI codepage specified in \ansicpg
// charset NULL means 'no conversion', ie direct 1-to-1 translation to UNICODE 
struct _rtfcharset {
	int id;
	char *charset;
} RTFCHARSET[] = {
	{0, ""}, // This is actually the codepage specified in \ansicpg
	{1, ""}, // Unknown, probably also the codepage in \ansicpg
	{2, NULL}, // This is SYMBOL, but in practice, we can just send the data
			 // one-on-one down the line, as the actual character codes
			 // don't change when converted to unicode (because the other
			 // side will also be using the MS Symbol font to display)
	{3, NULL}, // 'Invalid'
	{77, "MAC"}, // Unsure if this is the correct charset
	{128,"SJIS"}, // OR cp 932 ?
	{129,NULL}, // 'Hangul' ?
	{130,"JOHAB"},
	{134,"GB2312"},
	{136,"BIG5"},
	{161,"windows-1253"},
	{162,"windows-1254"}, // 'Turkish'
	{163,"windows-1258"}, // Vietnamese
	{177,"windows-1255"},
	{178,"windows-1256"}, // Arabic
	{179,"windows-1256"}, // Arabic traditional
	{180,"windows-1256"}, // Arabic user
	{181,"windows-1255"}, // Hebrew user
	{186,"windows-1257"},
	{204,"windows-1251"}, // Cyrillic for russian
	{222,NULL},		// Thai ?
	{238,"windows-1250"}, // Eastern European
	{254,"IBM437"},
	{255,NULL} 		// OEM
};

struct RTFSTATE {
	int ulFont;
	char *szCharset;
	bool bInFontTbl;
	bool bInColorTbl;
	bool bInSkipTbl;
	std::string output;
	bool bRTFOnly;
};

struct RTFFONT {
	int ulCharset;
};

#define RTF_MAXSTATE 	256
#define RTF_MAXFONT 	256
#define RTF_MAXCMD		64

// DECODERTFHTML -- Given an uncompressed RTF body of the message,
// and assuming that it contains encoded-html, this function
// turns it onto regular html.
// [in] (buf,*len) indicate the start and length of the uncompressed RTF body.
// [out] the buffer is overwritten with the HTML version, null-terminated,
// and *len indicates the length of this HTML.
//
// Notes: (1) because of how the encoding works, the HTML version is necessarily
// shorter than the encoded version. That's why it's safe for the function to
// place the decoded html in the same buffer that formerly held the encoded stuff.
// (2) Some messages include characters \'XX, where XX is a hexedecimal number.
// This function simply converts this into ASCII. The conversion will only make
// sense if the right code-page is being used. I don't know how rtf specifies which
// code page it wants.
// (3) By experiment, I discovered that \pntext{..} and \liN and \fi-N are RTF
// markup that should be removed. There might be other RTF markup that should
// also be removed. But I don't know what else.
//
void decodertfhtml(char *buf,unsigned int *len,unsigned int *lpulCodepage)
{ // c -- pointer to where we're reading from
  // d -- pointer to where we're writing to. Invariant: d<c
  // max -- how far we can read from (i.e. to the end of the original rtf)
  // ignore_tag -- stores 'N': after \mhtmlN, we will ignore the subsequent \htmlN.
  char *c=buf, *max=buf+*len, *d=buf; int ignore_tag=-1;
  char *ansicpg = NULL;
  unsigned int ulCodepage = 0;
  
  // First, search for the codepage
  if(lpulCodepage) {
      ansicpg = strstr(buf, "\\ansicpg");
      if(ansicpg == NULL) 
          ulCodepage = 0;
      else {      
          ansicpg += 8;
          while(ansicpg<max && *ansicpg > '0' && *ansicpg <'9') { ulCodepage *= 10; ulCodepage += (*ansicpg - '0'); ansicpg++; }
      }
      *lpulCodepage = ulCodepage;
  }
  
  // Now, we skip forwards to the first \htmltag.
  while (c<max && strncmp(c,"{\\*\\htmltag",11)!=0) c++;
  //
  // Now work through the document. Our plan is as follows:
  // * Ignore { and }. These are part of RTF markup.
  // * Ignore \htmlrtf...\htmlrtf0. This is how RTF keeps its equivalent markup separate from the html.
  // * Ignore \r and \n. The real carriage returns are stored in \par tags.
  // * Ignore \pntext{..} and \liN and \fi-N. These are RTF junk.
  // * Convert \par and \tab into \r\n and \t
  // * Convert \'XX into the ascii character indicated by the hex number XX
  // * Convert \{ and \} into { and }. This is how RTF escapes its curly braces.
  // * When we get \*\mhtmltagN, keep the tag, but ignore the subsequent \*\htmltagN
  // * When we get \*\htmltagN, keep the tag as long as it isn't subsequent to a \*\mhtmltagN
  // * All other text should be kept as it is.
  while (c<max)
  { if (*c=='{') c++;
    else if (*c=='}') c++;
    else if (strncmp(c,"\\*\\htmltag",10)==0)
    { c+=10; int tag=0; while (*c>='0' && *c<='9') {tag=tag*10+*c-'0'; c++;}
      if (*c==' ') c++;
      if (tag==ignore_tag) {while (c<max && *c!='}') c++; if (*c=='}') c++;}
      ignore_tag=-1;
    }
    else if (strncmp(c,"\\*\\mhtmltag",11)==0)
    { c+=11; int tag=0; while (*c>='0' && *c<='9') {tag=tag*10+*c-'0'; c++;}
      if (*c==' ') c++;
      ignore_tag=tag;
    }
    else if (strncmp(c,"\\par",4)==0) {strcpy(d,"\r\n"); d+=2; c+=4; if (*c==' ') c++;}
    else if (strncmp(c,"\\tab",4)==0) {strcpy(d,"   "); d+=3; c+=4; if (*c==' ') c++;}
    else if (strncmp(c,"\\li",3)==0)
    { c+=3; while (*c>='0' && *c<='9') c++; if (*c==' ') c++;
    }
    else if (strncmp(c,"\\fi-",4)==0)
    { c+=4; while (*c>='0' && *c<='9') c++; if (*c==' ') c++;
    }
    else if (strncmp(c,"\\'",2)==0)
    { unsigned int hi=c[2], lo=c[3];
      if (hi>='0' && hi<='9') hi-='0'; else if (hi>='A' && hi<='Z') hi=hi-'A'+10; else if (hi>='a' && hi<='z') hi=hi-'a'+10;
      if (lo>='0' && lo<='9') lo-='0'; else if (lo>='A' && lo<='Z') lo=lo-'A'+10; else if (lo>='a' && lo<='z') lo=lo-'a'+10;
      *((unsigned char*)d) = (unsigned char)(hi*16+lo);
      c+=4; d++;
    }
    else if (strncmp(c,"\\pntext",7)==0) {c+=7; while (c<max && *c!='}') c++;}
    else if (strncmp(c,"\\htmlrtf",8)==0)
    { c++; while (c<max && strncmp(c,"\\htmlrtf0",9)!=0) c++;
      if (c<max) c+=9; if (*c==' ') c++;
    }
    else if (*c=='\r' || *c=='\n') c++;
    else if (strncmp(c,"\\{",2)==0) {*d='{'; d++; c+=2;}
    else if (strncmp(c,"\\}",2)==0) {*d='}'; d++; c+=2;}
    else if (strncmp(c,"\\\\",2)==0) {*d='\\'; d++; c+=2;}
    else {*d=*c; c++; d++;}
  }
  *d=0; d++;
  *len = (unsigned int)(d-buf);
}

HRESULT HrGetCharsetByRTFID(int id, char **lpszCharset)
{
	unsigned int i = 0;
	
	for (i=0; i<sizeof(RTFCHARSET)/sizeof(RTFCHARSET[0]); i++) {
		if(RTFCHARSET[i].id == id) {
			*lpszCharset = RTFCHARSET[i].charset;
			return hrSuccess;
		}
	}
	return MAPI_E_NOT_FOUND;
}

// RTF ignore Commando's
bool isRTFIgnoreCommand(char* lpCommand)
{
	bool bIgnore = false;

	if(lpCommand == NULL)
		goto exit;

	if(strcmp(lpCommand,"stylesheet") == 0 ||
		strcmp(lpCommand,"revtbl") == 0 || 
		strcmp(lpCommand,"xmlnstbl") == 0 ||
		strcmp(lpCommand,"rsidtbl") == 0 ||
		strcmp(lpCommand,"fldinst") == 0 ||
		strcmp(lpCommand,"shpinst") == 0 ||
		strcmp(lpCommand,"wgrffmtfilter") == 0 ||
		strcmp(lpCommand,"pnseclvl") == 0 ||
		strcmp(lpCommand,"atrfstart") == 0 ||
		strcmp(lpCommand,"atrfend") == 0 ||
		strcmp(lpCommand,"atnauthor") == 0 ||
		strcmp(lpCommand,"annotation") == 0 ||
		strcmp(lpCommand,"sp") == 0 ||
		strcmp(lpCommand,"atnid") == 0
		//strcmp(lpCommand,"fldrslt") == 0
		)
	{
			bIgnore = true;		
	}

exit:
	return bIgnore;
}

// This function extracts the HTML from an RTF stream, with support for UNICODE character output. It automatically
// decodes the localized chars in the RTF into UNICODE chars.
HRESULT HrExtractHTMLFromRTF(std::string &lpStrRTFIn, ucs2string &lpStrHTMLOut)
{
	HRESULT hr = hrSuccess;
	char *szInput = (char *)lpStrRTFIn.c_str();
	char *szANSICharset = "windows-1252";
	int ulState = 0;
	RTFSTATE sState[RTF_MAXSTATE];	
	RTFFONT sFont[RTF_MAXFONT];
	
    lpStrHTMLOut.resize(0,0);
    
    // Find \\htmltag, if there is none we can't extract HTML
    if(strstr(szInput, "{\\*\\htmltag") == NULL) {
    	hr = MAPI_E_NOT_FOUND;
    	goto exit;
	}

	memset(sFont, 0, sizeof(sFont));
    
	sState[0].bInSkipTbl = false;
	sState[0].bInFontTbl = false;
	sState[0].bInColorTbl = false;
	sState[0].szCharset = "iso-8859-1";
	sState[0].bRTFOnly = false;
	sState[0].ulFont = 0;

	while(*szInput) {
		if(strncmp(szInput,"\\*",2) == 0) {
			szInput+=2;
		} else if(*szInput == '\\') {
			// Command
			char szCommand[RTF_MAXCMD];
			char *szCmdOutput;
			int lArg = -1;
			bool bNeg = false;
			
			szInput++;
			if(isalpha(*szInput)) {
				szCmdOutput = szCommand;
				
				while(isalpha(*szInput) && szCmdOutput < szCommand+RTF_MAXCMD-1) *szCmdOutput++ = *szInput++;
				*szCmdOutput = 0;
				
				if(*szInput == '-') {
					bNeg = true;
					szInput++;
				}
				
				if(isdigit(*szInput)) {
					lArg = 0;
					while(isdigit(*szInput)) { lArg = lArg*10 + *szInput - '0'; szInput++; }
					if(bNeg) lArg = -lArg;
				}
				if(*szInput == ' ')
					szInput++;
				
				// szCommand is the command, lArg is the argument.
				if(strcmp(szCommand,"fonttbl") == 0) {
					sState[ulState].bInFontTbl = true;
				} else if(strcmp(szCommand,"colortbl") == 0) {
					sState[ulState].bInColorTbl = true;
				} else if(strcmp(szCommand,"pntext") == 0) { // pntext is the plaintext alternative, ignore it.
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"ansicpg") == 0) {
					if(HrGetCharsetByCP(lArg, &szANSICharset) != hrSuccess)
						szANSICharset = "iso-8859-1";
                    sState[ulState].szCharset = szANSICharset;
				} else if(strcmp(szCommand,"fcharset") == 0) {
					if(sState[ulState].bInFontTbl) {
						sFont[sState[ulState].ulFont].ulCharset = lArg;
					}
				} else if(strcmp(szCommand,"htmltag") == 0) {
				} else if(strcmp(szCommand,"mhtmltag") == 0) {
				} else if(strcmp(szCommand,"pard") == 0 || strcmp(szCommand,"par") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,'\r');
						sState[ulState].output.append(1,'\n');
					}
				} else if(strcmp(szCommand,"tab") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
					}
				} else if(strcmp(szCommand,"f") == 0) {
					if(lArg >= RTF_MAXFONT)
						continue;
					
					sState[ulState].ulFont = lArg;
					
					if(!sState[ulState].bInFontTbl) {
						// Output any data before this point
						ucs2string wstrUnicode;
						
						HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
						lpStrHTMLOut.append(wstrUnicode);

						sState[ulState].output = "";

						// Set new charset			
						HrGetCharsetByRTFID(sFont[lArg].ulCharset, &sState[ulState].szCharset);
						if(sState[ulState].szCharset == NULL) {
							sState[ulState].szCharset = "windows-1252";
						} else if(sState[ulState].szCharset[0] == 0) {
							sState[ulState].szCharset = szANSICharset;
						}
						
					} 
					// ignore error
				} else if(strcmp(szCommand,"htmlrtf") == 0) {
					if(lArg != 0) {
						// \\htmlrtf
						sState[ulState].bRTFOnly = true;
					} else {
						// \\htmlrtf0
						sState[ulState].bRTFOnly = false;
					}
				}else if(isRTFIgnoreCommand(szCommand)) {
					sState[ulState].bInSkipTbl = true;
				}

			}
			// Non-alnum after '\'
			else if(*szInput == '\\') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'\\');
			}
			else if(*szInput == '{') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'{');
			}
			else if(*szInput == '}') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'}');
			} 
			else if(*szInput == '\'') {
				/* Apparently, escaped data is back in the charset0 charset (ansicpg)
				 * context. This is rather strange, but seems to be the way! Also, it could be surrounded
				 * by more data which is not escaped, like in SJIS: \'1BB$!G\'1B(G .. this is strange because
				 * this only works of the current charset == ansicpg .. so if \'XX occurs inside a different
				 * charset, we have to output everything until then, but if \'XX occurs inside the ansicpg charset
				 * then we can just keep reading data in the current charset.
				 */
				unsigned int ulChar;
				char szHex[] = "0123456789ABCDEF";
				
				// Dump output data until now, if we're switching charsets
				if(szANSICharset == NULL || strcmp(sState[ulState].szCharset, szANSICharset) != 0) {
    				ucs2string wstrUnicode;
	    			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
		    		lpStrHTMLOut.append(wstrUnicode);
			    	sState[ulState].output = "";
                }
				
				while(*szInput == '\'')
				{
					ulChar = 0;
					szInput++;

					if(*szInput) {
						ulChar = (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						ulChar = ulChar << 4;
						szInput++;
					}
					
					if(*szInput) {
						ulChar += (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						szInput++;
					}

					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,ulChar);
					}

					if(*szInput == '\\' && *(szInput+1) == '\'')
						szInput++;
				}				

   				// Dump escaped data in charset 0 (ansicpg), if we had to switch charsets
				if(szANSICharset == NULL || strcmp(sState[ulState].szCharset, szANSICharset) != 0) {
    				ucs2string wstrUnicode;
	    			HrConvertToWString(sState[ulState].output, szANSICharset == NULL ? "windows-1252" : szANSICharset, wstrUnicode);
		    		lpStrHTMLOut.append(wstrUnicode);
			    	sState[ulState].output = "";
                }
				
			} else {
				szInput++; // skip single character after '\'
			}
		} // Non-command
		else if(*szInput == '{') {
			// Dump output data
			ucs2string wstrUnicode;
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
			lpStrHTMLOut.append(wstrUnicode);
			
			sState[ulState].output = "";
			
			ulState++;
			if(ulState >= RTF_MAXSTATE) {
				hr = MAPI_E_NOT_ENOUGH_MEMORY;
				goto exit;
			}
			sState[ulState] = sState[ulState-1];
			sState[ulState].output = "";
			szInput++;
		} else if(*szInput == '}') {
			// Dump output data
			ucs2string wstrUnicode;
			
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
				
			// Append
			lpStrHTMLOut.append(wstrUnicode);
		
			if(ulState > 0)
				ulState--;
			szInput++;
		} else if(*szInput == '\r' || *szInput == '\n') {
			szInput++;
		} else {
			if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
				sState[ulState].output.append(1,*szInput);
			}
			szInput++;
		}
	}
	
	if(sState[ulState].output.length() > 0) {
		ucs2string wstrUnicode;
		HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
		lpStrHTMLOut.append(wstrUnicode);
	}		
	
exit:
	return hr;
	
}

HRESULT HrExtractHTMLFromTextRTF(std::string &lpStrRTFIn, ucs2string &lpStrHTMLOut)
{
	HRESULT hr = hrSuccess;
	ucs2string wstrUnicodeTmp;
	char *szInput = (char *)lpStrRTFIn.c_str();
	char *szANSICharset = "windows-1252";
	int ulState = 0;
	bool bPar = false;
	bool bNewLine = false;
	int nLineChar=0;
	RTFSTATE sState[RTF_MAXSTATE];	
	RTFFONT sFont[RTF_MAXFONT];
	
    lpStrHTMLOut.resize(0,0);
    
	string tmp;

	tmp =	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\r\n" \
			"<HTML>\r\n" \
			"<HEAD>\r\n" \
			"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=windows-1252\">\r\n" \
			"<META NAME=\"Generator\" CONTENT=\"Zarafa text/HTML builder 1.0\">\r\n" \
			"<TITLE></TITLE>\r\n" \
			"</HEAD>\r\n" \
			"<BODY>\r\n" \
			"<!-- Converted from text/plain format -->\r\n" \
			"\r\n"; //FIXME create title on the fly ?

	wstrUnicodeTmp.resize(0,0);
	HrConvertToWString(tmp, "windows-1252", wstrUnicodeTmp);
	lpStrHTMLOut.append(wstrUnicodeTmp);

	memset(sFont, 0, sizeof(sFont));
    
	sState[0].bInSkipTbl = false;
	sState[0].bInFontTbl = false;
	sState[0].bInColorTbl = false;
	sState[0].szCharset = "iso-8859-1";
	sState[0].bRTFOnly = false;
	sState[0].ulFont = 0;

	while(*szInput) {
		if(strncmp(szInput,"\\*",2) == 0) {
			szInput+=2;
		} else if(*szInput == '\\') {
			// Command
			char szCommand[RTF_MAXCMD];
			char *szCmdOutput;
			int lArg = -1;
			bool bNeg = false;
			
			szInput++;
			if(isalpha(*szInput)) {
				szCmdOutput = szCommand;
				
				while(isalpha(*szInput) && szCmdOutput < szCommand+RTF_MAXCMD-1) *szCmdOutput++ = *szInput++;
				*szCmdOutput = 0;
				
				if(*szInput == '-') {
					bNeg = true;
					szInput++;
				}
				
				if(isdigit(*szInput)) {
					lArg = 0;
					while(isdigit(*szInput)) { lArg = lArg*10 + *szInput - '0'; szInput++; }
					if(bNeg) lArg = -lArg;
				}
				if(*szInput == ' ')
					szInput++;
				
				// szCommand is the command, lArg is the argument.
				if(strcmp(szCommand,"fonttbl") == 0) {
					sState[ulState].bInFontTbl = true;
				} else if(strcmp(szCommand,"colortbl") == 0) {
					sState[ulState].bInColorTbl = true;
				} else if(strcmp(szCommand,"pntext") == 0) { // pntext is the plaintext alternative, ignore it.
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"ansicpg") == 0) {
					if(HrGetCharsetByCP(lArg, &szANSICharset) != hrSuccess)
						szANSICharset = "iso-8859-1";
                    sState[ulState].szCharset = szANSICharset;
				} else if(strcmp(szCommand,"fcharset") == 0) {
					if(sState[ulState].bInFontTbl) {
						sFont[sState[ulState].ulFont].ulCharset = lArg;
					}
				} else if(strcmp(szCommand,"htmltag") == 0) {
				} else if(strcmp(szCommand,"mhtmltag") == 0) {
				} else if(strcmp(szCommand,"pard") == 0 || strcmp(szCommand,"par") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {

						if(bNewLine == true && nLineChar >0 && bPar == true){
							sState[ulState].output.append("</P>\r\n\r\n", 8);
							bPar = false;
							nLineChar = 0;
						}else if(bNewLine == false && nLineChar >0)
							sState[ulState].output.append("</FONT>\r\n", 9);
						else
							sState[ulState].output.append("\r\n", 2);
					}
				} else if(strcmp(szCommand,"tab") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
					}
				} else if(strcmp(szCommand,"f") == 0) {
					if(lArg >= RTF_MAXFONT)
						continue;
					
					sState[ulState].ulFont = lArg;
					
					if(!sState[ulState].bInFontTbl) {
						// Output any data before this point
						ucs2string wstrUnicode;
						
						HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
						lpStrHTMLOut.append(wstrUnicode);

						sState[ulState].output = "";

						// Set new charset			
						HrGetCharsetByRTFID(sFont[lArg].ulCharset, &sState[ulState].szCharset);
						if(sState[ulState].szCharset == NULL) {
							sState[ulState].szCharset = "windows-1252";
						} else if(sState[ulState].szCharset[0] == 0) {
							sState[ulState].szCharset = szANSICharset;
						}
						
					} 
					// ignore error
				}
				else if(strcmp(szCommand,"htmlrtf") == 0) {
					if(lArg != 0) {
						// \\htmlrtf
						sState[ulState].bRTFOnly = true;
					} else {
						// \\htmlrtf0
						sState[ulState].bRTFOnly = false;
					}
				} else if(isRTFIgnoreCommand(szCommand)) {
					sState[ulState].bInSkipTbl = true;
				}
			}
			// Non-alnum after '\'
			else if(*szInput == '\\') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'\\');
			}
			else if(*szInput == '{') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'{');
			}
			else if(*szInput == '}') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'}');
			} 
			else if(*szInput == '\'') {
				unsigned int ulChar;
				char szHex[] = "0123456789ABCDEF";
				
				// Dump output data until now, if we're switching charsets
				if(szANSICharset == NULL || strcmp(sState[ulState].szCharset, szANSICharset) != 0) {
    				ucs2string wstrUnicode;
	    			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
		    		lpStrHTMLOut.append(wstrUnicode);
			    	sState[ulState].output = "";
                }
				
				while(*szInput == '\'')
				{
					ulChar = 0;
					szInput++;
					
					if(*szInput) {
						ulChar = (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						ulChar = ulChar << 4;
						szInput++;
					}
					
					if(*szInput) {
						ulChar += (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						szInput++;
					}
					
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl) {		
						sState[ulState].output.append(1,ulChar);
					}

					if(*szInput == '\\' && *(szInput+1) == '\'')
						szInput++;
				}

   				// Dump escaped data in charset 0 (ansicpg), if we had to switch charsets
				if(szANSICharset == NULL || strcmp(sState[ulState].szCharset, szANSICharset) != 0) {
    				ucs2string wstrUnicode;
    				HrConvertToWString(sState[ulState].output, szANSICharset == NULL ? "windows-1252" : szANSICharset, wstrUnicode);
	    			lpStrHTMLOut.append(wstrUnicode);
		    		sState[ulState].output = "";
                }
				
			} else {
				szInput++; // skip single character after '\'
			}
		} // Non-command
		else if(*szInput == '{') {
			// Dump output data
			ucs2string wstrUnicode;
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
			lpStrHTMLOut.append(wstrUnicode);
			
			sState[ulState].output = "";
			
			ulState++;
			if(ulState >= RTF_MAXSTATE) {
				hr = MAPI_E_NOT_ENOUGH_MEMORY;
				goto exit;
			}
			sState[ulState] = sState[ulState-1];
			sState[ulState].output = "";
			szInput++;
		} else if(*szInput == '}') {
			// Dump output data
			ucs2string wstrUnicode;
			
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
				
			// Append
			lpStrHTMLOut.append(wstrUnicode);
		
			if(ulState > 0)
				ulState--;
			szInput++;
		} else if(*szInput == '\r' || *szInput == '\n') {
			bNewLine = true;
			szInput++;
		} else {
			if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
				if(bPar == false){
					sState[ulState].output.append("<P>", 3);
					bPar = true;
				}else if(bNewLine == true && bPar == true)
					sState[ulState].output.append("\r\n<BR>", 6);

				if(bNewLine == true){
					sState[ulState].output.append("<FONT SIZE=2>", 13);
					bNewLine = false;
				}
				
				// Change space to &nbsp; . The last space is a real space like "&nbsp;&nbsp; " or " "
				if(*szInput == ' ') {
					szInput++;
					
					while(*szInput == ' ') {
						sState[ulState].output.append("&nbsp;", 6);
						szInput++;
					}
					
					sState[ulState].output.append(1, ' ');
				}else {
					sState[ulState].output += Util::HtmlEntityFromChar(*szInput);
					szInput++;
				}
				
				nLineChar++;
			} else {
				szInput++;
			}
		}
	}
	
	if(sState[ulState].output.length() > 0) {
		ucs2string wstrUnicode;
		HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
		lpStrHTMLOut.append(wstrUnicode);
	}		
	
	tmp =	"\r\n" \
			"</BODY>\r\n" \
			"</HTML>\r\n";

	wstrUnicodeTmp.resize(0,0);
	HrConvertToWString(tmp, "windows-1252", wstrUnicodeTmp);
	lpStrHTMLOut.append(wstrUnicodeTmp);
	
	
exit:
	return hr;
	
}

//FIXME: Export the right HTML tags, now only plain stuff
HRESULT HrExtractHTMLFromRealRTF(std::string &lpStrRTFIn, ucs2string &lpStrHTMLOut)
{
	HRESULT hr = hrSuccess;
	ucs2string wstrUnicodeTmp;
	char *szInput = (char *)lpStrRTFIn.c_str();
	char *szANSICharset = "windows-1252";
	int ulState = 0;
	RTFSTATE sState[RTF_MAXSTATE];	
	RTFFONT sFont[RTF_MAXFONT];
	
    lpStrHTMLOut.resize(0,0);
    
	string tmp;

	tmp =	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\r\n" \
			"<HTML>\r\n" \
			"<HEAD>\r\n" \
			"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=windows-1252\">\r\n" \
			"<META NAME=\"Generator\" CONTENT=\"Zarafa rtf/HTML builder 1.0\">\r\n" \
			"<TITLE></TITLE>\r\n" \
			"</HEAD>\r\n" \
			"<BODY>\r\n" \
			"<!-- Converted from text/rtf format -->\r\n" \
			"\r\n"; //FIXME create title on the fly ?

	wstrUnicodeTmp.resize(0,0);
	HrConvertToWString(tmp, "windows-1252", wstrUnicodeTmp);
	lpStrHTMLOut.append(wstrUnicodeTmp);

	memset(sFont, 0, sizeof(sFont));
    
	sState[0].bInSkipTbl = false;
	sState[0].bInFontTbl = false;
	sState[0].bInColorTbl = false;
	sState[0].szCharset = "iso-8859-1";
	sState[0].bRTFOnly = false;
	sState[0].ulFont = 0;

	while(*szInput) {
		if(strncmp(szInput,"\\*",2) == 0) {
			szInput+=2;
		} else if(*szInput == '\\') {
			// Command
			char szCommand[RTF_MAXCMD];
			char *szCmdOutput;
			int lArg = -1;
			bool bNeg = false;
			
			szInput++;
			if(isalpha(*szInput)) {
				szCmdOutput = szCommand;
				
				while(isalpha(*szInput) && szCmdOutput < szCommand+RTF_MAXCMD-1) *szCmdOutput++ = *szInput++;
				*szCmdOutput = 0;
				
				if(*szInput == '-') {
					bNeg = true;
					szInput++;
				}
				
				if(isdigit(*szInput)) {
					lArg = 0;
					while(isdigit(*szInput)) { lArg = lArg*10 + *szInput - '0'; szInput++; }
					if(bNeg) lArg = -lArg;
				}
				if(*szInput == ' ')
					szInput++;
				
				// szCommand is the command, lArg is the argument.
				if(strcmp(szCommand,"fonttbl") == 0) {
					sState[ulState].bInFontTbl = true;
				} else if(strcmp(szCommand,"colortbl") == 0) {
					sState[ulState].bInColorTbl = true;
				} else if(strcmp(szCommand,"pntext") == 0) { // pntext is the plaintext alternative, ignore it.
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"ansicpg") == 0) {
					if(HrGetCharsetByCP(lArg, &szANSICharset) != hrSuccess)
						szANSICharset = "iso-8859-1";
				} else if(strcmp(szCommand,"fcharset") == 0) {
					if(sState[ulState].bInFontTbl) {
						sFont[sState[ulState].ulFont].ulCharset = lArg;
					}
				} else if(strcmp(szCommand,"htmltag") == 0) {
				} else if(strcmp(szCommand,"latentstyles") == 0) {
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"datastore") == 0) {
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"mhtmltag") == 0) {
				} else if(strcmp(szCommand,"pard") == 0 || strcmp(szCommand,"par") == 0 || strcmp(szCommand,"line") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append("<br>\r\n", 6);
					}
				} else if(strcmp(szCommand,"tab") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
						sState[ulState].output.append(1,' ');
					}
				} else if(strcmp(szCommand,"f") == 0) {
					if(lArg >= RTF_MAXFONT)
						continue;
					
					sState[ulState].ulFont = lArg;
					
					if(!sState[ulState].bInFontTbl) {
						// Output any data before this point
						ucs2string wstrUnicode;
						
						HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
						lpStrHTMLOut.append(wstrUnicode);

						sState[ulState].output = "";

						// Set new charset			
						HrGetCharsetByRTFID(sFont[lArg].ulCharset, &sState[ulState].szCharset);
						if(sState[ulState].szCharset == NULL) {
							sState[ulState].szCharset = "windows-1252";
						} else if(sState[ulState].szCharset[0] == 0) {
							sState[ulState].szCharset = szANSICharset;
						}
						
					} 
					// ignore error
				}
				else if(strcmp(szCommand,"htmlrtf") == 0) {
					if(lArg != 0) {
						// \\htmlrtf
						sState[ulState].bRTFOnly = true;
					} else {
						// \\htmlrtf0
						sState[ulState].bRTFOnly = false;
					}
				}
				/*else if(strcmp(szCommand,"b") == 0) {
					if( lArg == -1)
						sState[ulState].output.append("<b>", 3);
					else
						sState[ulState].output.append("</b>", 4);
				}else if(strcmp(szCommand,"i") == 0) {
					if( lArg == -1)
						sState[ulState].output.append("<i>", 3);
					else
						sState[ulState].output.append("</i>", 4);
				}else if(strcmp(szCommand,"ul") == 0) {
					if( lArg == -1)
						sState[ulState].output.append("<u>", 3);
					else
						sState[ulState].output.append("</u>", 4);
				}else if(strcmp(szCommand,"ulnone") == 0) {
					sState[ulState].output.append("</u>", 4);
				}*/
				else if(strcmp(szCommand,"generator") == 0){
					 while(*szInput != ';' && *szInput) szInput++;

					 if(*szInput) 
						 szInput++;
				}
				else if(strcmp(szCommand,"bkmkstart") == 0 || strcmp(szCommand,"bkmkend") == 0){
					// skip bookmark name
					while (*szInput && isalnum(*szInput)) szInput++;

					sState[ulState].bInSkipTbl = true;
				} else if(isRTFIgnoreCommand(szCommand)) {
					sState[ulState].bInSkipTbl = true;
				}
				
			}
			// Non-alnum after '\'
			else if(*szInput == '\\') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'\\');
			}
			else if(*szInput == '{') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'{');
			}
			else if(*szInput == '}') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'}');
			} 
			else if(*szInput == '\'') {
				/* Apparently, escaped data is back in the charset0 charset (ansicpg)
				 * context. This is rather strange, but seems to be the way!
				 */
				unsigned int ulChar;
				char szHex[] = "0123456789ABCDEF";
				
				// Dump output data until now
				ucs2string wstrUnicode;
				HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
				lpStrHTMLOut.append(wstrUnicode);
				sState[ulState].output = "";
				
				while(*szInput == '\'')
				{
					ulChar = 0;
					szInput++;
					
					if(*szInput) {
						ulChar = (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						ulChar = ulChar << 4;
						szInput++;
					}
					
					if(*szInput) {
						ulChar += (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
						szInput++;
					}
					
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append(1,ulChar);
					}

					if(*szInput == '\\' && *(szInput+1) == '\'')
						szInput++;
				}

				// Dump escaped data in charset 0 (ansicpg)
				HrConvertToWString(sState[ulState].output, szANSICharset == NULL ? "windows-1252" : szANSICharset, wstrUnicode);
				lpStrHTMLOut.append(wstrUnicode);
				sState[ulState].output = "";
				
			} else {
				szInput++; // skip single character after '\'
			}
		} // Non-command
		else if(*szInput == '{') {
			// Dump output data
			ucs2string wstrUnicode;
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
			lpStrHTMLOut.append(wstrUnicode);
			
			sState[ulState].output = "";
			
			ulState++;
			if(ulState >= RTF_MAXSTATE) {
				hr = MAPI_E_NOT_ENOUGH_MEMORY;
				goto exit;
			}
			sState[ulState] = sState[ulState-1];
			sState[ulState].output = "";
			szInput++;
		} else if(*szInput == '}') {
			// Dump output data
			ucs2string wstrUnicode;
			
			HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
				
			// Append
			lpStrHTMLOut.append(wstrUnicode);
		
			if(ulState > 0)
				ulState--;
			szInput++;
		} else if(*szInput == '\r' || *szInput == '\n') {
			szInput++;
		} else {
			if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
				sState[ulState].output.append(1,*szInput);
			}
			szInput++;
		}
	}
	
	if(sState[ulState].output.length() > 0) {
		ucs2string wstrUnicode;
		HrConvertToWString(sState[ulState].output, sState[ulState].szCharset, wstrUnicode);
		lpStrHTMLOut.append(wstrUnicode);
	}		
	
	tmp =	"\r\n" \
			"</BODY>\r\n" \
			"</HTML>\r\n";

	wstrUnicodeTmp.resize(0,0);
	HrConvertToWString(tmp, "windows-1252", wstrUnicodeTmp);
	lpStrHTMLOut.append(wstrUnicodeTmp);
	
	
exit:
	return hr;
	
}

bool isrtfhtml(const char *buf,unsigned int len)
{ // We look for the words "\f\romhtml" somewhere in the file.
  // If the rtf encodes text rather than html, then instead
  // it will only find "\fromtext".
  for (const char *c=buf; c<buf+len; c++)
  { if (strncmp(c,"\\from",5)==0) return strncmp(c,"\\fromhtml",9)==0;
  } return false;
}

bool isrtftext(const char *buf,unsigned int len)
{ // We look for the words "\f\romtext" somewhere in the file.
  // If the rtf encodes text rather than text, then instead
  // it will only find "\fromhtml".
  for (const char *c=buf; c<buf+len; c++)
  { if (strncmp(c,"\\from",5)==0) return strncmp(c,"\\fromtext",9)==0;
  } return false;
}

int encapHTMLInRTF(string *lpStrBuffIn, string *lpStrBuffOut, unsigned int ulCodepage)
{
	ostringstream s;	
	const char *lpC = lpStrBuffIn->c_str();
	int			iLen = lpStrBuffIn->length();
	char		szHex[3];
	
	if(ulCodepage == 0)
	    ulCodepage = 1252;

	// start rtf
	s << "{\\rtf1\\ansi\\ansicpg" << ulCodepage << "\\fromhtml1 \\deff1{\\fonttbl\n"
		"{\\f0\\fswiss\\fcharset204 Arial;}\n"
		"{\\f1\\fmodern Courier New;}\n"
		"{\\f2\\fnil\\fcharset2 Symbol;}\n"
		"{\\f3\\fmodern\\fcharset0 Courier New;}}\n";
	s << "{\\*\\htmltag243 ";

	// do the wrapping	
	int i = 0;
	while (i < iLen) {
		if (lpC[i] == '{') {
			s << "\\{";
		} else if (lpC[i] == '}') {
			s << "\\}";
		} else if (lpC[i] == '\\') {
			s << "\\\\";
		} else if (strncmp(&lpC[i],"\r\n", 2) == 0) {
			s << "\\par\n";
			i++;
		} else if (strncmp(&lpC[i],"\r", 1) == 0) {
		} else if (strncmp(&lpC[i],"\n", 1) == 0) {
			s << "\\par\n";
		} else if (strncmp(&lpC[i],"\t", 1) == 0) {
			s << "\\tab ";
		} else if ((unsigned char)lpC[i] < 32 || (unsigned char)lpC[i] > 127) {
			snprintf(szHex,3,"%X", ((unsigned char *)lpC)[i]);
			s << "\\'";
			s << szHex;
		} else {
		// or just add whats in input
			s << lpC[i];
		}
		// increase counter
		i++;
	}
	// end rtf
	s << "}\n}\n";

	lpStrBuffOut->assign(s.str());
	
	return 0;
}

HRESULT HrExtractBODYFromTextRTF(std::string &lpStrRTFIn, std::string &strBodyOut)
{
	HRESULT hr = hrSuccess;
	char *szInput = (char *)lpStrRTFIn.c_str();
	char *szANSICharset = "windows-1252";
	int ulState = 0;
	RTFSTATE sState[RTF_MAXSTATE];	
	RTFFONT sFont[RTF_MAXFONT];
	
    strBodyOut.resize(0,0);
   
	memset(sFont, 0, sizeof(sFont));
    
	sState[0].bInSkipTbl = false;
	sState[0].bInFontTbl = false;
	sState[0].bInColorTbl = false;
	sState[0].szCharset = "iso-8859-1";
	sState[0].bRTFOnly = false;
	sState[0].ulFont = 0;

	while(*szInput) {
		if(strncmp(szInput,"\\*",2) == 0) {
			szInput+=2;
		} else if(*szInput == '\\') {
			// Command
			char szCommand[RTF_MAXCMD];
			char *szCmdOutput;
			int lArg = -1;
			bool bNeg = false;
			
			szInput++;
			if(isalpha(*szInput)) {
				szCmdOutput = szCommand;
				
				while(isalpha(*szInput) && szCmdOutput < szCommand+RTF_MAXCMD-1) *szCmdOutput++ = *szInput++;
				*szCmdOutput = 0;
				
				if(*szInput == '-') {
					bNeg = true;
					szInput++;
				}
				
				if(isdigit(*szInput)) {
					lArg = 0;
					while(isdigit(*szInput)) { lArg = lArg*10 + *szInput - '0'; szInput++; }
					if(bNeg) lArg = -lArg;
				}
				if(*szInput == ' ')
					szInput++;
				
				// szCommand is the command, lArg is the argument.
				if(strcmp(szCommand,"fonttbl") == 0) {
					sState[ulState].bInFontTbl = true;
				} else if(strcmp(szCommand,"colortbl") == 0) {
					sState[ulState].bInColorTbl = true;
				} else if(strcmp(szCommand,"pntext") == 0) { // pntext is the plaintext alternative, ignore it.
					sState[ulState].bRTFOnly = true;
				} else if(strcmp(szCommand,"ansicpg") == 0) {
					if(HrGetCharsetByCP(lArg, &szANSICharset) != hrSuccess)
						szANSICharset = "iso-8859-1";
				} else if(strcmp(szCommand,"fcharset") == 0) {
					if(sState[ulState].bInFontTbl) {
						sFont[sState[ulState].ulFont].ulCharset = lArg;
					}
				} else if(strcmp(szCommand,"htmltag") == 0) {
				} else if(strcmp(szCommand,"mhtmltag") == 0) {
				} else if(strcmp(szCommand,"par") == 0 || strcmp(szCommand,"line") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append("\r\n", 2);
					}
				} else if(strcmp(szCommand,"tab") == 0) {
					if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
						sState[ulState].output.append("\t", 1);
					}
				} else if(strcmp(szCommand,"f") == 0) {
					if(lArg >= RTF_MAXFONT)
						continue;
					
					sState[ulState].ulFont = lArg;
					
					if(!sState[ulState].bInFontTbl) {
						// Output any data before this point
						strBodyOut.append(sState[ulState].output);

						sState[ulState].output = "";

						// Set new charset			
						HrGetCharsetByRTFID(sFont[lArg].ulCharset, &sState[ulState].szCharset);
						if(sState[ulState].szCharset == NULL) {
							sState[ulState].szCharset = "windows-1252";
						} else if(sState[ulState].szCharset[0] == 0) {
							sState[ulState].szCharset = szANSICharset;
						}
						
					} 
					// ignore error
				}
				else if(strcmp(szCommand,"htmlrtf") == 0) {
					if(lArg != 0) {
						// \\htmlrtf
						sState[ulState].bRTFOnly = true;
					} else {
						// \\htmlrtf0
						sState[ulState].bRTFOnly = false;
					}
				}
				else if(strcmp(szCommand,"generator") == 0){
					 while(*szInput != ';' && *szInput) szInput++;

					 if(*szInput) 
						 szInput++;
				}
				else if(strcmp(szCommand,"bkmkstart") == 0 || strcmp(szCommand,"bkmkend") == 0){
					// skip bookmark name
					while (*szInput && isalnum(*szInput)) szInput++;

					sState[ulState].bInSkipTbl = true;
				} else if(isRTFIgnoreCommand(szCommand)) {
					sState[ulState].bInSkipTbl = true;
				}				
				
			}
			// Non-alnum after '\'
			else if(*szInput == '\\') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'\\');
			}
			else if(*szInput == '{') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'{');
			}
			else if(*szInput == '}') {
				szInput++;
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) 
					sState[ulState].output.append(1,'}');
			} 
			else if(*szInput == '\'') {
				/* Apparently, escaped data is back in the charset0 charset (ansicpg)
				 * context. This is rather strange, but seems to be the way!
				 */
				unsigned int ulChar = 0;
				char szHex[] = "0123456789ABCDEF";
				
				// Dump output data until now
				strBodyOut.append(sState[ulState].output);

				sState[ulState].output = "";
				
				szInput++;
				
				if(*szInput) {
					ulChar = (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
					ulChar = ulChar << 4;
					szInput++;
				}
				
				if(*szInput) {
					ulChar += (unsigned int) (strchr(szHex, toupper(*szInput)) == NULL ? 0 : (strchr(szHex, toupper(*szInput)) - szHex));
					szInput++;
				}
				
				if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
					if(ulChar > 255)
						sState[ulState].output.append(1, '?');
					else
						sState[ulState].output.append(1, ulChar);
				}
				
				// Dump output
				strBodyOut.append(sState[ulState].output);
				sState[ulState].output = "";
				
			} else {
				szInput++; // skip single character after '\'
			}
		} // Non-command
		else if(*szInput == '{') {
			// Dump output data
			strBodyOut.append(sState[ulState].output);
			
			sState[ulState].output = "";
			
			ulState++;
			if(ulState >= RTF_MAXSTATE) {
				hr = MAPI_E_NOT_ENOUGH_MEMORY;
				goto exit;
			}
			sState[ulState] = sState[ulState-1];
			sState[ulState].output = "";
			szInput++;
		} else if(*szInput == '}') {
			// Dump output data
			strBodyOut.append(sState[ulState].output);
		
			if(ulState > 0)
				ulState--;
			szInput++;
		} else if(*szInput == '\r' || *szInput == '\n') {
			szInput++;
		} else {
			if(!sState[ulState].bInFontTbl && !sState[ulState].bRTFOnly && !sState[ulState].bInColorTbl && !sState[ulState].bInSkipTbl) {		
				sState[ulState].output.append(1,*szInput);
			}
			szInput++;
		}
	}
	
	if(sState[ulState].output.length() > 0) {
		strBodyOut.append(sState[ulState].output);
	}		
	
	
exit:
	return hr;
}
