/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <algorithm>

#include <mapi.h>
#include <mapix.h>
#include <mapicode.h>
#include <mapidefs.h>
#include <mapiutil.h>

#include <CommonUtil.h>
#include <ECTags.h>
#include <inetmapi.h>
#include <mapiext.h>

#include "POP3.h"
using namespace std;

POP3::POP3(ECChannel *lpChan, char *szServerPath, ECLogger *lpLog, ECConfig *lpConf) {
	lpChannel = lpChan;
	lpLogger = lpLog;
	lpConfig = lpConf;

	lpSession = NULL;
	lpStore = NULL;
	lpInbox = NULL;
	lpAddrBook = NULL;

	szPath = szServerPath;
}

POP3::~POP3() {
	for (vector<MailListItem>::iterator i = lstMails.begin(); i != lstMails.end(); i++) {
		delete [] i->sbEntryID.lpb;
	}

	if (lpInbox)
		lpInbox->Release();

	if (lpStore)
		lpStore->Release();

	if (lpSession)
		lpSession->Release();

	if (lpAddrBook)
		lpAddrBook->Release();
}

bool POP3::isLoggedIn() {
	return (lpStore != NULL);
}

// Convert string to all uppercase
void POP3::ToUpper(string &strString) {
	transform(strString.begin(), strString.end(), strString.begin(), ::toupper);
}

HRESULT POP3::HrSplitInput(string &strInput, vector<string> &lstWords) {
	HRESULT hr = hrSuccess;
	string::size_type beginPos = 0;
	string::size_type findPos;

	while ( (findPos = strInput.find(" ", beginPos)) != string::npos) {
		lstWords.push_back(strInput.substr(beginPos, findPos - beginPos));
		beginPos = findPos + 1;
	}
	if (beginPos < strInput.size()) {
		lstWords.push_back(strInput.substr(beginPos));
	}

	return hr;
}


// Parse POP3 Command
HRESULT POP3::HrGetCommand(string &strCommand, POP3_Command &eCommand) {
	HRESULT hr = hrSuccess;

	ToUpper(strCommand);
	if (strCommand.compare("USER") == 0) {
		eCommand = POP3_CMD_USER;
	} else if (strCommand.compare("PASS") == 0) {
		eCommand = POP3_CMD_PASS;
	} else if (strCommand.compare("STAT") == 0) {
		eCommand = POP3_CMD_STAT;
	} else if (strCommand.compare("LIST") == 0) {
		eCommand = POP3_CMD_LIST;
	} else if (strCommand.compare("RETR") == 0) {
		eCommand = POP3_CMD_RETR;
	} else if (strCommand.compare("DELE") == 0) {
		eCommand = POP3_CMD_DELE;
	} else if (strCommand.compare("NOOP") == 0) {
		eCommand = POP3_CMD_NOOP;
	} else if (strCommand.compare("RSET") == 0) {
		eCommand = POP3_CMD_RSET;
	} else if (strCommand.compare("TOP") == 0) {
		eCommand = POP3_CMD_TOP;
	} else if (strCommand.compare("UIDL") == 0) {
		eCommand = POP3_CMD_UIDL;
	} else if (strCommand.compare("QUIT") == 0) {
		eCommand = POP3_CMD_QUIT;
	} else {
		hr = MAPI_E_CALL_FAILED;
	}
	return hr;
}

// Server response +OK ...
HRESULT POP3::HrResponseOk(string strResponse) {
	string strRe("+OK ");
	strRe += strResponse;
	return lpChannel->HrWriteLine(strRe);
}

// Server response -ERR ...
HRESULT POP3::HrResponseErr(string strResponse) {
	string strRe("-ERR ");
	lpLogger->Log(EC_LOGLEVEL_DEBUG, "-ERR %s", strResponse.c_str());
	strRe += strResponse;
	return lpChannel->HrWriteLine(strRe);
}

HRESULT POP3::HrCmdUser(string &strUser) {
	if (lpStore != NULL) {
		HrResponseErr("Can't login twice");
	} else if (strUser.length() > POP3_MAX_RESPONSE_LENGTH) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Username too long: %d > %d", (int)strUser.length(), POP3_MAX_RESPONSE_LENGTH);
		HrResponseErr("Username to long");
	} else {
		szUser = strUser;
		HrResponseOk("Waiting for password");
	}
	return hrSuccess;
}

HRESULT POP3::HrCmdPass(string &strPass) {
	HRESULT hr = hrSuccess;

	if (lpStore != NULL) {
		HrResponseErr("Can't login twice");
	} else if (strPass.length() > POP3_MAX_RESPONSE_LENGTH) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Password too long: %d > %d", (int)strPass.length(), POP3_MAX_RESPONSE_LENGTH);
		HrResponseErr("Password to long");
	} else if (szUser.length() == 0) {
		HrResponseErr("Give username first");
	} else {
		szPass = strPass;

		hr = this->HrLogin();
		if (hr != hrSuccess) {
			HrResponseErr("Wrong username or password");
			goto exit;
		}

		hr = this->HrMakeMailList();
		if (hr != hrSuccess) {
			HrResponseErr("Can't get mail list");
			goto exit;
		}

		HrResponseOk("Username and password accepted");
	}

exit:
	return hr;
}

HRESULT POP3::HrCmdStat() {
	HRESULT hr = hrSuccess;
	ULONG ulSize = 0;

	for (size_t i = 0; i < lstMails.size(); i++) {
		ulSize += lstMails[i].ulSize;
	}

	char szResponse[POP3_MAX_RESPONSE_LENGTH];

	snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u %u", (ULONG)lstMails.size(), ulSize);
	HrResponseOk(szResponse);

	return hr;
}

HRESULT POP3::HrCmdList() {
	HRESULT hr = hrSuccess;
	char szResponse[POP3_MAX_RESPONSE_LENGTH];

	snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u messages", (ULONG)lstMails.size());
	HrResponseOk(szResponse);

	for (size_t i = 0; i < lstMails.size(); i++) {
		snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u %u", (ULONG)i + 1, lstMails[i].ulSize);
		lpChannel->HrWriteLine(szResponse);
	}

	lpChannel->HrWriteLine(".");

	return hr;
}

HRESULT POP3::HrCmdList(unsigned int ulMailNr) {
	HRESULT hr = hrSuccess;
	char szResponse[POP3_MAX_RESPONSE_LENGTH];

	if (ulMailNr > lstMails.size() || ulMailNr < 1) {
		HrResponseErr("Wrong mail number");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u %u", ulMailNr, lstMails[ulMailNr - 1].ulSize);
	HrResponseOk(szResponse);

exit:
	return hr;
}

HRESULT POP3::HrCmdRetr(unsigned int ulMailNr) {
	HRESULT hr = hrSuccess;
	sending_options sopt;

	imopt_default_sending_options(&sopt);
	sopt.no_recipients_workaround = true;	// do not stop processing mail on empty recipient table
	sopt.add_received_date = true;			// add Received header (outlook uses this)

	LPMESSAGE lpMessage = NULL;
	ULONG ulObjType;
	string strMessage;
	char *szMessage = NULL;
	char szResponse[POP3_MAX_RESPONSE_LENGTH];

	if (ulMailNr < 1 || ulMailNr > lstMails.size()) {
		HrResponseErr("mail nr not found");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpStore->OpenEntry(lstMails[ulMailNr - 1].sbEntryID.cb, (LPENTRYID) lstMails[ulMailNr - 1].sbEntryID.lpb, &IID_IMessage, 0,
							&ulObjType, (LPUNKNOWN *) &lpMessage);
	if (hr != hrSuccess) {
		HrResponseErr("Failing to open entry");
		goto exit;
	}

	hr = IMToINet(lpSession, lpAddrBook, lpMessage, &szMessage, sopt, lpLogger);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error converting MAPI to MIME");
		HrResponseErr("Converting MAPI to MIME error");
		goto exit;
	}

	strMessage = DotFilter(szMessage);

	snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u octets", (ULONG)strMessage.length());
	HrResponseOk(szResponse);

	lpChannel->HrWriteLine(strMessage);
	lpChannel->HrWriteLine(".");

exit:
	if (lpMessage)
		lpMessage->Release();

	if (szMessage)
		delete [] szMessage;

	return hr;
}

HRESULT POP3::HrCmdDele(unsigned int ulMailNr) {
	HRESULT hr = hrSuccess;

	if (ulMailNr < 1 || ulMailNr > lstMails.size()) {
		HrResponseErr("mail nr not found");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	lstMails[ulMailNr - 1].bDeleted = true;
	HrResponseOk("mail deleted");

exit:
	return hr;
}

HRESULT POP3::HrCmdNoop() {
	//HRESULT hr = hrSuccess;
	HrResponseOk("");
//exit:
	return hrSuccess;
}

HRESULT POP3::HrCmdRset() {
	HRESULT hr = hrSuccess;

	for (vector<MailListItem>::iterator i = lstMails.begin(); i != lstMails.end(); i++) {
		i->bDeleted = false;
	}

	HrResponseOk("Undeleted mails");

	return hr;
}

HRESULT POP3::HrCmdQuit() {
	HRESULT hr = hrSuccess;
	unsigned int DeleteCount = 0;
	SBinaryArray ba;
	vector<MailListItem>::iterator i;

	ba.lpbin = NULL;

	for (i = lstMails.begin(); i != lstMails.end(); i++) {
		if (i->bDeleted) {
			DeleteCount++;
		}
	}
	if (DeleteCount) {
		ba.cValues = DeleteCount;
		ba.lpbin = new SBinary[DeleteCount];
		DeleteCount = 0;

		for (i = lstMails.begin(); i != lstMails.end(); i++) {
			if (i->bDeleted) {
				ba.lpbin[DeleteCount] = i->sbEntryID;
				DeleteCount++;
			}
		}

		hr = lpInbox->DeleteMessages(&ba, 0, NULL, 0);
	}

	HrResponseOk("Bye");

	if (ba.lpbin)
		delete [] ba.lpbin;

	return hr;
}

HRESULT POP3::HrCmdUidl() {
	HRESULT hr = hrSuccess;
	char szResponse[POP3_MAX_RESPONSE_LENGTH];
	string strResponse;

	lpChannel->HrWriteLine("+OK");

	for (size_t i = 0; i < lstMails.size(); i++) {
		snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u ", (ULONG)i + 1);
		strResponse = szResponse;
		HrHexFromBin(lstMails[i].sbEntryID.lpb, lstMails[i].sbEntryID.cb, strResponse);
		lpChannel->HrWriteLine(strResponse);
	}

	lpChannel->HrWriteLine(".");

	return hr;
}

HRESULT POP3::HrCmdUidl(unsigned int ulMailNr) {
	HRESULT hr = hrSuccess;
	string strResponse;
	char szResponse[POP3_MAX_RESPONSE_LENGTH];

	if (ulMailNr < 1 || ulMailNr > lstMails.size()) {
		HrResponseErr("mail nr not found");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	snprintf(szResponse, POP3_MAX_RESPONSE_LENGTH, "%u ", ulMailNr);
	strResponse = szResponse;
	HrHexFromBin(lstMails[ulMailNr - 1].sbEntryID.lpb, lstMails[ulMailNr - 1].sbEntryID.cb, strResponse);
	HrResponseOk(strResponse);

exit:
	return hr;
}

HRESULT POP3::HrCmdTop(unsigned int ulMailNr, unsigned int ulLines) {
	HRESULT hr = hrSuccess;
	LPMESSAGE lpMessage = NULL;
	ULONG ulObjType;
	char *szMessage = NULL;
	sending_options sopt;
	string strMessage;
	string::size_type ulPos;
	ULONG ulLineCount = 0;

	imopt_default_sending_options(&sopt);

	if (ulMailNr < 1 || ulMailNr > lstMails.size()) {
		HrResponseErr("mail nr not found");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpStore->OpenEntry(lstMails[ulMailNr - 1].sbEntryID.cb, (LPENTRYID) lstMails[ulMailNr - 1].sbEntryID.lpb, &IID_IMessage, 0,
							&ulObjType, (LPUNKNOWN *) &lpMessage);
	if (hr != hrSuccess) {
		HrResponseErr("Failing to open entry");
		goto exit;
	}

	hr = IMToINet(lpSession, lpAddrBook, lpMessage, &szMessage, sopt, lpLogger);
	if (hr != hrSuccess) {
		HrResponseErr("Converting MAPI to MIME error");
		goto exit;
	}

	strMessage = szMessage;
	ulPos = strMessage.find("\r\n\r\n", 0);

	while (ulPos != string::npos && ulLineCount <= ulLines) {
		ulPos = strMessage.find("\r\n", ulPos + 1);
		ulLineCount++;
	}

	if (ulPos != string::npos) {
		strMessage = strMessage.substr(0, ulPos);
	}

	strMessage = DotFilter(strMessage.c_str());

	HrResponseOk("");
	lpChannel->HrWriteLine(strMessage);
	lpChannel->HrWriteLine(".");

exit:
	if (lpMessage)
		lpMessage->Release();

	if (szMessage)
		delete [] szMessage;

	return hr;
}

HRESULT POP3::HrLogin() {
	HRESULT hr = hrSuccess;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;
	ULONG ulObjType = 0;

	hr = HrOpenECSession(&lpSession, (char*)szUser.c_str(), (char*)szPass.c_str(), (char*)szPath.c_str(),
						 EC_PROFILE_FLAGS_NO_NOTIFICATIONS, NULL, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to login from %s with invalid username \"%s\" or wrong password. Error: 0x%X",
					  lpChannel->GetIPAddress().c_str(), szUser.c_str(), hr);
		goto exit;
	}

	hr = HrOpenDefaultStore(lpSession, &lpStore);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open default store");
		goto exit;
	}

	hr = lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &lpAddrBook);
	if (hr != hrSuccess)
		lpAddrBook = NULL;

	hr = lpStore->GetReceiveFolder("IPM", 0, &cbEntryID, &lpEntryID, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to find receive folder of store");
		goto exit;
	}

	hr = lpStore->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType, (LPUNKNOWN *) &lpInbox);
	if (ulObjType != MAPI_FOLDER)
		hr = MAPI_E_NOT_FOUND;

	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open receive folder");
		goto exit;
	}

	lpLogger->Log(EC_LOGLEVEL_ERROR, "POP3 Login from %s for user %s", lpChannel->GetIPAddress().c_str(), szUser.c_str());

exit:
	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	if (hr != hrSuccess) {
		if (lpInbox) {
			lpInbox->Release();
			lpInbox = NULL;
		}
		if (lpStore) {
			lpStore->Release();
			lpStore = NULL;
		}
	}

	return hr;
}

HRESULT POP3::HrMakeMailList() {
	HRESULT hr = hrSuccess;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	MailListItem sMailListItem;
	enum { EID, SIZE, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID, PR_MESSAGE_SIZE} };
	const static SizedSSortOrderSet(1, tableSort) =
		{ 1, 0, 0,
		  {
			  { PR_CREATION_TIME, TABLE_SORT_ASCEND }
		  }
		};

	hr = lpInbox->GetContentsTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray) &spt, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SortTable((LPSSortOrderSet)&tableSort, 0);
	if (hr != hrSuccess) {
		goto exit;
	}

	hr = lpTable->QueryRows(-1, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	lstMails.clear();
	for (ULONG i = 0; i < lpRows->cRows; i++) {
		if (PROP_TYPE(lpRows->aRow[i].lpProps[EID].ulPropTag) == PT_ERROR) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Missing EntryID in message table for message %d", i);
			continue;
		}

		if (PROP_TYPE(lpRows->aRow[i].lpProps[SIZE].ulPropTag) == PT_ERROR) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Missing size in message table for message %d", i);
			continue;
		}

		sMailListItem.sbEntryID.cb = lpRows->aRow[i].lpProps[EID].Value.bin.cb;
		sMailListItem.sbEntryID.lpb = new BYTE[lpRows->aRow[i].lpProps[EID].Value.bin.cb];
		memcpy(sMailListItem.sbEntryID.lpb, lpRows->aRow[i].lpProps[EID].Value.bin.lpb, lpRows->aRow[i].lpProps[EID].Value.bin.cb);
		sMailListItem.bDeleted = false;
		sMailListItem.ulSize = lpRows->aRow[i].lpProps[SIZE].Value.l;
		lstMails.push_back(sMailListItem);
	}

exit:
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	return hr;
}

HRESULT POP3::HrHexFromBin(LPBYTE lpBin, ULONG ulSize, string &strHex) {
	//HRESULT hr = hrSuccess;
	char szBuffer[5];

	for (ULONG i = 0; i < ulSize; i++) {
		snprintf(szBuffer, 5, "%02X", lpBin[i]);
		strHex += szBuffer;
	}

//exit:
	return hrSuccess;
}

string POP3::DotFilter(const char *input) {
	string output;
	ULONG i = 0;

	while (input[i] != '\0') {
		if (input[i] == '.' && input[i-1] == '\n')
			output += '.';
		output += input[i++];
	}
	return output;
}
