/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// based on htmlTextPart, but with additions
// we cannot use a class derived from htmlTextPart, since that class has alot of privates

#include "mapiTextPart.h"
#include "vmime/exception.hpp"

#include "vmime/emptyContentHandler.hpp"
#include "vmime/stringContentHandler.hpp"


namespace vmime
{


mapiTextPart::mapiTextPart()
	: m_plainText(new emptyContentHandler),
	  m_text(new emptyContentHandler),
	  m_otherText(new emptyContentHandler)
{
	m_bHaveOtherCharset = false;
}


mapiTextPart::~mapiTextPart()
{
    free_container (m_objects);
    
	delete (m_plainText);
	delete (m_text);
	delete (m_otherText);
}


const mediaType mapiTextPart::getType() const
{
	// TODO: fixme?
	if (m_text->isEmpty())
		return (mediaType(mediaTypes::TEXT, mediaTypes::TEXT_PLAIN));
	else
		return (mediaType(mediaTypes::TEXT, mediaTypes::TEXT_HTML));
}


const int mapiTextPart::getPartCount() const
{
	int count = 0;
	if (!m_plainText->isEmpty())
		count++;
	if (!m_text->isEmpty())
		count++;
	if (!m_otherText->isEmpty())
		count++;
	return count;
}


void mapiTextPart::generateIn(bodyPart& /* message */, bodyPart& parent) const
{
	// Plain text
	if (!m_plainText->isEmpty())
	{
		// -- Create a new part
		bodyPart* plainPart = new bodyPart();
		parent.getBody()->appendPart(plainPart);

		// -- Set header fields
		plainPart->getHeader()->ContentType().setValue(mediaType(mediaTypes::TEXT, mediaTypes::TEXT_PLAIN));
		plainPart->getHeader()->ContentType().setCharset(m_charset);
		plainPart->getHeader()->ContentTransferEncoding().setValue(encoding(encodingTypes::QUOTED_PRINTABLE));

		// -- Set contents
		plainPart->getBody()->setContents(*m_plainText);
	}

	// HTML text
	// -- Create a new part
	if (!m_text->isEmpty())
	{
		bodyPart* htmlPart = new bodyPart();

		// -- Set header fields
		htmlPart->getHeader()->ContentType().setValue(mediaType(mediaTypes::TEXT, mediaTypes::TEXT_HTML));
		htmlPart->getHeader()->ContentType().setCharset(m_charset);
		htmlPart->getHeader()->ContentTransferEncoding().setValue(encoding(encodingTypes::QUOTED_PRINTABLE));

		// -- Set contents
		htmlPart->getBody()->setContents(*m_text);

		// Handle the case we have embedded objects
		if (!m_objects.empty())
		{
			// Create a "multipart/related" body part
			bodyPart* relPart = new bodyPart();
			parent.getBody()->appendPart(relPart);

			relPart->getHeader()->ContentType().
				setValue(mediaType(mediaTypes::MULTIPART, mediaTypes::MULTIPART_RELATED));

			// Add the HTML part into this part
			relPart->getBody()->appendPart(htmlPart);

			// Also add objects into this part
			for (std::vector <embeddedObject*>::const_iterator it = m_objects.begin() ;
				 it != m_objects.end() ; ++it)
			{
				bodyPart* objPart = new bodyPart();
				relPart->getBody()->appendPart(objPart);

				string id = (*it)->getId();
				string name = (*it)->getName();

				if (id.substr(0, 4) == "CID:")
					id = id.substr(4);

				// throw an error when id and location are empty?

				objPart->getHeader()->ContentType().setValue((*it)->getType());
				if (!id.empty())
					objPart->getHeader()->ContentId().setValue(messageId("<" + id + ">"));
				objPart->getHeader()->ContentDisposition().setValue(contentDisposition(contentDispositionTypes::INLINE));
				objPart->getHeader()->ContentTransferEncoding().setValue((*it)->getEncoding());
				if (!(*it)->getLocation().empty())
					objPart->getHeader()->ContentLocation().setValue((*it)->getLocation());
				//encoding(encodingTypes::BASE64);

				if (!name.empty())
					objPart->getHeader()->ContentDisposition().setFilename(name);

				objPart->getBody()->setContents((*it)->getData());
			}
		}
		else
		{
			// Add the HTML part into the parent part
			parent.getBody()->appendPart(htmlPart);
		}
	}

	// Other text
	if (!m_otherText->isEmpty())
	{
		// -- Create a new part
		bodyPart* otherPart = new bodyPart();
		parent.getBody()->appendPart(otherPart);

		// -- Set header fields
		otherPart->getHeader()->ContentType().setValue(m_otherMediaType);
		if (m_bHaveOtherCharset)
			otherPart->getHeader()->ContentType().setCharset(m_otherCharset);
		else
			otherPart->getHeader()->ContentType().setCharset(m_charset);
		otherPart->getHeader()->ContentTransferEncoding().setValue(m_otherEncoding);

		// used by ical
		if (!m_otherMethod.empty()) {
			vmime::parameter *p = vmime::parameterFactory::getInstance()->create("method");
			p->parse(m_otherMethod);
			otherPart->getHeader()->ContentType().appendParameter(p);
		}

		// -- Set contents
		otherPart->getBody()->setContents(*m_otherText);
	}

}


void mapiTextPart::findEmbeddedParts(const bodyPart& part,
	std::vector <const bodyPart*>& cidParts, std::vector <const bodyPart*>& locParts)
{
	for (int i = 0 ; i < part.getBody()->getPartCount() ; ++i)
	{
		const bodyPart& p = *part.getBody()->getPartAt(i);

		try
		{
			dynamic_cast<messageIdField&>(*p.getHeader()->findField(fields::CONTENT_ID));
			cidParts.push_back(&p);
		}
		catch (exceptions::no_such_field)
		{
			// No "Content-id" field.
			// Maybe there is a "Content-Location" field...
			try
			{
				dynamic_cast<messageIdField&>(*p.getHeader()->findField(fields::CONTENT_ID));
				locParts.push_back(&p);
			}
			catch (exceptions::no_such_field)
			{
				// No "Content-Location" field.
				// Cannot be an embedded object since it cannot be referenced in HTML text.
			}
		}

		findEmbeddedParts(p, cidParts, locParts);
	}
}


void mapiTextPart::addEmbeddedObject(const bodyPart& part, const string& id)
{
	mediaType type;

	try
	{
		const contentTypeField& ctf = dynamic_cast<contentTypeField&>
			(*part.getHeader()->findField(fields::CONTENT_TYPE));

		type = ctf.getValue();
	}
	catch (exceptions::no_such_field)
	{
		// No "Content-type" field: assume "application/octet-stream".
	}

	m_objects.push_back(new embeddedObject
						(part.getBody()->getContents(), part.getBody()->getEncoding(), id, type, string(), string()));
}


void mapiTextPart::parse(const bodyPart& message, const bodyPart& parent, const bodyPart& textPart)
{
	// Search for possible embedded objects in the _whole_ message.
	std::vector <const bodyPart*> cidParts;
	std::vector <const bodyPart*> locParts;

	findEmbeddedParts(message, cidParts, locParts);

	// Extract HTML text
	std::ostringstream oss;
	utility::outputStreamAdapter adapter(oss);

	textPart.getBody()->getContents().extract(adapter);

	const string data = oss.str();

	delete (m_text);
	m_text = textPart.getBody()->getContents().clone();

	try
	{
		const contentTypeField& ctf = dynamic_cast<contentTypeField&>
			(*textPart.getHeader()->findField(fields::CONTENT_TYPE));

		m_charset = ctf.getCharset();
	}
	catch (exceptions::no_such_field)
	{
		// No "Content-type" field.
	}
	catch (exceptions::no_such_parameter)
	{
		// No "charset" parameter.
	}

	// Extract embedded objects. The algorithm is quite simple: for each previously
	// found inline part, we check if its CID/Location is contained in the HTML text.
	for (std::vector <const bodyPart*>::const_iterator p = cidParts.begin() ; p != cidParts.end() ; ++p)
	{
		const messageIdField& midField = dynamic_cast<messageIdField&>
			(*(*p)->getHeader()->findField(fields::CONTENT_ID));

		const string searchFor("CID:" + midField.getValue().getId());

		if (data.find(searchFor) != string::npos)
		{
			// This part is referenced in the HTML text.
			// Add it to the embedded object list.
			addEmbeddedObject(**p, "CID:" + midField.getValue().getId());
		}
	}

	for (std::vector <const bodyPart*>::const_iterator p = locParts.begin() ; p != locParts.end() ; ++p)
	{
		const defaultField& locField = dynamic_cast<defaultField&>
			(*(*p)->getHeader()->findField(fields::CONTENT_LOCATION));

		if (data.find(locField.getValue()) != string::npos)
		{
			// This part is referenced in the HTML text.
			// Add it to the embedded object list.
			addEmbeddedObject(**p, locField.getValue());
		}
	}

	// Extract plain text, if any.
	if (!findPlainTextPart(message, parent, textPart))
	{
		delete (m_plainText);
		m_plainText = new emptyContentHandler();
	}
}


bool mapiTextPart::findPlainTextPart(const bodyPart& part, const bodyPart& parent, const bodyPart& textPart)
{
	// We search for the nearest "multipart/alternative" part.
	try
	{
		const contentTypeField& ctf = dynamic_cast<contentTypeField&>
			(*part.getHeader()->findField(fields::CONTENT_TYPE));

		if (ctf.getValue().getType() == mediaTypes::MULTIPART &&
		    ctf.getValue().getSubType() == mediaTypes::MULTIPART_ALTERNATIVE)
		{
			bodyPart const* foundPart = NULL;

			for (int i = 0 ; i < part.getBody()->getPartCount() ; ++i)
			{
				const bodyPart* p = part.getBody()->getPartAt(i);

				if (p == &parent ||     // if "text/html" is in "multipart/related"
				    p == &textPart)     // if not...
				{
					foundPart = &(*p);
				}
			}

			if (foundPart)
			{
				bool found = false;

				// Now, search for the alternative plain text part
				for (int i = 0 ; !found && i < part.getBody()->getPartCount() ; ++i)
				{
					const bodyPart& p = *part.getBody()->getPartAt(i);

					try
					{
						const contentTypeField& ctf = dynamic_cast<contentTypeField&>
							(*p.getHeader()->findField(fields::CONTENT_TYPE));

						if (ctf.getValue().getType() == mediaTypes::TEXT &&
						    ctf.getValue().getSubType() == mediaTypes::TEXT_PLAIN)
						{
							delete (m_plainText);
							m_plainText = p.getBody()->getContents().clone();
							found = true;
						}
					}
					catch (exceptions::no_such_field)
					{
						// No "Content-type" field.
					}
				}

				// If we don't have found the plain text part here, it means that
				// it does not exists (the MUA which built this message probably
				// did not include it...).
				return (found);
			}
		}
	}
	catch (exceptions::no_such_field)
	{
		// No "Content-type" field.
	}

	bool found = false;

	for (int i = 0 ; !found && i < part.getBody()->getPartCount() ; ++i)
	{
		found = findPlainTextPart(*part.getBody()->getPartAt(i), parent, textPart);
	}

	return (found);
}


const charset& mapiTextPart::getCharset() const
{
	return (m_charset);
}


void mapiTextPart::setCharset(const charset& ch)
{
	m_charset = ch;
}


const contentHandler& mapiTextPart::getPlainText() const
{
	return (*m_plainText);
}


void mapiTextPart::setPlainText(const contentHandler& plainText)
{
	delete (m_plainText);
	m_plainText = plainText.clone();
}


const contentHandler& mapiTextPart::getOtherText() const
{
	return (*m_otherText);
}


void mapiTextPart::setOtherText(const contentHandler& otherText)
{
	delete (m_otherText);
	m_otherText = otherText.clone();
}

void mapiTextPart::setOtherContentType(const mediaType& type)
{
	m_otherMediaType = type;
}

void mapiTextPart::setOtherContentEncoding(const encoding& enc)
{
	m_otherEncoding = enc;
}

void mapiTextPart::setOtherMethod(const string& method)
{
	m_otherMethod = method;
}

void mapiTextPart::setOtherCharset(const charset& ch)
{
	m_otherCharset = ch;
	m_bHaveOtherCharset = true;
}


const contentHandler& mapiTextPart::getText() const
{
	return (*m_text);
}


void mapiTextPart::setText(const contentHandler& text)
{
	delete (m_text);
	m_text = text.clone();
}


const int mapiTextPart::getObjectCount() const
{
	return (m_objects.size());
}


const mapiTextPart::embeddedObject* mapiTextPart::getObjectAt(const int pos) const
{
	return (m_objects[pos]);
}


const mapiTextPart::embeddedObject* mapiTextPart::findObject(const string& id) const
{
	for (std::vector <embeddedObject*>::const_iterator o = m_objects.begin() ;
	     o != m_objects.end() ; ++o)
	{
		if ((*o)->getId() == id)
			return (*o);
	}

	throw exceptions::no_object_found();
}


const bool mapiTextPart::hasObject(const string& id) const
{
	for (std::vector <embeddedObject*>::const_iterator o = m_objects.begin() ;
	     o != m_objects.end() ; ++o)
	{
		if ((*o)->getId() == id)
			return (true);
	}

	return (false);
}


const string mapiTextPart::addObject(const contentHandler& data,
	const vmime::encoding& enc, const mediaType& type)
{
	const messageId mid(messageId::generateId());
	const string id = "CID:" + mid.getId();

	return (addObject(data, enc, type, id));
}


const string mapiTextPart::addObject(const contentHandler& data, const mediaType& type)
{
	return (addObject(data, encoding::decide(data), type));
}


const string mapiTextPart::addObject(const string& data, const mediaType& type)
{
	stringContentHandler cts(data);
	return (addObject(cts, encoding::decide(cts), type));
}

const string mapiTextPart::addObject(const contentHandler& data,
									 const vmime::encoding& enc, const mediaType& type, const string& id, const string& name, const string& loc)
{
	m_objects.push_back(new embeddedObject(data, enc, id, type, name, loc));
	return (id);
}



//
// mapiTextPart::embeddedObject
//

mapiTextPart::embeddedObject::embeddedObject
	(const contentHandler& data, const encoding& enc,
	 const string& id, const mediaType& type, const string& name, const string& loc)
	: m_data(data.clone()), m_encoding(enc), m_id(id), m_type(type), m_name(name), m_loc(loc)
{
}

mapiTextPart::embeddedObject::~embeddedObject()
{
    if(m_data) delete m_data;
}

const contentHandler& mapiTextPart::embeddedObject::getData() const
{
	return (*m_data);
}


const vmime::encoding& mapiTextPart::embeddedObject::getEncoding() const
{
	return (m_encoding);
}


const string& mapiTextPart::embeddedObject::getId() const
{
	return (m_id);
}

const string& mapiTextPart::embeddedObject::getLocation() const
{
	return (m_loc);
}


const mediaType& mapiTextPart::embeddedObject::getType() const
{
	return (m_type);
}

const string& mapiTextPart::embeddedObject::getName() const
{
	return (m_name);
}

} // vmime namespace
