/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapicode.h>
#include <mapidefs.h>
#include <mapitags.h>
#include <mapiguid.h>
#include <mapiutil.h>

#include "Mem.h"
#include "ECMAPITable.h"
#include "guid.h"
#include "ECGuid.h"
#include "Util.h"

#include "ECDebug.h"
#include "ECDebugClient.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

ECMAPITable::ECMAPITable(ECNotifyClient *lpNotifyClient, ULONG ulFlags) : ECUnknown("IMAPITable")
{
	TRACE_MAPI(TRACE_ENTRY, "ECMAPITable::ECMAPITable","");

	this->lpNotifyClient = lpNotifyClient;
	
	if(this->lpNotifyClient)
		this->lpNotifyClient->AddRef();

	this->lpsSortOrderSet = NULL;
	this->lpsPropTags = NULL;
	this->ulFlags = ulFlags;
	this->lpTableOps = NULL;
	
	m_lpSetColumns = NULL;
	m_lpRestrict = NULL;
	m_lpSortTable = NULL;
	m_ulRowCount = 0;
	m_ulFlags = 0;
	m_ulDeferredFlags = 0;

	pthread_mutexattr_t mattr;
	pthread_mutexattr_init(&mattr);
	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&m_hLock, &mattr);
}

HRESULT ECMAPITable::FlushDeferred(LPSRowSet *lppRowSet)
{
    HRESULT hr = hrSuccess;
    
	hr = lpTableOps->HrOpenTable();
	if(hr != hrSuccess)
		goto exit;

    // No deferred calls -> nothing to do
    if(!IsDeferred())
        goto exit;
        
    hr = lpTableOps->HrMulti(m_ulDeferredFlags, m_lpSetColumns, m_lpRestrict, m_lpSortTable, m_ulRowCount, m_ulFlags, lppRowSet);

    // Reset deferred items
    if(m_lpSetColumns)
        MAPIFreeBuffer(m_lpSetColumns);
	m_lpSetColumns = NULL;
	if(m_lpRestrict)
	    MAPIFreeBuffer(m_lpRestrict);
	m_lpRestrict = NULL;
	if(m_lpSortTable)
	    MAPIFreeBuffer(m_lpSortTable);
	m_lpSortTable = NULL;
	m_ulRowCount = 0;
	m_ulFlags = 0;
	m_ulDeferredFlags = 0;

exit: 
    return hr;
}

BOOL ECMAPITable::IsDeferred()
{
    if(m_lpSetColumns == NULL && m_lpRestrict == NULL && m_lpSortTable == NULL && m_ulRowCount == 0 && m_ulFlags == 0 && m_ulDeferredFlags == 0)
        return false;
    return true;
}

ECMAPITable::~ECMAPITable()
{
	TRACE_MAPI(TRACE_ENTRY, "ECMAPITable::~ECMAPITable","");
	std::set<ULONG>::iterator iterMapInt;
	std::set<ULONG>::iterator iterMapIntDel;

	// Remove all advises	
	iterMapInt = m_ulConnectionList.begin();
	while( iterMapInt != m_ulConnectionList.end() )
	{
		iterMapIntDel = iterMapInt;
		iterMapInt++;
		Unadvise(*iterMapIntDel);
	}

	if(lpsPropTags)
		delete [] this->lpsPropTags;

	if(lpNotifyClient)
		lpNotifyClient->Release();

	if(lpTableOps)
		lpTableOps->Release();	// closes the table on the server too

	if(lpsSortOrderSet)
		delete [] lpsSortOrderSet;

	pthread_mutex_destroy(&m_hLock);
}

HRESULT ECMAPITable::Create(ECNotifyClient *lpNotifyClient, ULONG ulFlags, ECMAPITable **lppECMAPITable)
{
	HRESULT hr = hrSuccess;
	ECMAPITable *lpMAPITable = NULL;
	
	lpMAPITable = new ECMAPITable(lpNotifyClient, ulFlags);

	hr = lpMAPITable->QueryInterface(IID_ECMAPITable, (void **)lppECMAPITable);

	return hr;
}

HRESULT ECMAPITable::QueryInterface(REFIID refiid, void **lpvoid)
{
	if(refiid == IID_ECMAPITable || refiid == IID_ECUnknown) {
		AddRef();
		*lpvoid = this;
		return hrSuccess;
	}
	if(refiid == IID_IMAPITable || refiid == IID_IUnknown) {
		AddRef();
		*lpvoid = &this->m_xMAPITable;
		return hrSuccess;
	}
	return MAPI_E_INTERFACE_NOT_SUPPORTED;
}

HRESULT ECMAPITable::GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR *lppMAPIError)
{
	HRESULT hr = hrSuccess;
	
	pthread_mutex_lock(&m_hLock);

	hr = MAPI_E_NO_SUPPORT;

	pthread_mutex_unlock(&m_hLock);

	return hr;
}

HRESULT ECMAPITable::Advise(ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG * lpulConnection)
{
	HRESULT hr = hrSuccess;
	
	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	if(lpNotifyClient == NULL) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	if (lpulConnection == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpNotifyClient->Advise(4, (BYTE *)&lpTableOps->ulTableId, ulEventMask, lpAdviseSink, lpulConnection);
	if(hr != hrSuccess)
		goto exit;

	m_ulConnectionList.insert(*lpulConnection);

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

HRESULT ECMAPITable::Unadvise(ULONG ulConnection)
{
	HRESULT hr = hrSuccess;
	std::set<ULONG>::iterator iterMapInt;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	if(lpNotifyClient == NULL) {
		hr = MAPI_E_NO_SUPPORT;
		goto exit;
	}

	m_ulConnectionList.erase(ulConnection);
	lpNotifyClient->Unadvise(ulConnection);

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

// @fixme Do we need to lock here or just update the status?
HRESULT ECMAPITable::GetStatus(ULONG *lpulTableStatus, ULONG *lpulTableType)
{
	HRESULT hr = hrSuccess;

	*lpulTableStatus = TBLSTAT_COMPLETE;
	*lpulTableType = TBLTYPE_DYNAMIC;

	return hr;
}


HRESULT ECMAPITable::SetColumns(LPSPropTagArray lpPropTagArray, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	if(lpPropTagArray == NULL || lpPropTagArray->cValues == 0)
		return MAPI_E_INVALID_PARAMETER;

	pthread_mutex_lock(&m_hLock);

	if(lpsPropTags)
		delete [] this->lpsPropTags;

	lpsPropTags = (LPSPropTagArray) new BYTE[CbNewSPropTagArray(lpPropTagArray->cValues)];

	lpsPropTags->cValues = lpPropTagArray->cValues;
	memcpy(&lpsPropTags->aulPropTag, &lpPropTagArray->aulPropTag, lpPropTagArray->cValues * sizeof(ULONG));

    if(m_lpSetColumns)
        MAPIFreeBuffer(m_lpSetColumns);
        
    hr = MAPIAllocateBuffer(CbNewSPropTagArray(lpPropTagArray->cValues), (void **)&m_lpSetColumns);
    if(hr != hrSuccess)
        goto exit;
        
    m_lpSetColumns->cValues = lpPropTagArray->cValues;
    memcpy(&m_lpSetColumns->aulPropTag, &lpPropTagArray->aulPropTag, lpPropTagArray->cValues * sizeof(ULONG));

    if(!(ulFlags & TBL_BATCH)) {
        hr = FlushDeferred();
	    if(hr != hrSuccess)
	        goto exit;
    }
    
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

HRESULT ECMAPITable::QueryColumns(ULONG ulFlags, LPSPropTagArray *lppPropTagArray)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	// FIXME if the client has done SetColumns, we can handle this
	// call locally instead of querying the server (unless TBL_ALL_COLUMNS has been
	// specified)

	hr = this->lpTableOps->HrQueryColumns(ulFlags, lppPropTagArray);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

HRESULT ECMAPITable::GetRowCount(ULONG ulFlags, ULONG *lpulCount)
{
	HRESULT hr = hrSuccess;
	ULONG ulRow = 0; // discarded

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = this->lpTableOps->HrGetRowCount(lpulCount, &ulRow);

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::SeekRow(BOOKMARK bkOrigin, LONG lRowCount, LONG *lplRowsSought)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

    hr = FlushDeferred();
    if(hr != hrSuccess)
        goto exit;

    hr = this->lpTableOps->HrSeekRow(bkOrigin, lRowCount, lplRowsSought);

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::SeekRowApprox(ULONG ulNumerator, ULONG ulDenominator)
{
	HRESULT hr = hrSuccess;
	ULONG ulRows = 0;
	ULONG ulCurrent = 0;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrGetRowCount(&ulRows, &ulCurrent);

	if(hr != hrSuccess)
		goto exit;

	hr = SeekRow(BOOKMARK_BEGINNING, (ULONG)((double)ulRows * ((double)ulNumerator / ulDenominator)),NULL);
		
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::QueryPosition(ULONG *lpulRow, ULONG *lpulNumerator, ULONG *lpulDenominator)
{
	HRESULT hr = hrSuccess;
	ULONG ulRows = 0;
	ULONG ulCurrentRow = 0;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrGetRowCount(&ulRows, &ulCurrentRow);

	if(hr != hrSuccess)
		goto exit;

	*lpulRow = ulCurrentRow;
	*lpulNumerator = ulCurrentRow;
	*lpulDenominator = (ulRows == 0)?1:ulRows;

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::FindRow(LPSRestriction lpRestriction, BOOKMARK bkOrigin, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = this->lpTableOps->HrFindRow(lpRestriction, bkOrigin, ulFlags);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::Restrict(LPSRestriction lpRestriction, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

    if(m_lpRestrict)
        MAPIFreeBuffer(m_lpRestrict);
    if(lpRestriction) {
        MAPIAllocateBuffer(sizeof(SRestriction), (void **)&m_lpRestrict);
        
        hr = Util::HrCopySRestriction(m_lpRestrict, lpRestriction, m_lpRestrict);

		m_ulDeferredFlags &= ~TABLE_MULTI_CLEAR_RESTRICTION;
    } else {
        // FIXME setting the restriction to NULL is not the same as not setting the restriction at all
		m_ulDeferredFlags |= TABLE_MULTI_CLEAR_RESTRICTION;
        m_lpRestrict = NULL;
    }

    if(!(ulFlags & TBL_BATCH)) {
	    hr = FlushDeferred();
	    if(hr != hrSuccess)
	        goto exit;
    }

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::CreateBookmark(BOOKMARK* lpbkPosition)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = this->lpTableOps->CreateBookmark(lpbkPosition);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::FreeBookmark(BOOKMARK bkPosition)
{
	HRESULT hr = hrSuccess;
	
	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = this->lpTableOps->FreeBookmark(bkPosition);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::SortTable(LPSSortOrderSet lpSortCriteria, ULONG ulFlags)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	if (!lpSortCriteria) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpsSortOrderSet)
		delete [] lpsSortOrderSet;

	lpsSortOrderSet = (LPSSortOrderSet) new BYTE[CbSSortOrderSet(lpSortCriteria)];

	memcpy(lpsSortOrderSet, lpSortCriteria, CbSSortOrderSet(lpSortCriteria));

    if(m_lpSortTable)
        MAPIFreeBuffer(m_lpSortTable);
    MAPIAllocateBuffer(CbSSortOrderSet(lpSortCriteria), (void **) &m_lpSortTable);
    memcpy(m_lpSortTable, lpSortCriteria, CbSSortOrderSet(lpSortCriteria));

    if(!(ulFlags & TBL_BATCH)) {
	    hr = FlushDeferred();
	    if(hr != hrSuccess)
	        goto exit;
    }

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::QuerySortOrder(LPSSortOrderSet *lppSortCriteria)
{
	HRESULT hr = hrSuccess;
	LPSSortOrderSet lpSortCriteria = NULL;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	if(lpsSortOrderSet)
		hr = ECAllocateBuffer(CbSSortOrderSet(lpsSortOrderSet), (void **) &lpSortCriteria);
	else
		hr = ECAllocateBuffer(CbNewSSortOrderSet(0), (void **) &lpSortCriteria);

	if(hr != hrSuccess)
		goto exit;

	if(lpsSortOrderSet)
		memcpy(lpSortCriteria, lpsSortOrderSet, CbSSortOrderSet(lpsSortOrderSet));
	else
		memset(lpSortCriteria, 0, CbNewSSortOrderSet(0));

	*lppSortCriteria = lpSortCriteria;
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}
 

HRESULT ECMAPITable::Abort()
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;
exit:
	hr = S_OK; // Fixme: sent this call to the server, and breaks the search! // OLK 2007 request

	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::ExpandRow(ULONG cbInstanceKey, LPBYTE pbInstanceKey, ULONG ulRowCount, ULONG ulFlags, LPSRowSet * lppRows, ULONG *lpulMoreRows)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrExpandRow(cbInstanceKey, pbInstanceKey, ulRowCount, ulFlags, lppRows, lpulMoreRows);
exit:
	pthread_mutex_unlock(&m_hLock);
	return hr;
}


HRESULT ECMAPITable::CollapseRow(ULONG cbInstanceKey, LPBYTE pbInstanceKey, ULONG ulFlags, ULONG *lpulRowCount)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrCollapseRow(cbInstanceKey, pbInstanceKey, ulFlags, lpulRowCount);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

// @todo do we need lock here, currently we do. maybe we must return MAPI_E_TIMEOUT
HRESULT ECMAPITable::WaitForCompletion(ULONG ulFlags, ULONG ulTimeout, ULONG *lpulTableStatus)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	if(lpulTableStatus)
		*lpulTableStatus = S_OK;
exit:
	pthread_mutex_unlock(&m_hLock);
	return hr;
}


HRESULT ECMAPITable::GetCollapseState(ULONG ulFlags, ULONG cbInstanceKey, LPBYTE lpbInstanceKey, ULONG *lpcbCollapseState, LPBYTE *lppbCollapseState)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrGetCollapseState(lppbCollapseState,lpcbCollapseState, lpbInstanceKey, cbInstanceKey);
exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}


HRESULT ECMAPITable::SetCollapseState(ULONG ulFlags, ULONG cbCollapseState, LPBYTE pbCollapseState, BOOKMARK *lpbkLocation)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	hr = FlushDeferred();
	if(hr != hrSuccess)
	    goto exit;

	hr = lpTableOps->HrSetCollapseState(pbCollapseState, cbCollapseState, lpbkLocation);

	if(lpbkLocation)
		*lpbkLocation = BOOKMARK_BEGINNING;
exit:
	pthread_mutex_unlock(&m_hLock);
	return hr;
}

HRESULT ECMAPITable::HrSetTableOps(WSTableView *lpTableOps, bool fLoad)
{
	HRESULT hr = hrSuccess;
	SSortOrderSet sSort;

	sSort.cCategories = 0;
	sSort.cExpanded = 0;
	sSort.cSorts = 1;
	sSort.aSort[0].ulOrder = TABLE_SORT_ASCEND;
	sSort.aSort[0].ulPropTag = PR_DISPLAY_NAME;

	this->lpTableOps = lpTableOps;
	lpTableOps->AddRef();

	// Open the table on the server, ready for reading ..
	if(fLoad) {
    	hr = lpTableOps->HrOpenTable();

    	if(hr != hrSuccess)
		    goto exit;
    }

	lpTableOps->SetReloadCallback(Reload, this);

exit:
	return hr;
}

HRESULT ECMAPITable::QueryRows(LONG lRowCount, ULONG ulFlags, LPSRowSet *lppRows)
{
	HRESULT hr = hrSuccess;

	pthread_mutex_lock(&m_hLock);

	if(IsDeferred()) {
	    m_ulRowCount = lRowCount;
	    m_ulFlags = ulFlags;
	    
	    hr = FlushDeferred(lppRows);
    } else {
        
        // Send the request to the TableOps object, which will send the request to the server.
        hr = this->lpTableOps->HrQueryRows(lRowCount, ulFlags, lppRows);
    }
    
	if(hr != hrSuccess)
		goto exit;

exit:
	pthread_mutex_unlock(&m_hLock);

	return hr;
}

HRESULT ECMAPITable::Reload(void *lpParam)
{
	HRESULT hr = hrSuccess;
	ECMAPITable *lpThis = (ECMAPITable *)lpParam;
	std::set<ULONG>::iterator iter;

	pthread_mutex_lock(&lpThis->m_hLock);

	// The underlying data has been reloaded, therefore we must re-register the advises. This is called
	// after the transport has re-established its state
	for(iter = lpThis->m_ulConnectionList.begin(); iter != lpThis->m_ulConnectionList.end(); iter++) {
		hr = lpThis->lpNotifyClient->Reregister(*iter, 4, (BYTE *)&lpThis->lpTableOps->ulTableId);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	pthread_mutex_unlock(&lpThis->m_hLock);

	return hr;
}

ULONG ECMAPITable::xMAPITable::AddRef()
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::AddRef", "table=%d", pThis->lpTableOps->ulTableId);
	return pThis->AddRef();
}

ULONG ECMAPITable::xMAPITable::Release()
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::Release", "table=%d", pThis->lpTableOps->ulTableId);	
	return pThis->Release();
}

HRESULT ECMAPITable::xMAPITable::QueryInterface(REFIID refiid, void **pInterface)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::QueryInterface", "table=%d: %s", pThis->lpTableOps->ulTableId, DBGGUIDToString(refiid).c_str());
	HRESULT hr = pThis->QueryInterface(refiid, pInterface);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::QueryInterface", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::GetLastError(HRESULT hResult, ULONG ulFlags, LPMAPIERROR *lppMAPIError)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::GetLastError", "table=%d", pThis->lpTableOps->ulTableId);
	return pThis->GetLastError(hResult, ulFlags, lppMAPIError);
}

HRESULT ECMAPITable::xMAPITable::Advise(ULONG ulEventMask, LPMAPIADVISESINK lpAdviseSink, ULONG * lpulConnection)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::Advise", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr =  pThis->Advise(ulEventMask, lpAdviseSink, lpulConnection);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::Advise", "table=%d, connectionid=%d", pThis->lpTableOps->ulTableId, (lpulConnection != NULL)?*lpulConnection:-1);
	return hr;
}

HRESULT ECMAPITable::xMAPITable::Unadvise(ULONG ulConnection)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::Unadvise", "table=%d", pThis->lpTableOps->ulTableId);
	return pThis->Unadvise(ulConnection);
}

HRESULT ECMAPITable::xMAPITable::GetStatus(ULONG *lpulTableStatus, ULONG *lpulTableType)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::GetStatus", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->GetStatus(lpulTableStatus, lpulTableType);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::GetStatus", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::SetColumns(LPSPropTagArray lpPropTagArray, ULONG ulFlags)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::SetColumns", "table=%d: flags=0x%08X, proptag=%s", pThis->lpTableOps->ulTableId, ulFlags, PropNameFromPropTagArray(lpPropTagArray).c_str());
	HRESULT hr = pThis->SetColumns(lpPropTagArray, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::SetColumns", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::QueryColumns(ULONG ulFlags, LPSPropTagArray *lpPropTagArray)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::QueryColumns", "table=%d: flags=0x%08X", pThis->lpTableOps->ulTableId, ulFlags);
	HRESULT hr = pThis->QueryColumns(ulFlags, lpPropTagArray);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::QueryColumns", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::GetRowCount(ULONG ulFlags, ULONG *lpulCount)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::GetRowCount", "table=%d: flags=0x%08X", pThis->lpTableOps->ulTableId, ulFlags);
	HRESULT hr = pThis->GetRowCount(ulFlags, lpulCount);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::GetRowCount", "%s, count=%d", GetMAPIErrorDescription(hr).c_str(), (lpulCount)?*lpulCount:0);
	return hr;
}

HRESULT ECMAPITable::xMAPITable::SeekRow(BOOKMARK bkOrigin, LONG lRowCount, LONG *lplRowsSought)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::SeekRow", "table=%d: bkorgin=%d rowcount=%d", pThis->lpTableOps->ulTableId, bkOrigin, lRowCount);
	HRESULT hr = pThis->SeekRow(bkOrigin, lRowCount, lplRowsSought);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::SeekRow", "%s, lplRowsSought = %d", GetMAPIErrorDescription(hr).c_str(), lplRowsSought ? *lplRowsSought : -1);
	return hr;
}

HRESULT ECMAPITable::xMAPITable::SeekRowApprox(ULONG ulNumerator, ULONG ulDenominator)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::SeekRowApprox", "table=%d: Numerator=%d, Denominator=%d", pThis->lpTableOps->ulTableId, ulNumerator, ulDenominator);
	HRESULT hr = pThis->SeekRowApprox(ulNumerator, ulDenominator);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::SeekRowApprox", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::QueryPosition(ULONG *lpulRow, ULONG *lpulNumerator, ULONG *lpulDenominator)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);	
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::QueryPosition", "table=%d: ", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->QueryPosition(lpulRow, lpulNumerator, lpulDenominator);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::QueryPosition","lpulRow = %d, lpulNumerator = %d, lpulDenominator = %d", lpulRow ? *lpulRow : -1, lpulNumerator ? *lpulNumerator : -1, lpulDenominator ? *lpulDenominator : -1);
	return hr;
}

HRESULT ECMAPITable::xMAPITable::FindRow(LPSRestriction lpRestriction, BOOKMARK bkOrigin, ULONG ulFlags)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::FindRow", "table=%d: bookmark=%d,flags=%d, restriction: %s", pThis->lpTableOps->ulTableId, bkOrigin, ulFlags, RestrictionToString(lpRestriction).c_str());
	HRESULT hr = pThis->FindRow(lpRestriction, bkOrigin, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::FindRow", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::Restrict(LPSRestriction lpRestriction, ULONG ulFlags)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::Restrict", "table=%d: flags=0x%08X, Restrict=%s", pThis->lpTableOps->ulTableId, ulFlags, RestrictionToString(lpRestriction).c_str());
	HRESULT hr = pThis->Restrict(lpRestriction, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::Restrict", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::CreateBookmark(BOOKMARK* lpbkPosition)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::CreateBookmark", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->CreateBookmark(lpbkPosition);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::CreateBookmark", "%s, bkpos=%d", GetMAPIErrorDescription(hr).c_str(), (lpbkPosition)?*lpbkPosition:0);
	return hr;
}

HRESULT ECMAPITable::xMAPITable::FreeBookmark(BOOKMARK bkPosition)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::FreeBookmark", "table=%d, bkpos=%d", pThis->lpTableOps->ulTableId, bkPosition);
	HRESULT hr = pThis->FreeBookmark(bkPosition);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::FreeBookmark", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::SortTable(LPSSortOrderSet lpSortCriteria, ULONG ulFlags)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::SortTable", "table=%d: flags=0x%08X, sort=%s", pThis->lpTableOps->ulTableId, ulFlags, SortOrderSetToString(lpSortCriteria).c_str());
	HRESULT hr = pThis->SortTable(lpSortCriteria, ulFlags);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::SortTable", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::QuerySortOrder(LPSSortOrderSet *lppSortCriteria)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::QuerySortOrder", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->QuerySortOrder(lppSortCriteria);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::QuerySortOrder", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::QueryRows(LONG lRowCount, ULONG ulFlags, LPSRowSet *lppRows)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::QueryRows", "table=%d: flags=0x%08X, rowcount=%d", pThis->lpTableOps->ulTableId, ulFlags, lRowCount);
	HRESULT hr = pThis->QueryRows(lRowCount, ulFlags, lppRows);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::QueryRows", "%s , %d rows returned %s", GetMAPIErrorDescription(hr).c_str(), (*lppRows) ? (*lppRows)->cRows : -1, RowSetToString(*lppRows).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::Abort()
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::Abort", "table=%d: ", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->Abort();
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::Abort", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::ExpandRow(ULONG cbInstanceKey, LPBYTE pbInstanceKey, ULONG ulRowCount, ULONG ulFlags, LPSRowSet * lppRows, ULONG *lpulMoreRows)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::ExpandRow", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->ExpandRow(cbInstanceKey, pbInstanceKey, ulRowCount, ulFlags, lppRows, lpulMoreRows);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::ExpandRow", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::CollapseRow(ULONG cbInstanceKey, LPBYTE pbInstanceKey, ULONG ulFlags, ULONG *lpulRowCount)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::CollapseRow", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->CollapseRow(cbInstanceKey, pbInstanceKey, ulFlags, lpulRowCount);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::CollapseRow", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::WaitForCompletion(ULONG ulFlags, ULONG ulTimeout, ULONG *lpulTableStatus)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::WaitForCompletion", "table=%d: flags=0x%08X, Timeout=", pThis->lpTableOps->ulTableId, ulFlags, ulTimeout);
	HRESULT hr = pThis->WaitForCompletion(ulFlags, ulTimeout, lpulTableStatus);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::WaitForCompletion", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::GetCollapseState(ULONG ulFlags, ULONG cbInstanceKey, LPBYTE lpbInstanceKey, ULONG *lpcbCollapseState, LPBYTE *lppbCollapseState)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::GetCollapseState", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->GetCollapseState(ulFlags, cbInstanceKey, lpbInstanceKey, lpcbCollapseState, lppbCollapseState);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::GetCollapseState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

HRESULT ECMAPITable::xMAPITable::SetCollapseState(ULONG ulFlags, ULONG cbCollapseState, LPBYTE pbCollapseState, BOOKMARK *lpbkLocation)
{
	METHOD_PROLOGUE_(ECMAPITable, MAPITable);
	TRACE_MAPI(TRACE_ENTRY, "IMAPITable::SetCollapseState", "table=%d", pThis->lpTableOps->ulTableId);
	HRESULT hr = pThis->SetCollapseState(ulFlags, cbCollapseState, pbCollapseState, lpbkLocation);
	TRACE_MAPI(TRACE_RETURN, "IMAPITable::SetCollapseState", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

