/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include <mapi.h>
#include <mapispi.h>
#include <mapiutil.h>

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif

#include <string>
#include <cassert>

#include "Zarafa.h"
#include "Mem.h"

#include "DLLGlobal.h"
#include "ECMSProviderSwitch.h"
#include "ECXPProvider.h"
#include "ECABProviderSwitch.h"

#include <iostream>

#include "ECDebug.h"
#include <stringutil.h>

#include "ECGuid.h"
#include "edkmdb.h"
#include "edkguid.h"

#include "guid.h"

#include "SSLUtil.h"
#include "ClientUtil.h"
#include "SymmetricCrypt.h"

#include "EntryPoint.h"

using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;

#define DEBUG_WITH_MEMORY_DUMP 0 // Sure to dump memleaks before the dll is exit
#endif

// Client wide variable
std::string		g_strCommonFilesZarafa;
std::string		g_strUserLocalAppDataZarafa;
std::string		g_strZarafaDirectory;

std::basic_string<TCHAR>	g_strManufacturer;
std::basic_string<TCHAR>	g_strProductName;
std::basic_string<TCHAR>	g_strProductNameShort;
bool g_isOEM;

// Map of msprovider with Profilename as key
ECMapProvider	g_mapProviders;


class CZarafaApp {
public:
    CZarafaApp() {
        ssl_threading_setup();

		g_strManufacturer = _T("Zarafa");
		g_strProductName = _T("Zarafa Collaboration Platform");
		g_isOEM = false;

		// FIXME for offline
		// - g_strUserLocalAppDataZarafa = ~/zarafa ?
		// - g_strZarafaDirectory = /usr/bin/ ?
    }
    ~CZarafaApp() {
        ssl_threading_cleanup();

		RemoveAllProviders(&g_mapProviders);
    }
};

CZarafaApp theApp;


///////////////////////////////////////////////////////////////////
// Zarafa entrypoints
//

// Called by MAPI to return a MSProvider object when a user opens a store based on our service
extern "C" HRESULT __cdecl MSProviderInit(HINSTANCE hInstance, LPMALLOC pmalloc, LPALLOCATEBUFFER pfnAllocBuf, LPALLOCATEMORE pfnAllocMore, LPFREEBUFFER pfnFreeBuf, ULONG ulFlags, ULONG ulMAPIver, ULONG * lpulProviderVer, LPMSPROVIDER * ppmsp)
{
	TRACE_MAPI(TRACE_ENTRY, "MSProviderInit", "flags=%08X", ulFlags);

	HRESULT hr = hrSuccess;
	ECMSProviderSwitch *lpMSProvider = NULL;

	// Check the interface version is ok
	if(ulMAPIver != CURRENT_SPI_VERSION) {
		hr = MAPI_E_VERSION;
		goto exit;
	}

	*lpulProviderVer = CURRENT_SPI_VERSION;
	
	// Save the pointers for later use
	_pmalloc = pmalloc;
	_pfnAllocBuf = pfnAllocBuf;
	_pfnAllocMore = pfnAllocMore;
	_pfnFreeBuf = pfnFreeBuf;
	_hInstance = hInstance;

	// This object is created for the lifetime of the DLL and destroyed when the
	// DLL is closed (same on linux, but then for the shared library);
	hr = ECMSProviderSwitch::Create(ulFlags, &lpMSProvider);

	if(hr != hrSuccess)
		goto exit;

	hr = lpMSProvider->QueryInterface(IID_IMSProvider, (void **)ppmsp); 

exit:
	if (lpMSProvider)
		lpMSProvider->Release();

	TRACE_MAPI(TRACE_RETURN, "MSProviderInit", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}

/**
 * Get the service name from the provider admin
 *
 * The service name is the string normally passed to CreateMsgService, like "ZARAFA6" or "MSEMS".
 *
 * @param lpProviderAdmin[in] The ProviderAdmin object passed to MSGServiceEntry
 * @param lpServiceName[out] The name of the message service
 * @return HRESULT Result status
 */
HRESULT GetServiceName(IProviderAdmin *lpProviderAdmin, std::string *lpServiceName)
{
	lpServiceName->assign("ZARAFA6");
	return hrSuccess;
}

// Called by MAPI to configure, or create a service
extern "C" HRESULT __stdcall MSGServiceEntry(HINSTANCE hInst, LPMALLOC lpMalloc, LPMAPISUP psup, ULONG ulUIParam, ULONG ulFlags, ULONG ulContext, ULONG cvals, LPSPropValue pvals, LPPROVIDERADMIN lpAdminProviders, MAPIERROR **lppMapiError)
{
	TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "0x%08X", ulFlags);

	HRESULT			hr = erSuccess;
	std::string		strServerName;
	std::string		strUserName;
	std::string		strUserPassword;
	std::string		strServerPort;
	std::string		strDefaultOfflinePath;
	std::string		strType;
	std::string		strRedirServer;
	std::string		strDefStoreServer;
	sGlobalProfileProps	sProfileProps;
	std::string		strError;
	LPPROFSECT		lpGlobalProfSect = NULL;
	LPSPropTagArray	lpsPropTagArray = NULL;
	LPSPropValue	lpsPropValueProfSect = NULL;
	IMAPISession	*lpSession = NULL;
	WSTransport		*lpTransport = NULL;
	LPSPropValue	lpsPropValue = NULL;
	LPSPropValue	lpsPropValueFind = NULL;
	ULONG			cValues = 0;
	bool			bShowDialog = false;
	LPMAPITABLE		lpTable = NULL;
	LPSRowSet		lpsProviderRowSet = NULL;
	LPPROFSECT		lpProfSect = NULL;
	MAPIERROR		*lpMapiError = NULL;
	char			*lpszStoreName = NULL;
	ULONG			ulRowCount = 0;
	SPropValue		sPropVals[5]; 
	ULONG			cPropValue = 0;
	bool			bShowAllSettingsPages = false;
	LPBYTE			lpDeligateStores = NULL;
	ULONG			cDeligateStores = 0;
	PABEID			pABeid = NULL;
	BYTE *			lpPrivateStoreGUID;
	BYTE *			lpPublicStoreGUID;


	ULONG			cbEntryId = 0;
	ULONG			cbWrappedEntryId = 0;
	LPENTRYID		lpEntryId = NULL;
	LPENTRYID		lpWrappedEntryId = NULL;
	SPropValue		sProps[6];
	ULONG cValueIndex = 0;
	std::string		strServiceName;

	bool bGlobalProfileUpdate = false;
	bool bUpdatedPageConnection = false;
	bool bInitStores = true;


	_hInstance = hInst;

	if (psup) {
		hr = psup->GetMemAllocRoutines(&_pfnAllocBuf, &_pfnAllocMore, &_pfnFreeBuf);
		if(hr != hrSuccess) {
			ASSERT(FALSE);
		}
	} else {
		// Support object not available on linux at this time... TODO: fix mapi4linux?
		_pfnAllocBuf = MAPIAllocateBuffer;
		_pfnAllocMore = MAPIAllocateMore;
		_pfnFreeBuf = MAPIFreeBuffer;
	}

	hr = GetServiceName(lpAdminProviders, &strServiceName);
	if (hr != hrSuccess)
		goto exit;

	lpPrivateStoreGUID = (BYTE *)(strServiceName == "MSEMS" ? &MSEMS_SERVICE_GUID : &ZARAFA_SERVICE_GUID);
	lpPublicStoreGUID = (BYTE *)(strServiceName == "MSEMS" ? &MSEMS_PUBLIC_STORE_GUID : &ZARAFA_STORE_PUBLIC_GUID);


	// Logon defaults
	strType = "http";
	strServerName = "";
	strServerPort = "236";



	switch(ulContext) {
	case MSG_SERVICE_INSTALL:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_INSTALL");
		hr = hrSuccess;
		break;
	case MSG_SERVICE_UNINSTALL:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_UNINSTALL");
		hr = hrSuccess;
		break;
	case MSG_SERVICE_DELETE:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_DELETE");
		hr = hrSuccess;
		break;
	case MSG_SERVICE_PROVIDER_CREATE:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_PROVIDER_CREATE");
		if(cvals && pvals) {

			LPSPropValue lpsPropName = NULL;
			
			lpsPropValueFind = PpropFindProp(pvals, cvals, PR_PROVIDER_UID);
			if(lpsPropValueFind == NULL || lpsPropValueFind->Value.bin.cb == 0)
			{
				//FIXME: give the right error?
				hr = MAPI_E_UNCONFIGURED;
				goto exit;
			}

			// PR_EC_USERNAME is the user we're adding ...
			lpsPropName = PpropFindProp(pvals, cvals, PR_EC_USERNAME);
			if(lpsPropName == NULL || lpsPropName->Value.bin.cb == 0)
			{
				hr = MAPI_E_UNCONFIGURED;
				goto exit;
			}

			//Open profile section
			hr = lpAdminProviders->OpenProfileSection((MAPIUID *)lpsPropValueFind->Value.bin.lpb, NULL, MAPI_MODIFY, &lpProfSect);
			if(hr != hrSuccess)
				goto exit;
		
			hr = lpAdminProviders->OpenProfileSection((LPMAPIUID)pbGlobalProfileSectionGuid, NULL, MAPI_MODIFY , &lpGlobalProfSect);
			if(hr != hrSuccess)
				goto exit;

			// Get username/pass settings
			hr = ClientUtil::GetGlobalProfileProperties(lpGlobalProfSect, &sProfileProps);
			if(hr != hrSuccess)
				goto exit;

			if(sProfileProps.strUserName.empty() || sProfileProps.strServerPath.empty()) {
				hr = MAPI_E_UNCONFIGURED; // FIXME: check or this is the right error
				goto exit;
			}

			hr = WSTransport::Create(ulFlags & SERVICE_UI_ALLOWED ? 0 : MDB_NO_DIALOG, &lpTransport);
			if(hr != hrSuccess)
				goto exit;


			hr = lpTransport->HrLogon(sProfileProps);
			if(hr != hrSuccess)
				goto exit;

			hr = lpTransport->HrResolveUserStore(lpsPropName->Value.lpszA, 0, NULL, &cbEntryId, &lpEntryId, &strRedirServer);
			if (hr == MAPI_E_UNABLE_TO_COMPLETE)
			{
				lpTransport->HrLogOff();
				sProfileProps.strServerPath = strRedirServer;
				hr = lpTransport->HrLogon(sProfileProps);
				if (hr != hrSuccess)
					goto exit;
				
				hr = lpTransport->HrResolveUserStore(lpsPropName->Value.lpszA, 0, NULL, &cbEntryId, &lpEntryId);
			}
			if(hr != hrSuccess)
				goto exit;

			hr = lpTransport->HrGetStoreName(cbEntryId, lpEntryId, &lpszStoreName);
			if(hr != hrSuccess) 
				goto exit;


			hr = WrapStoreEntryID(0, ZARAFA_DLL_NAME, cbEntryId, lpEntryId, &cbWrappedEntryId, &lpWrappedEntryId);
			if(hr != hrSuccess) 
				goto exit;


			sProps[0].ulPropTag = PR_RECORD_KEY;
			sProps[0].Value.bin.cb = sizeof(MAPIUID);
//FIXME: special function needed to get guid
			sProps[0].Value.bin.lpb = (BYTE *)&((PEID)lpEntryId)->guid;

			sProps[1].ulPropTag = PR_ENTRYID;
			sProps[1].Value.bin.cb = cbWrappedEntryId;
			sProps[1].Value.bin.lpb = (BYTE *)lpWrappedEntryId;

			sProps[2].ulPropTag = PR_DISPLAY_NAME;
			sProps[2].Value.lpszA = lpszStoreName;

			lpProfSect->SetProps(3, sProps, NULL);

			hr = lpProfSect->SaveChanges(0);
			if(hr != hrSuccess)
				goto exit;
		
		}

		break;
	case MSG_SERVICE_PROVIDER_DELETE:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_PROVIDER_DELETE");
		hr = hrSuccess;
		
		//FIXME: delete Offline database

		break;
	case MSG_SERVICE_CONFIGURE:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_CONFIGURE");
		bShowAllSettingsPages = true;
	case MSG_SERVICE_CREATE:
		TRACE_MAPI(TRACE_ENTRY, "MSGServiceEntry", "MSG_SERVICE_CREATE");

		//Open global profile, add the store.(for show list, delete etc)
		hr = lpAdminProviders->OpenProfileSection((LPMAPIUID)pbGlobalProfileSectionGuid, NULL, MAPI_MODIFY , &lpGlobalProfSect);
		if(hr != hrSuccess)
			goto exit;

		if(cvals) {
			hr = lpGlobalProfSect->SetProps(cvals, pvals, NULL);

			if(hr != hrSuccess)
				goto exit;
		}

		hr = ClientUtil::GetGlobalProfileProperties(lpGlobalProfSect, &sProfileProps);

		if(sProfileProps.strServerPath.empty() || sProfileProps.strUserName.empty() || (sProfileProps.strPassword.empty() && sProfileProps.strSSLKeyFile.empty())) {
			bShowDialog = true;
		}
		//FIXME: check here offline path with the flags
		if(!sProfileProps.strServerPath.empty()) {
			strServerName = GetServerNameFromPath(sProfileProps.strServerPath.c_str());
			strServerPort = GetServerPortFromPath(sProfileProps.strServerPath.c_str());
			strType = GetServerTypeFromPath(sProfileProps.strServerPath.c_str());
		}

		// Get deligate stores, Ignore error
		ClientUtil::GetGlobalProfileDeligateStoresProp(lpGlobalProfSect, &cDeligateStores, &lpDeligateStores);

		// init defaults
		hr = WSTransport::Create(ulFlags & SERVICE_UI_ALLOWED ? 0 : MDB_NO_DIALOG, &lpTransport);
		if(hr != hrSuccess)
			goto exit;

		// Check the path, username and password
		while(1)
		{
			bGlobalProfileUpdate = false;
			bUpdatedPageConnection = false;

			if((bShowDialog && ulFlags & SERVICE_UI_ALLOWED) || ulFlags & SERVICE_UI_ALWAYS )
			{
				hr = MAPI_E_USER_CANCEL;

			}// if(bShowDialog...)

						
			if(!(ulFlags & SERVICE_UI_ALLOWED || ulFlags & SERVICE_UI_ALWAYS) && (strServerName.empty() || sProfileProps.strUserName.empty())){
				hr = MAPI_E_UNCONFIGURED;
				goto exit;
			}else if(!strServerName.empty() && !sProfileProps.strUserName.empty()) {

				//Logon the server
				hr = lpTransport->HrLogon(sProfileProps);
			}else{
				hr = MAPI_E_LOGON_FAILED;
			}

			if(hr == MAPI_E_LOGON_FAILED || hr == MAPI_E_NETWORK_ERROR || hr == MAPI_E_VERSION) {
				bShowDialog = true;
			} else if(hr != erSuccess){ // Big error?
				bShowDialog = true;
				ASSERT(FALSE);
			}else {
				//Update global profile
				if( bGlobalProfileUpdate == true) {

					cValues = 10;
					cValueIndex = 0;
					hr = MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void**)&lpsPropValue);
					if(hr != hrSuccess)
						goto exit;

					lpsPropValue[cValueIndex].ulPropTag	= PR_EC_PATH;
					lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strServerPath.c_str();

					lpsPropValue[cValueIndex].ulPropTag	= PR_EC_USERNAME;
					lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strUserName.c_str();

					lpsPropValue[cValueIndex].ulPropTag = PR_EC_USERPASSWORD;
					lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strPassword.c_str();

					lpsPropValue[cValueIndex].ulPropTag = PR_EC_FLAGS;
					lpsPropValue[cValueIndex++].Value.ul = sProfileProps.ulProfileFlags;

					if (bUpdatedPageConnection == true)
					{
						lpsPropValue[cValueIndex].ulPropTag = PR_EC_CONNECTION_TIMEOUT;
						lpsPropValue[cValueIndex++].Value.ul = sProfileProps.ulConnectionTimeOut;

						// Proxy settings
						lpsPropValue[cValueIndex].ulPropTag = PR_EC_PROXY_FLAGS;
						lpsPropValue[cValueIndex++].Value.ul = sProfileProps.ulProxyFlags;

						lpsPropValue[cValueIndex].ulPropTag = PR_EC_PROXY_PORT;
						lpsPropValue[cValueIndex++].Value.ul = sProfileProps.ulProxyPort;

						lpsPropValue[cValueIndex].ulPropTag = PR_EC_PROXY_HOST;
						lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strProxyHost.c_str();
							
						lpsPropValue[cValueIndex].ulPropTag = PR_EC_PROXY_USERNAME;
						lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strProxyUserName.c_str();

						lpsPropValue[cValueIndex].ulPropTag = PR_EC_PROXY_PASSWORD;
						lpsPropValue[cValueIndex++].Value.lpszA = (char *)sProfileProps.strProxyPassword.c_str();
					}

					hr = lpGlobalProfSect->SetProps(cValueIndex, lpsPropValue, NULL);
					if(hr != hrSuccess)
						goto exit;
					
					//Free allocated memory
					if(lpsPropValue){ MAPIFreeBuffer(lpsPropValue); lpsPropValue = NULL;}

				}
				break; // Everything is oke
			}
			

			// On incorrect password, and UI allowed, show incorrect password error
			if((ulFlags & SERVICE_UI_ALLOWED || ulFlags & SERVICE_UI_ALWAYS)) {
				// what do we do on linux?
				cout << "Access Denied: Incorrect username and/or password." << endl;
				hr = MAPI_E_UNCONFIGURED;
				goto exit;
			}else if(!(ulFlags & SERVICE_UI_ALLOWED || ulFlags & SERVICE_UI_ALWAYS)){
				// Do not reset the logon error from HrLogon()
				// The DAgent uses this value to determain if the delivery is fatal or not
				// 
				// Although this error is not in the online spec from MS, it should not really matter .... right?
				// hr = MAPI_E_UNCONFIGURED;
				goto exit;
			}

		}// while(1)

		// Get the provider table
		hr = lpAdminProviders->GetProviderTable(0, &lpTable);
		if(hr != hrSuccess)
			goto exit;

		// Get provider count
		// count almost 3, private store and public store and transport
		hr = lpTable->GetRowCount(0, &ulRowCount);
		if(hr != hrSuccess)
			goto exit;

		// Get the rows
		hr = lpTable->QueryRows(ulRowCount, 0, &lpsProviderRowSet);
		if(hr != hrSuccess)
			goto exit;

		//Check is exist one or more rows
		if(lpsProviderRowSet == NULL || lpsProviderRowSet->cRows < 1) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}

		// Scan the rows for message stores
		for(unsigned int i = 0; bInitStores == true && i < ulRowCount; i++)
		{
			// Check type of the provider
			lpsPropValueFind = PpropFindProp(lpsProviderRowSet->aRow[i].lpProps, lpsProviderRowSet->aRow[i].cValues, PR_RESOURCE_TYPE);

			if(lpsPropValueFind != NULL && lpsPropValueFind->Value.l == MAPI_STORE_PROVIDER)
			{
				//Free allocated memory

				//Get de UID of the provider
				lpsPropValueFind = PpropFindProp(lpsProviderRowSet->aRow[i].lpProps, lpsProviderRowSet->aRow[i].cValues, PR_PROVIDER_UID);
				if(lpsPropValueFind == NULL || lpsPropValueFind->Value.bin.cb == 0)
					continue;

				//Open profile section
				hr = lpAdminProviders->OpenProfileSection((MAPIUID *)lpsPropValueFind->Value.bin.lpb, NULL, MAPI_MODIFY, &lpProfSect);

				if(hr != hrSuccess)
					goto exit;

				// Check PR_MDB_PROVIDER
				cValues = 1;
				hr = MAPIAllocateBuffer(CbNewSPropTagArray(cValues), (void **)&lpsPropTagArray);
				if(hr != hrSuccess)
					goto exit;

				lpsPropTagArray->cValues = cValues;
				lpsPropTagArray->aulPropTag[0] = PR_MDB_PROVIDER;

				hr = lpProfSect->GetProps(lpsPropTagArray, 0, &cValues, &lpsPropValueProfSect);
				if(hr != hrSuccess) 
					goto exit;

				if(lpsPropTagArray) {
					MAPIFreeBuffer(lpsPropTagArray);
					lpsPropTagArray = NULL;
				}
			
				if(memcmp(lpsPropValueProfSect[0].Value.bin.lpb, lpPublicStoreGUID, sizeof(MAPIUID) ) == 0) {
					// Get the public store
					if (sProfileProps.ulProfileFlags & EC_PROFILE_FLAGS_NO_PUBLIC_STORE)
						hr = MAPI_E_INVALID_PARAMETER; // just to get to the DeleteProvider part
					else
						hr = lpTransport->HrGetPublicStore(&cbEntryId, &lpEntryId, &strRedirServer);
					if (hr == MAPI_E_UNABLE_TO_COMPLETE)
					{
						lpTransport->HrLogOff();
						sProfileProps.strServerPath = strRedirServer;
						hr = lpTransport->HrLogon(sProfileProps);
											
						if (hr == hrSuccess)
							hr = lpTransport->HrGetPublicStore(&cbEntryId, &lpEntryId);
					}
					if(hr != hrSuccess) {
						if(lpsPropValueFind != NULL)
							lpAdminProviders->DeleteProvider((MAPIUID *)lpsPropValueFind->Value.bin.lpb);
						hr = hrSuccess;
						goto skip;
					}
				} else if(memcmp(lpsPropValueProfSect[0].Value.bin.lpb, lpPrivateStoreGUID, sizeof(MAPIUID) ) == 0) {
					// Get the default store for this user
					hr = lpTransport->HrGetStore(0, NULL, &cbEntryId, &lpEntryId, 0, NULL, &strRedirServer);
					if (hr == MAPI_E_UNABLE_TO_COMPLETE)
					{
						lpTransport->HrLogOff();
						sProfileProps.strServerPath = strRedirServer;
						hr = lpTransport->HrLogon(sProfileProps);
						if (hr != hrSuccess)
							goto exit;

						hr = lpTransport->HrGetStore(0, NULL, &cbEntryId, &lpEntryId, 0, NULL);
						if (hr == hrSuccess)
						{
							// This should happen only once.
							assert(strDefStoreServer.empty());

							strDefStoreServer = strRedirServer;

							// This should be a real URL
							assert(strDefStoreServer.compare(0, 9, "pseudo://") != 0);
						}
					}
 					if(hr != hrSuccess) 
						goto exit;
					
				} else goto skip;

				hr = lpTransport->HrGetStoreName(cbEntryId, lpEntryId, &lpszStoreName);
				if(hr != hrSuccess) 
					goto exit;


				hr = WrapStoreEntryID(0, ZARAFA_DLL_NAME, cbEntryId, lpEntryId, &cbWrappedEntryId, &lpWrappedEntryId);
				if(hr != hrSuccess) 
					goto exit;

				cPropValue = 0;

				sPropVals[cPropValue].ulPropTag = PR_ENTRYID;
				sPropVals[cPropValue].Value.bin.cb = cbWrappedEntryId;
				sPropVals[cPropValue++].Value.bin.lpb = (LPBYTE) lpWrappedEntryId;

				sPropVals[cPropValue].ulPropTag = PR_RECORD_KEY;
				sPropVals[cPropValue].Value.bin.cb = sizeof(MAPIUID);
//FIXME: special function needed to get guid
				sPropVals[cPropValue++].Value.bin.lpb = (LPBYTE) &((PEID)lpEntryId)->guid;

				sPropVals[cPropValue].ulPropTag = PR_DISPLAY_NAME;
				sPropVals[cPropValue++].Value.lpszA = lpszStoreName;

				sPropVals[cPropValue].ulPropTag = PR_EC_PATH;
				sPropVals[cPropValue++].Value.lpszA = "Server";
								
				hr = lpProfSect->SetProps(cPropValue, sPropVals, NULL);
				if(hr != hrSuccess)
					goto exit;

				hr = lpProfSect->SaveChanges(0);
				if(hr != hrSuccess)
					goto exit;

			} else if(lpsPropValueFind != NULL && lpsPropValueFind->Value.l == MAPI_AB_PROVIDER){
				
				TRACE_MAPI(TRACE_RETURN, "MSGServiceEntry", "MAPI_AB_PROVIDER, MAPI_AB_PROVIDER, MAPI_AB_PROVIDER");

				//Get de UID of the provider
				lpsPropValueFind = PpropFindProp(lpsProviderRowSet->aRow[i].lpProps, lpsProviderRowSet->aRow[i].cValues, PR_PROVIDER_UID);
				if(lpsPropValueFind == NULL || lpsPropValueFind->Value.bin.cb == 0)
					continue;

				//Open profile section
				hr = lpAdminProviders->OpenProfileSection((MAPIUID *)lpsPropValueFind->Value.bin.lpb, NULL, MAPI_MODIFY, &lpProfSect);
				if(hr != hrSuccess)
					goto exit;

				cPropValue = 0;

				hr = MAPIAllocateBuffer(CbNewABEID(""), (void**)&pABeid);
				if(hr != hrSuccess)
					goto exit;

				memset(pABeid, 0, CbNewABEID(""));

				memcpy(&pABeid->guid, &MUIDECSAB, sizeof(GUID));
				pABeid->ulType = MAPI_ABCONT;

				sPropVals[cPropValue].ulPropTag = PR_ENTRYID;
				sPropVals[cPropValue].Value.bin.cb = CbNewABEID("");
				sPropVals[cPropValue++].Value.bin.lpb = (LPBYTE) pABeid;

				sPropVals[cPropValue].ulPropTag = PR_RECORD_KEY;
				sPropVals[cPropValue].Value.bin.cb = sizeof(MAPIUID);
				sPropVals[cPropValue++].Value.bin.lpb = (LPBYTE)&MUIDECSAB;

				sPropVals[cPropValue].ulPropTag = PR_DISPLAY_NAME;
				sPropVals[cPropValue++].Value.lpszA = "Zarafa Addressbook";

				sPropVals[cPropValue].ulPropTag = PR_EC_PATH;
				sPropVals[cPropValue++].Value.lpszA = "Zarafa Addressbook";

				sPropVals[cPropValue].ulPropTag = PR_PROVIDER_DLL_NAME_A;
				sPropVals[cPropValue++].Value.lpszA = ZARAFA_DLL_NAME;

				hr = lpProfSect->SetProps(cPropValue, sPropVals, NULL);
				if(hr != hrSuccess)
					goto exit;

				hr = lpProfSect->SaveChanges(0);
				if(hr != hrSuccess)
					goto exit;
				
			}
		skip:
			//Free allocated memory
			if(lpProfSect){ lpProfSect->Release(); lpProfSect = NULL; }
			if(pABeid){ MAPIFreeBuffer(pABeid); pABeid = NULL; }
			if(lpEntryId) { MAPIFreeBuffer(lpEntryId); lpEntryId = NULL;}
			if(lpWrappedEntryId) { MAPIFreeBuffer(lpWrappedEntryId); lpWrappedEntryId = NULL;}
			if(lpsPropValueProfSect){ MAPIFreeBuffer(lpsPropValueProfSect); lpsPropValueProfSect = NULL; }
			if(lpszStoreName){ ECFreeBuffer(lpszStoreName); lpszStoreName = NULL;}
		}// for(unsigned int i = 0; i < ulRowCount; i++)


		break;
	
	} // switch(ulContext)

	if (!strDefStoreServer.empty()) {
		cValues = 1;
		cValueIndex = 0;
		hr = MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void**)&lpsPropValue);
		if(hr != hrSuccess)
			goto exit;

		lpsPropValue[cValueIndex].ulPropTag	= PR_EC_PATH;
		lpsPropValue[cValueIndex++].Value.lpszA = (char *)strDefStoreServer.c_str();

		hr = lpGlobalProfSect->SetProps(cValueIndex, lpsPropValue, NULL);
		if(hr != hrSuccess)
			goto exit;

		//Free allocated memory
		if(lpsPropValue){ MAPIFreeBuffer(lpsPropValue); lpsPropValue = NULL;}
	}


exit:
	if (lppMapiError) {
		
		*lppMapiError = NULL;

		if(hr != hrSuccess) {
			// Set Error
			strError = _T("EntryPoint: ");
			strError += Util::HrMAPIErrorToText(hr);

			// Some outlook 2007 clients can't allocate memory so check it
			if(MAPIAllocateBuffer(sizeof(MAPIERROR), (void**)&lpMapiError) == hrSuccess) { 

				memset(lpMapiError, 0, sizeof(MAPIERROR));				

				MAPIAllocateMore(strError.size()+1, lpMapiError, (void**)&lpMapiError->lpszError);
				strcpy(lpMapiError->lpszError, strError.c_str());
				
				MAPIAllocateMore(sizeof(std::basic_string<TCHAR>::value_type) * (g_strProductName.size()+1), lpMapiError, (void**)&lpMapiError->lpszComponent);
				memcpy(lpMapiError->lpszComponent, g_strProductName.c_str(), sizeof(std::basic_string<TCHAR>::value_type) * (g_strProductName.size()+1));
			
			
				lpMapiError->ulVersion = 0;
				lpMapiError->ulLowLevelError = 0;
				lpMapiError->ulContext = 0;

				*lppMapiError = lpMapiError;
			}
		}
	}

	if (pABeid)
		MAPIFreeBuffer(pABeid);

	if(lpDeligateStores)
		MAPIFreeBuffer(lpDeligateStores);

	if(lpGlobalProfSect)
		lpGlobalProfSect->Release();

	if(lpTransport)
		lpTransport->Release();

	if(lpSession)
		lpSession->Release();

	if(lpTable)
		lpTable->Release();
		
    if(lpsPropValue)
        MAPIFreeBuffer(lpsPropValue);

	if(lpsPropValueProfSect)
		MAPIFreeBuffer(lpsPropValueProfSect);

	if(lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if(lpsProviderRowSet)
		FreeProws(lpsProviderRowSet);
	
	if(lpszStoreName)
		ECFreeBuffer(lpszStoreName);


	if(lpProfSect)
		lpProfSect->Release();

	if(lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	if(lpWrappedEntryId)
		MAPIFreeBuffer(lpWrappedEntryId);


	TRACE_MAPI(TRACE_RETURN, "MSGServiceEntry", "%s", GetMAPIErrorDescription(hr).c_str());
	return hr;
}


extern "C" HRESULT __cdecl XPProviderInit(HINSTANCE hInstance, LPMALLOC lpMalloc, LPALLOCATEBUFFER lpAllocateBuffer, LPALLOCATEMORE lpAllocateMore, LPFREEBUFFER lpFreeBuffer, ULONG ulFlags, ULONG ulMAPIVer, ULONG * lpulProviderVer, LPXPPROVIDER * lppXPProvider)
{
	TRACE_MAPI(TRACE_ENTRY, "XPProviderInit", "");

	HRESULT hr = hrSuccess;
	ECXPProvider	*pXPProvider = NULL;

    if (ulMAPIVer < CURRENT_SPI_VERSION)
    {
        hr = MAPI_E_VERSION;
		goto exit;
    }

	*lpulProviderVer = CURRENT_SPI_VERSION;

	// Save the pointer to the allocation routines in global variables
	_pmalloc = lpMalloc;
	_pfnAllocBuf = lpAllocateBuffer;
	_pfnAllocMore = lpAllocateMore;
	_pfnFreeBuf = lpFreeBuffer;
	_hInstance = hInstance;

	hr = ECXPProvider::Create(&pXPProvider);
	if(hr != hrSuccess)
		goto exit;

	hr = pXPProvider->QueryInterface(IID_IXPProvider, (void **)lppXPProvider);

exit:
	if(pXPProvider)
		pXPProvider->Release();


	return hr;
}


extern "C" HRESULT  __cdecl ABProviderInit(HINSTANCE hInstance, LPMALLOC lpMalloc, LPALLOCATEBUFFER lpAllocateBuffer, LPALLOCATEMORE lpAllocateMore, LPFREEBUFFER lpFreeBuffer, ULONG ulFlags, ULONG ulMAPIVer, ULONG * lpulProviderVer, LPABPROVIDER * lppABProvider)
{
	TRACE_MAPI(TRACE_ENTRY, "ABProviderInit", "");

	HRESULT hr = hrSuccess;
	ECABProviderSwitch	*lpABProvider = NULL;

	if (ulMAPIVer < CURRENT_SPI_VERSION)
	{
		hr = MAPI_E_VERSION;
		goto exit;
	}

	*lpulProviderVer = CURRENT_SPI_VERSION;

	// Save the pointer to the allocation routines in global variables
	_pmalloc = lpMalloc;
	_pfnAllocBuf = lpAllocateBuffer;
	_pfnAllocMore = lpAllocateMore;
	_pfnFreeBuf = lpFreeBuffer;
	_hInstance = hInstance;

	hr = ECABProviderSwitch::Create(&lpABProvider);
	if(hr != hrSuccess)
		goto exit;

	hr = lpABProvider->QueryInterface(IID_IABProvider, (void **)lppABProvider);

exit:
	if(lpABProvider)
		lpABProvider->Release();


	return hr;
}

