/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <platform.h>

#include "StreamUtil.h"
#include "StorageUtil.h"
#include "ECAttachmentStorage.h"
#include "ECSecurity.h"
#include "ECSessionManager.h"
#include "ECGenProps.h"
#include "ECICS.h"

#include <ECFifoBuffer.h>
#include <ECSerializer.h>
#include <stringutil.h>
#include <mapitags.h>
#include <mapiext.h>
#include <mapidefs.h>
#include <edkmdb.h>
#include <assert.h>

#include <set>
#include <map>
#include <string>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*
	streams are as follows:
	  <version>
	  <message>

	messages are as follows:
	  <# of props>:32bit
	    <property>
	  <# of sub objects>:32bit
	    <subobject>


	properties are as follows:
	  <prop tag>:32bit
	  [<length in bytes>:32bit]
	  <data bytes>
	  [<guid, kind, [id|bytes+string>]


	subobjects are as follows:
	  <type>:32bit
	  <id>:32bit
	  <message>
*/

// External objects
extern ECSessionManager *g_lpSessionManager;	// ECServerEntrypoint.cpp

static inline bool operator<(const GUID &lhs, const GUID &rhs) {
	return memcmp(&lhs, &rhs, sizeof(GUID)) < 0;
}

// Helper class for mapping named properties from the stream to local proptags
class NamedPropertyMapper
{
public:
	NamedPropertyMapper(ECDatabase *lpDatabase);

	ECRESULT GetId(const GUID &guid, unsigned int ulNameId, unsigned int *lpId);
	ECRESULT GetId(const GUID &guid, const std::string &strNameString, unsigned int *lpId);

private:
	typedef std::pair<GUID,unsigned int> nameidkey_t;
	typedef std::pair<GUID,std::string> namestringkey_t;
	typedef std::map<nameidkey_t,unsigned int> nameidmap_t;
	typedef std::map<namestringkey_t,unsigned int> namestringmap_t;

	ECDatabase *m_lpDatabase;
	nameidmap_t m_mapNameIds;
	namestringmap_t m_mapNameStrings;
};


NamedPropertyMapper::NamedPropertyMapper(ECDatabase *lpDatabase)
	: m_lpDatabase(lpDatabase) 
{
	ASSERT(lpDatabase != NULL);
}

ECRESULT NamedPropertyMapper::GetId(const GUID &guid, unsigned int ulNameId, unsigned int *lpulId)
{
	ECRESULT er = erSuccess;

	std::string strQuery;
	DB_RESULT lpResult = NULL;
	DB_ROW lpRow = NULL;

	nameidkey_t key(guid, ulNameId);
	nameidmap_t::const_iterator i = m_mapNameIds.find(key);

	if (i != m_mapNameIds.end()) {
		*lpulId = i->second;
		goto exit;
	}

	// Check the database
	strQuery = 
		"SELECT id FROM names "
			"WHERE nameid=" + stringify(ulNameId) +
			" AND guid=" + m_lpDatabase->EscapeBinary((unsigned char*)&guid, sizeof(guid));

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		goto exit;

	if ((lpRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if (lpRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		*lpulId = atoui((char*)lpRow[0]) + 0x8501;
	} else {
		// Create the named property
		strQuery = 
			"INSERT INTO names (nameid, guid) "
				"VALUES(" + stringify(ulNameId) + "," + m_lpDatabase->EscapeBinary((unsigned char*)&guid, sizeof(guid)) + ")";
		
		er = m_lpDatabase->DoInsert(strQuery, lpulId);
		if (er != erSuccess)
			goto exit;

		*lpulId += 0x8501;
	}

	// *lpulId now contains the local propid, update the cache
	m_mapNameIds.insert(nameidmap_t::value_type(key, *lpulId));


exit:
	if (lpResult)
		m_lpDatabase->FreeResult(lpResult);

	return er;
}

ECRESULT NamedPropertyMapper::GetId(const GUID &guid, const std::string &strNameString, unsigned int *lpulId)
{
	ECRESULT er = erSuccess;

	std::string strQuery;
	DB_RESULT lpResult = NULL;
	DB_ROW lpRow = NULL;

	namestringkey_t key(guid, strNameString);
	namestringmap_t::const_iterator i = m_mapNameStrings.find(key);

	if (i != m_mapNameStrings.end()) {
		*lpulId = i->second;
		goto exit;
	}

	// Check the database
	strQuery = 
		"SELECT id FROM names "
			"WHERE namestring='" + m_lpDatabase->Escape(strNameString) + "'"
			" AND guid=" + m_lpDatabase->EscapeBinary((unsigned char*)&guid, sizeof(guid));

	er = m_lpDatabase->DoSelect(strQuery, &lpResult);
	if (er != erSuccess)
		goto exit;

	if ((lpRow = m_lpDatabase->FetchRow(lpResult)) != NULL) {
		if (lpRow[0] == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		*lpulId = atoui((char*)lpRow[0]) + 0x8501;
	} else {
		// Create the named property
		strQuery = 
			"INSERT INTO names (namestring, guid) "
				"VALUES('" + m_lpDatabase->Escape(strNameString) + "'," + m_lpDatabase->EscapeBinary((unsigned char*)&guid, sizeof(guid)) + ")";
		
		er = m_lpDatabase->DoInsert(strQuery, lpulId);
		if (er != erSuccess)
			goto exit;

		*lpulId += 0x8501;
	}

	// *lpulId now contains the local propid, update the cache
	m_mapNameStrings.insert(namestringmap_t::value_type(key, *lpulId));


exit:
	if (lpResult)
		m_lpDatabase->FreeResult(lpResult);

	return er;
}


// Utility Functions
ECRESULT SerializeDatabasePropVal(DB_ROW lpRow, DB_LENGTHS lpLen, ECSerializer *lpSink)
{
	ECRESULT er = erSuccess;
	unsigned int type = atoi(lpRow[FIELD_NR_TYPE]);
	unsigned int ulPropTag = PROP_TAG(atoi(lpRow[FIELD_NR_TYPE]),atoi(lpRow[FIELD_NR_TAG]));
	unsigned int ulCount;
	unsigned int ulLen;
	unsigned int ulLastPos;
	unsigned int ulLastPos2;
	std::string	strData;

	unsigned int ulKind;
	unsigned int ulNameId;
	std::string strNameString;
	locale_t loc = createlocale(LC_NUMERIC, "C");

	short i;
	unsigned int ul;
	float flt;
	unsigned char b;
	double dbl;
	hiloLong hilo;
	long long li;

	er = lpSink->Write(&ulPropTag, sizeof(ulPropTag), 1);
	if (er != erSuccess)
		goto exit;

	switch (type) {
	case PT_I2:
		if (lpRow[FIELD_NR_ULONG] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		i = (short)atoi(lpRow[FIELD_NR_ULONG]);
		er = lpSink->Write(&i, sizeof(i), 1);
		break;
	case PT_LONG:
		if (lpRow[FIELD_NR_ULONG] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ul = atoui(lpRow[FIELD_NR_ULONG]);
		er = lpSink->Write(&ul, sizeof(ul), 1);
		break;
	case PT_R4:
		if (lpRow[FIELD_NR_DOUBLE] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		flt = (float)strtod_l(lpRow[FIELD_NR_DOUBLE], NULL, loc);
		er = lpSink->Write(&flt, sizeof(flt), 1);
		break;
	case PT_BOOLEAN:
		if (lpRow[FIELD_NR_ULONG] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		b = (atoi(lpRow[FIELD_NR_ULONG]) ? 1 : 0);
		er = lpSink->Write(&b, sizeof(b), 1);
		break;
	case PT_DOUBLE:
	case PT_APPTIME:
		if (lpRow[FIELD_NR_DOUBLE] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		dbl = strtod_l(lpRow[FIELD_NR_DOUBLE], NULL, loc);
		er = lpSink->Write(&dbl, sizeof(dbl), 1);
		break;
	case PT_CURRENCY:
	case PT_SYSTIME:
		if (lpRow[FIELD_NR_HI] == NULL || lpRow[FIELD_NR_LO] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		hilo.hi = atoi(lpRow[FIELD_NR_HI]);
		hilo.lo = atoui(lpRow[FIELD_NR_LO]);
		er = lpSink->Write(&hilo.hi, sizeof(hilo.hi), 1);
		if (er == erSuccess)
			er = lpSink->Write(&hilo.lo, sizeof(hilo.lo), 1);
		break;
	case PT_I8:
		if (lpRow[FIELD_NR_LONGINT] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		li = _atoi64(lpRow[FIELD_NR_LONGINT]);
		er = lpSink->Write(&li, sizeof(li), 1);
		break;
	case PT_STRING8:
	case PT_UNICODE:
		if (lpRow[FIELD_NR_STRING] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ulLen = lpLen[FIELD_NR_STRING];
		er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
		if (er == erSuccess)
			er = lpSink->Write(lpRow[FIELD_NR_STRING], 1, lpLen[FIELD_NR_STRING]);
		break;
	case PT_CLSID:
	case PT_BINARY:
		if (lpRow[FIELD_NR_BINARY] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ulLen = lpLen[FIELD_NR_BINARY];
		er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
		if (er == erSuccess)
			er = lpSink->Write(lpRow[FIELD_NR_BINARY], 1, lpLen[FIELD_NR_BINARY]);
		break;
	case PT_MV_I2:
		if (lpRow[FIELD_NR_ULONG] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0 ;er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_ULONG], lpLen[FIELD_NR_ULONG], &ulLastPos, &strData);
			i = (short)atoi(strData.c_str());
			er = lpSink->Write(&i, sizeof(i), 1);
		}
		break;
	case PT_MV_LONG:
		if (lpRow[FIELD_NR_ULONG] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_ULONG], lpLen[FIELD_NR_ULONG], &ulLastPos, &strData);
			ul = atoui((char*)strData.c_str());
			er = lpSink->Write(&ul, sizeof(ul), 1);
		}
		break;
	case PT_MV_R4:
		if (lpRow[FIELD_NR_DOUBLE] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_DOUBLE], lpLen[FIELD_NR_DOUBLE], &ulLastPos, &strData);
			flt = (float)strtod_l(strData.c_str(), NULL, loc);
			er = lpSink->Write(&flt, sizeof(flt), 1);
		}
		break;
	case PT_MV_DOUBLE:
	case PT_MV_APPTIME:
		if (lpRow[FIELD_NR_DOUBLE] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_DOUBLE], lpLen[FIELD_NR_DOUBLE], &ulLastPos, &strData);
			dbl = strtod_l(strData.c_str(), NULL, loc);
			er = lpSink->Write(&dbl, sizeof(flt), 1);
		}
		break;
	case PT_MV_CURRENCY:
	case PT_MV_SYSTIME:
		if (lpRow[FIELD_NR_HI] == NULL || lpRow[FIELD_NR_LO] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		ulLastPos2 = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_LO], lpLen[FIELD_NR_LO], &ulLastPos, &strData);
			hilo.lo = atoui((char*)strData.c_str());
			ParseMVProp(lpRow[FIELD_NR_HI], lpLen[FIELD_NR_HI], &ulLastPos2, &strData);
			hilo.hi = atoi((char*)strData.c_str());
			er = lpSink->Write(&hilo.hi, sizeof(hilo.hi), 1);
			if (er == erSuccess)
				er = lpSink->Write(&hilo.lo, sizeof(hilo.lo), 1);
		}
		break;
	case PT_MV_BINARY:
	case PT_MV_CLSID:
		if (lpRow[FIELD_NR_BINARY] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_BINARY], lpLen[FIELD_NR_BINARY], &ulLastPos, &strData);
			ulLen = (unsigned int)strData.size();
			er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er == erSuccess)
				er = lpSink->Write(strData.c_str(), 1, ulLen);
		}
		break;
	case PT_MV_STRING8:
	case PT_MV_UNICODE:
		if (lpRow[FIELD_NR_STRING] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_STRING], lpLen[FIELD_NR_STRING], &ulLastPos, &strData);
			ulLen = (unsigned int)strData.size();
			er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er == erSuccess)
				er = lpSink->Write(strData.c_str(), 1, ulLen);
		}
		break;
	case PT_MV_I8:
		if (lpRow[FIELD_NR_LONGINT] == NULL) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}

		ulCount = atoi(lpRow[FIELD_NR_ID]);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		ulLastPos = 0;
		for (unsigned x = 0; er == erSuccess && x < ulCount; x++)
		{
			ParseMVProp(lpRow[FIELD_NR_LONGINT], lpLen[FIELD_NR_LONGINT], &ulLastPos, &strData);
			li = _atoi64(strData.c_str());
			er = lpSink->Write(&li, sizeof(li), 1);
		}
		break;
	default:
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	// If property is named property in the dynamic range we need to add some extra info
	if (PROP_ID(ulPropTag) >= 0x8500) {
		// Send out the GUID.
		er = lpSink->Write(lpRow[FIELD_NR_MAX + 2], 1, lpLen[FIELD_NR_MAX + 2]);

		if (er == erSuccess && lpRow[FIELD_NR_MAX + 0] != NULL) {
			ulKind = MNID_ID;
			ulNameId = atoui((char*)lpRow[FIELD_NR_MAX + 0]);

			er = lpSink->Write(&ulKind, sizeof(ulKind), 1);
			if (er == erSuccess)
				er = lpSink->Write(&ulNameId, sizeof(ulNameId), 1);

		} else if (er == erSuccess && lpRow[FIELD_NR_MAX + 1] != NULL) {
			ulKind = MNID_STRING;
			ulLen = lpLen[FIELD_NR_MAX + 1];

			er = lpSink->Write(&ulKind, sizeof(ulKind), 1);
			if (er == erSuccess)
				er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er == erSuccess)
				er = lpSink->Write(lpRow[FIELD_NR_MAX + 1], 1, ulLen);

		} else if (er == erSuccess)
			er = ZARAFA_E_INVALID_TYPE;
	}

exit:
	freelocale(loc);
	return er;
}

ECRESULT SerializePropVal(const struct propVal &sPropVal, ECSerializer *lpSink)
{
	ECRESULT er = erSuccess;
	unsigned int type = PROP_TYPE(sPropVal.ulPropTag);
	unsigned int ulLen;
	unsigned char b;
	std::string	strData;

	if (PROP_ID(sPropVal.ulPropTag) >= 0x8500) {
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	er = lpSink->Write(&sPropVal.ulPropTag, sizeof(sPropVal.ulPropTag), 1);
	if (er != erSuccess)
		goto exit;

	switch (type) {
	case PT_I2:
		er = lpSink->Write(&sPropVal.Value.i, sizeof(sPropVal.Value.i), 1);
		break;
	case PT_LONG:
		er = lpSink->Write(&sPropVal.Value.ul, sizeof(sPropVal.Value.ul), 1);
		break;
	case PT_R4:
		er = lpSink->Write(&sPropVal.Value.flt, sizeof(sPropVal.Value.flt), 1);
		break;
	case PT_BOOLEAN:
		b = (sPropVal.Value.b ? 1 : 0);
		er = lpSink->Write(&b, sizeof(b), 1);
		break;
	case PT_DOUBLE:
	case PT_APPTIME:
		er = lpSink->Write(&sPropVal.Value.dbl, sizeof(sPropVal.Value.dbl), 1);
		break;
	case PT_CURRENCY:
	case PT_SYSTIME:
		er = lpSink->Write(&sPropVal.Value.hilo->hi, sizeof(sPropVal.Value.hilo->hi), 1);
		if (er == erSuccess)
			er = lpSink->Write(&sPropVal.Value.hilo->lo, sizeof(sPropVal.Value.hilo->lo), 1);
		break;
	case PT_I8:
		er = lpSink->Write(&sPropVal.Value.li, sizeof(sPropVal.Value.li), 1);
		break;
	case PT_STRING8:
	case PT_UNICODE:
		ulLen = (unsigned)strlen(sPropVal.Value.lpszA);
		er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
		if (er == erSuccess)
			er = lpSink->Write(sPropVal.Value.lpszA, 1, ulLen);
		break;
	case PT_CLSID:
	case PT_BINARY:
		er = lpSink->Write(&sPropVal.Value.bin->__size, sizeof(sPropVal.Value.bin->__size), 1);
		if (er == erSuccess)
			er = lpSink->Write(sPropVal.Value.bin->__ptr, 1, sPropVal.Value.bin->__size);
		break;
	case PT_MV_I2:
		er = lpSink->Write(&sPropVal.Value.mvi.__size, sizeof(sPropVal.Value.mvi.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvi.__size; x++)
			er = lpSink->Write(&sPropVal.Value.mvi.__ptr[x], sizeof(sPropVal.Value.mvi.__ptr[x]), 1);
		break;
	case PT_MV_LONG:
		er = lpSink->Write(&sPropVal.Value.mvl.__size, sizeof(sPropVal.Value.mvl.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvl.__size; x++)
			er = lpSink->Write(&sPropVal.Value.mvl.__ptr[x], sizeof(sPropVal.Value.mvl.__ptr[x]), 1);
		break;
	case PT_MV_R4:
		er = lpSink->Write(&sPropVal.Value.mvflt.__size, sizeof(sPropVal.Value.mvflt.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvflt.__size; x++)
			er = lpSink->Write(&sPropVal.Value.mvflt.__ptr[x], sizeof(sPropVal.Value.mvflt.__ptr[x]), 1);
		break;
	case PT_MV_DOUBLE:
	case PT_MV_APPTIME:
		er = lpSink->Write(&sPropVal.Value.mvdbl.__size, sizeof(sPropVal.Value.mvdbl.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvdbl.__size; x++)
			er = lpSink->Write(&sPropVal.Value.mvdbl.__ptr[x], sizeof(sPropVal.Value.mvdbl.__ptr[x]), 1);
		break;
	case PT_MV_CURRENCY:
	case PT_MV_SYSTIME:
		er = lpSink->Write(&sPropVal.Value.mvhilo.__size, sizeof(sPropVal.Value.mvhilo.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvhilo.__size; x++) {
			er = lpSink->Write(&sPropVal.Value.mvhilo.__ptr[x].hi, sizeof(sPropVal.Value.mvhilo.__ptr[x].hi), 1);
			if (er == erSuccess)
				er = lpSink->Write(&sPropVal.Value.mvhilo.__ptr[x].lo, sizeof(sPropVal.Value.mvhilo.__ptr[x].lo), 1);
		}
		break;
	case PT_MV_BINARY:
	case PT_MV_CLSID:
		er = lpSink->Write(&sPropVal.Value.mvbin.__size, sizeof(sPropVal.Value.mvbin.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvbin.__size; x++) {
			er = lpSink->Write(&sPropVal.Value.mvbin.__ptr[x].__size, sizeof(sPropVal.Value.mvbin.__ptr[x].__size), 1);
			if (er == erSuccess)
				er = lpSink->Write(&sPropVal.Value.mvbin.__ptr[x].__ptr, 1, sPropVal.Value.mvbin.__ptr[x].__size);
		}
		break;
	case PT_MV_STRING8:
	case PT_MV_UNICODE:
		er = lpSink->Write(&sPropVal.Value.mvszA.__size, sizeof(sPropVal.Value.mvszA.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvszA.__size; x++) {
			ulLen = (unsigned)strlen(sPropVal.Value.mvszA.__ptr[x]);
			er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er == erSuccess)
				er = lpSink->Write(sPropVal.Value.mvszA.__ptr[x], 1, ulLen);
		}
		break;
	case PT_MV_I8:
		er = lpSink->Write(&sPropVal.Value.mvli.__size, sizeof(sPropVal.Value.mvli.__size), 1);
		for (int x = 0; er == erSuccess && x < sPropVal.Value.mvli.__size; x++)
			er = lpSink->Write(&sPropVal.Value.mvli.__ptr[x], sizeof(sPropVal.Value.mvli.__ptr[x]), 1);
		break;

	default:
		er = ZARAFA_E_INVALID_TYPE;
	}

exit:
	return er;
}

ECRESULT SerializeProps(ECSession *lpecSession, ECAttachmentStorage *lpAttachmentStorage, unsigned int ulObjId, unsigned int ulObjType, unsigned int ulStoreId, GUID *lpsGuid, ULONG ulFlags, ECSerializer *lpSink)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulRows = 0;
	unsigned int	ulRowsMV = 0;
	unsigned int	ulRowsNamed = 0;
	unsigned int	ulRowsNamedMV = 0;
	unsigned int	ulCount = 0;
	unsigned int	ulLen = 0;

	unsigned int	ulPropTag = 0;
	unsigned int	ulBodyFlags = 0;
	unsigned int	ulUseBody = 0;
	struct soap		*soap = NULL;
	struct propVal	sPropVal = {0};

	ECDatabase*		lpDatabase = NULL;
	DB_ROW 			lpDBRow = NULL;
	DB_LENGTHS		lpDBLen = NULL;
	DB_RESULT		lpDBResult = NULL;
	DB_RESULT		lpDBResultMV = NULL;
	DB_RESULT		lpDBResultNamed = NULL;
	DB_RESULT		lpDBResultNamedMV = NULL;
	std::string		strQuery;
	unsigned int	ulBodyTrans = 0;

	std::list<struct propVal> sPropValList;

	enum {
		HAVE_PLAINTEXT = 1,
		HAVE_HTML = 2,
		HAVE_RTF = 4
	};

	static struct {
		unsigned int ulBestBody;	// The index in ulBodyIdx for the best body
		unsigned int ulBodyCount;	// The total number of bodies found
		unsigned int ulBodyTrans;	// The number of bodies that will be send (when using the best body)
		unsigned int ulWorstBody;	// THe index in ulBodyIdx for the wordt body
	} ulaBodySelect[8] = {
		{0,					0, 0, 0},					// No bodies found, should not happen
		{PR_BODY,			1, 1, PR_BODY},				// Only plain text
		{0x10130102,		1, 1, 0x10130102},			// Only HTML
		{0x10130102,		2, 2, PR_BODY},				// Plain text and HTML
		{PR_RTF_COMPRESSED, 1, 1, PR_RTF_COMPRESSED},	// Only RTF
		{PR_RTF_COMPRESSED, 2, 2, PR_BODY},				// Plain text and RTF
		{PR_RTF_COMPRESSED, 2, 1, 0x10130102},			// HTML and RTF
		{PR_RTF_COMPRESSED, 3, 2, PR_BODY}				// Plain text, HTML and RTF
	};

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (!lpAttachmentStorage) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// We'll (ab)use a soap structure as a memory pool.
	soap = soap_new();

	
	// PR_SOURCE_KEY
	if (ECGenProps::GetPropComputed(soap, lpecSession, PR_SOURCE_KEY, ulObjId, 0, 0, ulObjType, 0, ulStoreId, NULL, &sPropVal) == erSuccess)
		sPropValList.push_back(sPropVal);

	// PR_MESSAGE_FLAGS | PR_LAST_MODIFICATION_TIME
	strQuery = "SELECT flags, UNIX_TIMESTAMP(modtime) FROM hierarchy WHERE id=" + stringify(ulObjId);
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	lpDBRow = lpDatabase->FetchRow(lpDBResult);
	if (lpDBRow && lpDBRow[0]) {
		sPropVal.ulPropTag = PR_MESSAGE_FLAGS;
		sPropVal.__union = SOAP_UNION_propValData_ul;
		sPropVal.Value.ul = atoui(lpDBRow[0]);
		sPropValList.push_back(sPropVal);
	}
	if (lpDBRow && lpDBRow[1]) {
		FILETIME ftTmp;
		UnixTimeToFileTime(atoui(lpDBRow[1]), &ftTmp);
		sPropVal.ulPropTag = PR_LAST_MODIFICATION_TIME;
		sPropVal.__union = SOAP_UNION_propValData_hilo;
		sPropVal.Value.hilo = s_alloc<struct hiloLong>(soap);
		sPropVal.Value.hilo->hi = ftTmp.dwHighDateTime;
		sPropVal.Value.hilo->lo = ftTmp.dwLowDateTime;
		sPropValList.push_back(sPropVal);
	}
	lpDatabase->FreeResult(lpDBResult);
	lpDBResult = NULL;


	// Named properties need to be handled differently as we can't just send the proptag. However the
	// range 0x8000 - 0x84FF is handled client side and will be equal on all systems. So we handle them
	// as if they are regular properties.
	strQuery = "SELECT " + (std::string)PROPCOLORDER + " FROM properties WHERE hierarchyid=" + stringify(ulObjId) + " AND tag<0x8500";
	er = lpDatabase->DoSelect(strQuery, &lpDBResult);
	if (er != erSuccess)
		goto exit;

	ulRows = lpDatabase->GetNumRows(lpDBResult);


	strQuery = 
		"SELECT " + (std::string)PROPCOLORDER + ",n.nameid,n.namestring,n.guid "
			"FROM properties JOIN names as n on (properties.tag-0x8501)=n.id "
			"WHERE properties.hierarchyid=" + stringify(ulObjId) + " AND properties.tag>=0x8500";

	er = lpDatabase->DoSelect(strQuery, &lpDBResultNamed);
	if (er != erSuccess)
		goto exit;

	ulRowsNamed = lpDatabase->GetNumRows(lpDBResultNamed);


	// Same for MV properties.
	strQuery = "SELECT " + (std::string)MVPROPCOLORDER + " FROM mvproperties WHERE hierarchyid=" + stringify(ulObjId) + " AND tag<0x8500 GROUP BY hierarchyid, tag";
	er = lpDatabase->DoSelect(strQuery, &lpDBResultMV);
	if (er != erSuccess)
		goto exit;

	ulRowsMV = lpDatabase->GetNumRows(lpDBResultMV);


	strQuery = 
		"SELECT " + (std::string)MVPROPCOLORDER + ",n.nameid,n.namestring,n.guid "
			"FROM mvproperties JOIN names as n on (mvproperties.tag-0x8501)=n.id "
			"WHERE mvproperties.hierarchyid=" + stringify(ulObjId) + " AND mvproperties.tag>=0x8500 GROUP BY hierarchyid, tag";

	er = lpDatabase->DoSelect(strQuery, &lpDBResultNamedMV);
	if (er != erSuccess)
		goto exit;

	ulRowsNamedMV = lpDatabase->GetNumRows(lpDBResultNamedMV);

	
	// See which bodies are available and select the best one.
	for (unsigned i = 0; i < ulRows; ++i) {
		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);
		if (lpDBRow == NULL || lpDBLen == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		ulPropTag = PROP_TAG(atoi(lpDBRow[FIELD_NR_TYPE]), atoi(lpDBRow[FIELD_NR_TAG]));
		switch (ulPropTag) {
		case PR_BODY:
			ulBodyFlags |= HAVE_PLAINTEXT;
			break;

		case 0x10130102:	/* PR_HTML */
			ulBodyFlags |= HAVE_HTML;
			break;

		case PR_RTF_COMPRESSED:
			ulBodyFlags |= HAVE_RTF;
			break;

		default:
			break;
		}
	}

	assert(ulBodyFlags >= 0 && ulBodyFlags < 8);
	
	if (ulFlags & (SYNC_LIMITED_IMESSAGE))
		ulBodyTrans = ulaBodySelect[ulBodyFlags].ulWorstBody ? 1 : 0; // Sending only one body
	else if(ulFlags & (SYNC_BEST_BODY))
		ulBodyTrans = ulaBodySelect[ulBodyFlags].ulBodyTrans; // Sending best body + plain so either 1 or 2 bodies (1 if best == plain)
	else
		ulBodyTrans = ulaBodySelect[ulBodyFlags].ulBodyCount; // Sending all bodies

	ulCount = ulRows + ulRowsNamed + ulRowsMV + ulRowsNamedMV - ulaBodySelect[ulBodyFlags].ulBodyCount + ulBodyTrans + sPropValList.size();
	er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
	if (er != erSuccess)
		goto exit;

	// Regular properties
	lpDatabase->ResetResult(lpDBResult);
	for (unsigned i = 0; i < ulRows; ++i) {
		lpDBRow = lpDatabase->FetchRow(lpDBResult);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);
		if (lpDBRow == NULL || lpDBLen == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		// Filter out the unwanted body properties
		if (ulFlags & (SYNC_BEST_BODY | SYNC_LIMITED_IMESSAGE)) {
			/* We only want the best body (RTF) plus plain text */
			if (ulFlags & SYNC_BEST_BODY)
				ulUseBody = ulaBodySelect[ulBodyFlags].ulBestBody;
	
			/* We only want the simplest body (plain text) */
			if (ulFlags & SYNC_LIMITED_IMESSAGE)
				ulUseBody = ulaBodySelect[ulBodyFlags].ulWorstBody;

			ulPropTag = PROP_TAG(atoi(lpDBRow[FIELD_NR_TYPE]), atoi(lpDBRow[FIELD_NR_TAG]));
			if ((ulPropTag == 0x10130102 || ulPropTag == PR_RTF_COMPRESSED) && (ulPropTag != ulUseBody))
				continue;
		}

		er = SerializeDatabasePropVal(lpDBRow, lpDBLen, lpSink);
		if (er != erSuccess)
			goto exit;
	}

	// Named properties
	for (unsigned i = 0; i < ulRowsNamed; ++i) {
		lpDBRow = lpDatabase->FetchRow(lpDBResultNamed);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResultNamed);
		if (lpDBRow == NULL || lpDBLen == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		er = SerializeDatabasePropVal(lpDBRow, lpDBLen, lpSink);
		if (er != erSuccess)
			goto exit;
	}

	// Regular MV properties
	for (unsigned i = 0; i < ulRowsMV; ++i) {
		lpDBRow = lpDatabase->FetchRow(lpDBResultMV);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResultMV);
		if (lpDBRow == NULL || lpDBLen == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		er = SerializeDatabasePropVal(lpDBRow, lpDBLen, lpSink);
		if (er != erSuccess)
			goto exit;
	}

	// Named MV properties
	for (unsigned i = 0; i < ulRowsNamedMV; ++i) {
		lpDBRow = lpDatabase->FetchRow(lpDBResultNamedMV);
		lpDBLen = lpDatabase->FetchRowLengths(lpDBResultNamedMV);
		if (lpDBRow == NULL || lpDBLen == NULL) {
			er = ZARAFA_E_DATABASE_ERROR;
			goto exit;
		}

		er = SerializeDatabasePropVal(lpDBRow, lpDBLen, lpSink);
		if (er != erSuccess)
			goto exit;
	}

	for (std::list<struct propVal>::const_iterator it = sPropValList.begin(); it != sPropValList.end(); ++it) {
		er = SerializePropVal(*it, lpSink);
		if (er != erSuccess)
			goto exit;
	}

	if (ulObjType == MAPI_ATTACH) {
		
		if (lpAttachmentStorage->ExistAttachment(ulObjId)) {
			er = lpAttachmentStorage->GetSize(ulObjId, (ULONG*)&ulLen);
			if (er != erSuccess)
				goto exit;

			er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er != erSuccess)
				goto exit;

			er = lpAttachmentStorage->LoadAttachment(ulObjId, (int*)&ulLen, lpSink);
			if (er != erSuccess)
				goto exit;
		} else {
			ulLen = 0;
			er = lpSink->Write(&ulLen, sizeof(ulLen), 1);
			if (er != erSuccess)
				goto exit;
		}
	}



exit:
	if (lpDatabase) {
		if (lpDBResultNamedMV)
			lpDatabase->FreeResult(lpDBResultNamedMV);
		if (lpDBResultMV)
			lpDatabase->FreeResult(lpDBResultMV);
		if (lpDBResultNamed)
			lpDatabase->FreeResult(lpDBResultNamed);
		if (lpDBResult)
			lpDatabase->FreeResult(lpDBResult);
	}
	
	if (soap) {
		soap_end(soap);
		soap_free(soap);
	}

	return er;
}

ECRESULT SerializeObject(ECSession *lpecSession, ECAttachmentStorage *lpAttachmentStorage, unsigned int ulObjId, unsigned int ulStoreId, GUID *lpsGuid, ULONG ulFlags, ECSerializer *lpSink, bool bTopLevel)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulStreamVersion = 0;
	unsigned int	ulObjType = 0;
	unsigned int	ulSubObjId = 0;
	unsigned int	ulSubObjType = 0;
	unsigned int	ulCount = 0;

	ECDatabase*		lpDatabase = NULL;
	DB_ROW 			lpDBRow = NULL;
	DB_LENGTHS		lpDBLen = NULL;
	DB_RESULT		lpDBResult = NULL;
	std::string		strQuery;

	lpDatabase = lpecSession->GetDatabase();
	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, NULL, NULL, NULL, &ulObjType);
	if (er != erSuccess)
		goto exit;

	if (ulObjType != MAPI_MESSAGE && ulObjType != MAPI_ATTACH && ulObjType != MAPI_MAILUSER && ulObjType != MAPI_DISTLIST) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	if (bTopLevel) {
		er = lpSink->Write(&ulStreamVersion, sizeof(ulStreamVersion), 1);
		if (er != erSuccess)
			goto exit;
	}

	er = SerializeProps(lpecSession, lpAttachmentStorage, ulObjId, ulObjType, ulStoreId, lpsGuid, ulFlags, lpSink);
	if (er != erSuccess)
		goto exit;

	if (ulObjType == MAPI_MESSAGE || ulObjType == MAPI_ATTACH) {
		// Serialize sub objects
		strQuery = "SELECT id,type FROM hierarchy WHERE parent=" + stringify(ulObjId);
		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if (er != erSuccess)
			goto exit;

		ulCount = lpDatabase->GetNumRows(lpDBResult);
		er = lpSink->Write(&ulCount, sizeof(ulCount), 1);
		if (er != erSuccess)
			goto exit;

		for (unsigned i = 0; i < ulCount; ++i) {
			lpDBRow = lpDatabase->FetchRow(lpDBResult);
			lpDBLen = lpDatabase->FetchRowLengths(lpDBResult);

			if (lpDBRow == NULL || lpDBLen == NULL) {
				er = ZARAFA_E_DATABASE_ERROR;
				goto exit;
			}

			ulSubObjType = atoi(lpDBRow[1]);
			er = lpSink->Write(&ulSubObjType, sizeof(ulSubObjType), 1);
			if (er != erSuccess)
				goto exit;

			ulSubObjId = atoi(lpDBRow[0]);
			er = lpSink->Write(&ulSubObjId, sizeof(ulSubObjId), 1);
			if (er != erSuccess)
				goto exit;

			er = SerializeObject(lpecSession, lpAttachmentStorage, ulSubObjId, ulStoreId, lpsGuid, ulFlags, lpSink);
			if (er != erSuccess)
				goto exit;
		}
	}


exit:
	if (lpDatabase) {
		if (lpDBResult)
			lpDatabase->FreeResult(lpDBResult);
	}

	return er;
}

ECRESULT DeserializePropVal(struct soap *soap, NamedPropertyMapper &namedPropertyMapper, propVal **lppsPropval, ECSerializer *lpSource)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulCount;
	unsigned int	ulLen;
	propVal			*lpsPropval = NULL;
	unsigned char	b;

	GUID			guid = {0};
	unsigned int	ulKind = 0;
	unsigned int	ulNameId = 0;
	std::string		strNameString;
	unsigned int	ulLocalId = 0;

	lpsPropval = s_alloc<propVal>(soap);
	er = lpSource->Read(&lpsPropval->ulPropTag, sizeof(lpsPropval->ulPropTag), 1);
	if (er != erSuccess)
		goto exit;

	switch (PROP_TYPE(lpsPropval->ulPropTag)) {
	case PT_I2:
		lpsPropval->__union = SOAP_UNION_propValData_i;
		er = lpSource->Read(&lpsPropval->Value.i, sizeof(lpsPropval->Value.i), 1);
		break;
	case PT_LONG:
		lpsPropval->__union = SOAP_UNION_propValData_ul;
		er = lpSource->Read(&lpsPropval->Value.ul, sizeof(lpsPropval->Value.ul), 1);
		break;
	case PT_R4:
		lpsPropval->__union = SOAP_UNION_propValData_flt;
		er = lpSource->Read(&lpsPropval->Value.flt, sizeof(lpsPropval->Value.flt), 1);
		break;
	case PT_BOOLEAN:
		lpsPropval->__union = SOAP_UNION_propValData_b;
		er = lpSource->Read(&b, sizeof(b), 1);
		lpsPropval->Value.b = (b != 0);
		break;
	case PT_DOUBLE:
	case PT_APPTIME:
		lpsPropval->__union = SOAP_UNION_propValData_dbl;
		er = lpSource->Read(&lpsPropval->Value.dbl, sizeof(lpsPropval->Value.dbl), 1);
		break;
	case PT_CURRENCY:
	case PT_SYSTIME:
		lpsPropval->__union = SOAP_UNION_propValData_hilo;
		lpsPropval->Value.hilo = s_alloc<hiloLong>(soap);
		er = lpSource->Read(&lpsPropval->Value.hilo->hi, sizeof(lpsPropval->Value.hilo->hi), 1);
		if (er == erSuccess)
			er = lpSource->Read(&lpsPropval->Value.hilo->lo, sizeof(lpsPropval->Value.hilo->lo), 1);
		break;
	case PT_I8:
		lpsPropval->__union = SOAP_UNION_propValData_li;
		er = lpSource->Read(&lpsPropval->Value.li, sizeof(lpsPropval->Value.li), 1);
		break;
	case PT_STRING8:
	case PT_UNICODE:
		lpsPropval->__union = SOAP_UNION_propValData_lpszA;
		er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
		if (er == erSuccess) {
			lpsPropval->Value.lpszA = s_alloc<char>(soap, ulLen + 1);
			memset(lpsPropval->Value.lpszA, 0, ulLen + 1);
			er = lpSource->Read(lpsPropval->Value.lpszA, 1, ulLen);
		}
		break;
	case PT_CLSID:
	case PT_BINARY:
		lpsPropval->__union = SOAP_UNION_propValData_bin;
		er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
		if (er == erSuccess) {
			lpsPropval->Value.bin = s_alloc<xsd__base64Binary>(soap);
			lpsPropval->Value.bin->__size = ulLen;
			lpsPropval->Value.bin->__ptr = s_alloc<unsigned char>(soap, ulLen);
			er = lpSource->Read(lpsPropval->Value.bin->__ptr, 1, ulLen);
		}
		break;
	case PT_MV_I2:
		lpsPropval->__union = SOAP_UNION_propValData_mvi;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvi.__size = ulCount;
			lpsPropval->Value.mvi.__ptr = s_alloc<short>(soap, ulCount);
			er = lpSource->Read(lpsPropval->Value.mvi.__ptr, sizeof *lpsPropval->Value.mvi.__ptr, ulCount);
		}
		break;
	case PT_MV_LONG:
		lpsPropval->__union = SOAP_UNION_propValData_mvl;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvl.__size = ulCount;
			lpsPropval->Value.mvl.__ptr = s_alloc<unsigned int>(soap, ulCount);
			er = lpSource->Read(lpsPropval->Value.mvl.__ptr, sizeof *lpsPropval->Value.mvl.__ptr, ulCount);
		}
		break;
	case PT_MV_R4:
		lpsPropval->__union = SOAP_UNION_propValData_mvflt;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvflt.__size = ulCount;
			lpsPropval->Value.mvflt.__ptr = s_alloc<float>(soap, ulCount);
			er = lpSource->Read(lpsPropval->Value.mvflt.__ptr, sizeof *lpsPropval->Value.mvflt.__ptr, ulCount);
		}
		break;
	case PT_MV_DOUBLE:
	case PT_MV_APPTIME:
		lpsPropval->__union = SOAP_UNION_propValData_mvdbl;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvdbl.__size = ulCount;
			lpsPropval->Value.mvdbl.__ptr = s_alloc<double>(soap, ulCount);
			er = lpSource->Read(lpsPropval->Value.mvdbl.__ptr, sizeof *lpsPropval->Value.mvdbl.__ptr, ulCount);
		}
		break;
	case PT_MV_CURRENCY:
	case PT_MV_SYSTIME:
		lpsPropval->__union = SOAP_UNION_propValData_mvhilo;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvhilo.__size = ulCount;
			lpsPropval->Value.mvhilo.__ptr = s_alloc<hiloLong>(soap, ulCount);
			for (unsigned x = 0; er == erSuccess &&  x < ulCount; x++) {
				er = lpSource->Read(&lpsPropval->Value.mvhilo.__ptr[x].hi, sizeof(lpsPropval->Value.mvhilo.__ptr[x].hi), ulCount);
				if (er == erSuccess)
					er = lpSource->Read(&lpsPropval->Value.mvhilo.__ptr[x].lo, sizeof(lpsPropval->Value.mvhilo.__ptr[x].lo), ulCount);
			}
		}
		break;
	case PT_MV_BINARY:
	case PT_MV_CLSID:
		lpsPropval->__union = SOAP_UNION_propValData_mvbin;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvbin.__size = ulCount;
			lpsPropval->Value.mvbin.__ptr = s_alloc<xsd__base64Binary>(soap, ulCount);
			for (unsigned x = 0; er == erSuccess && x < ulCount; x++) {
				er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
				if (er == erSuccess) {
					lpsPropval->Value.mvbin.__ptr[x].__size = ulLen;
					lpsPropval->Value.mvbin.__ptr[x].__ptr = s_alloc<unsigned char>(soap, ulLen);
					er = lpSource->Read(lpsPropval->Value.mvbin.__ptr[x].__ptr, 1, ulLen);
				}
			}
		}
		break;
	case PT_MV_STRING8:
	case PT_MV_UNICODE:
		lpsPropval->__union = SOAP_UNION_propValData_mvszA;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvszA.__size = ulCount;
			lpsPropval->Value.mvszA.__ptr = s_alloc<char*>(soap, ulCount);
			for (unsigned x = 0; er == erSuccess && x < ulCount; x++) {
				er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
				if (er == erSuccess) {
					lpsPropval->Value.mvszA.__ptr[x] = s_alloc<char>(soap, ulLen + 1);
					memset(lpsPropval->Value.mvszA.__ptr[x], 0, ulLen + 1);
					er = lpSource->Read(lpsPropval->Value.mvszA.__ptr[x], 1, ulLen);
				}
			}
		}
		break;
	case PT_MV_I8:
		lpsPropval->__union = SOAP_UNION_propValData_mvli;
		er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
		if (er == erSuccess) {
			lpsPropval->Value.mvli.__size = ulCount;
			lpsPropval->Value.mvli.__ptr = s_alloc<LONG64>(soap, ulCount);
			er = lpSource->Read(lpsPropval->Value.mvli.__ptr, sizeof *lpsPropval->Value.mvli.__ptr, ulCount);
		}
		break;
	default:
		er = ZARAFA_E_INVALID_TYPE;
		goto exit;
	}

	// If the proptag is in the dynamic named property range, we need to get the correct local proptag
	if (PROP_ID(lpsPropval->ulPropTag) >= 0x8500) {
		er = lpSource->Read(&guid, 1, sizeof(guid));
		if (er == erSuccess)
			er = lpSource->Read(&ulKind, sizeof(ulKind), 1);
		if (er == erSuccess && ulKind == MNID_ID) {
			er = lpSource->Read(&ulNameId, sizeof(ulNameId), 1);
			if (er == erSuccess)
				er = namedPropertyMapper.GetId(guid, ulNameId, &ulLocalId);
		} else if (er == erSuccess && ulKind == MNID_STRING) {
			er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
			if (er == erSuccess) {
				strNameString.resize(ulLen, 0);
				er = lpSource->Read((void*)strNameString.data(), 1, ulLen);
			}
			if (er == erSuccess)
				er = namedPropertyMapper.GetId(guid, strNameString, &ulLocalId);
		}
		if (er == erSuccess)
			lpsPropval->ulPropTag = PROP_TAG(PROP_TYPE(lpsPropval->ulPropTag), ulLocalId);
	}

	*lppsPropval = lpsPropval;

exit:
	return er;
}

ECRESULT DeserializeProps(ECSession *lpecSession, ECDatabase *lpDatabase, ECAttachmentStorage *lpAttachmentStorage, unsigned int ulObjId, unsigned int ulObjType, unsigned int ulStoreId, GUID *lpsGuid, bool bNewItem, ECSerializer *lpSource, struct propValArray **lppPropValArray)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulCount = 0;
	unsigned int	ulFlags = 0;
	unsigned int	ulParentId = 0;
	unsigned int	ulOwner = 0;
	unsigned int	ulParentType = 0;
	unsigned int	nMVItems = 0;
	unsigned int	ulAffected = 0;
	unsigned int	ulColId = 0;
	unsigned int	ulPropInserts = 0;
	unsigned int	ulLen = 0;
	propVal			*lpsPropval = NULL;
	struct soap		*soap = NULL;
	struct propValArray *lpPropValArray = NULL;
	NamedPropertyMapper namedPropertyMapper(lpDatabase);

	std::string		strQuery;
	std::string		strInsertQuery;
	std::string		strColData;
	std::string		strColName;

	SOURCEKEY		sSourceKey;
	unsigned long long	ullIMAP = 0;

	DB_RESULT		lpDBResult = NULL;
	DB_ROW			lpDBRow = NULL;

	std::set<unsigned int>				setInserted;
	std::set<unsigned int>::iterator	iterInserted;

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	if (!lpAttachmentStorage) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, &ulParentId, &ulOwner, &ulFlags, NULL);
	if (er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentId, NULL, NULL, NULL, &ulParentType);
	if (er != erSuccess)
		goto exit;

	// Get the number of properties
	er = lpSource->Read(&ulCount, sizeof(ulCount), 1);
	if (er != erSuccess)
		goto exit;
		
	if (lppPropValArray) {
		// If requested we can store upto ulCount properties. Currently we won't store them all though.
		// Note that we test on lppPropValArray but allocate lpPropValArray. We'll assign that to
		// *lppPropValArray later if all went well.
		lpPropValArray = s_alloc<struct propValArray>(NULL);
		
		lpPropValArray->__ptr = s_alloc<struct propVal>(NULL, ulCount);
		memset(lpPropValArray->__ptr, 0, sizeof(struct propVal) * ulCount);
		
		lpPropValArray->__size = 0;
	}

	for (unsigned i = 0; i < ulCount; ++i) {
		// We'll (ab)use a soap structure as a memory pool.
		ASSERT(soap == NULL);
		soap = soap_new();

		er = DeserializePropVal(soap, namedPropertyMapper, &lpsPropval, lpSource);
		if (er != erSuccess)
			goto exit;

		iterInserted = setInserted.find(lpsPropval->ulPropTag);
		if (iterInserted != setInserted.end())
			goto next_property;

		if (ECGenProps::IsPropRedundant(lpsPropval->ulPropTag, ulObjType) == erSuccess)
			goto next_property;

		if (lpsPropval->ulPropTag == PR_LAST_MODIFICATION_TIME) {
			if (!bNewItem)
				goto next_property;
			if (lpsPropval->Value.hilo == NULL)
				goto next_property;

			strQuery = "UPDATE hierarchy SET modtime='" + stringify_datetime(FileTimeToUnixTime(lpsPropval->Value.hilo->hi, lpsPropval->Value.hilo->lo)) + "' WHERE id=" + stringify(ulObjId);
			er = lpDatabase->DoUpdate(strQuery);
			if (er != erSuccess)
				goto exit;

			goto next_property;
		}

		if (lpsPropval->ulPropTag == PR_MESSAGE_FLAGS) {
			// Store this property for later use.
			if (lppPropValArray) {
				lpPropValArray->__ptr[(lpPropValArray)->__size] = *lpsPropval;
				lpPropValArray->__size++;
			}
		
			if (!bNewItem)
				goto next_property;

			ulFlags |= lpsPropval->Value.ul & (MSGFLAG_SETTABLE_BY_USER | MSGFLAG_SETTABLE_BY_SPOOLER);

			strQuery = "UPDATE hierarchy SET flags=" + stringify(ulFlags) + " WHERE id=" + stringify(ulObjId);
			er = lpDatabase->DoUpdate(strQuery);
			if (er != erSuccess)
				goto exit;

			goto next_property;
		}

		if (lpsPropval->ulPropTag == PR_SEARCH_KEY) {
			if (!bNewItem)
				goto next_property;
		}

		if (lpsPropval->ulPropTag == PR_SOURCE_KEY) {
			if (!(bNewItem && ulParentType == MAPI_FOLDER))
				goto next_property;

			// don't use the sourcekey if found.
			// Don't query the cache as that can be out of sync with the db in rare occasions.
			strQuery = 
				"SELECT hierarchyid FROM indexedproperties "
					"WHERE tag=" + stringify(PROP_ID(PR_SOURCE_KEY)) + 
					" AND val_binary=" + lpDatabase->EscapeBinary(lpsPropval->Value.bin->__ptr, lpsPropval->Value.bin->__size);
			
			er = lpDatabase->DoSelect(strQuery, &lpDBResult);
			if(er != erSuccess)
				goto exit;

			lpDBRow = lpDatabase->FetchRow(lpDBResult);
			lpDatabase->FreeResult(lpDBResult); 
			lpDBResult = NULL;

			// We can't use lpDBRow here except for checking if it was NULL.
			if (lpDBRow != NULL)
				continue;


			strQuery = "REPLACE INTO indexedproperties(hierarchyid,tag,val_binary) VALUES (" + stringify(ulObjId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(lpsPropval->Value.bin->__ptr, lpsPropval->Value.bin->__size) + ")";
			er = lpDatabase->DoInsert(strQuery);
			if (er != erSuccess)
				goto exit;

			setInserted.insert(lpsPropval->ulPropTag);

			g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), lpsPropval->Value.bin->__size, lpsPropval->Value.bin->__ptr, ulObjId);
			goto next_property;
		}

		if ((PROP_TYPE(lpsPropval->ulPropTag) & MV_FLAG) == MV_FLAG) {
			if (!bNewItem) {
				strQuery = "DELETE FROM mvproperties WHERE hierarchyid=" + stringify(ulObjId) + " AND tag=" + stringify(PROP_ID(lpsPropval->ulPropTag)) + " AND type=" + stringify(PROP_TYPE(lpsPropval->ulPropTag));
				er = lpDatabase->DoDelete(strQuery);
				if (er != erSuccess)
					goto exit;
			}

			nMVItems = GetMVItemCount(lpsPropval);
			for (unsigned j = 0; j < nMVItems; ++j) {
				er = CopySOAPPropValToDatabaseMVPropVal(lpsPropval, j, strColName, strColData, lpDatabase);
				if (er != erSuccess) {
					er = erSuccess;
					goto next_property;
				}

				strQuery = "INSERT INTO mvproperties(hierarchyid,orderid,tag,type," + strColName + ") VALUES(" + stringify(ulObjId) + "," + stringify(j) + "," + stringify(PROP_ID(lpsPropval->ulPropTag)) + "," + stringify(PROP_TYPE(lpsPropval->ulPropTag)) + "," + strColData + ")";
				er = lpDatabase->DoInsert(strQuery, NULL, &ulAffected);
				if (er != erSuccess)
					goto exit;
				if (ulAffected != 1) {
					er = ZARAFA_E_DATABASE_ERROR;
					goto exit;
				}
			}
		} else {
			er = CopySOAPPropValToDatabasePropVal(lpsPropval, &ulColId, strColData, lpDatabase);
			if (er != erSuccess) {
				er = erSuccess;
				goto next_property;
			}

			if (!strInsertQuery.empty() && strInsertQuery.size() + strColData.size() * 2 > lpDatabase->GetMaxAllowedPacket()) {
				er = lpDatabase->DoInsert(strInsertQuery);
				if (er != erSuccess)
					goto exit;

				strInsertQuery.clear();
			}

			if (strInsertQuery.empty()) {
				strInsertQuery = "REPLACE INTO properties (hierarchyid,tag,type,storeid," + (std::string)PROPCOLVALUEORDER(properties) + ") VALUES";
				ulPropInserts = 0;
			}

			if (ulPropInserts > 0)
				strInsertQuery += ",";

			strInsertQuery += "(" + stringify(ulObjId) + "," + stringify(PROP_ID(lpsPropval->ulPropTag)) + "," + stringify(PROP_TYPE(lpsPropval->ulPropTag)) + "," + stringify(ulStoreId) + ",";
			for (unsigned j = 0; j < VALUE_NR_MAX; ++j) {
				if (j == ulColId) {
					strInsertQuery += strColData;
				} else {
					if (j == VALUE_NR_HILO)
						strInsertQuery += "null,null";
					else
						strInsertQuery += "null";
				}
				if (j != VALUE_NR_MAX - 1)
					strInsertQuery += ",";
			}

			strInsertQuery += ")";
			ulPropInserts++;
		}

		setInserted.insert(lpsPropval->ulPropTag);

next_property:
		soap_end(soap);
		soap_free(soap);
		soap = NULL;
	}

	if (ulPropInserts > 0) {
		er = lpDatabase->DoInsert(strInsertQuery);
		if (er != erSuccess)
			goto exit;
	}

	if (bNewItem && ulObjType == MAPI_MESSAGE && ulParentType == MAPI_FOLDER) {
		iterInserted = setInserted.find(PR_SOURCE_KEY);
		if (iterInserted == setInserted.end()) {
			er = lpecSession->GetNewSourceKey(&sSourceKey);
			if (er != erSuccess)
				goto exit;

			strQuery = "INSERT INTO indexedproperties(hierarchyid,tag,val_binary) VALUES(" + stringify(ulObjId) + "," + stringify(PROP_ID(PR_SOURCE_KEY)) + "," + lpDatabase->EscapeBinary(sSourceKey, sSourceKey.size()) + ")";
			er = lpDatabase->DoInsert(strQuery);
			if (er != erSuccess)
				goto exit;

			g_lpSessionManager->GetCacheManager()->SetObjectProp(PROP_ID(PR_SOURCE_KEY), sSourceKey.size(), sSourceKey, ulObjId);
		}

		er = g_lpSessionManager->GetNewSequence(ECSessionManager::SEQ_IMAP, &ullIMAP);
		if (er != erSuccess)
			goto exit;

		strQuery = "INSERT INTO properties(hierarchyid, tag, type, storeid, val_ulong) VALUES(" +
						stringify(ulObjId) + "," +
						stringify(PROP_ID(PR_EC_IMAP_ID)) + "," +
						stringify(PROP_TYPE(PR_EC_IMAP_ID)) + "," +
						stringify(ulStoreId) + "," +
						stringify_int64(ullIMAP) + ")";
		er = lpDatabase->DoInsert(strQuery);
		if (er != erSuccess)
			goto exit;
	}

	if (ulObjType == MAPI_ATTACH) {

		er = lpSource->Read(&ulLen, sizeof(ulLen), 1);
		if (er != erSuccess)
			goto exit;

		er = lpAttachmentStorage->SaveAttachment(ulObjId, true, ulLen, lpSource, NULL);	//FIXME: What to do with the instance id???
		if (er != erSuccess)
			goto exit;
	}

	if (!bNewItem)
		g_lpSessionManager->GetCacheManager()->Update(fnevObjectModified, ulObjId);

	g_lpSessionManager->GetCacheManager()->SetObject(ulObjId, ulParentId, ulOwner, ulFlags, ulObjType);
	
	if (lpPropValArray) {
		ASSERT(lppPropValArray != NULL);
		*lppPropValArray = lpPropValArray;
		lpPropValArray = NULL;
	}

exit:
	if (lpPropValArray)
		FreePropValArray(lpPropValArray, true);

	if (soap) {
		soap_end(soap);
		soap_free(soap);
	}

	return er;
}

ECRESULT DeserializeObject(ECSession *lpecSession, ECDatabase *lpDatabase, ECAttachmentStorage *lpAttachmentStorage, unsigned int ulObjId, unsigned int ulStoreId, GUID *lpsGuid, bool bNewItem, ECSerializer *lpSource, bool bTopLevel, struct propValArray **lppPropValArray)
{
	ECRESULT		er = erSuccess;
	unsigned int	ulStreamVersion = 0;
	unsigned int	ulObjType = 0;
	unsigned int	ulSubObjCount = 0;
	unsigned int	ulSubObjId = 0;
	unsigned int	ulSubObjType = 0;
	unsigned int	ulParentId = 0;
	unsigned int	ulParentType = 0;
	unsigned int	ulSize =0 ;
	struct propValArray *lpPropValArray = NULL;

	if (!lpDatabase) {
		er = ZARAFA_E_DATABASE_ERROR;
		goto exit;
	}

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulObjId, &ulParentId, NULL, NULL, &ulObjType);
	if (er != erSuccess)
		goto exit;

	er = g_lpSessionManager->GetCacheManager()->GetObject(ulParentId, NULL, NULL, NULL, &ulParentType);
	if (er != erSuccess)
		goto exit;

	if (ulObjType != MAPI_MESSAGE && ulObjType != MAPI_ATTACH && ulObjType != MAPI_MAILUSER && ulObjType != MAPI_DISTLIST) {
		er = ZARAFA_E_NO_SUPPORT;
		goto exit;
	}

	if (bTopLevel) {
		er = lpSource->Read(&ulStreamVersion, sizeof(ulStreamVersion), 1);
		if (er != erSuccess)
			goto exit;

		if (ulStreamVersion != 0) {
			er = ZARAFA_E_NO_SUPPORT;
			goto exit;
		}
	}

	er = DeserializeProps(lpecSession, lpDatabase, lpAttachmentStorage, ulObjId, ulObjType, ulStoreId, lpsGuid, bNewItem, lpSource, lppPropValArray ? &lpPropValArray : NULL);
	if (er != erSuccess)
		goto exit;


	if (ulObjType == MAPI_MESSAGE || ulObjType == MAPI_ATTACH) {
		er = lpSource->Read(&ulSubObjCount, sizeof(ulSubObjCount), 1);
		if (er != erSuccess)
			goto exit;

		for (unsigned i = 0; i < ulSubObjCount; ++i) {
			er = lpSource->Read(&ulSubObjType, sizeof(ulSubObjType), 1);
			if (er != erSuccess)
				goto exit;

			er = lpSource->Read(&ulSubObjId, sizeof(ulSubObjId), 1);
			if (er != erSuccess)
				goto exit;

			if (bNewItem) {
				/**
				 * For new items we're not interested in the ulSubObjId from the stream, we do need
				 * to create the object with the current object as its parent
				 **/
				er = CreateObject(lpecSession, lpDatabase, ulObjId, ulSubObjType, 0, &ulSubObjId);
				if (er != erSuccess)
					goto exit;
			} else {
				/**
				 * We need to map the object id from the stream to the local id (or change the protocol)
				 **/
			}

			er = DeserializeObject(lpecSession, lpDatabase, lpAttachmentStorage, ulSubObjId, ulStoreId, lpsGuid, bNewItem, lpSource, false, NULL);
			if (er != erSuccess)
				goto exit;
		}

		// Calc size of object, now that all children are saved.

		// Remove old size
		if (bNewItem == false && ulObjType == MAPI_MESSAGE && ulParentType == MAPI_FOLDER) {
			if (GetObjectSize(lpDatabase, ulStoreId, ulObjId, &ulSize) == erSuccess)
				UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, false, UPDATE_SUB, ulSize);
		}

		// Add new size
		if (CalculateObjectSize(lpDatabase, ulObjId, ulObjType, ulStoreId, &ulSize) == erSuccess) {
			UpdateObjectSize(lpDatabase, ulStoreId, ulObjId, ulObjType, bNewItem, UPDATE_NEW, ulSize);

			if (ulObjType == MAPI_MESSAGE && ulParentType == MAPI_FOLDER) {
				UpdateObjectSize(lpDatabase, ulStoreId, ulStoreId, MAPI_STORE, false, UPDATE_ADD, ulSize);
			}

		} else {
			ASSERT(FALSE);
		}
	}
	
	if (lpPropValArray) {
		ASSERT(lppPropValArray != NULL);
		*lppPropValArray = lpPropValArray;
		lpPropValArray = NULL;
	}

exit:
	if (lpPropValArray)
		FreePropValArray(lpPropValArray, true);

	return er;
}
