/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

// STL defines
#include <set>
#include <string>
#include <list>
#include <map>

// mapi defines
#include <mapidefs.h>
#include <mapitags.h>
#include "mapiext.h"
#include <EMSAbTag.h>
#include <edkmdb.h>
#include "ECMAPI.h"

#include "soapH.h"

#include "ECSessionManager.h"
#include "ECSecurity.h"
#include "ZarafaICS.h"
#include "ECICS.h"
#include "StorageUtil.h"
#include "ECAttachmentStorage.h"

#include "ZarafaCmdUtil.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern ECSessionManager*    g_lpSessionManager; // FIXME: remove this global and change the depended source code!

ECRESULT GetSourceKey(unsigned int ulObjId, SOURCEKEY *lpSourceKey)
{
	ECRESULT er = erSuccess;
	unsigned char *lpData = NULL;
	unsigned int cbData = 0;

	er = g_lpSessionManager->GetCacheManager()->GetPropFromObject(PROP_ID(PR_SOURCE_KEY), ulObjId, NULL, &cbData, &lpData);
	if(er != erSuccess)
		goto exit;

	*lpSourceKey = SOURCEKEY(cbData, (char *)lpData);

exit:
	if(lpData)
		delete [] lpData;
	return er;
}

/*
 * This is a generic delete function that is called from
 *
 * ns__deleteObjects
 * ns__emptyFolder
 * ns__deleteFolder
 * purgeSoftDelete
 *
 * Functions which using sub set of the delete system are:
 * ns__saveObject
 * importMessageFromStream
 *
 * It does a recursive delete of objects in the hierarchytable, according to the flags given
 * which is any combination of
 *
 * EC_DELETE_FOLDERS		- Delete subfolders
 * EC_DELETE_MESSAGES		- Delete messages
 * EC_DELETE_RECIPIENTS		- Delete recipients of messages
 * EC_DELETE_ATTACHMENTS	- Delete attachments of messages
 * EC_DELETE_CONTAINER		- Delete the container specified in the first place (otherwise only subobjects)
 *
 * This is done by first recusively retrieving the object ID's, then checking the types of objects in that
 * list. If there is any subobject that has *not* been specified for deletion, the function fails. Else,
 * all properties in the subobjects and the subobjects themselves are deleted. If EC_DELETE_CONTAINER
 * is specified, then the objects passed in lpEntryList are also deleted (together with their properties).
 *
 */

/**
 * Validate permissions and match object type
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] bCheckPermission Check if the object folder or message has delete permissions.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted. 
 * @param[in] sItem Reference to a DELETEITEM structure that contains object information which identifying the folder, message, reciptient and attachment.
 *
 * @return Zarafa error code
 */
ECRESULT ValidateDeleteObject(ECSession *lpSession, bool bCheckPermission, unsigned int ulFlags, DELETEITEM &sItem)
{
	ECRESULT er = erSuccess;

	if (lpSession == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	// Check permission for each folder and messages
	if (bCheckPermission && ((sItem.ulObjType == MAPI_MESSAGE && sItem.ulParentType == MAPI_FOLDER) || sItem.ulObjType == MAPI_FOLDER)) {
		er = lpSession->GetSecurity()->CheckPermission(sItem.ulId, ecSecurityDelete);
		if(er != erSuccess)
			goto exit;
	}

	if (sItem.fRoot == true)
		goto exit; // Not for a root

	switch(sItem.ulObjType) {
	case MAPI_MESSAGE:
		if(! (ulFlags & EC_DELETE_MESSAGES) ) {
			er = ZARAFA_E_HAS_MESSAGES;
			goto exit;
		}
		break;
	case MAPI_FOLDER:
		if(! (ulFlags & EC_DELETE_FOLDERS) ) {
			er = ZARAFA_E_HAS_FOLDERS;
			goto exit;
		}
		break;
	case MAPI_MAILUSER:
	case MAPI_DISTLIST:
		if(! (ulFlags & EC_DELETE_RECIPIENTS) ) {
			er = ZARAFA_E_HAS_RECIPIENTS;
			goto exit;
		}
		break;
	case MAPI_ATTACH:
		if(! (ulFlags & EC_DELETE_ATTACHMENTS) ) {
			er = ZARAFA_E_HAS_ATTACHMENTS;
			goto exit;
		}
		break;
	case MAPI_STORE: // only admins can delete a store, rights checked in ns__removeStore
		if(! (ulFlags & EC_DELETE_STORES) ) {
			er = ZARAFA_E_NOT_FOUND;
			goto exit;
		}
		break;
	default:
		// Unknown object type? We'll delete it anyway so we don't get frustrating non-deletable items
		ASSERT(FALSE); // Only frustrating developers!
		break;
	}

exit:
	return er;
}

/**
 * Expand a list of objects, validate permissions and object types
 * This function returns a list of all items that need to be
 * deleted. This may or may not include the given list in
 * lpsObjectList, because of EC_DELETE_CONTAINER.
 *
 * If the ulFLags includes EC_DELETE_NOT_ASSOCIATED_MSG only the associated messages for the container folder is 
 * not deleted. If ulFlags EC_DELETE_CONTAINER included, the EC_DELETE_NOT_ASSOCIATED_MSG flag will be ignored.
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] lpDatabase Reference to a database object; cannot be NULL.
 * @param[in] lpsObjectList Reference to a list of objects that contains itemid and must be expanded.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] bCheckPermission Check the objects delete permissions.
 * @param[out] lplstDeleteItems Recursive list with objects
 *
 * @return Zarafa error code
 */
ECRESULT ExpandDeletedItems(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt *lpsObjectList, unsigned int ulFlags, bool bCheckPermission, ECListDeleteItems *lplstDeleteItems)
{
	ECRESULT er = erSuccess;
	
	ECListIntIterator	iListObjectId;

	DB_RESULT lpDBResult = NULL;
	DB_ROW lpDBRow = NULL;
	std::string strQuery;
	std::set<unsigned int> setIDs;
	ECListDeleteItems lstDeleteItems;
	ECListDeleteItems::iterator iterDeleteItems;
	ECListDeleteItems lstContainerItems;
	DELETEITEM sItem;
	ECSessionManager *lpSessionManager = NULL;
	ECCacheManager *lpCacheManager = NULL;
	
	if (lpSession == NULL || lpDatabase == NULL || lpsObjectList == NULL || lplstDeleteItems == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpSessionManager = lpSession->GetSessionManager();
	lpCacheManager = lpSessionManager->GetCacheManager();

	// First, put all the root objects in the list
	for(iListObjectId = lpsObjectList->begin(); iListObjectId != lpsObjectList->end(); iListObjectId++) {

		sItem.fRoot = true;
		
		//Free database results
		if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

		strQuery = "SELECT h.id, h.parent, h.type, h.flags, p.type FROM hierarchy AS h LEFT JOIN hierarchy AS p ON h.parent=p.id WHERE ";
		if((ulFlags & EC_DELETE_CONTAINER) == 0)
			strQuery += "h.parent=" + stringify(*iListObjectId);
		else
			strQuery += "h.id=" + stringify(*iListObjectId);

		if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE)
			strQuery += " AND (h.flags&"+stringify(MSGFLAG_DELETED)+") !="+stringify(MSGFLAG_DELETED);

		if ((ulFlags & EC_DELETE_CONTAINER) == 0 && (ulFlags & EC_DELETE_NOT_ASSOCIATED_MSG) == EC_DELETE_NOT_ASSOCIATED_MSG)
			strQuery += " AND (h.flags&"+stringify(MSGFLAG_ASSOCIATED)+") !="+stringify(MSGFLAG_ASSOCIATED);

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		while ( (lpDBRow = lpDatabase->FetchRow(lpDBResult)) )
		{
			// No type or flags exist
			if(lpDBRow[2] == NULL || lpDBRow[3] == NULL) {
				//er = ZARAFA_E_DATABASE_ERROR;
				//goto exit;
				continue;
			}

			// When you delete a store the parent id is NULL, object type must be MAPI_STORE
			if(lpDBRow[1] == NULL && atoi(lpDBRow[2]) != MAPI_STORE) {
				//er = ZARAFA_E_DATABASE_ERROR;
				//goto exit;
				continue;
			}

			// Loop protection, don't insert duplicates.
			if (setIDs.insert(atoui(lpDBRow[0])).second == false)
				continue;
		
			sItem.ulId = atoui(lpDBRow[0]);
			sItem.ulParent = (lpDBRow[1])?atoui(lpDBRow[1]) : 0;
			sItem.ulObjType = atoi(lpDBRow[2]);
			sItem.ulFlags = atoui(lpDBRow[3]);
			sItem.ulObjSize = 0;
			sItem.ulStoreId = 0;
			sItem.ulParentType = (lpDBRow[4])?atoui(lpDBRow[4]) : 0; 
			sItem.sEntryId.__size = 0;
			sItem.sEntryId.__ptr = NULL;

			// Validate deleted object, if no valid, break directly
			er = ValidateDeleteObject(lpSession, bCheckPermission, ulFlags, sItem);
			if (er != erSuccess)
				goto exit;

			// Get extended data
			if(sItem.ulObjType == MAPI_STORE || sItem.ulObjType == MAPI_FOLDER || sItem.ulObjType == MAPI_MESSAGE) {
			
				lpCacheManager->GetStore(sItem.ulId, &sItem.ulStoreId , NULL); //CHECKme:"oude gaf geen errors
			
				if (!(sItem.ulFlags&MSGFLAG_DELETED) ) {
					GetObjectSize(lpDatabase, sItem.ulStoreId, sItem.ulId, &sItem.ulObjSize);
				}

				lpCacheManager->GetEntryIdFromObject(sItem.ulId, NULL, &sItem.sEntryId);//CHECKme:"oude gaf geen errors

				GetSourceKey(sItem.ulId, &sItem.sSourceKey);
				GetSourceKey(sItem.ulParent, &sItem.sParentSourceKey);
			}

			lstDeleteItems.push_back(sItem);
		}
	}

	// Now, run through the list, adding children to the bottom of the list. This means
	// we're actually running width-first, and don't have to do anything recursive.
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) 
	{

		// Free database results
		if(lpDBResult) { lpDatabase->FreeResult(lpDBResult); lpDBResult = NULL; }

		strQuery = "SELECT id, type, flags FROM hierarchy WHERE parent=" + stringify(iterDeleteItems->ulId);
		if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE)
			strQuery += " AND (flags&"+stringify(MSGFLAG_DELETED)+") !="+stringify(MSGFLAG_DELETED);

		er = lpDatabase->DoSelect(strQuery, &lpDBResult);
		if(er != erSuccess)
			goto exit;

		while((lpDBRow = lpDatabase->FetchRow(lpDBResult)) != NULL )
		{

			// No id, type or flags exist
			if(lpDBRow[0] == NULL || lpDBRow[1] == NULL || lpDBRow[2] == NULL)
				continue;

			// Loop protection, don't insert duplicates.
			if (setIDs.insert(atoui(lpDBRow[0])).second == false)
				continue;

			// Add this object as a node to the end of the list
			sItem.fRoot = false;
			sItem.ulObjSize = 0;
			sItem.ulStoreId = 0;
			sItem.sEntryId.__size = 0;
			sItem.sEntryId.__ptr = NULL;

			sItem.ulId = atoui(lpDBRow[0]);
			sItem.ulParent = iterDeleteItems->ulId;
			sItem.ulParentType = iterDeleteItems->ulObjType;
			sItem.ulObjType = atoi(lpDBRow[1]);
			sItem.ulFlags = atoui(lpDBRow[2]) | (iterDeleteItems->ulFlags&MSGFLAG_DELETED); // Add the parent delete flag, because only the top-level object is marked for deletion
			

			// Validate deleted object, if no valid, break directly
			er = ValidateDeleteObject(lpSession, bCheckPermission, ulFlags, sItem);
			if (er != erSuccess)
				goto exit;

			if(sItem.ulObjType == MAPI_STORE || sItem.ulObjType == MAPI_FOLDER || (sItem.ulObjType == MAPI_MESSAGE && sItem.ulParentType == MAPI_FOLDER) ) {
			
				lpCacheManager->GetStore(sItem.ulId, &sItem.ulStoreId , NULL); //CHECKme:"oude gaf geen errors

				if (!(sItem.ulFlags&MSGFLAG_DELETED) ) {
					GetObjectSize(lpDatabase, sItem.ulStoreId, sItem.ulId, &sItem.ulObjSize);
				}

				lpCacheManager->GetEntryIdFromObject(sItem.ulId, NULL, &sItem.sEntryId);//CHECKme:"oude gaf geen errors

				GetSourceKey(sItem.ulId, &sItem.sSourceKey);
				GetSourceKey(sItem.ulParent, &sItem.sParentSourceKey);
			}
			
			lstDeleteItems.push_back(sItem);
		}
	}
	
	// Move list
	std::swap(lstDeleteItems, *lplstDeleteItems);

exit:
	FreeDeletedItems(&lstDeleteItems);

	if(lpDBResult)
		lpDatabase->FreeResult(lpDBResult);

	return er;
}

/*
 * Add changes into the ICS system.
 * 
 * This adds a change for each removed folder and message. The change could be a soft- or hard delete.
 * It is possible to gives a list with different deleted objects, all not supported object types will be skipped.
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleted List with deleted objects
 * @param[in] ulSyncId ???
 *
 */
ECRESULT DeleteObjectUpdateICS(ECSession *lpSession, unsigned int ulFlags, ECListDeleteItems &lstDeleted, unsigned int ulSyncId)
{
	ECRESULT er = erSuccess;
	ECListDeleteItems::iterator iterDeleteItems;

	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++)
	{
		// ICS update
		if (iterDeleteItems->ulObjType == MAPI_MESSAGE && iterDeleteItems->ulParentType == MAPI_FOLDER)
			AddChange(lpSession, ulSyncId, iterDeleteItems->sSourceKey, iterDeleteItems->sParentSourceKey, ulFlags & EC_DELETE_HARD_DELETE ? ICS_MESSAGE_HARD_DELETE : ICS_MESSAGE_SOFT_DELETE);
		else if (iterDeleteItems->ulObjType == MAPI_FOLDER && !(iterDeleteItems->ulFlags & FOLDER_SEARCH)) {
			AddChange(lpSession, ulSyncId, iterDeleteItems->sSourceKey, iterDeleteItems->sParentSourceKey, ulFlags & EC_DELETE_HARD_DELETE ? ICS_FOLDER_HARD_DELETE : ICS_FOLDER_SOFT_DELETE);
		}
	}

	return er;
}

/*
 * Calculate and update the store size for deleted items
 *
 * The DeleteObjectStoreSize methode calculate and update the store size. Only top-level messages will
 * be calculate, all other objects are not supported and will be skipped. If a message has the 
 * MSGFLAG_DELETED flag, the size will ignored because it is already subtract from the store size.
 * The deleted object list may include more than one store.
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] lpDatabase Reference to a database object; cannot be NULL.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleted List with deleted objects
 */
ECRESULT DeleteObjectStoreSize(ECSession *lpSession, ECDatabase *lpDatabase, unsigned int ulFlags, ECListDeleteItems &lstDeleted)
{
	ECRESULT er = erSuccess;
	std::map<unsigned int, long long> mapStoreSize;

	ECListDeleteItems::iterator iterDeleteItems;
	std::map<unsigned int, long long>::iterator iterStoreSizeItems;

//TODO: check or foldersize also is used

	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++) {
		// Get size of all the messages
		if( iterDeleteItems->ulObjType == MAPI_MESSAGE && 
			iterDeleteItems->ulParentType == MAPI_FOLDER && 
			(iterDeleteItems->ulFlags&MSGFLAG_DELETED) != MSGFLAG_DELETED)
		{
			ASSERT(iterDeleteItems->ulStoreId != 0);

			if (mapStoreSize.find(iterDeleteItems->ulStoreId) != mapStoreSize.end() )
				mapStoreSize[iterDeleteItems->ulStoreId] += iterDeleteItems->ulObjSize;
			else
				mapStoreSize[iterDeleteItems->ulStoreId] = iterDeleteItems->ulObjSize;
		}
	}

	// Update store size for each store
	for(iterStoreSizeItems=mapStoreSize.begin(); iterStoreSizeItems != mapStoreSize.end(); iterStoreSizeItems++) {
		if ( iterStoreSizeItems->second > 0)	
			UpdateObjectSize(lpDatabase, iterStoreSizeItems->first, iterStoreSizeItems->first, MAPI_STORE, false, UPDATE_SUB, iterStoreSizeItems->second);
	}

exit:
	return er;
}

/*
 * Soft delete objects, mark the root objects as deleted
 *
 * Mark the root objects as deleted, add deleted on date on the root objects.
 * Since this is a fairly simple operation, we are doing soft deletes in a single transaction. Once the SQL has gone OK, 
 * we know that all items were successfully mark as deleted and we can therefore add all soft-deleted items into 
 * the 'lstDeleted' list at once
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] lpDatabase Reference to a database object; cannot be NULL.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleteItems List with objecta which must be deleted.
 * @param[out] lstDeleted List with deleted objects.
 *
 */
ECRESULT DeleteObjectSoft(ECSession *lpSession, ECDatabase *lpDatabase, unsigned int ulFlags, ECListDeleteItems &lstDeleteItems, ECListDeleteItems &lstDeleted)
{
	ECRESULT er = erSuccess;
	ECListDeleteItems::iterator iterDeleteItems;

	FILETIME ft;
	std::string strInclauseOQQ;
	std::string strInclause;
	std::string strQuery;

	// Build where condition
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {

		if( (iterDeleteItems->ulObjType == MAPI_MESSAGE && iterDeleteItems->ulParentType == MAPI_FOLDER) || 
			  iterDeleteItems->ulObjType == MAPI_FOLDER  || iterDeleteItems->ulObjType == MAPI_STORE) 
		{
			if ( (iterDeleteItems->ulFlags&MSGFLAG_SUBMIT) && iterDeleteItems->ulObjType == MAPI_MESSAGE) {
				if(!strInclauseOQQ.empty())
					strInclauseOQQ += ",";

				strInclauseOQQ += stringify(iterDeleteItems->ulId);
			}
			
			if (iterDeleteItems->fRoot == true)
			{
				if(!strInclause.empty())
					strInclause += ",";

				strInclause += stringify(iterDeleteItems->ulId);
			}
		}
	}

	// Mark all items as deleted, if a item in the outgoingqueue and remove the submit flag
	if (!strInclauseOQQ.empty())
	{
		// Remove any entries in the outgoing queue for deleted items
		strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id IN ( " + strInclauseOQQ + ")";
		er = lpDatabase->DoDelete(strQuery);
		if(er!= erSuccess)
			goto exit;
		
		// Mark item as deleted and remove the submit flag
		strQuery = "UPDATE hierarchy SET flags=flags&~" + stringify(MSGFLAG_SUBMIT)+" WHERE id IN(" + strInclauseOQQ + ")";
		er = lpDatabase->DoUpdate(strQuery);
		if(er!= erSuccess)
			goto exit;
	}

	if(!strInclause.empty())
	{
		// Mark item as deleted
		strQuery = "UPDATE hierarchy SET flags=flags|"+stringify(MSGFLAG_DELETED)+" WHERE id IN(" + strInclause + ")";
		er = lpDatabase->DoUpdate(strQuery);
		if(er!= erSuccess)
			goto exit;

	}

	// Add properties: PR_DELETED_ON
	GetSystemTimeAsFileTime(&ft);

	// TODO: move PR_DELETED_ON to another new table
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++)
	{
		if( iterDeleteItems->fRoot == true && 
			( (iterDeleteItems->ulObjType == MAPI_MESSAGE && iterDeleteItems->ulParentType == MAPI_FOLDER) || 
			iterDeleteItems->ulObjType == MAPI_FOLDER  || iterDeleteItems->ulObjType == MAPI_STORE) )
		{
			strQuery = "REPLACE INTO properties(hierarchyid, tag, type, storeid, val_lo, val_hi) VALUES("+stringify(iterDeleteItems->ulId)+","+stringify(PROP_ID(PR_DELETED_ON))+","+stringify(PROP_TYPE(PR_DELETED_ON))+","+stringify(iterDeleteItems->ulStoreId)+","+stringify(ft.dwLowDateTime)+","+stringify(ft.dwHighDateTime)+")";
			er = lpDatabase->DoUpdate(strQuery);
			if(er!= erSuccess)
				goto exit;
		}
	}

	lstDeleted = lstDeleteItems;

exit:

	return er;
}

/**
 * Hard delete objects, remove the data from storage
 *
 * This means we should be really deleting the actual data from the database and storage. This will be done in 
 * bachtches of 32 items each because deleteing records is generally fairly slow. Also, very large delete batches 
 * can taking up to more than an hour to process. We don't want to have a transaction lasting an hour because it 
 * would cause lots of locking problems. Also, each item successfully deleted and committed to the database will 
 * added into a list. So, If something fails we notify the items in the 'deleted items list' only.
 * 
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] lpDatabase Reference to a database object; cannot be NULL.
 * @param[in] lpAttachmentStorage Reference to an Attachment object. could not NULL if bNoTransaction is true.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleteItems List with objects which must be deleted.
 * @param[in] bNoTransaction Disable the database transactions.
 * @param[in] lstDeleted List with deleted objects.
 *
 * @return Zarafa error code
 */
ECRESULT DeleteObjectHard(ECSession *lpSession, ECDatabase *lpDatabase, ECAttachmentStorage *lpAttachmentStorage, unsigned int ulFlags, ECListDeleteItems &lstDeleteItems, bool bNoTransaction, ECListDeleteItems &lstDeleted)
{
	ECRESULT er = erSuccess;
	ECAttachmentStorage *lpInternalAttachmentStorage = NULL;
	std::list<ULONG> lstDeleteAttachments;
	std::string strInclause;
	std::string strQuery;
	ECListDeleteItems lstToBeDeleted;
	ECListDeleteItems::iterator iterDeleteItems;

	int i;

	if(!(ulFlags & EC_DELETE_HARD_DELETE)) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (bNoTransaction && lpAttachmentStorage == NULL) {
		ASSERT(FALSE);
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if (!lpAttachmentStorage) {
		er = CreateAttachmentStorage(lpDatabase, &lpInternalAttachmentStorage);
		if (er != erSuccess)
			goto exit;

		lpAttachmentStorage = lpInternalAttachmentStorage;
	}

// @todo use reverse loop, this is much better because lost data
	for (iterDeleteItems = lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); )
	{
		strInclause.clear();
		lstDeleteAttachments.clear();
		lstToBeDeleted.clear();
		i = 0;

		// Delete max 32 items per query
		while (i<32 && iterDeleteItems != lstDeleteItems.end())
		{
			if(!strInclause.empty())
				strInclause += ",";

			strInclause += stringify(iterDeleteItems->ulId);

			// make new list for attachment deletes
			if (iterDeleteItems->ulObjType == MAPI_ATTACH)
				lstDeleteAttachments.push_back(iterDeleteItems->ulId);

			lstToBeDeleted.push_back(*iterDeleteItems);

			i++;
			iterDeleteItems++;
		}

		// Start transaction
		if (!bNoTransaction) {
			er = lpAttachmentStorage->Begin();
			if (er != erSuccess)
				goto exit;

			er = lpDatabase->Begin();
			if (er != erSuccess)
				goto exit;
		}

		if(!strInclause.empty()) {
			// First, Remove any entries in the outgoing queue for deleted items
			strQuery = "DELETE FROM outgoingqueue WHERE hierarchy_id IN ( " + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er!= erSuccess)
				goto exit;
			
			// Then, the hierarchy entries of all the objects
			strQuery = "DELETE FROM hierarchy WHERE id IN (" + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er!= erSuccess)
				goto exit;

			// Then, the properties for the objects we just deleted
			strQuery = "DELETE FROM properties WHERE hierarchyid IN (" + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er!= erSuccess)
				goto exit;

			// Then, the MVproperties for the objects we just deleted
			strQuery = "DELETE FROM mvproperties WHERE hierarchyid IN (" + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er!= erSuccess)
				goto exit;

			// Then, the acls for the objects we just deleted (if exist)
			strQuery = "DELETE FROM acl WHERE hierarchy_id IN (" + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er != erSuccess)
				goto exit;

			// remove indexedproperties
			strQuery = "DELETE FROM indexedproperties WHERE hierarchyid IN (" + strInclause + ")";
			er = lpDatabase->DoDelete(strQuery);
			if(er != erSuccess)
				goto exit;
		}
		// list may contain non-attachment object id's!
		if (!lstDeleteAttachments.empty()) {
			er = lpAttachmentStorage->DeleteAttachments(lstDeleteAttachments);
			if (er != erSuccess)
				goto exit;
		}

		// Commit the transaction
		if (!bNoTransaction) {
			er = lpAttachmentStorage->Commit();
			if (er != erSuccess)
				goto exit;

			er = lpDatabase->Commit();
			if(er != erSuccess)
				goto exit;
		}

		// Deletes have been committed, add the deleted items to the list of items we have deleted
		lstDeleted.splice(lstDeleted.end(),lstToBeDeleted);

	} // while iterDeleteItems != end()

exit:
	if (er != erSuccess && !bNoTransaction) {
		if(lpInternalAttachmentStorage)
			lpInternalAttachmentStorage->Rollback();

		lpDatabase->Rollback();
	}

	if (lpInternalAttachmentStorage)
		lpInternalAttachmentStorage->Release();

	return er;
}

/*
 * Deleted object cache updates
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleted List with deleted objects. 
 */
ECRESULT DeleteObjectCacheUpdate(ECSession *lpSession, unsigned int ulFlags, ECListDeleteItems &lstDeleted)
{
	ECRESULT er = erSuccess;
	ECSessionManager *lpSessionManager = NULL;
	ECCacheManager *lpCacheManager = NULL;
	ECListDeleteItems::iterator iterDeleteItems;

	if (lpSession == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpSessionManager = lpSession->GetSessionManager();
	lpCacheManager = lpSessionManager->GetCacheManager();

	// Remove items from cache and update the outgoing queue
	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++) {

		// update the cache
		lpCacheManager->Update(fnevObjectDeleted, iterDeleteItems->ulId);

		if (iterDeleteItems->fRoot)
			lpCacheManager->Update(fnevObjectModified, iterDeleteItems->ulParent);

		// Update cache, Remove index properties
		if((ulFlags & EC_DELETE_HARD_DELETE) == EC_DELETE_HARD_DELETE)
			lpCacheManager->RemoveIndexData(iterDeleteItems->ulId);
	}

exit:
	
	return er;
}

/*
 * Deleted object notifications
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] lstDeleted List with deleted objects.
 */

ECRESULT DeleteObjectNotifications(ECSession *lpSession, unsigned int ulFlags, ECListDeleteItems &lstDeleted)
{
	ECRESULT er = erSuccess;
	ECSessionManager *lpSessionManager = NULL;
	ECListDeleteItems::iterator iterDeleteItems;

	std::list<unsigned int> lstParent;
	std::list<unsigned int>::iterator iterParent;
	ECMapTableChangeNotifications mapTableChangeNotifications;
	//std::set<unsigned int>	setFolderParents;
	size_t cDeleteditems = lstDeleted.size();
	unsigned int ulGrandParent = 0;

	if (lpSession == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpSessionManager = lpSession->GetSessionManager();

	// Now, send the notifications for MAPI_MESSAGE and MAPI_FOLDER types
	for(iterDeleteItems=lstDeleted.begin(); iterDeleteItems != lstDeleted.end(); iterDeleteItems++)
	{
		// Update the outgoing queue
		// Remove the item from both the local and master outgoing queues
		if( (iterDeleteItems->ulFlags & MSGFLAG_SUBMIT) && iterDeleteItems->ulParentType == MAPI_FOLDER && iterDeleteItems->ulObjType == MAPI_MESSAGE) {
			lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, iterDeleteItems->ulStoreId, iterDeleteItems->ulId, EC_SUBMIT_LOCAL, MAPI_MESSAGE);
			lpSessionManager->UpdateOutgoingTables(ECKeyTable::TABLE_ROW_DELETE, iterDeleteItems->ulStoreId, iterDeleteItems->ulId, EC_SUBMIT_MASTER, MAPI_MESSAGE);
		}

		if( (iterDeleteItems->ulParentType == MAPI_FOLDER && iterDeleteItems->ulObjType == MAPI_MESSAGE) || 
			  iterDeleteItems->ulObjType == MAPI_FOLDER ) 
		{
			// Notify that the message has been deleted
			lpSessionManager->NotificationDeleted(iterDeleteItems->ulObjType, iterDeleteItems->ulId, iterDeleteItems->ulStoreId, &iterDeleteItems->sEntryId, iterDeleteItems->ulParent, iterDeleteItems->ulFlags& (MSGFLAG_ASSOCIATED|MSGFLAG_DELETED) );

			// Update all tables viewing this message
			if(cDeleteditems < EC_TABLE_CHANGE_THRESHOLD) {
				lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_DELETE, (iterDeleteItems->ulFlags & (MSGFLAG_ASSOCIATED|MSGFLAG_DELETED)), iterDeleteItems->ulParent, iterDeleteItems->ulId, iterDeleteItems->ulObjType);
		
				if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE) {
					// Update all tables viewing this message
					lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_ADD, MSGFLAG_DELETED, iterDeleteItems->ulParent, iterDeleteItems->ulId, iterDeleteItems->ulObjType);
				}
			} else {
				// We need to send a table change notifications later on
				mapTableChangeNotifications[iterDeleteItems->ulParent].insert(TABLECHANGENOTIFICATION(iterDeleteItems->ulObjType, iterDeleteItems->ulFlags & MSGFLAG_NOTIFY_FLAGS));
				if((ulFlags & EC_DELETE_HARD_DELETE) != EC_DELETE_HARD_DELETE) {
					mapTableChangeNotifications[iterDeleteItems->ulParent].insert(TABLECHANGENOTIFICATION(iterDeleteItems->ulObjType, (iterDeleteItems->ulFlags & MSGFLAG_NOTIFY_FLAGS) | MSGFLAG_DELETED));
				}
			}

			// @todo: Is this correct ???
			if (iterDeleteItems->fRoot)
				 lstParent.push_back(iterDeleteItems->ulParent);
		}
	}

	// We have a list of all the folders in which something was deleted, so get a unique list
	lstParent.sort();
	lstParent.unique();

	// Now, send each parent folder a notification that it has been altered and send
	// its parent a notification (ie the grandparent of the deleted object) that its
	// hierarchy table has been changed.
	for(iterParent = lstParent.begin(); iterParent != lstParent.end(); iterParent++) {

		if(cDeleteditems >= EC_TABLE_CHANGE_THRESHOLD) {

			// Find the set of notifications to send for the current parent.
			ECMapTableChangeNotifications::const_iterator iParentNotifications = mapTableChangeNotifications.find(*iterParent);
			if (iParentNotifications != mapTableChangeNotifications.end()) {
				ECSetTableChangeNotifications::const_iterator iNotification;

				// Iterate the set and send notifications.
				for (iNotification = iParentNotifications->second.begin(); iNotification != iParentNotifications->second.end(); ++iNotification)
					lpSessionManager->UpdateTables(ECKeyTable::TABLE_CHANGE, iNotification->ulFlags, *iterParent, 0, iNotification->ulType);
			}
		}

		lpSessionManager->NotificationModified(MAPI_FOLDER, *iterParent);

		if(lpSessionManager->GetCacheManager()->GetParent(*iterParent, &ulGrandParent) == erSuccess)
			lpSessionManager->UpdateTables(ECKeyTable::TABLE_ROW_MODIFY, 0, ulGrandParent, *iterParent, MAPI_FOLDER);
	}

exit:
	return er;
}



/*
 * Delete objects from different stores.
 *
 * Delete a store, folders, messages, reciepints and attachments.
 *
 * @param[in] lpSession Reference to a session object; cannot be NULL.
 * @param[in] lpDatabase Reference to a database object; cannot be NULL.
 * @param[in] lpsObjectList Reference to a list of objects that contains itemid and must be expanded.
 * @param[in] ulFlags Bitmask of flags that controls how the objects will deleted.
 * @param[in] ulSyncId ??????
 * @param[in] bNoTransaction Disable the database transactions.
 * @param[in] bCheckPermission Check the objects delete permissions.
 *
 * @return Zarafa error code
 */
ECRESULT DeleteObjects(ECSession *lpSession, ECDatabase *lpDatabase, ECListInt *lpsObjectList, unsigned int ulFlags, unsigned int ulSyncId, bool bNoTransaction, bool bCheckPermission)
{
	ECRESULT er = erSuccess;
	ECListDeleteItems lstDeleteItems;
	ECListDeleteItems lstDeleted;
	ECListDeleteItems::iterator iterDeleteItems;
	ECSearchFolders *lpSearchFolders = NULL;
	ECSessionManager *lpSessionManager = NULL;
	
	if (lpSession == NULL || lpDatabase == NULL || lpsObjectList == NULL) {
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	lpSessionManager = lpSession->GetSessionManager();
	lpSearchFolders = lpSessionManager->GetSearchFolders();

	if (bNoTransaction && (ulFlags & EC_DELETE_HARD_DELETE)) {
		ASSERT(FALSE); // This means that the caller has a transaction but that's not allowed
		er = ZARAFA_E_INVALID_PARAMETER;
		goto exit;
	}

	if(!(ulFlags & EC_DELETE_HARD_DELETE) && !bNoTransaction) {
		er = lpDatabase->Begin();
		if (er != erSuccess)
			goto exit;
	}

	// Collect recursive parent objects, validate item and check the permissions
	er = ExpandDeletedItems(lpSession, lpDatabase, lpsObjectList, ulFlags, bCheckPermission, &lstDeleteItems);
	if (er != erSuccess) {
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Error while expanding delete item list, error code: %u", er);
		goto exit;
	}

	// Remove search results for deleted folders
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {
		if(iterDeleteItems->ulObjType == MAPI_FOLDER && iterDeleteItems->ulFlags == FOLDER_SEARCH) {
			lpSearchFolders->RemoveSearchFolder(iterDeleteItems->ulStoreId, iterDeleteItems->ulId);
		}
	}
	
	// Mark or delete objects
	if(ulFlags & EC_DELETE_HARD_DELETE)
		er = DeleteObjectHard(lpSession, lpDatabase, NULL, ulFlags, lstDeleteItems, bNoTransaction, lstDeleted);
	else
		er = DeleteObjectSoft(lpSession, lpDatabase, ulFlags, lstDeleteItems, lstDeleted);
	if (er != erSuccess) {
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Error while expanding delete item list, error code: %u", er);
		goto exit;
	}

	// Update store size
	er = DeleteObjectStoreSize(lpSession, lpDatabase, ulFlags, lstDeleted);
	if(er!= erSuccess) {
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Error while updating store sizes after delete, error code: %u", er);
		goto exit;
	}

	// Update ICS
	er = DeleteObjectUpdateICS(lpSession, ulFlags, lstDeleted, ulSyncId);
	if (er != erSuccess) {
		g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Error while updating ICS after delete, error code: %u", er);
		goto exit;
	}

	// Update local commit time on top level folders
	for(iterDeleteItems=lstDeleteItems.begin(); iterDeleteItems != lstDeleteItems.end(); iterDeleteItems++) {
		if( !(ulFlags & EC_DELETE_HARD_DELETE) && iterDeleteItems->fRoot &&
		    iterDeleteItems->ulParentType == MAPI_FOLDER && iterDeleteItems->ulObjType == MAPI_MESSAGE)
		{
			// directly hard-delete the item is not supported for updating PR_LOCAL_COMMIT_TIME_MAX
			er = WriteLocalCommitTimeMax(NULL, lpDatabase, iterDeleteItems->ulStoreId, iterDeleteItems->ulParent, NULL);
			if (er != erSuccess) {
				g_lpSessionManager->GetLogger()->Log(EC_LOGLEVEL_INFO, "Error while updating folder access time after delete, error code: %u", er);
				goto exit;
			}
			// the folder will receive a changed notification anyway, since items are being deleted from it
		}
	}
	
	// Finish transaction
	if(!(ulFlags & EC_DELETE_HARD_DELETE) && !bNoTransaction) {
		er = lpDatabase->Commit();
		if (er != erSuccess)
			goto exit;
	}

	// Update cache
	DeleteObjectCacheUpdate(lpSession, ulFlags, lstDeleted);

	// Send notifications
	DeleteObjectNotifications(lpSession, ulFlags, lstDeleted);

exit:
	if(er != erSuccess && !bNoTransaction && !(ulFlags & EC_DELETE_HARD_DELETE))
		lpDatabase->Rollback();

	FreeDeletedItems(&lstDeleteItems);

	return er;
}

/** 
 * Update PR_LOCAL_COMMIT_TIME_MAX property on a folder which contents has changed.
 *
 * This function should be called when the contents of a folder change
 * Affected: saveObject, emptyFolder, deleteObjects, (not done: copyObjects, copyFolder)
 * 
 * @param[in]  soap soap struct used for allocating memory for return value, can be NULL
 * @param[in]  lpDatabase database pointer, should be in transaction already
 * @param[in]  ulStoreId store id of folder
 * @param[in]  ulFolderId folder to update property in
 * @param[out] ppvTime time property that was written on the folder, can be NULL
 * 
 * @return Zarafa error code
 * @retval ZARAFA_E_DATABASE_ERROR database could not be updated
 */
ECRESULT WriteLocalCommitTimeMax(struct soap *soap, ECDatabase *lpDatabase, unsigned int ulStoreId, unsigned int ulFolderId, propVal **ppvTime)
{
	ECRESULT er = erSuccess;
	FILETIME ftNow;
	std::string strQuery;
	propVal *pvTime = NULL;

	GetSystemTimeAsFileTime(&ftNow);

	if (soap && ppvTime) {
		pvTime = s_alloc<propVal>(soap);
		pvTime->ulPropTag = PR_LOCAL_COMMIT_TIME_MAX;
		pvTime->__union = SOAP_UNION_propValData_hilo;
		pvTime->Value.hilo = s_alloc<hiloLong>(soap);
		pvTime->Value.hilo->hi = ftNow.dwHighDateTime;
		pvTime->Value.hilo->lo = ftNow.dwLowDateTime;
	}

	strQuery = "REPLACE INTO properties (storeid, hierarchyid, tag, type, val_hi, val_lo) VALUES ("
		+stringify(ulStoreId)+","+stringify(ulFolderId)+","+stringify(PROP_ID(PR_LOCAL_COMMIT_TIME_MAX))+","+stringify(PROP_TYPE(PR_LOCAL_COMMIT_TIME_MAX))+","
		+stringify(ftNow.dwHighDateTime)+","+stringify(ftNow.dwLowDateTime)+")";

	er = lpDatabase->DoInsert(strQuery);
	if (er != erSuccess)
		goto exit;

	if (ppvTime)
		*ppvTime = pvTime;

exit:
	return er;
}

void FreeDeleteItem(DELETEITEM *src)
{
	if(src->sEntryId.__ptr)
		delete [] src->sEntryId.__ptr;
}

void FreeDeletedItems(ECListDeleteItems *lplstDeleteItems)
{
	ECListDeleteItems::iterator iterDeleteItems;
	
	for(iterDeleteItems=lplstDeleteItems->begin(); iterDeleteItems != lplstDeleteItems->end(); iterDeleteItems++) {
		FreeDeleteItem(&(*iterDeleteItems));
	}

	lplstDeleteItems->clear();
}

